/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Sosftware
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include <stdio.h>
#include <string.h>

#include "dicom.h"
#include "lst.h"
#include "dicom_objects.h"

#include "dcmdict.h"
#include "CDcmScanner.h"

#ifdef DCM_UNSPECIFIEDLENGTH
#undef DCM_UNSPECIFIEDLENGTH
const unsigned long DCM_UNSPECIFIEDLENGTH = (unsigned long)0xFFFFFFFF;
#endif

//--------------------------------------
//
//--------------------------------------
void tDcmElementHdr::ReadHdr(const char* const ptr, const bool explicitVR)
{
	m_tag[0] = *reinterpret_cast<const unsigned short*>(ptr);
	m_tag[1] = *reinterpret_cast<const unsigned short*>(ptr + 2);

	if (m_tag[0] == 0xFFFF && m_tag[1] == 0xFFFF) {
		m_length = 0;
		m_hsize = 4;
	} else if (explicitVR) {
		if (m_tag[0] != DCM_GROUPDELIMITER) {
			m_vr = ptr + 4;
			if (isSQ() || isBLE()) {
				m_length = *reinterpret_cast<const unsigned long*>(ptr + 8);
				m_hsize = 12;
			} else {
				m_length = static_cast<unsigned long>(*reinterpret_cast<const unsigned short*>(ptr + 6));
				m_hsize = 8;
			}
		} else {
			const char* const del = "DLM";
			m_vr = del;
			m_length = *reinterpret_cast<const unsigned long*>(ptr + 4);
			m_hsize = 8;
		}
	} else {
		m_length = *reinterpret_cast<const unsigned long*>(ptr + 4);
		m_hsize = 8;
	}
}

//--------------------------------------
//
//--------------------------------------
template<class T, class R>
int get_value_from_binary(const tDcmElementHdr* hdr, const char* ptr, int nth)
{
	int count = hdr->value_length() / sizeof(T);
	if (nth >= count)
		return 0;
	T* va = (T*)hdr->value_ptr(ptr);
	return (R)va[nth];
}
//--------------------------------------
//
//--------------------------------------
int tDcmElementHdr::get_integer(const char* ptr, int nth)
{
	char str[256];

	if (nth < 0)
		return 0;

	if (this->m_vr[0] == 'I' && this->m_vr[1] == 'S') {	// IS
		this->get_string(ptr, str, sizeof(str), nth, true);
		return (int)atol(str);
	} else if (this->m_vr[0] == 'D' && this->m_vr[1] == 'A') {	// DA
		this->get_string(ptr, str, sizeof(str), nth, true);
		char* sp = str;
		char* dp = str;
		char* ep = sp + strlen(str) - 1;
		while (sp < ep) {
			if (isdigit(*sp)) {
				if (dp != sp)
					*dp = *sp;
				dp++;
			}
			sp++;
		}
		*dp = '\0';
		return (int)atol(str);	
	} else if (this->m_vr[0] == 'S' && this->m_vr[1] == 'S') {	// SS
		return get_value_from_binary<signed short, int>(this, ptr, nth);
	} else if (this->m_vr[0] == 'S' && this->m_vr[1] == 'L') {	// SL
		return get_value_from_binary<signed long, int>(this, ptr, nth);
	} else if (this->m_vr[0] == 'U' && this->m_vr[1] == 'S') {	// US
		return get_value_from_binary<unsigned short, int>(this, ptr, nth);
	} else if ((this->m_vr[0] == 'U' && this->m_vr[1] == 'L')	// UL
	        || (this->m_vr[0] == 'A' && this->m_vr[1] == 'T')) {	// AT
		return get_value_from_binary<unsigned long, int>(this, ptr, nth);
	} else if ((this->m_vr[0] == 'D' && this->m_vr[1] == 'S')	// DS
	        || (this->m_vr[0] == 'F' && this->m_vr[1] == 'L')	// FL
	        || (this->m_vr[0] == 'F' && this->m_vr[1] == 'D')	// FD
	        || (this->m_vr[0] == 'D' && this->m_vr[1] == 'T')	// DT
	        || (this->m_vr[0] == 'T' && this->m_vr[1] == 'M')) {	// TM
		return (int)this->get_real(ptr);
	} else {
		return -1;
	}
}
//--------------------------------------
//
//--------------------------------------
double tDcmElementHdr::get_real(const char* ptr, int nth)
{
	if (nth < 0)
		return 0;

	char str[256];

	if ((this->m_vr[0] == 'A' && this->m_vr[1] == 'T')	// AT
	 || (this->m_vr[0] == 'I' && this->m_vr[1] == 'S')	// IS
	 || (this->m_vr[0] == 'D' && this->m_vr[1] == 'A')	// DA
	 || (this->m_vr[0] == 'S' && this->m_vr[1] == 'S')	// SS
	 || (this->m_vr[0] == 'S' && this->m_vr[1] == 'L')	// SL
	 || (this->m_vr[0] == 'U' && this->m_vr[1] == 'S')	// US
	 || (this->m_vr[0] == 'U' && this->m_vr[1] == 'L')) {	// UL
		return (double)this->get_integer(ptr, nth);
	} else if (this->m_vr[0] == 'D' && this->m_vr[1] == 'S') {	// DS
		this->get_string(ptr, str, sizeof(str), nth, true);
		return strtod(str, NULL);
	} else if (this->m_vr[0] == 'T' && this->m_vr[1] == 'M') {	// TM
		this->get_string(ptr, str, sizeof(str), nth, true);
		char* sp = str;
		char* dp = str;
		char* ep = sp + strlen(str) - 1;
		while (sp < ep) {
			if (isdigit(*sp) || *sp == '.') {
				if (dp != sp)
					*dp = *sp;
				dp++;
			}
			sp++;
		}
		*dp = '\0';
		return strtod(str, NULL);
	} else if (this->m_vr[0] == 'D' && this->m_vr[1] == 'T') {	// DT
		this->get_string(ptr, str, sizeof(str), nth, true);
		char* sp = str;
		char* dp = str;
		char* ep = sp + strlen(str) - 1;
		while (sp < ep && *sp != '+' && *sp != '-') {
			if (isdigit(*sp) || *sp == '.') {
				if (dp != sp)
					*dp = *sp;
				dp++;
			}
			sp++;
		}
		*dp = '\0';
		return strtod(str, NULL);
	} else if (this->m_vr[0] == 'F' && this->m_vr[1] == 'L') {	// FL
		return get_value_from_binary<unsigned long, float>(this, ptr, nth);
	} else if (this->m_vr[0] == 'F' && this->m_vr[1] == 'D') {	// FL
		return get_value_from_binary<unsigned long, double>(this, ptr, nth);
	} else {
		return 0;
	}
}
//--------------------------------------
//
//--------------------------------------
void tDcmElementHdr::get_string(const char* ptr, char* str_buf, size_t str_buf_len, int nth, bool trimSpace)
{
	const char* cp = this->value_ptr(ptr);
	const char* ep = cp + this->value_length();
	if (nth > 0) {
		for (int i = 0; i < nth; i++) {
			while (cp < ep && *cp != '\\')
				cp++;
			if (cp == ep) {
				*str_buf = '\0';
				return;
			}
			cp++;
		}
		const char* tp = cp;
		while (tp < ep && *tp != '\\')
			tp++;
		if (tp != ep)
			ep = tp;
	}

	if (trimSpace) {
		while (*cp <= ' ')
			cp++;
		while (*(ep - 1) <= ' ')
			ep--;
	}

	size_t len = ep - cp;
	if (len > str_buf_len - 1)
		len = str_buf_len - 1;

	CopyMemory(str_buf, cp, len);
	str_buf[len] = '\0';
}

//--------------------------------------
//
//--------------------------------------
void CDcmScannerR::ParseImpl(
		const char* const block_begin, const char* const block_end,
		const bool explicitVR,
		const bool skip_sq,
		unsigned long& advance,
		const bool stop_at_delim)
{
	const char* cp = block_begin;

	while (cp < block_end) {
		tDcmElementHdr hdr;
		hdr.ReadHdr(cp, explicitVR);

		if (stop_at_delim && hdr.ltag() == DCM_DLMITEMDELIMITATIONITEM) {
			cp += hdr.m_hsize;
			break;
		}

		if (hdr.m_tag[1] == 0) {
			const unsigned long group_length = *reinterpret_cast<const unsigned long*>(cp + hdr.m_hsize);
			bool skip_group = false;
			Event_GroupLength(hdr, group_length, skip_group);
			if (skip_group) {
				cp = hdr.next_ptr(cp);
				cp += group_length;
				continue;
			}
		}

		const DCMDICT* dict = NULL;
		if (!explicitVR) {
			dict = m_dcmdict.Lookup(hdr.ltag());
			if (dict != NULL) {
				hdr.m_vr = DcmGetVRStr(dict->representation);
			} else {
				const char* const pr = "private";
				hdr.m_vr = pr;
			}
		}
		
		if (m_sq_level == 0 || !skip_sq) {
			bool tobe_stop = false;
			Event_DataElement(hdr, hdr.value_ptr(cp), block_end, explicitVR, dict, tobe_stop);
			if (tobe_stop)
				break;
		}

		if (!hdr.isSQ()) {
			cp = hdr.next_ptr(cp);
			continue;
		}

		bool local_skip_sq = skip_sq;
		if (!local_skip_sq)
			Event_EnterSQ(hdr, local_skip_sq);

		cp += hdr.m_hsize;
		const bool sq_un_len = hdr.m_length == DCM_UNSPECIFIEDLENGTH;
		if (local_skip_sq && !sq_un_len) {
			cp += hdr.m_length;
			continue;
		}
		const char* const sq_end = sq_un_len ? block_end : cp + hdr.m_length;

		while (cp < sq_end) {
			hdr.ReadHdr(cp, explicitVR);
			cp += hdr.m_hsize;

			if (hdr.ltag() != DCM_DLMITEM) {
				if (sq_un_len && hdr.ltag() == DCM_DLMSEQUENCEDELIMITATIONITEM)
					break;
				return; // ERROR
			}

			const bool item_un_len = hdr.m_length == DCM_UNSPECIFIEDLENGTH;
			if (local_skip_sq && !item_un_len) {
				cp += hdr.m_length;
				continue;
			}
			const char* const item_end = item_un_len ? sq_end : cp + hdr.m_length;

			if (!local_skip_sq)
				Event_EnterSQItem(hdr);

			unsigned long adv;
			m_sq_level++;
			ParseImpl(cp, item_end, explicitVR, local_skip_sq, adv, item_un_len);
			m_sq_level--;
			cp += adv;
			

			if (!local_skip_sq)
				Event_LeaveSQItem();
		}
		if (!local_skip_sq)
			Event_LeaveSQ();

	}

	advance = cp - block_begin; 
}

//--------------------------------------
//
//--------------------------------------
int CDcmScannerF::NextElement(const char* const cp, tDcmElementHdr& hdr, const bool explicitVR)
{
	if (cp >= m_end)
		return E_EOF;

	if (*(const DCM_TAG*)cp == 0xFFFFFFFF)
		return E_EOF;

	hdr.ReadHdr(cp, explicitVR);

	if (hdr.m_length != DCM_UNSPECIFIEDLENGTH) {
		return S_CONTINUE;
	} else if (hdr.ltag() == DCM_PXLPIXELDATA) {
		return S_UNKNOWN_LENGTH_PIXELDATA;
	} else {
		const char* const sq_begin = cp + hdr.m_hsize;
		const char* sq_end = sq_begin;
		int ret = skip_unspecified_length_sq(sq_begin, sq_end, explicitVR);
		if (ret != E_EOF)
			return ret;
		hdr.m_length = sq_end - sq_begin;
		return S_CONTINUE;
	}
}
//--------------------------------------
//
//--------------------------------------
int CDcmScannerF::skip_unspecified_length_sq(const char* const sq_begin, const char*& sq_end, const bool explicitVR)
{
	const char* cp = sq_begin;
	while (cp < m_end) {
		tDcmElementHdr hdr;
		hdr.ReadHdr(cp, explicitVR);
		cp += hdr.m_hsize;
		if (hdr.m_tag[0] != DCM_GROUPDELIMITER)
			return E_SQ_ERROR;
		if (hdr.m_tag[1] == DCM_TAG_ELEMENT(DCM_DLMSEQUENCEDELIMITATIONITEM)) {
			sq_end = cp;
			return E_EOF;
		}
		if (hdr.m_tag[1] != DCM_TAG_ELEMENT(DCM_DLMITEM)) {
			return E_SQ_ERROR;
		}
		if (hdr.m_length != DCM_UNSPECIFIEDLENGTH) {
			cp += hdr.m_length;
			continue;
		} else {
			while (cp < m_end) {
				hdr.ReadHdr(cp, explicitVR);
				cp += hdr.m_hsize;
				if (hdr.ltag() == DCM_DLMITEMDELIMITATIONITEM) {
					break;
				}
				if (hdr.m_length != DCM_UNSPECIFIEDLENGTH) {
					cp += hdr.m_length;
				} else {
					const char* const sq_begin = cp;
					const char* sq_end = sq_begin;
					int ret = skip_unspecified_length_sq(sq_begin, sq_end, explicitVR);
					if (ret != E_EOF)
						return ret;
					cp = sq_end;
				}
			}
		}
	}
	return E_SQ_ERROR;
}

//--------------------------------------
//
//--------------------------------------
int CDcmScannerS::Read(const char* buf, unsigned long buf_count, const bool explicitVR)
{
	if (m_error != S_CONTINUE)
		return m_error;

	while (buf_count > 0) {

		if (!m_hdr_complete) {
			//
			// wb_ǂł
			//
	
			if (m_hdr_buf_count < m_hdr_expect_length) {
				const unsigned long remain_count = m_hdr_expect_length - m_hdr_buf_count;
				const unsigned long copy_count = buf_count > remain_count ? remain_count : buf_count;
				memcpy(m_hdr_buf + m_hdr_buf_count, buf, copy_count);
				m_hdr_buf_count += copy_count;
				buf += copy_count;
				buf_count -= copy_count;
			}

			if (m_hdr_buf_count == 8) {
				m_hdr.m_tag[0] = *reinterpret_cast<const unsigned short*>(m_hdr_buf);
				m_hdr.m_tag[1] = *reinterpret_cast<const unsigned short*>(m_hdr_buf + 2);
				if (explicitVR) {
					if (m_hdr.m_tag[0] != DCM_GROUPDELIMITER) {
						m_hdr.m_vr = m_hdr_buf + 4;
						if (m_hdr.isSQ() || m_hdr.isBLE()) {
							m_hdr_expect_length = 12;
							continue;
						}
						m_hdr.m_length = *reinterpret_cast<const unsigned short*>(m_hdr_buf + 6);
					} else {
						const char* const del = "DLM";
						m_hdr.m_vr = del;
						m_hdr.m_length = *reinterpret_cast<const unsigned long*>(m_hdr_buf + 4);
					}
				} else {
					m_hdr.m_length = *reinterpret_cast<const unsigned long*>(m_hdr_buf + 4);
				}
				m_hdr.m_hsize = 8;
			} else if (m_hdr_buf_count == 12) {
				m_hdr.m_length = *reinterpret_cast<const unsigned long*>(m_hdr_buf + 8);
				m_hdr.m_hsize = 12;
				m_hdr_expect_length = 8;
			} else {
				return S_CONTINUE;
			}

			//
			// wb_ǂݐ؂
			//
			m_hdr_buf_count = 0;
			m_hdr_complete = true;

			if (m_hdr.m_tag[0] == DCM_GROUPDELIMITER) {

				if (m_sq == NULL)
					return m_error = E_ILLEAGAL_DELIMITER;

				m_sq->sq_read_length += m_hdr.m_hsize;

				if (m_hdr.m_tag[1] == DCM_TAG_ELEMENT(DCM_DLMITEM)) {
					Event_EnterSQItem(m_hdr);

					m_sq->item_length = m_hdr.m_length;
					if (m_sq->item_length == 0)
						Event_LeaveSQItem();
					else
						m_sq->item_read_length = 0;
					
				} else if (m_hdr.m_tag[1] == DCM_TAG_ELEMENT(DCM_DLMITEMDELIMITATIONITEM)) {
					Event_LeaveSQItem();

					while (m_sq->sq_length != DCM_UNSPECIFIEDLENGTH
					    && m_sq->sq_length == m_sq->sq_read_length) {
						Event_LeaveSQ();
						if (m_sq == m_sq_tbl) {
							m_sq = NULL;
							break;
						} else {
							const unsigned long read_length = m_sq->sq_read_length;
							m_sq--;
							m_sq->sq_read_length += read_length;
						}
					}

				} else if (m_hdr.m_tag[1] == DCM_TAG_ELEMENT(DCM_DLMSEQUENCEDELIMITATIONITEM)) {
					Event_LeaveSQ();

					if (m_sq == m_sq_tbl) {
						m_sq = NULL;
					} else {
						const unsigned long read_len = m_sq->sq_read_length;
						m_sq--;
						m_sq->sq_read_length += read_len;
						while (m_sq->sq_length != DCM_UNSPECIFIEDLENGTH
						    && m_sq->sq_length == m_sq->sq_read_length) {
							Event_LeaveSQ();
							if (m_sq == m_sq_tbl) {
								m_sq = NULL;
								break;
							} else {
								const unsigned long read_length = m_sq->sq_read_length;
								m_sq--;
								m_sq->sq_read_length += read_length;
							}
						}
					}
				}

				m_hdr_complete = false;
				continue;
			}

			if (!explicitVR) {
				m_dict = m_dcmdict.Lookup(m_hdr.ltag());
				if (m_dict != NULL) {
					m_hdr.m_vr = DcmGetVRStr(m_dict->representation);
				} else {
					const char* const pr = "private";
					m_hdr.m_vr = pr;
				}
			}

			if (m_hdr.isSQ()) {
				Event_DataElement(m_hdr, NULL, 0, 0, explicitVR, m_dict);

				if (m_sq == NULL) {
					m_sq = m_sq_tbl;
				} else {
					m_sq->sq_read_length += m_hdr.m_hsize;
					if (m_sq == m_sq_tbl + K_MAX_SQ_LEVEL - 1)
						return m_error = E_SQ_LEVEL_OVERFLOW;
					m_sq++;
				}

				Event_EnterSQ(m_hdr);

				m_sq->sq_length = m_hdr.m_length;
				m_sq->sq_read_length = 0;

				m_hdr_complete = false;
				continue;
			}
		}

		if (buf_count == 0)
			return S_CONTINUE;
			// 傤ǃwb_Ńobt@ǂ݂Ă܂΂B

		if (m_hdr.m_length == DCM_UNSPECIFIEDLENGTH) {
			if (m_hdr.ltag() != DCM_PXLPIXELDATA)
				return m_error = E_UNSPECIFIED_LENGTH;
				// sNZf[^ȊO 0xFFFFFFFF ͂肦Ȃ
				// SQ łɂ͂Ȃ
				
			Event_DataElement(m_hdr, buf, buf_count, m_value_count, explicitVR, m_dict);
			m_value_count += buf_count;
			return S_CONTINUE;
		}

		const unsigned long value_expected = m_hdr.m_length - m_value_count;
		const unsigned long value_avail = value_expected > buf_count ? buf_count : value_expected;
		Event_DataElement(m_hdr, buf, value_avail, m_value_count, explicitVR, m_dict);
		buf += value_avail;
		buf_count -= value_avail;
		m_value_count += value_avail;

		if (m_value_count < m_hdr.m_length)
			return S_CONTINUE;

		//
		// GgSɓǂݐ؂
		//
		m_value_count = 0;
		m_hdr_complete = false;
	
		if (m_sq != NULL) {
			const unsigned long element_length = m_hdr.m_hsize + m_hdr.m_length;

			m_sq->item_read_length += element_length;
			if (m_sq->item_length != DCM_UNSPECIFIEDLENGTH
			 && m_sq->item_length == m_sq->item_read_length)
				Event_LeaveSQItem();

			m_sq->sq_read_length += element_length;
			while (m_sq->sq_length != DCM_UNSPECIFIEDLENGTH
			    && m_sq->sq_length == m_sq->sq_read_length) {
				Event_LeaveSQ();
				if (m_sq == m_sq_tbl) {
					m_sq = NULL;
					break;
				} else {
					const unsigned long read_length = m_sq->sq_read_length;
					m_sq--;
					m_sq->sq_read_length += read_length;
				}
			}
		}
	}

	return S_CONTINUE;
}

