/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef _5F178AD5_DACC_4444_B87E_D0F9148FD67D_
#define _5F178AD5_DACC_4444_B87E_D0F9148FD67D_

#pragma warn_unusedarg off

class CDcmScannerR;
class CDcmScannerS;
class CDcmScannerF;

//======================================
//
//======================================
class tDcmElementHdr {
public:
	unsigned short m_tag[2];
	unsigned long m_length;
	const char* m_vr;
	unsigned long m_hsize;

	void ReadHdr(const char* const ptr, const bool explicitVR);

	tDcmElementHdr()
		: m_vr(NULL) {}

	bool isSQ(void) const {
		return m_vr[0] == 'S' && m_vr[1] == 'Q';
	}
	bool isBLE(void) const { // is Binary Large Element
		return (m_vr[0] == 'O' && (m_vr[1] == 'B' || m_vr[1] == 'W' || m_vr[1] == 'T'))
		    || (m_vr[0] == 'U' && m_vr[1] == 'N');
	}
	bool isString(void) const {
		return (m_vr[0] == 'A' && ( m_vr[1] == 'E' || m_vr[1] == 'S'))
		    || (m_vr[0] == 'C' && m_vr[1] == 'S')
		    || (m_vr[0] == 'D' && ( m_vr[1] == 'A' || m_vr[1] == 'S' || m_vr[1] == 'T'))
		    || (m_vr[0] == 'I' && m_vr[1] == 'S')
		    || (m_vr[0] == 'L' && ( m_vr[1] == 'O' || m_vr[1] == 'T'))
		    || (m_vr[0] == 'P' && m_vr[1] == 'N')
		    || (m_vr[0] == 'S' && ( m_vr[1] == 'H' || m_vr[1] == 'T'))
		    || (m_vr[0] == 'T' && m_vr[1] == 'M')
		    || (m_vr[0] == 'U' && ( m_vr[1] == 'I' || m_vr[1] == 'T'));
	}
	bool isShort(void) const {
		return (m_vr[0] == 'S' || m_vr[0] == 'U') && m_vr[1] == 'S';
	}
	bool isLong(void) const {
		return (m_vr[0] == 'S' || m_vr[0] == 'U') && m_vr[1] == 'L';
	}

	const char* value_ptr(const char* ptr) const { return ptr + m_hsize; }
	const char* next_ptr(const char* ptr) const { return ptr + m_hsize + m_length; }
	unsigned long header_length(void) const { return m_hsize; }
	unsigned long value_length(void) const { return m_length; }
	unsigned long element_length(void) const { return m_hsize + m_length; }
	unsigned short gtag(void) const { return m_tag[0]; }
	unsigned short etag(void) const { return m_tag[1]; }
	DCM_TAG ltag(void) const { return DCM_MAKETAG(m_tag[0], m_tag[1]); }

	int get_integer(const char* ptr, int nth = 0);
	double get_real(const char* ptr, int nth = 0);
	void get_string(const char* ptr, char* str_buf, size_t str_buf_len, int nth = -1, bool trimSpace = true);
};


//======================================
//
//======================================
//
// SAX like Event Driven Model
// Scan from a continues buffer
//
// ̃NX͎gȂ߃eiX󂯂ĂȂ߁Ap~Ƃ
//
class CDcmScannerR {
private:
	CDcmDict m_dcmdict;
	unsigned int m_sq_level;

	void ParseImpl(
			const char* const block_begin, const char* const block_end,
			const bool explictVR,
			const bool skip_sq,
			unsigned long& advance,
			const bool stop_at_delim
	);

	CDcmScannerR() : m_dcmdict(), m_sq_level(0) {} // ̃NXgƂĂG[ɂȂ

public:

	void Parse(
			const char* const block_begin, const char* const block_end,
			const bool explicitVR,
			const bool skip_sq = false,
			unsigned long* advance = NULL
			)
	{
		unsigned long adv;
		if (advance == NULL)
			advance = &adv;
		ParseImpl(block_begin, block_end, explicitVR, skip_sq, *advance, false);
	}

protected:
	unsigned int sq_level(void) { return m_sq_level; }

	virtual void Event_DataElement(
			const tDcmElementHdr& hdr,
			const char* const vp,
			const char* const ep,
			const bool explicitVR,
			const DCMDICT* const dict,
			bool& tobe_stop) {}

	virtual void Event_GroupLength(
			const tDcmElementHdr& hdr,
			const unsigned long group_length,
			bool& tobe_skip) {}

	virtual void Event_EnterSQ(
			const tDcmElementHdr& hdr,
			bool& tobe_skip) {}
	virtual void Event_LeaveSQ(void) {}
	virtual void Event_EnterSQItem(const tDcmElementHdr& hdr) {}
	virtual void Event_LeaveSQItem(void) {}
};

//======================================
//
//======================================
//
// Simple Forward Fetch Model
// skips SQ Element
//
class CDcmScannerF {
	const char* const m_begin;
	const char* const m_end;

	int skip_unspecified_length_sq(const char* const sq_begin, const char*& sq_end, const bool explicitVR);
	
public:
	CDcmScannerF(const char* const block_begin, const char* const block_end)
		: m_begin(block_begin), m_end(block_end) 
	{
	}

	enum {
		S_UNKNOWN_LENGTH_PIXELDATA = 1,
		S_CONTINUE = 0,
		E_EOF = -1,
		E_SQ_ERROR = -2
	}; 

	int NextElement(const char* const cp, tDcmElementHdr& hdr, const bool explicitVR);
};

//======================================
//
//======================================
//
// SAX like Event Driven Model
// Read from framgented stream
//
class CDcmScannerS {
	char m_hdr_buf[12];
	unsigned long m_hdr_buf_count;
	unsigned long m_hdr_expect_length;
	bool m_hdr_complete;
	tDcmElementHdr m_hdr;
	const DCMDICT* m_dict;
	unsigned long m_value_count;

	CDcmDict m_dcmdict;

	enum {
		K_MAX_SQ_LEVEL = 16
	};
	struct sq_t {
		unsigned long sq_length;
		unsigned long sq_read_length;
		unsigned long item_length;
		unsigned long item_read_length;
	} m_sq_tbl[K_MAX_SQ_LEVEL], *m_sq;

public:

	enum {
		S_CONTINUE = 0,
		E_ILLEAGAL_DELIMITER = -1,
		E_SQ_LEVEL_OVERFLOW = -2,
		E_UNSPECIFIED_LENGTH = -3
	};

	int m_error;

	CDcmScannerS()
		: m_hdr_buf_count(0)
		, m_hdr_expect_length(8)
		, m_hdr_complete(false)
		, m_value_count(0), m_dict(NULL)
		, m_sq(NULL)
		, m_error(S_CONTINUE) {}

	int Read(const char* buf, const unsigned long buf_count, const bool explicitVR);

protected:
	virtual void Event_DataElement(
			const tDcmElementHdr& hdr,
			const char* const vp,
			const unsigned long value_avail,
			const unsigned long value_offset,
			const bool explicitVR,
			const DCMDICT* const dict) {}

	virtual void Event_EnterSQ(const tDcmElementHdr& hdr) {}
	virtual void Event_LeaveSQ(void) {}
	virtual void Event_EnterSQItem(const tDcmElementHdr& hdr) {}
	virtual void Event_LeaveSQItem(void) {}

	int sq_level(void) {
		if (m_sq == NULL)
			return 0;
		return (m_sq - m_sq_tbl) + 1;
	}
};

#pragma warn_unusedarg reset

#endif
