
#include <stdio.h>
#include <string.h>

#include "jpeglib.h"

//========================================================================================
//
//========================================================================================
size_t djpeg_12_ctx_size(void);
void init_djpeg_12(void* ctx);
void setup_djpeg_12(void* ctx, void* jpegData, size_t jpegSize, unsigned short scale_denom, unsigned short* width, unsigned short* height);
int continue_djpeg_12(void* ctx);
void proc_djpeg_12(void* ctx, void* row_pointer, unsigned int rows);
void finish_djpeg_12(void*ctx);
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
class expJPEG {
public:
	char msg[200];

	expJPEG(char* message) {
		strcpy(msg, message);	
	}
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
typedef struct {
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	struct jpeg_source_mgr src;
} tDContext;

//----------------------------------------------------------------------------------------
static void init_source(j_decompress_ptr cinfo)
{
	return;
}
//----------------------------------------------------------------------------------------
static boolean fill_input_buffer(j_decompress_ptr cinfo)
{
	return TRUE;
}
//----------------------------------------------------------------------------------------
static void skip_input_data(j_decompress_ptr cinfo, long num_bytes)
{
	cinfo->src->next_input_byte += num_bytes;
	cinfo->src->bytes_in_buffer -= num_bytes;
	return;
}
//----------------------------------------------------------------------------------------
static void term_source(j_decompress_ptr cinfo)
{
	return;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void djpeg_12_output_message (j_common_ptr cinfo)
{
	/* Create the message */
	tDContext* ctx = (tDContext*)cinfo;
	unsigned long len;
	char buffer[JMSG_LENGTH_MAX];

	(*cinfo->err->format_message) (cinfo, buffer);

	FILE* fp = fopen("c:\\temp\\dio-image-error.log","a");
	fputs(buffer, fp);
	fclose(fp);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void djpeg_12_error_exit (j_common_ptr cinfo)
{
	tDContext* ctx = (tDContext*)cinfo;

	char buffer[JMSG_LENGTH_MAX];
	(*cinfo->err->format_message) (cinfo, buffer);

	throw expJPEG(buffer);
}

//========================================================================================
//
//========================================================================================
size_t djpeg_12_ctx_size(void)
{
	return sizeof(tDContext);
}
//========================================================================================
//
//========================================================================================
void init_djpeg_12(void* vctx)
{
	tDContext* ctx = (tDContext*)vctx;

	/* Initialize JPEG parameters.
	* Much of this may be overridden later.
	* In particular, we don't yet know the input file's color space,
	* but we need to provide some value for jpeg_set_defaults() to work.
	*/

	ctx->cinfo.err = jpeg_std_error(&ctx->jerr);
	ctx->jerr.output_message = djpeg_12_output_message;
	ctx->jerr.error_exit = djpeg_12_error_exit;
	jpeg_create_decompress(&ctx->cinfo);

	ctx->cinfo.src = &ctx->src;
	ctx->src.init_source = init_source;
	ctx->src.fill_input_buffer = fill_input_buffer;
	ctx->src.skip_input_data = skip_input_data;
	ctx->src.resync_to_restart = jpeg_resync_to_restart;
	ctx->src.term_source = term_source;

	ctx->cinfo.dct_method = JDCT_FLOAT;
}

//========================================================================================
//
//========================================================================================
void setup_djpeg_12(void* vctx, void* jpegData, size_t jpegSize, unsigned short scale_denom, unsigned short* width, unsigned short* height)
{
	tDContext* ctx = (tDContext*)vctx;

	ctx->src.next_input_byte = (JOCTET*)jpegData;
	ctx->src.bytes_in_buffer = jpegSize;

	/* Read file header, set default decompression parameters */
 	(void) jpeg_read_header(&ctx->cinfo, TRUE);

	if (scale_denom != 1) {
		ctx->cinfo.scale_denom = scale_denom;
		jpeg_calc_output_dimensions(&ctx->cinfo);
	}

	/* Start decompressor */
	(void) jpeg_start_decompress(&ctx->cinfo);

	*width = ctx->cinfo.output_width;
	*height = ctx->cinfo.output_height;
}

//========================================================================================
//
//========================================================================================
int continue_djpeg_12(void* vctx)
{
	tDContext* ctx = (tDContext*)vctx;

	return ctx->cinfo.output_scanline < ctx->cinfo.output_height;
}

//========================================================================================
//
//========================================================================================
void proc_djpeg_12(void* vctx, void* row_pointer, unsigned int rows)
{
	tDContext* ctx = (tDContext*)vctx;

	jpeg_read_scanlines(&ctx->cinfo, (JSAMPROW*)row_pointer, rows);
}

//========================================================================================
//
//========================================================================================
void finish_djpeg_12(void* vctx)
{
	tDContext* ctx = (tDContext*)vctx;

	jpeg_finish_decompress(&ctx->cinfo);
	jpeg_destroy_decompress(&ctx->cinfo);
}

