extern "C" {
#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "jmemsys.h"		/* import the system-dependent declarations */
}

#include "../codec.h"
#include "mbuf.h"

#include "libjpeg-8.h"

//========================================================================================
//
// Compressor
//
//========================================================================================

//----------------------------------------------------------------------------------------
typedef struct {
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	struct jpeg_destination_mgr dest;
	size_t streamSize;
	mbuf_t streamBufHead;
	mbuf_t* streamBufTail;
	codec_mem_t m_cm;
} tCContext;

//----------------------------------------------------------------------------------------
//
// jpeg data stream output manager
//
//----------------------------------------------------------------------------------------

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void cjpeg_8_init_destination (j_compress_ptr cinfo)
{
	tCContext* ctx = (tCContext*)cinfo;

	ctx->streamSize = 0;

	mbuf_t* const mbuf = &ctx->streamBufHead;
	mbuf->next = (mbuf_t*)NULL;
	mbuf->size = 0;

	ctx->streamBufTail = mbuf;

	ctx->dest.next_output_byte = mbuf->buf;
	ctx->dest.free_in_buffer = sizeof(mbuf->buf);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static boolean cjpeg_8_empty_output_buffer (j_compress_ptr cinfo)
{
	tCContext* ctx = (tCContext*)cinfo;

	mbuf_t* const buf1 = ctx->streamBufTail;

	buf1->size = sizeof(buf1->buf);
	ctx->streamSize += sizeof(buf1->buf);

	mbuf_t* buf2 = mbuf_malloc();
	if (buf2 == NULL)
		return FALSE;

	buf1->next = buf2;
	ctx->streamBufTail = buf2;

	ctx->dest.next_output_byte = buf2->buf;
	ctx->dest.free_in_buffer = sizeof(buf2->buf);

	return TRUE;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void cjpeg_8_term_destination (j_compress_ptr cinfo)
{
	tCContext* ctx = (tCContext*)cinfo;

	mbuf_t* const mbuf = ctx->streamBufTail;

	const size_t size = sizeof(mbuf->buf) - ctx->dest.free_in_buffer;

	mbuf->size = size;
	ctx->streamSize += size;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void cjpeg_8_output_message (j_common_ptr cinfo)
{
	/* Create the message */
	tCContext* ctx = (tCContext*)cinfo;
	char buffer[JMSG_LENGTH_MAX];

	(*cinfo->err->format_message) (cinfo, buffer);

	codec_log_message(buffer);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void cjpeg_8_error_exit (j_common_ptr cinfo) throw (expCODEC)
{
	tCContext* ctx = (tCContext*)cinfo;

	char buffer[JMSG_LENGTH_MAX];
	(*cinfo->err->format_message) (cinfo, buffer);

	throw expCODEC(buffer);
}

//========================================================================================
//
//========================================================================================
size_t cjpeg_8_ctx_size(void)
{
	return sizeof(tCContext);
}

//========================================================================================
//
//========================================================================================
void init_cjpeg_8(void* vctx) throw (expCODEC)
{
	tCContext* ctx = (tCContext*)vctx;

	ctx->m_cm.m_allocated_size = 0;
	ctx->cinfo.client_data = (void*)&ctx->m_cm;

	/* Initialize the JPEG compression object with default error handling. */
	ctx->cinfo.err = jpeg_std_error(&ctx->jerr);
	ctx->jerr.output_message = cjpeg_8_output_message;
	ctx->jerr.error_exit = cjpeg_8_error_exit;
	jpeg_create_compress(&ctx->cinfo);

	/* Specify data destination for compression */
	ctx->cinfo.dest = &ctx->dest;
	ctx->dest.init_destination = cjpeg_8_init_destination;
	ctx->dest.empty_output_buffer = cjpeg_8_empty_output_buffer;
	ctx->dest.term_destination = cjpeg_8_term_destination;
}

//========================================================================================
//
//========================================================================================
void setup_cjpeg_8(void* vctx, const unsigned short width, const unsigned short height, const bool isColor, const int comp_level, const int precision) throw (expCODEC)
{
	tCContext* ctx = (tCContext*)vctx;

	/* Initialize JPEG parameters.
	* Much of this may be overridden later.
	* In particular, we don't yet know the input file's color space,
	* but we need to provide some value for jpeg_set_defaults() to work.
	*/
	ctx->cinfo.image_width = width;
	ctx->cinfo.image_height = height;

	if (isColor) {
		ctx->cinfo.input_components = 3;
		ctx->cinfo.in_color_space = JCS_RGB;
	} else {	
		ctx->cinfo.input_components = 1;
		ctx->cinfo.in_color_space = JCS_GRAYSCALE;
	}
	
	jpeg_set_defaults(&ctx->cinfo);

	if (precision == 1)
		ctx->cinfo.dct_method = JDCT_FLOAT;
	else
		ctx->cinfo.dct_method = JDCT_IFAST;

	if (comp_level >= 0)
		jpeg_set_quality(&ctx->cinfo, comp_level, TRUE);

	/* Start compressor */
	jpeg_start_compress(&ctx->cinfo, TRUE);
}

//========================================================================================
//
//========================================================================================
void proc_cjpeg_8(void* vctx, void* row_pointer, unsigned int rows) throw (expCODEC)
{
	tCContext* ctx = (tCContext*)vctx;

	jpeg_write_scanlines(&ctx->cinfo, (JSAMPROW*)row_pointer, rows);
}

//========================================================================================
//
//========================================================================================
size_t finish_cjpeg_8(void* vctx) throw (expCODEC)
{
	tCContext* ctx = (tCContext*)vctx;

	jpeg_finish_compress(&ctx->cinfo);

	return ctx->streamSize;
}
//========================================================================================
//
//========================================================================================
void flush_cjpeg_8(void* vctx, void* pECB) throw (expCODEC)
{
	tCContext* ctx = (tCContext*)vctx;

	mbuf_t* mbuf = &ctx->streamBufHead;
	codec_write_client(pECB, mbuf->buf, mbuf->size);
	mbuf = mbuf->next;
	while (mbuf != NULL) {
		codec_write_client(pECB, mbuf->buf, mbuf->size);
		mbuf_t* next = mbuf->next;
		mbuf_free(mbuf);
		mbuf = next;
	}
}

//========================================================================================
//
//========================================================================================
void destroy_cjpeg_8(void* vctx) throw (expCODEC)
{
	tCContext* ctx = (tCContext*)vctx;

	jpeg_destroy_compress(&ctx->cinfo);
}

//========================================================================================
//
// Decompressor
//
//========================================================================================

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
typedef struct {
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	struct jpeg_source_mgr src;
	f_decode_input_func input_func;
	void* input_ctx;
} tDContext;

//----------------------------------------------------------------------------------------
static void init_source(j_decompress_ptr cinfo)
{
	tDContext* ctx = (tDContext*)cinfo;

	ctx->input_func(ctx->input_ctx, 0, (const void**)&cinfo->src->next_input_byte, &cinfo->src->bytes_in_buffer, 0);

	return;
}
//----------------------------------------------------------------------------------------
static boolean fill_input_buffer(j_decompress_ptr cinfo)
{
	tDContext* ctx = (tDContext*)cinfo;

	ctx->input_func(ctx->input_ctx, 1, (const void**)&cinfo->src->next_input_byte, &cinfo->src->bytes_in_buffer, 0);

	return TRUE;
}
//----------------------------------------------------------------------------------------
static void skip_input_data(j_decompress_ptr cinfo, long num_bytes)
{
	tDContext* ctx = (tDContext*)cinfo;

	ctx->input_func(ctx->input_ctx, 2, (const void**)&cinfo->src->next_input_byte, &cinfo->src->bytes_in_buffer, num_bytes);

	return;
}
//----------------------------------------------------------------------------------------
static void term_source(j_decompress_ptr cinfo)
{
	return;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void djpeg_8_output_message (j_common_ptr cinfo)
{
	/* Create the message */
	tDContext* ctx = (tDContext*)cinfo;
	char buffer[JMSG_LENGTH_MAX];

	(*cinfo->err->format_message) (cinfo, buffer);

	codec_log_message(buffer);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void djpeg_8_error_exit (j_common_ptr cinfo) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)cinfo;

	char buffer[JMSG_LENGTH_MAX];
	(*cinfo->err->format_message) (cinfo, buffer);

	throw expCODEC(buffer);
}

//========================================================================================
//
//========================================================================================
size_t djpeg_8_ctx_size(void)
{
	return sizeof(tDContext);
}

//========================================================================================
//
//========================================================================================
void init_djpeg_8(void* vctx, f_decode_input_func input_func, void* input_ctx) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	ctx->cinfo.client_data = NULL;

	/* Initialize JPEG parameters.
	* Much of this may be overridden later.
	* In particular, we don't yet know the input file's color space,
	* but we need to provide some value for jpeg_set_defaults() to work.
	*/

	ctx->cinfo.err = jpeg_std_error(&ctx->jerr);
	ctx->jerr.output_message = djpeg_8_output_message;
	ctx->jerr.error_exit = djpeg_8_error_exit;
	jpeg_create_decompress(&ctx->cinfo);

	ctx->cinfo.src = &ctx->src;
	ctx->src.init_source = init_source;
	ctx->src.fill_input_buffer = fill_input_buffer;
	ctx->src.skip_input_data = skip_input_data;
	ctx->src.resync_to_restart = jpeg_resync_to_restart;
	ctx->src.term_source = term_source;

	ctx->input_func = input_func;
	ctx->input_ctx = input_ctx;

	ctx->cinfo.dct_method = JDCT_FLOAT;
}

//========================================================================================
//
//========================================================================================
void setup_djpeg_8(void* vctx, unsigned short scale_denom, unsigned short* width, unsigned short* height) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	/* Read file header, set default decompression parameters */
 	(void) jpeg_read_header(&ctx->cinfo, TRUE);

	if (scale_denom != 1) {
		ctx->cinfo.scale_denom = scale_denom;
		jpeg_calc_output_dimensions(&ctx->cinfo);
	}

	/* Start decompressor */
	(void) jpeg_start_decompress(&ctx->cinfo);

	*width = ctx->cinfo.output_width;
	*height = ctx->cinfo.output_height;
}

//========================================================================================
//
//========================================================================================
int continue_djpeg_8(void* vctx) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	return ctx->cinfo.output_scanline < ctx->cinfo.output_height;
}

//========================================================================================
//
//========================================================================================
void proc_djpeg_8(void* vctx, void* row_pointer, unsigned int rows) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	jpeg_read_scanlines(&ctx->cinfo, (JSAMPROW*)row_pointer, rows);
}

//========================================================================================
//
//========================================================================================
void finish_djpeg_8(void* vctx) throw (expCODEC)
{
	tDContext* ctx = (tDContext*)vctx;

	jpeg_finish_decompress(&ctx->cinfo);
	jpeg_destroy_decompress(&ctx->cinfo);
}
