/*
 * my_fdct.c
 *
 * Copyright (C) 1999, Takaya Sakusabe
 * This file is based on the Independent JPEG Group's software.
 *
 * This file contains the forward-DCT and quantizer tuned Intel Architecture
 */

#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "jdct.h"		/* Private declarations for DCT subsystem */

void my_fdct_mmx(JSAMPARRAY sample_data, JDIMENSION start_col, DCTELEM* divisors, JCOEFPTR coef_ptr);
void my_fdct_fpu(JSAMPARRAY sample_data, JDIMENSION start_col, FAST_FLOAT* divisors, JCOEFPTR coef_ptr);

typedef float WORK_FLOAT;

static WORK_FLOAT WF_0_707106781 = 0.707106781;
static WORK_FLOAT WF_0_382683433 = 0.382683433;
static WORK_FLOAT WF_0_541196100 = 0.541196100;
static WORK_FLOAT WF_1_30656296 = 1.30656296;

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline void fdct_float_row_float(int* itmp, WORK_FLOAT* dataptr)
{
	int itmp2[4];

	itmp2[0] = itmp[0] + itmp[6];
	itmp2[1] = itmp[0] - itmp[6];
	itmp2[2] = itmp[2] + itmp[4];
	itmp2[3] = itmp[2] - itmp[4];

	dataptr[0] = (WORK_FLOAT)(itmp2[0] + itmp2[2]);
	dataptr[4] = (WORK_FLOAT)(itmp2[0] - itmp2[2]);

	FAST_FLOAT ftmp[4];

	ftmp[0] = (WORK_FLOAT)(itmp2[3] + itmp2[1]) * (WORK_FLOAT)0.707106781;

	dataptr[2] = (WORK_FLOAT)itmp2[1] + ftmp[0];
	dataptr[6] = (WORK_FLOAT)itmp2[1] - ftmp[0];

	itmp2[0] = itmp[7] + itmp[5];
	itmp2[2] = itmp[5] + itmp[3];
	itmp2[3] = itmp[3] + itmp[1];

	ftmp[0] = (WORK_FLOAT)(itmp2[0] - itmp2[3]) * (WORK_FLOAT)0.382683433;
	ftmp[2] = (WORK_FLOAT)0.541196100 * (WORK_FLOAT)itmp2[0] + ftmp[0];
	ftmp[3] = (WORK_FLOAT)1.306562965 * (WORK_FLOAT)itmp2[3] + ftmp[0];

	ftmp[0] = (WORK_FLOAT)itmp2[2] * (WORK_FLOAT)0.707106781;

	ftmp[1] = (WORK_FLOAT)itmp[1] - ftmp[0];
	dataptr[5] = ftmp[1] + ftmp[2];
	dataptr[3] = ftmp[1] - ftmp[2];

	ftmp[1] = (WORK_FLOAT)itmp[1] + ftmp[0];
	dataptr[1] = ftmp[1] + ftmp[3];
	dataptr[7] = ftmp[1] - ftmp[3];
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void c_fdct_float_row_float_____(int* itmp, WORK_FLOAT* dataptr)
{
	int itmp2[4];

	itmp2[0] = itmp[0] + itmp[6];
	itmp2[1] = itmp[0] - itmp[6];
	itmp2[2] = itmp[2] + itmp[4];
	itmp2[3] = itmp[2] - itmp[4];

	dataptr[0] = (WORK_FLOAT)(itmp2[0] + itmp2[2]);
	dataptr[4] = (WORK_FLOAT)(itmp2[0] - itmp2[2]);

	FAST_FLOAT ftmp[7];

	ftmp[0] = (WORK_FLOAT)(itmp2[3] + itmp2[1]) * (WORK_FLOAT)0.707106781;

	dataptr[2] = (WORK_FLOAT)itmp2[1] + ftmp[0];
	dataptr[6] = (WORK_FLOAT)itmp2[1] - ftmp[0];

	itmp2[0] = itmp[7] + itmp[5];
	itmp2[2] = itmp[5] + itmp[3];
	itmp2[3] = itmp[3] + itmp[1];

	ftmp[1] = (WORK_FLOAT)(itmp2[0] - itmp2[3]) * (WORK_FLOAT)0.382683433;
	ftmp[2] = (WORK_FLOAT)0.541196100 * (WORK_FLOAT)itmp2[0] + ftmp[1];
	ftmp[3] = (WORK_FLOAT)1.306562965 * (WORK_FLOAT)itmp2[3] + ftmp[1];

	ftmp[4] = (WORK_FLOAT)itmp2[2] * (WORK_FLOAT)0.707106781;

	ftmp[5] = (WORK_FLOAT)itmp[1] + ftmp[4];
	ftmp[6] = (WORK_FLOAT)itmp[1] - ftmp[4];

	dataptr[5] = ftmp[6] + ftmp[2];
	dataptr[3] = ftmp[6] - ftmp[2];
	dataptr[1] = ftmp[5] + ftmp[3];
	dataptr[7] = ftmp[5] - ftmp[3];
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline void fdct_float_row(JSAMPROW elemptr, WORK_FLOAT* dataptr)
{
	int data0 = GETJSAMPLE(elemptr[0]) - CENTERJSAMPLE;
	int data1 = GETJSAMPLE(elemptr[1]) - CENTERJSAMPLE;
	int data2 = GETJSAMPLE(elemptr[2]) - CENTERJSAMPLE;
	int data3 = GETJSAMPLE(elemptr[3]) - CENTERJSAMPLE;
	int data4 = GETJSAMPLE(elemptr[4]) - CENTERJSAMPLE;
	int data5 = GETJSAMPLE(elemptr[5]) - CENTERJSAMPLE;
	int data6 = GETJSAMPLE(elemptr[6]) - CENTERJSAMPLE;
	int data7 = GETJSAMPLE(elemptr[7]) - CENTERJSAMPLE;

	int itmp[8];

	itmp[0] = data0 + data7;
	itmp[1] = data0 - data7;
	itmp[2] = data1 + data6;
	itmp[3] = data1 - data6;
	itmp[4] = data2 + data5;
	itmp[5] = data2 - data5;
	itmp[6] = data3 + data4;
	itmp[7] = data3 - data4;

	fdct_float_row_float(itmp, dataptr);

}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void c_fdct_float_row(JSAMPROW elemptr, WORK_FLOAT* dataptr)
{
	int data0 = GETJSAMPLE(elemptr[0]) - CENTERJSAMPLE;
	int data1 = GETJSAMPLE(elemptr[1]) - CENTERJSAMPLE;
	int data2 = GETJSAMPLE(elemptr[2]) - CENTERJSAMPLE;
	int data3 = GETJSAMPLE(elemptr[3]) - CENTERJSAMPLE;
	int data4 = GETJSAMPLE(elemptr[4]) - CENTERJSAMPLE;
	int data5 = GETJSAMPLE(elemptr[5]) - CENTERJSAMPLE;
	int data6 = GETJSAMPLE(elemptr[6]) - CENTERJSAMPLE;
	int data7 = GETJSAMPLE(elemptr[7]) - CENTERJSAMPLE;

	int itmp0 = data0 + data7;
	int itmp7 = data0 - data7;
	int itmp1 = data1 + data6;
	int itmp6 = data1 - data6;
	int itmp2 = data2 + data5;
	int itmp5 = data2 - data5;
	int itmp3 = data3 + data4;
	int itmp4 = data3 - data4;

	int itmp10 = itmp0 + itmp3;	/* phase 2 */
	int itmp13 = itmp0 - itmp3;
	int itmp11 = itmp1 + itmp2;
	int itmp12 = itmp1 - itmp2;

	/* Even part */

	dataptr[0] = (WORK_FLOAT)(itmp10 + itmp11); /* phase 3 */
	dataptr[4] = (WORK_FLOAT)(itmp10 - itmp11);

	FAST_FLOAT z1 = (WORK_FLOAT)(itmp12 + itmp13) * (WORK_FLOAT)0.707106781; /* c4 */
	dataptr[2] = (WORK_FLOAT)itmp13 + z1;	/* phase 5 */
	dataptr[6] = (WORK_FLOAT)itmp13 - z1;

	/* Odd part */

	itmp10 = itmp4 + itmp5;	/* phase 2 */
	itmp11 = itmp5 + itmp6;
	itmp12 = itmp6 + itmp7;

	/* The rotator is modified from fig 4-8 to avoid extra negations. */
	WORK_FLOAT z5 = (WORK_FLOAT)(itmp10 - itmp12) * (WORK_FLOAT)0.382683433; /* c6 */
	WORK_FLOAT z2 = (WORK_FLOAT)0.541196100 * (WORK_FLOAT)itmp10 + z5; /* c2-c6 */
	WORK_FLOAT z4 = (WORK_FLOAT)1.306562965 * (WORK_FLOAT)itmp12 + z5; /* c2+c6 */
	WORK_FLOAT z3 = (WORK_FLOAT)itmp11 * (WORK_FLOAT)0.707106781; /* c4 */

	WORK_FLOAT z11 = (WORK_FLOAT)itmp7 + z3;		/* phase 5 */
	WORK_FLOAT z13 = (WORK_FLOAT)itmp7 - z3;

	dataptr[5] = z13 + z2;	/* phase 6 */
	dataptr[3] = z13 - z2;
	dataptr[1] = z11 + z4;
	dataptr[7] = z11 - z4;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static asm void a_calc_col1(WORK_FLOAT* dataptr, WORK_FLOAT* ptr)
{ asm {
	push ebx
	mov eax, dword ptr [ESP][4 * (1 + 1)]
	mov ebx, dword ptr [ESP][4 * (1 + 2)]

	// WORK_FLOAT ftmp0 = dataptr[DCTSIZE*0] + dataptr[DCTSIZE*7];
	// WORK_FLOAT ftmp7 = dataptr[DCTSIZE*0] - dataptr[DCTSIZE*7];
	FLD dword ptr [eax][4 * DCTSIZE * 0]
	FLD dword ptr [eax][4 * DCTSIZE * 7]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [ebx][4 * 0]
	FSUBP ST(1), ST
	FSTP dword ptr [ebx][4 * 7]

	// WORK_FLOAT ftmp1 = dataptr[DCTSIZE*1] + dataptr[DCTSIZE*6];
	// WORK_FLOAT ftmp6 = dataptr[DCTSIZE*1] - dataptr[DCTSIZE*6];
	FLD dword ptr [eax][4 * DCTSIZE * 1]
	FLD dword ptr [eax][4 * DCTSIZE * 6]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [ebx][4 * 1]
	FSUBP ST(1), ST
	FSTP dword ptr [ebx][4 * 6]

	// WORK_FLOAT ftmp2 = dataptr[DCTSIZE*2] + dataptr[DCTSIZE*5];
	// WORK_FLOAT ftmp5 = dataptr[DCTSIZE*2] - dataptr[DCTSIZE*5];
	FLD dword ptr [eax][4 * DCTSIZE * 2]
	FLD dword ptr [eax][4 * DCTSIZE * 5]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [ebx][4 * 2]
	FSUBP ST(1), ST
	FSTP dword ptr [ebx][4 * 5]

	// WORK_FLOAT ftmp3 = dataptr[DCTSIZE*3] + dataptr[DCTSIZE*4];
	// WORK_FLOAT ftmp4 = dataptr[DCTSIZE*3] - dataptr[DCTSIZE*4];
	FLD dword ptr [eax][4 * DCTSIZE * 3]
	FLD dword ptr [eax][4 * DCTSIZE * 4]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [ebx][4 * 3]
	FSUBP ST(1), ST
	FSTP dword ptr [ebx][4 * 4]

	pop ebx
	ret

} }

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static asm void a_calc_col2(WORK_FLOAT* dataptr, WORK_FLOAT* ftmp1, WORK_FLOAT* ftmp2)
{ asm {
	push ebx
	sub esp, 4
	mov eax, dword ptr [ESP][4 * (2 + 1)]
	mov ebx, dword ptr [ESP][4 * (2 + 2)]
	
	// ftmp2[7] -- [ESP][0] // [edi][4 * 7]
	FLDZ	// ftmp2[5] -- ST(4 + ?) // [edi][4 * 5]
	FLDZ	// ftmp2[3] -- ST(3 + ?)
	FLDZ	// ftmp2[2] -- ST(2 + ?)
	FLDZ	// ftmp2[1] -- ST(1 + ?)
	FLDZ	// ftmp2[0] -- ST(0 + ?)

	// ftmp2[0] = ftmp1[0] + ftmp1[3]; /* #01 */
	// ftmp2[1] = ftmp1[0] - ftmp1[3]; /* #02 */
	FLD dword ptr [ebx][4 * 0]
	FLD dword ptr [ebx][4 * 3]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP ST(3) // dword ptr [edi][4 * 0]
	FSUBP ST(1), ST
	FSTP ST(2) // dword ptr [edi][4 * 1]

	// ftmp2[2] = ftmp1[1] + ftmp1[2]; /* #03 */
	// ftmp2[3] = ftmp1[1] - ftmp1[2]; /* #04 */
	FLD dword ptr [ebx][4 * 1]
	FLD dword ptr [ebx][4 * 2]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP ST(5) // dword ptr [edi][4 * 2]
	FSUBP ST(1), ST
	FSTP ST(4) // dword ptr [edi][4 * 3]

	// dataptr[DCTSIZE*0] = ftmp2[0] + ftmp2[2]; /* #05 */
	// dataptr[DCTSIZE*4] = ftmp2[0] - ftmp2[2]; /* #06 */
	FLD ST(0) // dword ptr [edi][4 * 0]
	FLD ST(3) // dword ptr [edi][4 * 2]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 0]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 4]

	// ftmp2[0] = (ftmp2[3] + ftmp2[1]) * (WORK_FLOAT)0.707106781;	/* #07 */
	fld ST(3) // dword ptr [edi][4 * 3];
	fadd ST(2) // dword ptr [edi][4 * 1];
	fmul WF_0_707106781
	fstp ST(1) // dword ptr [edi][4 * 0]

	// dataptr[DCTSIZE*2] = ftmp2[1] + ftmp2[0]; /* #08 */
	// dataptr[DCTSIZE*6] = ftmp2[1] - ftmp2[0]; /* #09 */
	FLD ST(1) // dword ptr [edi][4 * 1]
	FLD ST(1) // dword ptr [edi][4 * 0]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 2]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 6]

	// ftmp2[0] = ftmp1[4] + ftmp1[5];	/* #10 */
	// ftmp2[2] = ftmp1[5] + ftmp1[6];	/* #11 */
	// ftmp2[3] = ftmp1[6] + ftmp1[7];	/* #12 */
	FLD dword ptr [ebx][4 * 5]
	FLD dword ptr [ebx][4 * 4]
	FADD ST(0), ST(1)
	FSTP ST(2) // dword ptr [edi][4 * 0]
	FLD dword ptr [ebx][4 * 6]
	FXCH
	FADD ST(0), ST(1)
	FSTP ST(4) // dword ptr [edi][4 * 2]
	FLD dword ptr [ebx][4 * 7]
	FADDP ST(1), ST(0)
	FSTP ST(4) // dword ptr [edi][4 * 3]

	// ftmp2[5] = (ftmp2[0] - ftmp2[3]) * (WORK_FLOAT)0.382683433;	/* #13 */
	// ftmp2[1] = (WORK_FLOAT)0.541196100 * ftmp2[0] + ftmp2[5];	/* #14 */
	// ftmp2[7] = (WORK_FLOAT)1.306562965 * ftmp2[3] + ftmp2[5];	/* #15 */
	FLD ST(3) // dword ptr [edi][4 * 3]
	FLD ST(1) // dword ptr [edi][4 * 0]
	FLD ST(0)
	FSUB ST(0), ST(2)
	FMUL WF_0_382683433
	FSTP ST(7) // dword ptr [edi][4 * 5]
	FLD ST(6) // dword ptr [edi][4 * 5]
	FXCH
	FMUL WF_0_541196100
	FADD ST, ST(1)
	FSTP ST(4) // dword ptr [edi][4 * 1]
	FXCH
	FMUL WF_1_30656296
	FADDP ST(1), ST
	FSTP dword ptr [ESP][0] // [edi][4 * 7]

	// ftmp2[5] = ftmp2[2] * (WORK_FLOAT)0.707106781;	/* #16 */
	FLD ST(2) // dword ptr [edi][4 * 2]
	fmul WF_0_707106781
	FSTP ST(5) // dword ptr [edi][4 * 5]

	// ftmp2[2] = ftmp1[7] + ftmp2[5]; /* #17 */
	// ftmp2[0] = ftmp1[7] - ftmp2[5]; /* #18 */
	FLD dword ptr [ebx][4 * 7]
	FLD ST(5) // dword ptr [edi][4 * 5]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP ST(5) //dword ptr [edi][4 * 2]
	FSUBP ST(1), ST
	FSTP ST(1) // dword ptr [edi][4 * 0]

	// dataptr[DCTSIZE*5] = ftmp2[0] + ftmp2[1]; /* #19 */
	// dataptr[DCTSIZE*3] = ftmp2[0] - ftmp2[1]; /* #20 */
	FLD ST(0) // dword ptr [edi][4 * 0]
	FLD ST(2) // dword ptr [edi][4 * 1]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 5]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 3]

	// ataptr[DCTSIZE*1] = ftmp2[2] + ftmp2[7]; /* #21 */
	// dataptr[DCTSIZE*7] = ftmp2[2] - ftmp2[7]; /* #22 */
	FLD ST(2) // dword ptr [edi][4 * 2]
	FLD dword ptr [ESP][0] // [edi][4 * 7]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 1]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 7]

	FSTP ST(0)
	FSTP ST(0)
	FSTP ST(0)
	FSTP ST(0)
	FSTP ST(0)

	add esp, 4
	pop ebx
	ret
} }

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static asm void a_calc_col2____(WORK_FLOAT* dataptr, WORK_FLOAT* ftmp1, WORK_FLOAT* ftmp2)
{ asm {
	push ebx
	push edi
	mov eax, dword ptr [ESP][4 * (2 + 1)]
	mov ebx, dword ptr [ESP][4 * (2 + 2)]
	mov edi, dword ptr [ESP][4 * (2 + 3)]

	// ftmp2[0] = ftmp1[0] + ftmp1[3]; /* #01 */
	// ftmp2[1] = ftmp1[0] - ftmp1[3]; /* #02 */
	FLD dword ptr [ebx][4 * 0]
	FLD dword ptr [ebx][4 * 3]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [edi][4 * 0]
	FSUBP ST(1), ST
	FSTP dword ptr [edi][4 * 1]

	// ftmp2[2] = ftmp1[1] + ftmp1[2]; /* #03 */
	// ftmp2[3] = ftmp1[1] - ftmp1[2]; /* #04 */
	FLD dword ptr [ebx][4 * 1]
	FLD dword ptr [ebx][4 * 2]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [edi][4 * 2]
	FSUBP ST(1), ST
	FSTP dword ptr [edi][4 * 3]

	// dataptr[DCTSIZE*0] = ftmp2[0] + ftmp2[2]; /* #05 */
	// dataptr[DCTSIZE*4] = ftmp2[0] - ftmp2[2]; /* #06 */
	FLD dword ptr [edi][4 * 0]
	FLD dword ptr [edi][4 * 2]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 0]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 4]

	// ftmp2[0] = (ftmp2[3] + ftmp2[1]) * (WORK_FLOAT)0.707106781;	/* #07 */
	fld dword ptr [edi][4 * 3];
	fadd dword ptr [edi][4 * 1];
	fmul WF_0_707106781
	fstp dword ptr [edi][4 * 0]

	// dataptr[DCTSIZE*2] = ftmp2[1] + ftmp2[0]; /* #08 */
	// dataptr[DCTSIZE*6] = ftmp2[1] - ftmp2[0]; /* #09 */
	FLD dword ptr [edi][4 * 1]
	FLD dword ptr [edi][4 * 0]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 2]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 6]

	// ftmp2[0] = ftmp1[4] + ftmp1[5];	/* #10 */
	// ftmp2[2] = ftmp1[5] + ftmp1[6];	/* #11 */
	// ftmp2[3] = ftmp1[6] + ftmp1[7];	/* #12 */
	FLD dword ptr [ebx][4 * 5]
	FLD dword ptr [ebx][4 * 4]
	FADD ST(0), ST(1)
	FSTP dword ptr [edi][4 * 0]
	FLD dword ptr [ebx][4 * 6]
	FXCH
	FADD ST(0), ST(1)
	FSTP dword ptr [edi][4 * 2]
	FLD dword ptr [ebx][4 * 7]
	FADDP ST(1), ST(0)
	FSTP dword ptr [edi][4 * 3]

	// ftmp2[5] = (ftmp2[0] - ftmp2[3]) * (WORK_FLOAT)0.382683433;	/* #13 */
	// ftmp2[1] = (WORK_FLOAT)0.541196100 * ftmp2[0] + ftmp2[5];	/* #14 */
	// ftmp2[7] = (WORK_FLOAT)1.306562965 * ftmp2[3] + ftmp2[5];	/* #15 */
	FLD dword ptr [edi][4 * 3]
	FLD dword ptr [edi][4 * 0]
	FLD ST(0)
	FSUB ST(0), ST(2)
	FMUL WF_0_382683433
	FSTP dword ptr [edi][4 * 5]
	FLD dword ptr [edi][4 * 5]
	FXCH
	FMUL WF_0_541196100
	FADD ST, ST(1)
	FSTP dword ptr [edi][4 * 1]
	FXCH
	FMUL WF_1_30656296
	FADDP ST(1), ST
	FSTP dword ptr [edi][4 * 7]

	// ftmp2[5] = ftmp2[2] * (WORK_FLOAT)0.707106781;	/* #16 */
	FLD dword ptr [edi][4 * 2]
	fmul WF_0_707106781
	FSTP dword ptr [edi][4 * 5]

	// ftmp2[2] = ftmp1[7] + ftmp2[5]; /* #17 */
	// ftmp2[0] = ftmp1[7] - ftmp2[5]; /* #18 */
	FLD dword ptr [ebx][4 * 7]
	FLD dword ptr [edi][4 * 5]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [edi][4 * 2]
	FSUBP ST(1), ST
	FSTP dword ptr [edi][4 * 0]

	// dataptr[DCTSIZE*5] = ftmp2[0] + ftmp2[1]; /* #19 */
	// dataptr[DCTSIZE*3] = ftmp2[0] - ftmp2[1]; /* #20 */	
	FLD dword ptr [edi][4 * 0]
	FLD dword ptr [edi][4 * 1]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 5]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 3]

	// dataptr[DCTSIZE*1] = ftmp2[2] + ftmp2[7]; /* #21 */
	// dataptr[DCTSIZE*7] = ftmp2[2] - ftmp2[7]; /* #22 */
	FLD dword ptr [edi][4 * 2]
	FLD dword ptr [edi][4 * 7]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 1]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 7]

	pop edi
	pop ebx
	ret
} }

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline void c_calc_col2_____(WORK_FLOAT* dataptr, WORK_FLOAT* ftmp1, WORK_FLOAT* ftmp2)
{
	ftmp2[0] = ftmp1[0] + ftmp1[3];	/* phase 2 */
	ftmp2[1] = ftmp1[0] - ftmp1[3];
	ftmp2[2] = ftmp1[1] + ftmp1[2];
	ftmp2[3] = ftmp1[1] - ftmp1[2];

	dataptr[DCTSIZE*0] = ftmp2[0] + ftmp2[2];
	dataptr[DCTSIZE*4] = ftmp2[0] - ftmp2[2];

	ftmp2[4] = (ftmp2[3] + ftmp2[1]) * (WORK_FLOAT)0.707106781; /* c4 */
	dataptr[DCTSIZE*2] = ftmp2[1] + ftmp2[4];
	dataptr[DCTSIZE*6] = ftmp2[1] - ftmp2[4];

	ftmp2[0] = ftmp1[4] + ftmp1[5];	/* phase 2 */
	ftmp2[2] = ftmp1[5] + ftmp1[6];
	ftmp2[3] = ftmp1[6] + ftmp1[7];

	/* The rotator is modified from fig 4-8 to avoid extra negations. */
	ftmp2[5] = (ftmp2[0] - ftmp2[3]) * (WORK_FLOAT)0.382683433; /* c6 */
	ftmp2[6] = (WORK_FLOAT)0.541196100 * ftmp2[0] + ftmp2[5]; /* c2-c6 */
	ftmp2[7] = (WORK_FLOAT)1.306562965 * ftmp2[3] + ftmp2[5]; /* c2+c6 */
	ftmp2[8] = ftmp2[2] * (WORK_FLOAT)0.707106781; /* c4 */

	ftmp2[9] = ftmp1[7] + ftmp2[8];		/* phase 5 */
	ftmp2[10] = ftmp1[7] - ftmp2[8];

	dataptr[DCTSIZE*5] = ftmp2[10] + ftmp2[6];
	dataptr[DCTSIZE*3] = ftmp2[10] - ftmp2[6];
	dataptr[DCTSIZE*1] = ftmp2[9] + ftmp2[7];
	dataptr[DCTSIZE*7] = ftmp2[9] - ftmp2[7];
}

inline void c_calc_col2(WORK_FLOAT* dataptr, WORK_FLOAT* ftmp1, WORK_FLOAT* ftmp2)
{
	ftmp2[0] = ftmp1[0] + ftmp1[3]; /* #01 */
	ftmp2[1] = ftmp1[0] - ftmp1[3]; /* #02 */
	ftmp2[2] = ftmp1[1] + ftmp1[2]; /* #03 */
	ftmp2[3] = ftmp1[1] - ftmp1[2]; /* #04 */

	dataptr[DCTSIZE*0] = ftmp2[0] + ftmp2[2]; /* #05 */
	dataptr[DCTSIZE*4] = ftmp2[0] - ftmp2[2]; /* #06 */

	ftmp2[0] = (ftmp2[3] + ftmp2[1]) * (WORK_FLOAT)0.707106781;	/* #07 */

	dataptr[DCTSIZE*2] = ftmp2[1] + ftmp2[0]; /* #08 */
	dataptr[DCTSIZE*6] = ftmp2[1] - ftmp2[0]; /* #09 */

	ftmp2[0] = ftmp1[4] + ftmp1[5];	/* #10 */
	ftmp2[2] = ftmp1[5] + ftmp1[6];	/* #11 */
	ftmp2[3] = ftmp1[6] + ftmp1[7];	/* #12 */

	ftmp2[5] = (ftmp2[0] - ftmp2[3]) * (WORK_FLOAT)0.382683433;	/* #13 */
	ftmp2[1] = (WORK_FLOAT)0.541196100 * ftmp2[0] + ftmp2[5];	/* #14 */
	ftmp2[7] = (WORK_FLOAT)1.306562965 * ftmp2[3] + ftmp2[5];	/* #15 */
	ftmp2[5] = ftmp2[2] * (WORK_FLOAT)0.707106781;	/* #16 */

	ftmp2[2] = ftmp1[7] + ftmp2[5];	/* #17 */
	ftmp2[0] = ftmp1[7] - ftmp2[5]; /* #18 */

	dataptr[DCTSIZE*5] = ftmp2[0] + ftmp2[1]; /* #19 */
	dataptr[DCTSIZE*3] = ftmp2[0] - ftmp2[1]; /* #20 */	
	dataptr[DCTSIZE*1] = ftmp2[2] + ftmp2[7]; /* #21 */
	dataptr[DCTSIZE*7] = ftmp2[2] - ftmp2[7]; /* #22 */
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline void fdct_float_col____(WORK_FLOAT* dataptr)
{
	WORK_FLOAT ftmp1[DCTSIZE];
	a_calc_col1(dataptr, ftmp1);

	WORK_FLOAT ftmp2[11];
	a_calc_col2(dataptr, ftmp1, ftmp2);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static asm void fdct_float_col(WORK_FLOAT* dataptr)
{ asm {
	mov eax, dword ptr [ESP][4 * 1]
	push ebx
	sub esp, 4 * DCTSIZE
	mov ebx, esp
	sub esp, 4
	
	// ftmp2[7] -- [ESP][0] // [edi][4 * 7]
	FLDZ	// ftmp2[5] -- ST(4 + ?) // [edi][4 * 5]
	FLDZ	// ftmp2[3] -- ST(3 + ?)
	FLDZ	// ftmp2[2] -- ST(2 + ?)
	FLDZ	// ftmp2[1] -- ST(1 + ?)
	FLDZ	// ftmp2[0] -- ST(0 + ?)

	// WORK_FLOAT ftmp0 = dataptr[DCTSIZE*0] + dataptr[DCTSIZE*7];
	// WORK_FLOAT ftmp7 = dataptr[DCTSIZE*0] - dataptr[DCTSIZE*7];
	FLD dword ptr [eax][4 * DCTSIZE * 0]
	FLD dword ptr [eax][4 * DCTSIZE * 7]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [ebx][4 * 0]
	FSUBP ST(1), ST
	FSTP dword ptr [ebx][4 * 7]

	// WORK_FLOAT ftmp1 = dataptr[DCTSIZE*1] + dataptr[DCTSIZE*6];
	// WORK_FLOAT ftmp6 = dataptr[DCTSIZE*1] - dataptr[DCTSIZE*6];
	FLD dword ptr [eax][4 * DCTSIZE * 1]
	FLD dword ptr [eax][4 * DCTSIZE * 6]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [ebx][4 * 1]
	FSUBP ST(1), ST
	FSTP dword ptr [ebx][4 * 6]

	// WORK_FLOAT ftmp2 = dataptr[DCTSIZE*2] + dataptr[DCTSIZE*5];
	// WORK_FLOAT ftmp5 = dataptr[DCTSIZE*2] - dataptr[DCTSIZE*5];
	FLD dword ptr [eax][4 * DCTSIZE * 2]
	FLD dword ptr [eax][4 * DCTSIZE * 5]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [ebx][4 * 2]
	FSUBP ST(1), ST
	FSTP dword ptr [ebx][4 * 5]

	// WORK_FLOAT ftmp3 = dataptr[DCTSIZE*3] + dataptr[DCTSIZE*4];
	// WORK_FLOAT ftmp4 = dataptr[DCTSIZE*3] - dataptr[DCTSIZE*4];
	FLD dword ptr [eax][4 * DCTSIZE * 3]
	FLD dword ptr [eax][4 * DCTSIZE * 4]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [ebx][4 * 3]
	FSUBP ST(1), ST
	FSTP dword ptr [ebx][4 * 4]

	// ftmp2[0] = ftmp1[0] + ftmp1[3]; /* #01 */
	// ftmp2[1] = ftmp1[0] - ftmp1[3]; /* #02 */
	FLD dword ptr [ebx][4 * 0]
	FLD dword ptr [ebx][4 * 3]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP ST(3) // dword ptr [edi][4 * 0]
	FSUBP ST(1), ST
	FSTP ST(2) // dword ptr [edi][4 * 1]

	// ftmp2[2] = ftmp1[1] + ftmp1[2]; /* #03 */
	// ftmp2[3] = ftmp1[1] - ftmp1[2]; /* #04 */
	FLD dword ptr [ebx][4 * 1]
	FLD dword ptr [ebx][4 * 2]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP ST(5) // dword ptr [edi][4 * 2]
	FSUBP ST(1), ST
	FSTP ST(4) // dword ptr [edi][4 * 3]

	// dataptr[DCTSIZE*0] = ftmp2[0] + ftmp2[2]; /* #05 */
	// dataptr[DCTSIZE*4] = ftmp2[0] - ftmp2[2]; /* #06 */
	FLD ST(0) // dword ptr [edi][4 * 0]
	FLD ST(3) // dword ptr [edi][4 * 2]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 0]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 4]

	// ftmp2[0] = (ftmp2[3] + ftmp2[1]) * (WORK_FLOAT)0.707106781;	/* #07 */
	fld ST(3) // dword ptr [edi][4 * 3];
	fadd ST(2) // dword ptr [edi][4 * 1];
	fmul WF_0_707106781
	fstp ST(1) // dword ptr [edi][4 * 0]

	// dataptr[DCTSIZE*2] = ftmp2[1] + ftmp2[0]; /* #08 */
	// dataptr[DCTSIZE*6] = ftmp2[1] - ftmp2[0]; /* #09 */
	FLD ST(1) // dword ptr [edi][4 * 1]
	FLD ST(1) // dword ptr [edi][4 * 0]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 2]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 6]

	// ftmp2[0] = ftmp1[4] + ftmp1[5];	/* #10 */
	// ftmp2[2] = ftmp1[5] + ftmp1[6];	/* #11 */
	// ftmp2[3] = ftmp1[6] + ftmp1[7];	/* #12 */
	FLD dword ptr [ebx][4 * 5]
	FLD dword ptr [ebx][4 * 4]
	FADD ST(0), ST(1)
	FSTP ST(2) // dword ptr [edi][4 * 0]
	FLD dword ptr [ebx][4 * 6]
	FXCH
	FADD ST(0), ST(1)
	FSTP ST(4) // dword ptr [edi][4 * 2]
	FLD dword ptr [ebx][4 * 7]
	FADDP ST(1), ST(0)
	FSTP ST(4) // dword ptr [edi][4 * 3]

	// ftmp2[5] = (ftmp2[0] - ftmp2[3]) * (WORK_FLOAT)0.382683433;	/* #13 */
	// ftmp2[1] = (WORK_FLOAT)0.541196100 * ftmp2[0] + ftmp2[5];	/* #14 */
	// ftmp2[7] = (WORK_FLOAT)1.306562965 * ftmp2[3] + ftmp2[5];	/* #15 */
	FLD ST(3) // dword ptr [edi][4 * 3]
	FLD ST(1) // dword ptr [edi][4 * 0]
	FLD ST(0)
	FSUB ST(0), ST(2)
	FMUL WF_0_382683433
	FSTP ST(7) // dword ptr [edi][4 * 5]
	FLD ST(6) // dword ptr [edi][4 * 5]
	FXCH
	FMUL WF_0_541196100
	FADD ST, ST(1)
	FSTP ST(4) // dword ptr [edi][4 * 1]
	FXCH
	FMUL WF_1_30656296
	FADDP ST(1), ST
	FSTP dword ptr [ESP][0] // [edi][4 * 7]

	// ftmp2[5] = ftmp2[2] * (WORK_FLOAT)0.707106781;	/* #16 */
	FLD ST(2) // dword ptr [edi][4 * 2]
	fmul WF_0_707106781
	FSTP ST(5) // dword ptr [edi][4 * 5]

	// ftmp2[2] = ftmp1[7] + ftmp2[5]; /* #17 */
	// ftmp2[0] = ftmp1[7] - ftmp2[5]; /* #18 */
	FLD dword ptr [ebx][4 * 7]
	FLD ST(5) // dword ptr [edi][4 * 5]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP ST(5) //dword ptr [edi][4 * 2]
	FSUBP ST(1), ST
	FSTP ST(1) // dword ptr [edi][4 * 0]

	// dataptr[DCTSIZE*5] = ftmp2[0] + ftmp2[1]; /* #19 */
	// dataptr[DCTSIZE*3] = ftmp2[0] - ftmp2[1]; /* #20 */
	FLD ST(0) // dword ptr [edi][4 * 0]
	FLD ST(2) // dword ptr [edi][4 * 1]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 5]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 3]

	// ataptr[DCTSIZE*1] = ftmp2[2] + ftmp2[7]; /* #21 */
	// dataptr[DCTSIZE*7] = ftmp2[2] - ftmp2[7]; /* #22 */
	FLD ST(2) // dword ptr [edi][4 * 2]
	FLD dword ptr [ESP][0] // [edi][4 * 7]
	FLD ST(1)
	FADD ST, ST(1)
	FSTP dword ptr [eax][4 * DCTSIZE * 1]
	FSUBP ST(1), ST
	FSTP dword ptr [eax][4 * DCTSIZE * 7]

	FSTP ST(0)
	FSTP ST(0)
	FSTP ST(0)
	FSTP ST(0)
	FSTP ST(0)

	add esp, 4 * (DCTSIZE + 1)
	pop ebx
	ret
} }

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline void c_fdct_float_col(WORK_FLOAT* dataptr)
{
	WORK_FLOAT ftmp0 = dataptr[DCTSIZE*0] + dataptr[DCTSIZE*7];
	WORK_FLOAT ftmp7 = dataptr[DCTSIZE*0] - dataptr[DCTSIZE*7];
	WORK_FLOAT ftmp1 = dataptr[DCTSIZE*1] + dataptr[DCTSIZE*6];
	WORK_FLOAT ftmp6 = dataptr[DCTSIZE*1] - dataptr[DCTSIZE*6];
	WORK_FLOAT ftmp2 = dataptr[DCTSIZE*2] + dataptr[DCTSIZE*5];
	WORK_FLOAT ftmp5 = dataptr[DCTSIZE*2] - dataptr[DCTSIZE*5];
	WORK_FLOAT ftmp3 = dataptr[DCTSIZE*3] + dataptr[DCTSIZE*4];
	WORK_FLOAT ftmp4 = dataptr[DCTSIZE*3] - dataptr[DCTSIZE*4];

	WORK_FLOAT ftmp10 = ftmp0 + ftmp3;	/* phase 2 */
	WORK_FLOAT ftmp13 = ftmp0 - ftmp3;
	WORK_FLOAT ftmp11 = ftmp1 + ftmp2;
	WORK_FLOAT ftmp12 = ftmp1 - ftmp2;

	dataptr[DCTSIZE*0] = ftmp10 + ftmp11;
	dataptr[DCTSIZE*4] = ftmp10 - ftmp11;

	WORK_FLOAT z1 = (ftmp12 + ftmp13) * (WORK_FLOAT)0.707106781; /* c4 */
	dataptr[DCTSIZE*2] = ftmp13 + z1;
	dataptr[DCTSIZE*6] = ftmp13 - z1;

	ftmp10 = ftmp4 + ftmp5;	/* phase 2 */
	ftmp11 = ftmp5 + ftmp6;
	ftmp12 = ftmp6 + ftmp7;

	/* The rotator is modified from fig 4-8 to avoid extra negations. */
	WORK_FLOAT z5 = (ftmp10 - ftmp12) * (WORK_FLOAT)0.382683433; /* c6 */
	WORK_FLOAT z2 = (WORK_FLOAT)0.541196100 * ftmp10 + z5; /* c2-c6 */
	WORK_FLOAT z4 = (WORK_FLOAT)1.306562965 * ftmp12 + z5; /* c2+c6 */
	WORK_FLOAT z3 = ftmp11 * (WORK_FLOAT)0.707106781; /* c4 */

	WORK_FLOAT z11 = ftmp7 + z3;		/* phase 5 */
	WORK_FLOAT z13 = ftmp7 - z3;

	dataptr[DCTSIZE*5] = z13 + z2;
	dataptr[DCTSIZE*3] = z13 - z2;
	dataptr[DCTSIZE*1] = z11 + z4;
}

const unsigned short ceil_cw = 0x0B7F;
const unsigned short floor_cw = 0x077F;
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
const WORK_FLOAT WF_16384_5 = 16384.5;
static asm void float_quantize(WORK_FLOAT* workspace, FAST_FLOAT* divisors, JCOEFPTR coef_ptr)
{ asm {
	push edi
	push esi
	push ebx
	push ecx
	push edx
	sub esp, 4
	mov esi, dword ptr [ESP][4 * (6 + 1)] // workspace
	mov edx, dword ptr [ESP][4 * (6 + 2)] // divisors
	mov edi, dword ptr [ESP][4 * (6 + 3)] // coef_ptr

	mov ax, -16384
	mov ebx, 4

	fnstcw word ptr [esp]
	fldcw floor_cw

	FLD WF_16384_5

	mov ecx, (DCTSIZE2 / 2) + 1
for_loop:
	dec ecx
	jz for_break

	FLD dword ptr [esi]
	add esi, ebx
	FLD dword ptr [esi]
	add esi, ebx
	
	FLD dword ptr [edx]
	add edx, ebx
	FLD dword ptr [edx]
	add edx, ebx

	FMULP ST(2), ST(0)
	FXCH
	FADD ST(3)

	FXCH ST(2)

	FMULP ST(1), ST(0)
	FADD ST(2)

	FISTP word ptr [edi]
	add word ptr [edi], ax
	add edi, 2
	
	FISTP word ptr [edi]
	add word ptr [edi], ax
	add edi, 2

	jmp for_loop

for_break:
	FSTP ST(0)

	fldcw word ptr [esp]
	add esp, 4
	pop edx
	pop ecx
	pop ebx
	pop esi
	pop edi
	ret
} }

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline void c_float_quantize(WORK_FLOAT* workspace, FAST_FLOAT* divisors, JCOEFPTR coef_ptr)
{
	FAST_FLOAT* divisor_ptr = divisors;
	WORK_FLOAT* dataptr = workspace;
	JCOEFPTR out_ptr = coef_ptr;
	for (int i = DCTSIZE2; i > 0; i--) {
		*out_ptr++ = (JCOEF)((int)((*dataptr++ * *divisor_ptr++) + (WORK_FLOAT)16384.5) - 16384);
	}
}

//========================================================================================
//
//========================================================================================
void my_fdct_fpu(JSAMPARRAY sample_data, JDIMENSION start_col, FAST_FLOAT* divisors, JCOEFPTR coef_ptr)
{
	WORK_FLOAT workspace[DCTSIZE2]; /* work area for FDCT subroutine */

	/* Pass 1: process rows. */

	WORK_FLOAT *dataptr = workspace;
//	for (int row = 0; row < DCTSIZE; row++, dataptr += DCTSIZE)
//		fdct_float_row(sample_data[row] + start_col, dataptr);
	
	fdct_float_row(sample_data[0] + start_col, dataptr); dataptr += DCTSIZE;
	fdct_float_row(sample_data[1] + start_col, dataptr); dataptr += DCTSIZE;
	fdct_float_row(sample_data[2] + start_col, dataptr); dataptr += DCTSIZE;
	fdct_float_row(sample_data[3] + start_col, dataptr); dataptr += DCTSIZE;
	fdct_float_row(sample_data[4] + start_col, dataptr); dataptr += DCTSIZE;
	fdct_float_row(sample_data[5] + start_col, dataptr); dataptr += DCTSIZE;
	fdct_float_row(sample_data[6] + start_col, dataptr); dataptr += DCTSIZE;
	fdct_float_row(sample_data[7] + start_col, dataptr); dataptr += DCTSIZE;

	/* Pass 2: process columns. */

	dataptr = workspace;
//	for (int i = DCTSIZE; i > 0; i--, dataptr++)
//		fdct_float_col(dataptr);
	fdct_float_col(dataptr); dataptr++;
	fdct_float_col(dataptr); dataptr++;
	fdct_float_col(dataptr); dataptr++;
	fdct_float_col(dataptr); dataptr++;
	fdct_float_col(dataptr); dataptr++;
	fdct_float_col(dataptr); dataptr++;
	fdct_float_col(dataptr); dataptr++;
	fdct_float_col(dataptr); dataptr++;

	float_quantize(workspace, divisors, coef_ptr);

//	FAST_FLOAT* divisor_ptr = divisors;
//	dataptr = workspace;
//	JCOEFPTR out_ptr = coef_ptr;
//	for (int i = DCTSIZE2; i > 0; i--)
//		*out_ptr++ = (JCOEF)((int)((*dataptr++ * *divisor_ptr++) + (WORK_FLOAT)16384.5) - 16384);
}

//========================================================================================
//
// MMX Forward DCT
//
//========================================================================================
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void mmx_prepare_block(JSAMPROW* row_tbl, int* iblock, const long long* const VW_CENTERJSAMPLE)
{
#if 1
	
//	for (int i = 0; i < DCTSIZE; i++) {
//		JSAMPROW row_ptr = row_tbl[i];
//		int* itmp = iblock + DCTSIZE * i;
	asm {
		mov esi, row_tbl
		mov edi, iblock
		mov ecx, 8
	for_loop:		

		// elmptr[0-7] -> xxx -> itmp
		mov ebx, dword ptr [esi]
		movq mm0, qword ptr [ebx]
		pxor mm2, mm2
		movq mm1, mm0
		punpcklbw mm0, mm2
		punpckhbw mm1, mm2
		mov eax, VW_CENTERJSAMPLE
		movq mm2, qword ptr [eax]
		psubsw mm0, mm2
		psubsw mm1, mm2
	// now - mm0: 3210, mm1: 7654

		// reverse mm0 -> mm7
		movq mm7, mm0
		movq mm2, mm0
		psllq mm7, 32
		psrlq mm2, 32
		por mm7, mm2
		movq mm2, mm7
		pslld mm7, 16
		psrld mm2, 16
		por mm7, mm2

		// reverse mm1 -> mm6
		movq mm6, mm1
		movq mm2, mm1
		psllq mm6, 32
		psrlq mm2, 32
		por mm6, mm2
		movq mm2, mm6
		pslld mm6, 16
		psrld mm2, 16
		por mm6, mm2

		// cascade
		paddsw mm0, mm6
		psubsw mm7, mm1

		punpcklwd mm2, mm0
		psrad mm2, 16
		movq qword ptr [edi][8 * 0], mm2

		punpckhwd mm2, mm0
		psrad mm2, 16
		movq qword ptr [edi][8 * 1], mm2

		punpcklwd mm2, mm7
		psrad mm2, 16
		movq qword ptr [edi][8 * 2], mm2

		punpckhwd mm2, mm7
		psrad mm2, 16
		movq qword ptr [edi][8 * 3], mm2

		add esi, 4
		add edi, DCTSIZE * 4
		dec cx
		jnz for_loop
	}
#else
	short stmp[8];
	for (int i = 0; i < 8; i++)
		stmp[i] = elemptr[i];// - 128;
	itmp[0] = stmp[0] + stmp[7];
	itmp[7] = stmp[0] - stmp[7];
	itmp[1] = stmp[1] + stmp[6];
	itmp[6] = stmp[1] - stmp[6];
	itmp[2] = stmp[2] + stmp[5];
	itmp[5] = stmp[2] - stmp[5];
	itmp[3] = stmp[3] + stmp[4];
	itmp[4] = stmp[3] - stmp[4];
#endif
//	tmp0 = dataptr[0] + dataptr[7];
//	tmp1 = dataptr[1] + dataptr[6];
//	tmp2 = dataptr[2] + dataptr[5];
//	tmp3 = dataptr[3] + dataptr[4];
//	tmp4 = dataptr[3] - dataptr[4];
//	tmp5 = dataptr[2] - dataptr[5];
//	tmp6 = dataptr[1] - dataptr[6];
//	tmp7 = dataptr[0] - dataptr[7];
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
#define CONST_BITS  8
#define FIX_0_382683433  ((INT32)   98)		/* FIX(0.382683433) */
#define FIX_0_541196100  ((INT32)  139)		/* FIX(0.541196100) */
#define FIX_0_707106781  ((INT32)  181)		/* FIX(0.707106781) */
#define FIX_1_306562965  ((INT32)  334)		/* FIX(1.306562965) */

#ifndef USE_ACCURATE_ROUNDING
#undef DESCALE
#define DESCALE(x,n)  RIGHT_SHIFT(x, n)
#endif

#define MULTIPLY(var,const)  ((DCTELEM) DESCALE((var) * (const), CONST_BITS))

inline void hacked_jpeg_fdct_ifast(int* data, int* itmp_ptr)
{
  DCTELEM tmp10, tmp11, tmp12, tmp13;
  DCTELEM z1, z2, z3, z4, z5, z11, z13;
  DCTELEM *dataptr;
  int ctr;
  SHIFT_TEMPS

  /* Pass 1: process rows. */

  dataptr = data;
  int* itmp = itmp_ptr;
  for (ctr = DCTSIZE-1; ctr >= 0; ctr--) {
    
    /* Even part */
    
    tmp10 = itmp[0] + itmp[3];	/* phase 2 */
    tmp13 = itmp[0] - itmp[3];
    tmp11 = itmp[1] + itmp[2];
    tmp12 = itmp[1] - itmp[2];
    
    dataptr[0] = tmp10 + tmp11; /* phase 3 */
    dataptr[4] = tmp10 - tmp11;
    
    z1 = MULTIPLY(tmp12 + tmp13, FIX_0_707106781); /* c4 */
    dataptr[2] = tmp13 + z1;	/* phase 5 */
    dataptr[6] = tmp13 - z1;
    
    /* Odd part */

    tmp10 = itmp[4] + itmp[5];	/* phase 2 */
    tmp11 = itmp[5] + itmp[6];
    tmp12 = itmp[6] + itmp[7];

    /* The rotator is modified from fig 4-8 to avoid extra negations. */
    z5 = MULTIPLY(tmp10 - tmp12, FIX_0_382683433); /* c6 */
    z2 = MULTIPLY(tmp10, FIX_0_541196100) + z5; /* c2-c6 */
    z4 = MULTIPLY(tmp12, FIX_1_306562965) + z5; /* c2+c6 */
    z3 = MULTIPLY(tmp11, FIX_0_707106781); /* c4 */

    z11 = itmp[7] + z3;		/* phase 5 */
    z13 = itmp[7] - z3;

    dataptr[5] = z13 + z2;	/* phase 6 */
    dataptr[3] = z13 - z2;
    dataptr[1] = z11 + z4;
    dataptr[7] = z11 - z4;

    dataptr += DCTSIZE;		/* advance pointer to next row */
    itmp += DCTSIZE;
  }

  /* Pass 2: process columns. */

  dataptr = data;
  itmp = itmp_ptr;
  for (ctr = DCTSIZE-1; ctr >= 0; ctr--) {
    itmp[0] = dataptr[DCTSIZE*0] + dataptr[DCTSIZE*7];
    itmp[7] = dataptr[DCTSIZE*0] - dataptr[DCTSIZE*7];
    itmp[1] = dataptr[DCTSIZE*1] + dataptr[DCTSIZE*6];
    itmp[6] = dataptr[DCTSIZE*1] - dataptr[DCTSIZE*6];
    itmp[2] = dataptr[DCTSIZE*2] + dataptr[DCTSIZE*5];
    itmp[5] = dataptr[DCTSIZE*2] - dataptr[DCTSIZE*5];
    itmp[3] = dataptr[DCTSIZE*3] + dataptr[DCTSIZE*4];
    itmp[4] = dataptr[DCTSIZE*3] - dataptr[DCTSIZE*4];
    
    /* Even part */
    
    tmp10 = itmp[0] + itmp[3];	/* phase 2 */
    tmp13 = itmp[0] - itmp[3];
    tmp11 = itmp[1] + itmp[2];
    tmp12 = itmp[1] - itmp[2];
    
    dataptr[DCTSIZE*0] = tmp10 + tmp11; /* phase 3 */
    dataptr[DCTSIZE*4] = tmp10 - tmp11;
    
    z1 = MULTIPLY(tmp12 + tmp13, FIX_0_707106781); /* c4 */
    dataptr[DCTSIZE*2] = tmp13 + z1; /* phase 5 */
    dataptr[DCTSIZE*6] = tmp13 - z1;
    
    /* Odd part */

    tmp10 = itmp[4] + itmp[5];	/* phase 2 */
    tmp11 = itmp[5] + itmp[6];
    tmp12 = itmp[6] + itmp[7];

    /* The rotator is modified from fig 4-8 to avoid extra negations. */
    z5 = MULTIPLY(tmp10 - tmp12, FIX_0_382683433); /* c6 */
    z2 = MULTIPLY(tmp10, FIX_0_541196100) + z5; /* c2-c6 */
    z4 = MULTIPLY(tmp12, FIX_1_306562965) + z5; /* c2+c6 */
    z3 = MULTIPLY(tmp11, FIX_0_707106781); /* c4 */

    z11 = itmp[7] + z3;		/* phase 5 */
    z13 = itmp[7] - z3;

    dataptr[DCTSIZE*5] = z13 + z2; /* phase 6 */
    dataptr[DCTSIZE*3] = z13 - z2;
    dataptr[DCTSIZE*1] = z11 + z4;
    dataptr[DCTSIZE*7] = z11 - z4;

    dataptr++;			/* advance pointer to next column */
  }
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline void int_quantize(int* workspace, int* divisors, short* coef_ptr)
{
	for (int i = 0; i < DCTSIZE2; i++) {
		int qval = divisors[i];
		int temp = workspace[i];
#ifdef FAST_DIVIDE
#	define DIVIDE_BY(a,b)	a /= b
#else
#	define DIVIDE_BY(a,b)	if (a >= b) a /= b; else a = 0
#endif
		if (temp < 0) {
			temp = -temp;
			temp += qval >> 1;	/* for rounding */
			DIVIDE_BY(temp, qval);
			temp = -temp;
		} else {
			temp += qval>>1;	/* for rounding */
			DIVIDE_BY(temp, qval);
		}
		coef_ptr[i] = (JCOEF)temp;
	}
}
//========================================================================================
//
//========================================================================================
void my_fdct_mmx(JSAMPARRAY sample_data, JDIMENSION start_col, int* divisors, JCOEFPTR coef_ptr)
{
	const static long long VW_CENTERJSAMPLE = 0x0080008000800080;

	JSAMPROW row_ptr[DCTSIZE];
	for (int elemr = 0; elemr < DCTSIZE; elemr++)
		row_ptr[elemr] = sample_data[elemr] + start_col;

	int itmp[DCTSIZE2];
	mmx_prepare_block(row_ptr, itmp, &VW_CENTERJSAMPLE);

	int workspace[DCTSIZE2];
	hacked_jpeg_fdct_ifast(workspace, itmp);

	int_quantize(workspace, divisors, coef_ptr);
}
