/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "codec.h"
#include "LosslessJPEGCodec.h"

/*
#pragma peephole on
#pragma opt_common_subs on
#pragma opt_dead_assignments on
#pragma opt_lifetimes on
#pragma opt_loop_invariants on
#pragma opt_propagation on
#pragma opt_strength_reduction on
#pragma opt_unroll_loops on
#pragma opt_vectorize_loops on
*/


void CJPEGStream::fill_input_buffer(const int flag)
{
	m_input_func(m_input_ctx, flag, (const void**)&m_data_ptr, &m_data_len, 0);
	m_cp = m_data_ptr;
	m_ep = m_cp + m_data_len;
	m_bitsInCacheByte = 0;
}

void CJPEGStream::read_cacheByte(void)
{
	if (m_ep == m_cp)
		fill_input_buffer();
	m_cacheByte = static_cast<unsigned int>(*m_cp++);
	if (m_cacheByte == 0xFF) {
		if (m_ep == m_cp)
			fill_input_buffer();
		m_cp++;
	}
	m_bitsInCacheByte = 8;
}

void CJPEGStream::init_source(f_decode_input_func input_func, void* input_ctx)
{
	m_input_func = input_func;
	m_input_ctx = input_ctx;
	fill_input_buffer(0);
}

unsigned int CJPEGStream::readMarker(void)
{
	const int remain = m_ep - m_cp;
	if (remain < 2) {
		if (remain == 1) {
			if (*m_cp != 0xFF)
				return 0xFF;
			fill_input_buffer();
			return static_cast<unsigned int>(*m_cp++);
		} else {
			fill_input_buffer();
		}
	}
	
	if (*m_cp != 0xFF)
		return 0xFF;
	m_cp++;
	return static_cast<unsigned int>(*m_cp++);
}

unsigned int CJPEGStream::readByte(void)
{
	if (m_cp == m_ep)
		fill_input_buffer();
	const unsigned int byte = static_cast<unsigned int>(*m_cp++);
	if (byte == 0xFF) {
		if (m_cp == m_ep)
			fill_input_buffer();
		m_cp++;
	}
	return byte;
}

unsigned int CJPEGStream::readWord(void)
{
	const unsigned int hi = readByte();
	const unsigned int lo = readByte();
	return (hi << 8) | lo;
}

const unsigned char* CJPEGStream::getBytes(unsigned int len)
{
	const int remain = m_ep - m_cp;
	if (len > remain)
		return NULL;
	const unsigned char* cp = m_cp;
	m_cp += len;
	return cp;
}

void CJPEGStream::readBytes(unsigned char* buf, unsigned int len)
{
	while (len > 0) {
		const int remain = m_ep - m_cp;
		if (len <= remain) {
			CopyMemory(buf, m_cp, len);
			m_cp += len;
			return;
		}
		CopyMemory(buf, m_cp, remain);
		len -= remain;
		buf += remain;
		fill_input_buffer();
	}
		
}

void CJPEGStream::skipBytes(unsigned int skip)
{
	while (skip > 0) {
		const int remain = m_ep - m_cp;
		if (remain > skip) {
			m_cp += skip;
			return;
		}
		skip -= remain;
		fill_input_buffer();
	}
	}

unsigned int CJPEGStream::readBit(void)
{
	if (m_bitsInCacheByte == 0)
		read_cacheByte();

	const unsigned int bit = (m_cacheByte >> 7) & 1;

	m_cacheByte <<= 1;
	m_bitsInCacheByte--;

	return bit;
}

int CJPEGStream::readBits(const unsigned int bitCount)
{
	if (bitCount == 0)
		return 0;
	unsigned int bitsRemain = bitCount;
	unsigned int bits = 0;
	while (bitsRemain > 0) {
		if (m_bitsInCacheByte == 0)
			read_cacheByte();
		const unsigned int copyCount = bitsRemain > m_bitsInCacheByte ? m_bitsInCacheByte : bitsRemain;
		bits |= m_cacheByte & 0x000000FF;
		bits <<= copyCount;
		m_cacheByte <<= copyCount;
		m_bitsInCacheByte -= copyCount;
		bitsRemain -= copyCount;
	}
	bits >>= 8;
	const unsigned int sign_mask = 1 << (bitCount - 1);
	if (bits & sign_mask)
		return static_cast<int>(bits);
	const unsigned int pad_mask = 0xFFFFFFFF << bitCount;
	return static_cast<int>(pad_mask | ++bits);
}

const unsigned char* frame_header_t::component_spec_t::read(const unsigned char* cp)
{
	m_componentIdentifier = static_cast<unsigned int>(*cp++);
	const unsigned int sf = static_cast<unsigned int>(*cp++);
	m_horizontalSamplingFactor = sf >> 4;
	m_verticalSamplingFactor = sf & 0x0F;
	m_quantization_table_selector = static_cast<unsigned int>(*cp++);
	return cp;
}

#pragma optimization_level 0
const unsigned char* frame_header_t::read(const unsigned char* cp)
{
	m_precision = static_cast<unsigned int>(*cp++);
	m_height = (static_cast<unsigned int>(*cp++) << 8) + static_cast<unsigned int>(*cp++);
	m_width = (static_cast<unsigned int>(*cp++) << 8) + static_cast<unsigned int>(*cp++);
	m_numComponents = static_cast<unsigned int>(*cp++);
	
	for (int i = 0; i < m_numComponents && i < 4; i++)
		cp = m_component_spec[i].read(cp);

	return cp;
}
#pragma optimization_level reset

const unsigned char* scan_header_t::component_spec_t::read(const unsigned char* cp)
{
	m_componentSelector = static_cast<unsigned int>(*cp++);
	unsigned int codingTableSelector = static_cast<unsigned int>(*cp++);
	m_dcCodingTableSelector = codingTableSelector >> 4;
	m_acCodingTableSelector = codingTableSelector & 0x0F;
	return cp;
}

const unsigned char* scan_header_t::read(const unsigned char* cp)
{
	m_numComponents = static_cast<unsigned int>(*cp++);
	for (int i = 0; i < m_numComponents; i++)
		cp = m_component_spec[i].read(cp);
	m_predictorSelection = static_cast<unsigned int>(*cp++);
	cp++;
	m_pointTransform = static_cast<unsigned int>(*cp++) & 0x0F;
	return cp;
}

const unsigned char* huffman_table_t::read(const unsigned char* cp)
{
	const unsigned int class_id = static_cast<unsigned int>(*cp++);
	m_class = class_id >> 4;
	m_id = class_id & 0x0F;

	unsigned int num_codes = 0;
	unsigned int code_length_table[16];
	for (int i = 0; i < 16; i++, cp++)
		num_codes += code_length_table[i] = static_cast<unsigned int>(*cp);

	struct huff_table_entry_t {
		unsigned char m_length;
		unsigned char m_value;
		unsigned short m_code;
	};
	huff_table_entry_t* const huff_table = (huff_table_entry_t*)_alloca(sizeof(huff_table_entry_t) * num_codes);
	const huff_table_entry_t* const tend = huff_table + num_codes;
	huff_table_entry_t* htp;

	htp = huff_table;
	for (int i = 0; i < 16; i++)
		for (int j = 0; j < code_length_table[i]; j++, htp++)
			htp->m_length = static_cast<unsigned char>(i + 1);

	htp = huff_table;
	unsigned int code = 0;
	unsigned int code_length = htp->m_length;
	while (true) {
		for (; htp->m_length == code_length; htp++, code++)
			htp->m_code = code;
		if (htp == tend)
			break;
		do {
			code <<= 1;
			code_length++;
		} while (htp->m_length != code_length);
	}

	for (htp = huff_table; htp < tend; htp++, cp++)
		htp->m_value = static_cast<unsigned int>(*cp);

	ZeroMemory(m_tree_box, sizeof(m_tree_box));
	unsigned int free_node_index = 1;
	for (htp = huff_table; htp < tend; htp++) {
		const unsigned int code = htp->m_code;
		huffman_tree_node_t* node = m_tree_box;
		for (unsigned int mask = 1 << (htp->m_length - 1); mask > 1; mask >>= 1) {
			const unsigned int bit = (code & mask) == 0 ? 0 : 1;
			if ((*node)[bit] == 0)
				(*node)[bit] = free_node_index++;
			node = m_tree_box + (*node)[bit];
		}
		(*node)[code & 1] = 0x80 | htp->m_value;
	}

	return cp;
}

unsigned int huffman_table_t::readCode(CJPEGStream& stream)
{
	huffman_tree_node_t* node = m_tree_box;
	while (true) {
		const unsigned int bit = stream.readBit();
		const unsigned int index = (*node)[bit];
		if (index & 0x80)
			return index & 0x7F;
		node = m_tree_box + index;
	}
}

void CLosslessJPEGDecoder::read_SOF(CLosslessJPEGDecoder* self, unsigned char marker)
{
	if (marker != 0xC3)
		return;

	const unsigned int len = self->m_stream.readWord() - 2;
	const unsigned char* segment_data = self->m_stream.getBytes(len);
	if (segment_data == NULL) {
		segment_data = (const unsigned char*)_alloca(len);
		self->m_stream.readBytes(const_cast<unsigned char*>(segment_data), len);
	}

	self->m_fh.read(segment_data);

	self->m_state = 2;
}

void CLosslessJPEGDecoder::read_SOS(CLosslessJPEGDecoder* self, unsigned char marker)
{
	const unsigned int len = self->m_stream.readWord() - 2;
	const unsigned char* segment_data = self->m_stream.getBytes(len);
	if (segment_data == NULL) {
		segment_data = (const unsigned char*)_alloca(len);
		self->m_stream.readBytes(const_cast<unsigned char*>(segment_data), len);
	}

	self->m_sh.read(segment_data);

	self->m_state = 3;
}

void CLosslessJPEGDecoder::read_DHT(CLosslessJPEGDecoder* self, unsigned char marker)
{
	const unsigned int len = self->m_stream.readWord() - 2;
	const unsigned char* segment_data = self->m_stream.getBytes(len);
	if (segment_data == NULL) {
		segment_data = (const unsigned char*)_alloca(len);
		self->m_stream.readBytes(const_cast<unsigned char*>(segment_data), len);
	}

	const unsigned char* cp = segment_data;
	const unsigned char* ep = cp + len;

	for (int i = 0; i < 4 && cp < ep; i++)
		cp = self->m_ht[i].read(cp);
}

void CLosslessJPEGDecoder::read_DRI(CLosslessJPEGDecoder* self, unsigned char marker)
{
	const unsigned int len = self->m_stream.readWord() - 2;
	self->m_restartInterval = self->m_stream.readWord();
}

void CLosslessJPEGDecoder::read_SOI(CLosslessJPEGDecoder* self, unsigned char marker)
{
	self->m_state = 1;
}

void CLosslessJPEGDecoder::nop(CLosslessJPEGDecoder* self, unsigned char marker)
{
}

void CLosslessJPEGDecoder::skip_segment(CLosslessJPEGDecoder* self, unsigned char marker)
{
	const unsigned int len = self->m_stream.readWord() - 2;
	self->m_stream.skipBytes(len);
}

void CLosslessJPEGDecoder::read_segment(void)
{
	const unsigned int marker = m_stream.readMarker();
	if (marker == 0xFF)
		return;
	if (marker >= 0xC0 && marker < 0xE0) {
		typedef void (*marker_reader_proc_t)(CLosslessJPEGDecoder* self, unsigned char marker);
		const static marker_reader_proc_t s_maker_reader_table[16 * 2] = {
			read_SOF, // FFC0
			read_SOF, // FFC1
			read_SOF, // FFC2
			read_SOF, // FFC3
			read_DHT, // FFC4
			read_SOF, // FFC5
			read_SOF, // FFC6
			read_SOF, // FFC7
			skip_segment, // FFC8 "JPG": Reserved for JPEG extensions
			read_SOF, // FFC9
			read_SOF, // FFCA
			read_SOF, // FFCB
			skip_segment, // FFCC "DAC": Define arithmetic coding conditionings
			read_SOF, // FFCD
			read_SOF, // FFCE
			read_SOF, // FFCF
			nop, // FFD0 "RST"
			nop, // FFD1 "RST"
			nop, // FFD2 "RST"
			nop, // FFD3 "RST"
			nop, // FFD4 "RST"
			nop, // FFD5 "RST"
			nop, // FFD6 "RST"
			nop, // FFD7 "RST"
			read_SOI, // FFD8 "SOI"
			nop, // FFD9 "EOI"
			read_SOS, // FFDA
			skip_segment, // FFDB "DQT"
			skip_segment, // FFDC "DNL"
			read_DRI, // FFDD
			skip_segment, // FFDE "DHP"
			skip_segment, // FFDF "EXP"
		};
		(s_maker_reader_table[marker - 0xC0])(this, marker);
	} else {
		skip_segment(this, marker);
	}
}

void CLosslessJPEGDecoder::init(f_decode_input_func input_func, void* input_ctx)
{
	ZeroMemory(this, sizeof(CLosslessJPEGDecoder));

	m_stream.init_source(input_func, input_ctx);
}

void CLosslessJPEGDecoder::setup(void)
{
	while (m_state < 3)
		read_segment();

	m_rowCounter = 0;
	m_restartCounter = 0;
	m_firstPixelPredict = 1 << (m_fh.m_precision - m_sh.m_pointTransform - 1);
	
}

void CLosslessJPEGDecoder::proc(void* buf)
{
	if (m_rowCounter == m_fh.m_height)
		return;

	if (m_restartInterval != 0 && m_restartCounter == m_restartInterval) {
		const unsigned int marker = m_stream.readMarker();
		m_restartCounter = 0;
		m_firstPixelPredict = 1 << (m_fh.m_precision - m_sh.m_pointTransform - 1);
	}
	
	unsigned short* cp = (unsigned short*)buf;
	unsigned int a = m_firstPixelPredict;

	const unsigned int code = m_ht[0].readCode(m_stream);
	const int diff = m_stream.readBits(code);
	*cp++ = m_firstPixelPredict = a = a + diff;

	for (int i = 1; i < m_fh.m_width; i++) {
		const unsigned int code = m_ht[0].readCode(m_stream);
		const int diff = m_stream.readBits(code);
		 *cp++ = a = a + diff;
	}

	m_restartCounter++;
	m_rowCounter++;
}
