/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

class CJPEGStream {
	
	f_decode_input_func m_input_func;
	void* m_input_ctx;

	const unsigned char* m_data_ptr;
	unsigned int m_data_len;

	const unsigned char* m_cp;
	const unsigned char* m_ep;
	unsigned int m_cacheByte;
	unsigned int m_bitsInCacheByte;

	void fill_input_buffer(const int flag = 1);
	void read_cacheByte();

public:
	CJPEGStream()
	{
	}

	// ԍŏ̃f[^̎擾
	// IJG  CODEC ł init_source() ɑB
	void init_source(f_decode_input_func input_func, void* input_ctx);

	// }[J[ZOgǂݍݐpC^tFCX
	
	// ZOg}[J[ǂށBāAFF ĂGXP[v͂ȂB
	// tɁA擪 FF łȂꍇ́A0xFF ԂA|C^i߂ȂB
	unsigned int readMarker(void);
	unsigned int readByte(void);
	unsigned int readWord(void);

	// w肳ꂽ̃obt@AĎȂAobt@̃AhXԂB
	// AAĂƂȂȂ NULL ԂÃC^tFCXB
	const unsigned char* getBytes(unsigned int len);
	// AĂƂȂ̂ŁAĂяopӂobt@ɃRs[ƂB
	void readBytes(unsigned char* buf, unsigned int len);

	void skipBytes(unsigned int skip);

	// Encoded Code Segment pǂݍ݃C^tFCX
	unsigned int readBit();
	int readBits(const unsigned int bitCount);
};

struct frame_header_t {
	unsigned int m_width;
	unsigned int m_height;
	unsigned int m_precision;
	unsigned int m_numComponents;

	struct component_spec_t {
		unsigned int m_componentIdentifier;
		unsigned int m_horizontalSamplingFactor;
		unsigned int m_verticalSamplingFactor;
		unsigned int m_quantization_table_selector;

		const unsigned char* read(const unsigned char* cp);
	}  m_component_spec[4];

	const unsigned char* read(const unsigned char* cp);
};

struct scan_header_t {
	unsigned int m_numComponents;

	struct component_spec_t {
		unsigned int m_componentSelector;
		unsigned int m_dcCodingTableSelector;
		unsigned int m_acCodingTableSelector;

		const unsigned char* read(const unsigned char* cp);
	} m_component_spec[4];

	unsigned int m_predictorSelection;
	unsigned int m_pointTransform;

	const unsigned char* read(const unsigned char* cp);
};

struct huffman_table_t {
	unsigned int m_class;
	unsigned int m_id;

	typedef unsigned char huffman_tree_node_t[2];
	huffman_tree_node_t m_tree_box[128];

	const unsigned char* read(const unsigned char* cp);

	unsigned int readCode(CJPEGStream& stream);
};

class CLosslessJPEGDecoder {
public:
	frame_header_t m_fh;
	scan_header_t m_sh;
	huffman_table_t m_ht[4];
	unsigned int m_restartInterval;

	CLosslessJPEGDecoder()
	{
	}

	void init(f_decode_input_func input_func, void* input_ctx);
	void setup(void);
	void proc(void* buf);

private:
	CJPEGStream m_stream;

	unsigned int m_state;
		// state of process
		// 0: before SOI
		// 1: after SOI, before SOF, in tables/misc segments for Image
		// 2: after SOF, before SOS Header, in tables/misc segments for Frame
		// 3: after SOS, in Encoded Code Segment

	unsigned int m_firstPixelPredict;
	unsigned int m_restartCounter;
	unsigned int m_rowCounter;

	static void read_SOF(CLosslessJPEGDecoder* self, unsigned char marker);
	static void read_SOS(CLosslessJPEGDecoder* self, unsigned char marker);
	static void read_DHT(CLosslessJPEGDecoder* self, unsigned char marker);
	static void read_DRI(CLosslessJPEGDecoder* self, unsigned char marker);
	static void read_SOI(CLosslessJPEGDecoder* self, unsigned char marker);
	static void nop(CLosslessJPEGDecoder* self, unsigned char marker);
	static void skip_segment(CLosslessJPEGDecoder* self, unsigned char marker);
	void read_segment(void);
};
