/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

static DWORD tlsSlot = TLS_OUT_OF_INDEXES;

void Init_CTN_HEAP()
{
	tlsSlot = TlsAlloc();
}

HANDLE Get_CTN_HEAP_HANDLE()
{
	if (tlsSlot == TLS_OUT_OF_INDEXES)
		return NULL;

	HANDLE hHeap = (HANDLE)TlsGetValue(tlsSlot);
	if (hHeap != NULL)
		return hHeap;

	hHeap = HeapCreate(HEAP_NO_SERIALIZE, 0x8000, 0);
	if (hHeap == NULL)
		return NULL;

	TlsSetValue(tlsSlot, hHeap);

	return hHeap;
}

void* CTN_MALLOC(size_t size)
{
	if (tlsSlot == TLS_OUT_OF_INDEXES)
		return GlobalAlloc(GMEM_FIXED, size);

	HANDLE hHeap = (HANDLE)TlsGetValue(tlsSlot);
	if (hHeap != NULL)
		return HeapAlloc(hHeap, HEAP_NO_SERIALIZE, size);

	hHeap = HeapCreate(HEAP_NO_SERIALIZE, 0x8000, 0);
	if (hHeap == NULL)
		return GlobalAlloc(GMEM_FIXED, size);

	TlsSetValue(tlsSlot, hHeap);

	return HeapAlloc(hHeap, HEAP_NO_SERIALIZE, size);
}

void CTN_FREE(void* ptr)
{
	if (tlsSlot == TLS_OUT_OF_INDEXES) {
		GlobalFree(ptr);
		return;
	}

	HANDLE hHeap = (HANDLE)TlsGetValue(tlsSlot);
	if (hHeap == NULL) {
		GlobalFree(ptr);
		return;
	}

	HeapFree(hHeap, HEAP_NO_SERIALIZE, ptr);
}
