/*
          Copyright (C) 1993, 1994, RSNA and Washington University

          The software and supporting documentation for the Radiological
          Society of North America (RSNA) 1993, 1994 Digital Imaging and
          Communications in Medicine (DICOM) Demonstration were developed
          at the
                  Electronic Radiology Laboratory
                  Mallinckrodt Institute of Radiology
                  Washington University School of Medicine
                  510 S. Kingshighway Blvd.
                  St. Louis, MO 63110
          as part of the 1993, 1994 DICOM Central Test Node project for, and
          under contract with, the Radiological Society of North America.

          THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER RSNA NOR
          WASHINGTON UNIVERSITY MAKE ANY WARRANTY ABOUT THE SOFTWARE, ITS
          PERFORMANCE, ITS MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR
          USE, FREEDOM FROM ANY COMPUTER DISEASES OR ITS CONFORMITY TO ANY
          SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF
          THE SOFTWARE IS WITH THE USER.

          Copyright of the software and supporting documentation is
          jointly owned by RSNA and Washington University, and free access
          is hereby granted as a license to use this software, copy this
          software and prepare derivative works based upon this software.
          However, any distribution of this software source code or
          supporting documentation or derivative works (source code and
          supporting documentation) must include the three paragraphs of
          the copyright notice.
*/
/* Copyright marker.  Copyright will be inserted above.  Do not remove */
/*
** @$=@$=@$=
*/
/*
**				DICOM 93
**		     Electronic Radiology Laboratory
**		   Mallinckrodt Institute of Radiology
**		Washington University School of Medicine
**
** Module Name(s):
**			DUL_InitializeNetwork
**			DUL_DropNetwork
**			DUL_RequestAssociation
**			DUL_ReceiveAssociationRQ
**			DUL_AcknowledgeAssociateRQ
**			DUL_RejectAssociateRQ
**			DUL_ReleaseAssociation
**			DUL_AbortAssociation
**			DUL_DropAssociation
**			DUL_WritePDVs
**			DUL_ReadPDVs
** Author, Date:	Stephen M. Moore, 14-Apr-93
** Intent:		This module contains the public entry points for the
**			DICOM Upper Layer (DUL) protocol package.
** Last Update:		$Author: scsb $, $Date: 2002/10/15 19:37:43 $
** Source File:		$RCSfile: scsb_dulprotocol.cpp,v $
** Revision:		$Revision: 1.1.1.1 $
** Status:		$State: Exp $
*/

#include "dicom.h"
#include "condition.h"
#include "lst.h"
#include "dicom_uids.h"
#include "ctnthread.h"
#include "dulprotocol.h"
#include "dulstructures.h"
#include "dulfsm.h"

static void dump_uid_ex(FILE* fp, char *UID, char *indent)
{
    CONDITION
    cond;
    UID_DESCRIPTION
	d;

    if (UID[0] == '\0') {
	fprintf(fp, indent, " ");
	fprintf(fp, "No UID\n");
    } else {
	cond = UID_Lookup(UID, &d);
	if (cond == UID_NORMAL) {
	    fprintf(fp, indent, " ");
	    fprintf(fp, "%s, %s\n", d.description, d.originator);
	} else {
	    (void) COND_PopCondition(FALSE);
	    fprintf(fp, indent, " ");
	    fprintf(fp, "Unknown UID\n");
	}
    }
}

typedef struct {
    DUL_SC_ROLE role;
    char *text;
}   SC_MAP;
static SC_MAP scMap[] = {
    {DUL_SC_ROLE_DEFAULT, "Default"},
    {DUL_SC_ROLE_SCU, "SCU"},
    {DUL_SC_ROLE_SCP, "SCP"},
    {DUL_SC_ROLE_SCUSCP, "SCUSCP"},
};

static void dump_presentation_ctx_ex(FILE* fp, LST_HEAD ** l)
{
    DUL_PRESENTATIONCONTEXT
    * ctx;
    DUL_TRANSFERSYNTAX
	* transfer;
    int
        index;

    if (*l == NULL)
	return;

    ctx = (DUL_PRESENTATIONCONTEXT*)LST_Head(l);
    if (ctx == NULL)
	return;

    (void) LST_Position(l, ctx);

    while (ctx != NULL) {
	fprintf(fp, "  Context ID:           %d\n", ctx->presentationContextID);
	fprintf(fp, "  Abstract Syntax:      %s\n", ctx->abstractSyntax);
	dump_uid_ex(fp, ctx->abstractSyntax, "%24s");
	fprintf(fp, "  Result field:         %d\n", (int) ctx->result);
	for (index = 0; index < (int) DIM_OF(scMap); index++) {
	    if (ctx->proposedSCRole == scMap[index].role)
		fprintf(fp, "  Proposed SCU/SCP Role:  %s\n", scMap[index].text);
	}
	for (index = 0; index < (int) DIM_OF(scMap); index++) {
	    if (ctx->acceptedSCRole == scMap[index].role)
		fprintf(fp, "  Accepted SCU/SCP Role:  %s\n", scMap[index].text);
	}
	fprintf(fp, "  Proposed Xfer Syntax(es)\n");
	if (ctx->proposedTransferSyntax != NULL) {
	    transfer = (DUL_TRANSFERSYNTAX*)LST_Head((void**)&ctx->proposedTransferSyntax);
	    if (transfer != NULL)
		(void) LST_Position(&ctx->proposedTransferSyntax, transfer);

	    while (transfer != NULL) {
		fprintf(fp, "                  %s\n", transfer->transferSyntax);
		dump_uid_ex(fp, transfer->transferSyntax, "%18s");
		transfer = (DUL_TRANSFERSYNTAX*)LST_Next((void**)&ctx->proposedTransferSyntax);
	    }
	}
	fprintf(fp, "  Accepted Xfer Syntax: %s\n", ctx->acceptedTransferSyntax);
	dump_uid_ex(fp, ctx->acceptedTransferSyntax, "%24s");
	ctx = (DUL_PRESENTATIONCONTEXT*)LST_Next(l);
    }

}

void DUL_DumpParamsEx(FILE* fp, DUL_ASSOCIATESERVICEPARAMETERS * params, int flag)
{
	fprintf(fp, "APP CTX NAME:%s\n", params->applicationContextName);
	dump_uid_ex(fp, params->applicationContextName, "%13s");
	fprintf(fp, "AP TITLE:     %s\n", params->callingAPTitle);
	fprintf(fp, "AP TITLE:     %s\n", params->calledAPTitle);
	fprintf(fp, "AP TITLE:     %s\n", params->respondingAPTitle);
	fprintf(fp, "MAX PDU:      %d\n", params->maxPDU);
	fprintf(fp, "Peer MAX PDU: %d\n", params->peerMaxPDU);
	fprintf(fp, "PRES ADDR:    %s\n", params->callingPresentationAddress);
	fprintf(fp, "PRES ADDR:    %s\n", params->calledPresentationAddress);
	fprintf(fp, "REQ IMP UID:  %s\n", params->callingImplementationClassUID);
	dump_uid_ex(fp, params->callingImplementationClassUID, "%13s");
	fprintf(fp, "REQ VERSION:  %s\n", params->callingImplementationVersionName);
	fprintf(fp, "ACC IMP UID:  %s\n", params->calledImplementationClassUID);
	dump_uid_ex(fp, params->calledImplementationClassUID, "%13s");
	fprintf(fp, "ACC VERSION:  %s\n", params->calledImplementationVersionName);
	if (flag & 1) {
		fprintf(fp, "Requested Presentation Ctx\n");
		dump_presentation_ctx_ex(fp, &params->requestedPresentationContext);
	}
	if (flag & 2) {
		fprintf(fp, "Accepted Presentation Ctx\n");
		dump_presentation_ctx_ex(fp, &params->acceptedPresentationContext);
	}
}

