/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

using System;
using System.Runtime.InteropServices;
using System.Collections;
using System.Net;

namespace DIOWave {

	//==========================================================================
	//
	//
	//
	//==========================================================================
	public enum VR {
		NONE,
		AE,
		AS,
		CS,
		DA,
		DS,
		DT,
		IS,
		LO,
		LT,
		PN,
		SH,
		ST,
		SL,
		SS,
		TM,
		UI,
		UT,
		UL,
		US,
		UN,
		CTX
	}

	//==========================================================================
	//
	//
	//
	//==========================================================================
	public class DataElementObject {
		public readonly UInt32 tag;
		public readonly VR vr;
		public readonly int vm;
		public readonly Object value;

		//======================================================================
		//
		//======================================================================
		public DataElementObject(UInt32 tag, VR vr, int vm, Object value) {
			this.tag = tag;
			this.vr = vr;
			this.vm = vm;
			this.value = value;
		}
	}

	//==========================================================================
	//
	//
	//
	//==========================================================================
	public class SOPInstanceObject {
		public readonly String path;
		public readonly ArrayList data;

		//======================================================================
		//
		//======================================================================
		public SOPInstanceObject(String path, ArrayList data) {
			this.path = path;
			this.data = data;
		}
	}

	//==========================================================================
	//
	//
	//
	//==========================================================================
	public class DicomStudy {
		public ArrayList el, fl;

		//======================================================================
		//
		//======================================================================
		public DicomStudy() {
			el = new ArrayList();
			fl = new ArrayList();
		}
	}

	//======================================================================
	//
	//======================================================================
	public class RawCache {
		public Byte[] dataElm;
		public Byte[] instDir;
	}

	//==========================================================================
	//
	//
	//
	//==========================================================================
	public class StudyCache {

		//======================================================================
		//
		//======================================================================
		public void Initialize() {
			InitializeStudyCacheBuffer(false);
		}

		//======================================================================
		//
		//======================================================================
		public void Uninitialize() {
			UninitializeStudyCacheBuffer();
		}

		//======================================================================
		//
		//======================================================================
		public DicomStudy GetDicomStudy(String location) {

			DicomStudy dicomstudy = new DicomStudy();

			unsafe {
				void* ctx = null;
				try {
					Byte* dataElmPtr, instDirPtr;
					UInt32 dataElmSize, instDirSize;
					ctx = MapStudyCacheBuffer(location, &dataElmPtr, &dataElmSize, &instDirPtr, &instDirSize);

					Hashtable hash = new Hashtable();
					ProcDataElementCache(dataElmPtr, dataElmSize, dicomstudy.el, hash);
					ProcInstanceDirectory(instDirPtr, instDirSize, dicomstudy.fl, hash);

				} finally {
					if (ctx != null)
						UnmapStudyCacheBuffer(ctx);
				}
			}

			return dicomstudy;
		}

		//======================================================================
		//
		//======================================================================
		public DicomStudy GetDicomStudy(String studyDir, int study_cache_flag) {

			DicomStudy dicomstudy = new DicomStudy();

			unsafe {
				void* ctx = null;
				try {
					Byte* dataElmPtr, instDirPtr;
					UInt32 dataElmSize, instDirSize;
					ctx = MapStudyCacheBuffer_(studyDir, study_cache_flag, &dataElmPtr, &dataElmSize, &instDirPtr, &instDirSize);

					Hashtable hash = new Hashtable();
					ProcDataElementCache(dataElmPtr, dataElmSize, dicomstudy.el, hash);
					ProcInstanceDirectory(instDirPtr, instDirSize, dicomstudy.fl, hash);

				} finally {
					if (ctx != null)
						UnmapStudyCacheBuffer(ctx);
				}
			}

			return dicomstudy;
		}

		//======================================================================
		//
		//======================================================================
		public DicomStudy GetRemoteDicomStudy(String location) {

			WebClient client = new WebClient();
			Byte[] rawCache = client.DownloadData(location);

			DicomStudy dicomstudy = new DicomStudy();
	
			unsafe {
				fixed (Byte* cachePtr = rawCache) {
					Byte* dataElmPtr, instDirPtr;
					UInt32 dataElmSize, instDirSize;
	
					Byte* ptr = cachePtr;
	
					dataElmSize = *(UInt32*)ptr;
					ptr += 4;
	
					dataElmPtr = ptr;
					ptr += dataElmSize;
	
					instDirSize = *(UInt32*)ptr;
					ptr += 4;
	
					instDirPtr = ptr;
	
					Hashtable hash = new Hashtable();
					ProcDataElementCache(dataElmPtr, dataElmSize, dicomstudy.el, hash);
					ProcInstanceDirectory(instDirPtr, instDirSize, dicomstudy.fl, hash);
				}
	
			}

			return dicomstudy;
		}

		//======================================================================
		//
		//======================================================================
		private static unsafe void ProcDataElementCache(Byte* dataElmPtr, UInt32 dataElmSize, ArrayList el, Hashtable hash) {

			Byte* cp = dataElmPtr;
			Byte* ep = cp + dataElmSize;

			CDicomDataElement elm = new CDicomDataElement(); 

			while (cp < ep) {
				elm.Parse(cp);

				UInt32 offset = (UInt32)(cp - dataElmPtr);

				DataElementObject e = elm.ToDataElementObject(cp);
				el.Add(e);
				hash.Add(offset, e);

				cp = elm.next_ptr(cp);
			}
		}

		//======================================================================
		//
		//======================================================================
		private static unsafe void ProcInstanceDirectory(Byte* instDirPtr, UInt32 instDirSize, ArrayList fl, Hashtable hash) {

			Byte* cp = instDirPtr;
			Byte* ep = cp + instDirSize;

			while (cp < ep) {

				UInt32 entry_size = *(UInt32*)(cp);
				UInt16 header_size = *(UInt16*)(cp + 4);

				Char* path = (Char*)(cp + 6);
				Int32 pathlen = 0;
				for (Char* pp = path; *pp != 0; pp++)
					pathlen++;

				ArrayList data = new ArrayList();

				Byte* dp = cp + header_size;
				Byte* de = cp + entry_size;
				while (dp < de) {

					UInt32 offset = *(UInt32*)(dp + 4);
					data.Add(hash[offset]);

					dp += 8;
				}

				fl.Add(new SOPInstanceObject(new String(path, 0, pathlen), data));

				cp += entry_size;
			}
		}
		//======================================================================
		//
		//
		//
		//======================================================================
		private unsafe struct CDicomDataElement {
			public UInt16 m_gtag;
			public UInt16 m_etag;
			public Byte* m_vr;
			public UInt32 m_length;
			public UInt32 m_hsize;

			private enum VARTYPE { VT_I2, VT_I4, VT_UI2, VT_UI4, VT_R4, VT_R8 }
	
			//==================================================================
			//
			//==================================================================
			public void Parse(Byte* ptr) {
				m_gtag = *(ushort*)(ptr + 0);
				m_etag = *(ushort*)(ptr + 2);
				m_vr = ptr + 4;
				if (isSQ() || isBLE()) {
					m_length = *(uint*)(ptr + 8);
					m_hsize = 12;
				} else {
					m_length = *(ushort*)(ptr + 6);
					m_hsize = 8;
				}
			}

			//==================================================================
			//
			//==================================================================
			private bool isBLE() { // is Binary Large Element
				return (m_vr[0] == 'O' && (m_vr[1] == 'B' || m_vr[1] == 'W' || m_vr[1] == 'T'))
				    || (m_vr[0] == 'U' && m_vr[1] == 'N');
			}
			//==================================================================
			//
			//==================================================================
			private bool isSQ() {
				return m_vr[0] == 'S' && m_vr[1] == 'Q';
			}

			//==================================================================
			//
			//==================================================================
			public Byte* next_ptr(Byte* ptr) { return ptr + m_hsize + m_length; }

			//==================================================================
			//
			//==================================================================
			public DataElementObject ToDataElementObject(Byte* ptr) {

				UInt32 tag = (UInt32)m_gtag * 0x10000 + (UInt32)m_etag;

				if (m_length == 0)
					return new DataElementObject(tag, VR.NONE, 0, null);

				VR vr = VR.NONE;

				ArrayList values = new ArrayList();

				Byte* value_ptr = ptr + m_hsize;

				if (m_vr[0] == 'A') {
					if (m_vr[1] == 'E') { // AS
						string_value(value_ptr, true, true, false, values);
						vr = VR.AE;
					} else if (m_vr[1] == 'S') { // AS
						string_value(value_ptr, true, true, false, values);
						vr = VR.AS;
					}
				} else if (m_vr[0] == 'C') {
					if (m_vr[1] == 'S') { // CS
						string_value(value_ptr, true, true, false, values);
						vr = VR.CS;
					} else if (m_vr[1] == 'T') { // CTX
						binary_numeric_value(value_ptr, VARTYPE.VT_I2, values);
						vr = VR.CTX;
					}
				} else if (m_vr[0] == 'D') {
					if (m_vr[1] == 'A') { // DA
						string_value(value_ptr, true, true, false, values);
						vr = VR.DA;
					} else if (m_vr[1] == 'S') { // DS
						string_value(value_ptr, true, true, false, values);
						vr = VR.DS;
					} else if (m_vr[1] == 'T') { // DT
						string_value(value_ptr, true, true, false, values);
						vr = VR.DT;
					}
				} else if (m_vr[0] == 'I') {
					if (m_vr[1] == 'S') { // IS
						string_value(value_ptr, true, true, false, values);
						vr = VR.IS;
					}
				} else if (m_vr[0] == 'L') {
					if (m_vr[1] == 'O') { // LO
						string_value(value_ptr, true, false, true, values);
						vr = VR.LO;
					} else if (m_vr[1] == 'T') { // LT
						string_value(value_ptr, false, false, true, values);
						vr = VR.LT;
					}
				} else if (m_vr[0] == 'P') {
					if (m_vr[1] == 'N') { // PN
						string_value(value_ptr, true, true, true, values);
						vr = VR.PN;
					}
				} else if (m_vr[0] == 'S') {
					if (m_vr[1] == 'H') { // SH
						string_value(value_ptr, true, true, true, values);
						vr = VR.SH;
					} else if (m_vr[1] == 'T') { // ST
						string_value(value_ptr, false, false, true, values);
						vr = VR.ST;
					} else if (m_vr[1] == 'L') { // SL
						binary_numeric_value(value_ptr, VARTYPE.VT_I4, values);
						vr = VR.SL;
					} else if (m_vr[1] == 'S') { // SS
						binary_numeric_value(value_ptr, VARTYPE.VT_I2, values);
						vr = VR.SS;
					}
				} else if (m_vr[0] == 'T') {
					if (m_vr[1] == 'M') { // TM
						string_value(value_ptr, true, true, false, values);
						vr = VR.TM;
					}
				} else if (m_vr[0] == 'U') {
					if (m_vr[1] == 'I') { // UI
						string_value(value_ptr, true, true, false, values);
						vr = VR.UI;
					} else if (m_vr[1] == 'T') { // UT
						string_value(value_ptr, false, false, true, values);
						vr = VR.UT;
					} else if (m_vr[1] == 'L') { // UL
						binary_numeric_value(value_ptr, VARTYPE.VT_UI4, values);
						vr = VR.UL;
					} else if (m_vr[1] == 'S') { // US
						binary_numeric_value(value_ptr, VARTYPE.VT_UI2, values);
						vr = VR.US;
					}
				}

				if (values.Count == 0)
					return new DataElementObject(tag, vr, 0, null);
				else if (values.Count == 1)
					return new DataElementObject(tag, vr, 1, values[0]);
				else
					return new DataElementObject(tag, vr, values.Count, values);
			}

			//==================================================================
			//
			//==================================================================
			private void string_value(Byte* value_ptr, bool multiValue, bool trimSpace, bool multiByte, ArrayList values)
			{
				Byte* cp = value_ptr;
				Byte* ep = cp + m_length;

				while (cp < ep) {
					if (trimSpace) {
						while (cp < ep && (!multiValue || *cp != '\\') && *cp <= ' ')
							cp++;
						if (cp == ep) {
							break;
						}
					}
					Byte* bp = cp;
					while (cp < ep && (!multiValue || *cp != '\\'))
						cp++;
					Byte* tp = cp;
					if (trimSpace) {
						while (tp > bp && *(tp - 1) <= ' ')
							tp--;
					}

					if (tp == bp) {
						values.Add(String.Empty);
					} else if (multiByte) {
						values.Add(new String((SByte*)bp, 0, (int)(tp - bp)));
					} else {
						values.Add(new String((SByte*)bp, 0, (int)(tp - bp)));
					}

					if (cp < ep && *cp == '\\')
						cp++;
				}
			}
	
			//==================================================================
			//
			//==================================================================
			private void binary_numeric_value(Byte* value_ptr, VARTYPE vt, ArrayList values)
			{
				Byte* cp = value_ptr;
				Byte* ep = cp + m_length;
	
				while (cp < ep) {
	
					switch (vt) {
					case VARTYPE.VT_I2:
						values.Add(*(Int16*)cp);
						cp += 2;
						break;
					case VARTYPE.VT_I4:
						values.Add(*(Int32*)cp);
						cp += 4;
						break;
					case VARTYPE.VT_UI2:
						values.Add(*(UInt16*)cp);
						cp += 2;
						break;
					case VARTYPE.VT_UI4:
						values.Add(*(UInt32*)cp);
						cp += 4;
						break;
					case VARTYPE.VT_R4:
						values.Add(*(Single*)cp);
						cp += 4;
						break;
					case VARTYPE.VT_R8:
						values.Add(*(Double*)cp);
						cp += 8;
						break;
					}
	
				}
			}

		}

		//======================================================================
		//
		//======================================================================
	    static unsafe void Copy(Byte* src, Byte[] dst, UInt32 count) {
			fixed (Byte* pDst = dst) {
				Byte* ps = src;
				Byte* pd = pDst;
				for (UInt32 n = count >> 2; n != 0; n--) {
					*((int*)pd) = *((int*)ps);
					pd += 4;
					ps += 4;
				}
				for (count &= 3; count != 0; count--) {
					*pd = *ps;
					pd++;
					ps++;
				}
			}
        }

		//======================================================================
		//
		//======================================================================
		[DllImport("dio-study.dll", CharSet=CharSet.Unicode, PreserveSig=false)]
		static extern unsafe void InitializeStudyCacheBuffer(bool logon);

		//======================================================================
		//
		//======================================================================
		[DllImport("dio-study.dll", CharSet=CharSet.Unicode, PreserveSig=false)]
		static extern unsafe void UninitializeStudyCacheBuffer();

		//======================================================================
		//
		//======================================================================
		[DllImport("dio-study.dll", CharSet=CharSet.Unicode, PreserveSig=false)]
		static extern unsafe void* MapStudyCacheBuffer(String studyLoc, Byte** dataElmPtr, UInt32* dataElmSize, Byte** instDirPtr, UInt32* instDirSize);

		//======================================================================
		//
		//======================================================================
		[DllImport("dio-study.dll", CharSet=CharSet.Unicode, PreserveSig=false)]
		static extern unsafe void UnmapStudyCacheBuffer(void* ctx);

		//======================================================================
		//
		//======================================================================
		[DllImport("dio-study.dll", CharSet=CharSet.Unicode, PreserveSig=false)]
		static extern unsafe void* MapStudyCacheBuffer_(String studyDir, int study_cache_flag, Byte** dataElmPtr, UInt32* dataElmSize, Byte** instDirPtr, UInt32* instDirSize);

	}
}
