#ifndef _735ED798_ED26_4ADA_972B_B6DA5844574E_
#define _735ED798_ED26_4ADA_972B_B6DA5844574E_

/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#pragma warn_unusedarg off

class CDataElementCache {
protected:
	struct avlnode_buffer_t;

	struct avlnode_t {
		avlnode_t *left;
		avlnode_t *right;
		DWORD skew;
		DWORD offset;
		LPCVOID d;

		enum AVLRES {
			AVL_HIT = 0,
			AVL_OK,
			AVL_BALANCE,
			AVL_ERROR = -1
		};
		enum AVLSKEW {
			AVL_NONE,	
			AVL_LEFT,
			AVL_RIGHT
		};

		static void avlrotright(avlnode_t **n);
		static void avlrotleft(avlnode_t **n);
		static AVLRES avlleftgrown(avlnode_t **n);
		static AVLRES avlrightgrown(avlnode_t **n);

		static AVLRES avlinsert(avlnode_t** n, const tDcmElementHdr& hdr, LPCVOID d, avlnode_buffer_t& avlbuf, CDataElementCache* elementcache, DWORD& offset);

	};

	struct avlnode_buffer_t {
		enum {
			kNodesInBlock = 1024
		};
		struct block_t {
			block_t* m_next;
			avlnode_t m_data[kNodesInBlock];

			block_t()
				: m_next(NULL)
			{
			}

			~block_t()
			{
			}
			static void* operator new(size_t size, void* ptr)
			{
				return ptr;
			}
			static void operator delete(void* deadObject)
			{
				return CoTaskMemFree(deadObject);
			}
		};

		block_t m_first_block;
		block_t* m_current_block;
		unsigned int m_num_nodes_in_current_block;
		unsigned int m_num_nodes;

		avlnode_buffer_t()
			: m_first_block(), m_current_block(&m_first_block), m_num_nodes_in_current_block(0), m_num_nodes(0)
		{
		}
		
		void free_blocks();

		~avlnode_buffer_t();

		avlnode_t* get_new_node(void);

		void recalc_pointer(const char* block_begin);
	};

	avlnode_t* m_avltree_root;
	avlnode_buffer_t m_avlnode_buffer;

public:
	struct block_t {
		block_t* m_next;
		DWORD m_bytesInBlock;
		BYTE m_data[1024 * 32];

		block_t()
			: m_next(NULL), m_bytesInBlock(0)
		{
		}

		~block_t()
		{
		}
		static void* operator new(size_t size, void* ptr)
		{
			return ptr;
		}
		static void operator delete(void* deadObject, size_t size)
		{
			return CoTaskMemFree(deadObject);
		}
	};

	block_t m_first_block;
	block_t* m_current_block;

	void free_blocks();

	DWORD m_dataSize;

	HRESULT append(const tDcmElementHdr& hdr, LPCVOID cp, DWORD& offset, LPCVOID& ptrInCache);

	HRESULT CheckinElement(const tDcmElementHdr& hdr, LPCVOID cp, DWORD& offset);

	HRESULT Dump(const wchar_t* path);
	HRESULT GetDOM(IXMLDOMDocument* dom);

	CDataElementCache()
		: m_first_block(), m_current_block(&m_first_block),
		  m_avltree_root(NULL),
		  m_dataSize(0)
	{}

	~CDataElementCache()
	{
		free_blocks();
	}

	static const char* sig;
	static const wchar_t* filename;
	struct header_t {
		char m_sig[4];
		DWORD m_foo;
		DWORD m_headerSize;
		DWORD m_dataSize;
	};
};

class CDataElementCacheFile : public CDataElementCache {
	HANDLE m_hFile;
	HANDLE m_hMap;
	LPVOID m_lpMap;

	HRESULT close(HRESULT hr = S_OK);

public:
	CDataElementCacheFile()
		: m_hFile(INVALID_HANDLE_VALUE), m_hMap(NULL), m_lpMap(NULL)
	{
	}

	~CDataElementCacheFile()
	{
		Close();
	}
	
	HRESULT Open(const wchar_t* path);
	HRESULT Commit(bool bClose = false);
	void Close(void);

	static HRESULT LoadDOM(const wchar_t* path, IXMLDOMDocument* dom);
};

#pragma warn_unusedarg reset

#endif
