/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "dcmdict.h"
#include "CDcmScanner.h"
#include "UtilClasses.h"

#include "DataElementCache.h"

#include "InstanceCache.h"

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
struct instance_entry_t {
	DWORD m_size;
	unsigned short m_header_size; // not a character count, size in bytes
	wchar_t m_path[1]; // null terminated.
	DWORD m_offset[1];
};

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
template < class BLOCK_T >
static void free_blocks_(BLOCK_T& first_block)
{
	BLOCK_T* mb = first_block.m_next;
	while (mb != NULL) {
		BLOCK_T* next = mb->m_next;
		delete mb;
		mb = next;
	}
	first_block.m_next = NULL;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static void hexwstr(unsigned int val, wchar_t* str, unsigned int len)
{
	wchar_t* cp = str + len - 1;
	do {
		unsigned int ch = val & 0xF;
		if (ch < 10)
			ch += L'0';
		else
			ch += L'A' - 10;
		*cp-- = static_cast<wchar_t>(ch);
		val = val >> 4;
	} while (cp >= str);

	*(str + len) = L'\0';
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static HRESULT append_data(DWORD ltag, DWORD id, IXMLDOMNode* fileElement, IXMLDOMElement* dataTemplate, IXMLDOMAttribute* tagAttr, IXMLDOMAttribute* refAttr, VARIANT& var, BSTR& valueStr)
{
	HRESULT hr;

	wchar_t tagStr[12];
	hexwstr(ltag, tagStr, 8);
	if (!SysReAllocString(&valueStr, tagStr))
		return E_OUTOFMEMORY;
	var.vt = VT_BSTR;
	var.bstrVal = valueStr;
	hr = tagAttr->put_value(var);
	if (FAILED(hr))
		return hr;

	var.vt = VT_UI4;
	var.ulVal = id;
	hr = refAttr->put_value(var);
	if (FAILED(hr))
		return hr;

	AutoComPtr<IXMLDOMNode> newData;
	hr = dataTemplate->cloneNode(VARIANT_TRUE, newData.r());
	if (FAILED(hr))
		return hr;
	
	hr = fileElement->appendChild(newData.p(), NULL);
	if (FAILED(hr))
		return hr;

	return S_OK;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
template < class BLOCK_T >
static HRESULT block_list_to_dom(BLOCK_T* bp, IXMLDOMDocument* dom)
{
	HRESULT hr;

	BSTR str1 = SysAllocStringLen(NULL, 1);
	AutoSysFreeString af_str1(str1);

	BSTR str2 = SysAllocStringLen(NULL, 2);
	AutoSysFreeString af_str2(str2);

	BSTR valueStr = SysAllocStringLen(NULL, MAX_PATH);
	AutoSysFreeString af_valueStr(valueStr);

	VARIANT var;
	VariantInit(&var);

	*str1 = L'F';
	AutoComPtr<IXMLDOMElement> fileTemplate;
	hr = dom->createElement(str1, fileTemplate.r());
	if (FAILED(hr))
		return hr;

	*str1 = L'l';
	AutoComPtr<IXMLDOMAttribute> locAttr;
	hr = dom->createAttribute(str1, locAttr.r());
	if (FAILED(hr))
		return hr;

	hr = fileTemplate->setAttributeNode(locAttr.p(), NULL);
	if (FAILED(hr))
		return hr;

	*str1 = L'd';
	AutoComPtr<IXMLDOMElement> dataTemplate;
	hr = dom->createElement(str1, dataTemplate.r());
	if (FAILED(hr))
		return hr;

	*str1 = L't';
	AutoComPtr<IXMLDOMAttribute> tagAttr;
	hr = dom->createAttribute(str1, tagAttr.r());
	if (FAILED(hr))
		return hr;

	hr = dataTemplate->setAttributeNode(tagAttr.p(), NULL);
	if (FAILED(hr))
		return hr;

	*str1 = L'r';
	AutoComPtr<IXMLDOMAttribute> refAttr;
	hr = dom->createAttribute(str1, refAttr.r());
	if (FAILED(hr))
		return hr;

	hr = dataTemplate->setAttributeNode(refAttr.p(), NULL);
	if (FAILED(hr))
		return hr;

	lstrcpy(str2, L"FL");
	AutoComPtr<IXMLDOMElement> listElm;
	hr = dom->createElement(str2, listElm.r());
	if (FAILED(hr))
		return hr;

	const char* block_begin = reinterpret_cast<const char*>(bp->m_data);
	const char* block_end = block_begin + bp->m_bytesInBlock;
	const char* cp = block_begin;

	while (true) {
		if (cp == block_end) {
			bp = bp->m_next;
			if (bp == NULL)
				break;
			block_begin = reinterpret_cast<const char*>(bp->m_data);
			block_end = block_begin + bp->m_bytesInBlock;
			cp = block_begin;
		}
			
		const instance_entry_t* entry = reinterpret_cast<const instance_entry_t*>(cp);
		DWORD count = entry->m_size - entry->m_header_size;
		cp += entry->m_header_size;
		
		if (!SysReAllocString(&valueStr, entry->m_path))
			return E_OUTOFMEMORY;
		var.vt = VT_BSTR;
		var.bstrVal = valueStr;
		hr = locAttr->put_value(var);
		if (FAILED(hr))
			return hr;

		AutoComPtr<IXMLDOMNode> newElement;
		hr = fileTemplate->cloneNode(VARIANT_TRUE, newElement.r());
		if (FAILED(hr))
			return hr;

		while (count != 0) {
			if (cp == block_end) {
				bp = bp->m_next;
				if (bp == NULL)
					return E_UNEXPECTED;
				block_begin = reinterpret_cast<const char*>(bp->m_data);
				block_end = block_begin + bp->m_bytesInBlock;
				cp = block_begin;
			}
			struct data_t {
				DWORD m_ltag;
				DWORD m_offset;
			};
			const data_t* data = reinterpret_cast<const data_t*>(cp);
			cp += sizeof(data_t);
			count -= sizeof(data_t);

			hr = ::append_data(data->m_ltag, data->m_offset, newElement.p(), dataTemplate.p(), tagAttr.p(), refAttr.p(), var, valueStr);
			if (FAILED(hr))
				return hr;
		}
	
		hr = listElm->appendChild(newElement.p(), NULL);
		if (FAILED(hr))
			return hr;
	}

	AutoComPtr<IXMLDOMElement> documentElement;
	hr = dom->get_documentElement(documentElement.r());
	if (FAILED(hr))
		return hr;

	hr = documentElement->appendChild(listElm.p(), NULL);
	if (FAILED(hr))
		return hr;
	
	return S_OK;
}

//======================================================================
//
//
//
//======================================================================
const char* CInstanceCache::sig = "IFCh";
const wchar_t* CInstanceCache::filename = L".InstDir";

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
void CInstanceCache::free_blocks()
{
	free_blocks_<block_t>(m_first_block);
	m_first_block.m_next = NULL;
	m_first_block.m_bytesInBlock = 0;
	m_current_block = &m_first_block;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
HRESULT CInstanceCache::append(LPCVOID ptr, DWORD size, LPVOID* dataInCache)
{
	const unsigned long block_data_size = sizeof(m_current_block->m_data);

	if (size > block_data_size)
		return E_UNEXPECTED;

	const DWORD bufRemain = block_data_size - m_current_block->m_bytesInBlock;
	if (bufRemain < size) {
		void* ptr = CoTaskMemAlloc(sizeof(block_t));
		if (ptr == NULL)
			return E_OUTOFMEMORY;
		m_current_block->m_next = new (ptr) block_t;
		m_current_block = m_current_block->m_next;
	}

	BYTE* dp = m_current_block->m_data + m_current_block->m_bytesInBlock;
	CopyMemory(dp, ptr, size);

	if (dataInCache != NULL)
		*dataInCache = dp;

	m_current_block->m_bytesInBlock += size;
	m_dataSize += size;

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT CInstanceCache::AppendInstance(const wchar_t* location, const char* scannedElements, size_t scannedElementsSize, const bool explicitVR, CDataElementCache& dataElementCache)
{
	if (check_existance(location))
		return S_FALSE;

	DWORD entry_size = sizeof(DWORD) + sizeof(unsigned short) + sizeof(wchar_t) * (lstrlen(location) + 1);
	unsigned int pad = entry_size & 3;
	if (pad != 0) {
		pad = 4 - pad;
		entry_size += pad;
	}

	instance_entry_t* entry = (instance_entry_t*)_alloca(entry_size);
	entry->m_size = 0;
	entry->m_header_size = entry_size;
	lstrcpy(entry->m_path, location);

	instance_entry_t* entryInBlock;
	HRESULT hr = append(entry, entry_size, (LPVOID*)&entryInBlock);
	if (FAILED(hr))
		return hr;

	const char* const block_begin = scannedElements;
	const char* const block_end = block_begin + scannedElementsSize;

	CDcmScannerF scanner(block_begin, block_end);
	tDcmElementHdr hdr;

	const char* cp = block_begin;

	CDcmDict dict;

	while (true) {
		int sr = scanner.NextElement(cp, hdr, explicitVR);
		if (sr != CDcmScannerF::S_CONTINUE) {
			break;
		}

		const unsigned int gtag = hdr.gtag();
		if (gtag == DCM_GROUPPIXEL) {
			break;
		} else if (gtag & 1) {
			cp = hdr.next_ptr(cp);
			continue;
		}

		if (!explicitVR) {
			const DCMDICT* const d = dict.Lookup(hdr.ltag());
			if (d == NULL) {
				cp = hdr.next_ptr(cp);
				continue;
			}
			hdr.m_vr = DcmGetVRStr(d->representation);
		}

		if (hdr.isBLE() || hdr.isSQ()) {
			cp = hdr.next_ptr(cp);
			continue;
		}

		struct data_t {
			DWORD m_ltag;
			DWORD m_offset;
		} data;

		data.m_ltag = hdr.ltag();

		HRESULT hr = dataElementCache.CheckinElement(hdr, static_cast<LPCVOID>(cp), data.m_offset);
		if (FAILED(hr))
			return hr;

		append(&data, sizeof(data));
		entry_size += sizeof(data);

		cp = hdr.next_ptr(cp);
	}

	entryInBlock->m_size = entry_size;

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT CInstanceCache::AppendInstance(const wchar_t* filePath, const wchar_t* location, CDataElementCache& dataElementCache)
{
	CStackFileMapping map(filePath);
	if (map.lastError() != NO_ERROR)
		return HRESULT_FROM_WIN32(map.lastError());

	const char* const block_begin = static_cast<const char*>(map.mapPointer());
	const char* const block_end = block_begin + map.mapSize();

	if (memcmp(block_begin + DCM_PREAMBLELENGTH, "DICM", 4) != 0)
		return E_UNEXPECTED;

	CDcmScannerF scanner(block_begin, block_end);
	tDcmElementHdr hdr;

	const char* cp = block_begin + DCM_PREAMBLELENGTH + 4;
	bool explicitVR;

	while (true) {
		int sr = scanner.NextElement(cp, hdr, true);
		if (sr != CDcmScannerF::S_CONTINUE) {
			return E_UNEXPECTED;
		}
		if (hdr.m_tag[0] != DCM_GROUPFILEMETA)
			break;
		if (hdr.m_tag[1] == DCM_TAG_ELEMENT(DCM_METATRANSFERSYNTAX)) {
			char transfer_syntax_uid[DICOM_UI_LENGTH + 1];
			memcpy(transfer_syntax_uid, cp + hdr.m_hsize, hdr.m_length);
			transfer_syntax_uid[hdr.m_length] = '\0';

			if (strcmp(transfer_syntax_uid, DICOM_TRANSFERLITTLEENDIAN) == 0) {
				explicitVR = false;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERLITTLEENDIANEXPLICIT) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGBASELINEPROCESS1) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGEXTENDEDPROC2AND4) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGLOSSLESSPROC14) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, DICOM_TRANSFERJPEGLOSSLESSPROCFIRSTORDERREDICT) == 0) {
				explicitVR = true;
			} else if (strcmp(transfer_syntax_uid, "PNG") == 0) {
				explicitVR = true;
			} else {
				return E_UNEXPECTED;
			}
		}
		cp = hdr.next_ptr(cp);
	}

	return AppendInstance(location, cp, block_end - cp, explicitVR, dataElementCache);
}

//======================================================================
//
//======================================================================
HRESULT CInstanceCache::Dump(const wchar_t* filePath)
{
	HANDLE hFile = CreateFile(filePath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoCloseFileHandle ac_hFile(hFile);

	const DWORD dwFileSize = GetFileSize(hFile, NULL);
	if (dwFileSize == 0) {
		header_t h;
		ZeroMemory(&h, sizeof(h));
		CopyMemory(h.m_sig, sig, 4);
		h.m_headerSize = sizeof(header_t);
		h.m_dataSize = 0;
		DWORD bytesWritten;
		if (!WriteFile(hFile, &h, h.m_headerSize, &bytesWritten, NULL))
			return HRESULT_FROM_WIN32(GetLastError());
		if (SetFilePointer(hFile, 0, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
			return HRESULT_FROM_WIN32(GetLastError());
	} else if (dwFileSize < sizeof(header_t)) {
		return E_UNEXPECTED;
	}

	OVERLAPPED ov = { 0, 0, 0, 0, 0 };
	if (!LockFileEx(hFile, LOCKFILE_EXCLUSIVE_LOCK, 0, sizeof(header_t), 0, &ov))
		return HRESULT_FROM_WIN32(GetLastError());
	AutoUnlockFileEx au(hFile, sizeof(header_t), 0, &ov);

	header_t header;

	DWORD bytesRead;
	if (!ReadFile(hFile, &header, sizeof(header), &bytesRead, NULL))
		return HRESULT_FROM_WIN32(GetLastError());

	if (SetFilePointer(hFile, 0, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
		return HRESULT_FROM_WIN32(GetLastError());

	header.m_dataSize = 0;

	DWORD bytesWritten;
	if (!WriteFile(hFile, &header, sizeof(header), &bytesWritten, NULL))
		return HRESULT_FROM_WIN32(GetLastError());

	block_t* bp = &m_first_block;
	do {
		DWORD bytesWritten;
		if (!WriteFile(hFile, bp->m_data, bp->m_bytesInBlock, &bytesWritten, NULL))
			return HRESULT_FROM_WIN32(GetLastError());

		header.m_dataSize += bytesWritten;
		bp = bp->m_next;
	} while (bp != NULL);

	if (!SetEndOfFile(hFile))
		return HRESULT_FROM_WIN32(GetLastError());

	if (SetFilePointer(hFile, 0, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
		return HRESULT_FROM_WIN32(GetLastError());
	
	if (!WriteFile(hFile, &header, sizeof(header), &bytesWritten, NULL))
		return HRESULT_FROM_WIN32(GetLastError());
	
	return S_OK;
}
//======================================================================
//
//======================================================================
HRESULT CInstanceCache::GetDOM(IXMLDOMDocument* dom)
{
	return block_list_to_dom<block_t>(&m_first_block, dom);
}

//======================================================================
//
//
//
//======================================================================
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
HRESULT CInstanceCacheFile::close(HRESULT hr)
{
	if (m_lpMap != NULL) {
		UnmapViewOfFile(m_lpMap);
		m_lpMap = NULL;
	}
	if (m_hMap != NULL) {
		CloseHandle(m_hMap);
		m_hMap = NULL;
	}
	if (m_hFile != INVALID_HANDLE_VALUE) {
		CloseHandle(m_hFile);
		m_hFile = INVALID_HANDLE_VALUE;
	}

	return hr;
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
bool CInstanceCacheFile::check_existance(const wchar_t* location)
{
	header_t* header = static_cast<header_t*>(m_lpMap);
	const char* const block_begin = static_cast<char*>(m_lpMap) + header->m_headerSize;
	const char* const block_end = block_begin + header->m_dataSize;

	const char* cp = block_begin;
	while(cp < block_end) {
		const instance_entry_t* entry = reinterpret_cast<const instance_entry_t*>(cp);
		if (lstrcmp(entry->m_path, location) == 0)
			return true;

		cp += entry->m_size;
	}

	return false;
}

//======================================================================
//
//======================================================================
HRESULT CInstanceCacheFile::Open(const wchar_t* filePath)
{
	m_dataSize = 0;

	m_hFile = CreateFile(filePath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (m_hFile == INVALID_HANDLE_VALUE)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	const DWORD dwFileSize = GetFileSize(m_hFile, NULL);
	if (dwFileSize == 0) {
		header_t h;
		CopyMemory(h.m_sig, sig, 4);
		h.m_headerSize = sizeof(header_t);
		h.m_dataSize = 0;
		DWORD bytesWritten;
		if (!WriteFile(m_hFile, &h, h.m_headerSize, &bytesWritten, NULL))
			return close(HRESULT_FROM_WIN32(GetLastError()));
	} else if (dwFileSize < sizeof(header_t)) {
		return close(E_UNEXPECTED);
	}

	m_hMap = CreateFileMapping(m_hFile, NULL, PAGE_READWRITE, 0, 0, NULL);
	if (m_hMap == NULL)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	m_lpMap = MapViewOfFile(m_hMap, FILE_MAP_ALL_ACCESS, 0, 0, 0);
	if (m_lpMap == NULL)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	if (dwFileSize == 0)
		return S_OK;

	const header_t* header = static_cast<header_t*>(m_lpMap);
	if (memcmp(header->m_sig, sig, sizeof(header->m_sig)) != 0)
		return close(E_UNEXPECTED);
	if (dwFileSize < header->m_headerSize + header->m_dataSize)
		return close(E_UNEXPECTED);

	m_dataSize = header->m_dataSize;

	return S_OK;
}


//======================================================================
//
//======================================================================
HRESULT CInstanceCacheFile::Commit(bool bClose)
{
	if (m_first_block.m_bytesInBlock == 0)
		return S_FALSE;

	OVERLAPPED ov = { 0, 0, 0, 0, 0 };
	if (!LockFileEx(m_hFile, LOCKFILE_EXCLUSIVE_LOCK, 0, sizeof(header_t), 0, &ov))
		return close(HRESULT_FROM_WIN32(GetLastError()));
	AutoUnlockFileEx au(m_hFile, sizeof(header_t), 0, &ov);

	header_t* header = static_cast<header_t*>(m_lpMap);
	const long pos = header->m_headerSize + header->m_dataSize;

	UnmapViewOfFile(m_lpMap); m_lpMap = NULL;
	CloseHandle(m_hMap); m_hMap = NULL;

	if (SetFilePointer(m_hFile, pos, NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	block_t* bp = &m_first_block;
	do {
		DWORD bytesWritten;
		if (!WriteFile(m_hFile, bp->m_data, bp->m_bytesInBlock, &bytesWritten, NULL)) {
			HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			Close();
			return hr;
		}
		bp = bp->m_next;
	} while (bp != NULL);

	free_blocks();

	if (!SetEndOfFile(m_hFile)) {
		HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
		Close();
		return hr;
	}

	if (bClose) {
		HRESULT hr = S_OK;
		if (SetFilePointer(m_hFile, offsetof(header_t, m_dataSize), NULL, FILE_BEGIN) == INVALID_SET_FILE_POINTER) {
			hr = HRESULT_FROM_WIN32(GetLastError());
		} else {
			DWORD bytesWritten;
			if (!WriteFile(m_hFile, &m_dataSize, sizeof(m_dataSize), &bytesWritten, NULL)) {
				hr = HRESULT_FROM_WIN32(GetLastError());
			}
		}
		Close();
		return hr;
	}

	m_hMap = CreateFileMapping(m_hFile, NULL, PAGE_READWRITE, 0, 0, NULL);
	if (m_hMap == NULL)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	m_lpMap = MapViewOfFile(m_hMap, FILE_MAP_ALL_ACCESS, 0, 0, 0);
	if (m_lpMap == NULL)
		return close(HRESULT_FROM_WIN32(GetLastError()));

	header = static_cast<header_t*>(m_lpMap);
	header->m_dataSize = m_dataSize;
	
	return S_OK;
}

//======================================================================
//
//======================================================================
void CInstanceCacheFile::Close()
{
	free_blocks();
	close(S_OK);
}


//======================================================================
//
//======================================================================
HRESULT CInstanceCacheFile::LoadDOM(const wchar_t* path, IXMLDOMDocument* dom)
{
	HANDLE hFile = CreateFile(path, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoCloseFileHandle ac_hFile(hFile);

	if (GetFileSize(hFile, NULL) < sizeof(header_t))
		return E_UNEXPECTED;

	OVERLAPPED ov = { 0, 0, 0, 0, 0 };
	if (!LockFileEx(hFile, 0, 0, sizeof(header_t), 0, &ov))
		return HRESULT_FROM_WIN32(GetLastError());
	AutoUnlockFileEx au(hFile, sizeof(header_t), 0, &ov);

	HANDLE hMap = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (hMap == NULL)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoCloseHandle ac_hMap(hMap);

	LPVOID lpMap = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
	if (lpMap == NULL)
		return HRESULT_FROM_WIN32(GetLastError());
	AutoUnmapViewOfFile au_lpMap(lpMap);

	const header_t* const header = static_cast<header_t*>(lpMap);
	if (memcmp(header->m_sig, sig, sizeof(header->m_sig)) != 0)
		return E_UNEXPECTED;
	if (GetFileSize(hFile, NULL) < header->m_headerSize + header->m_dataSize)
		return E_UNEXPECTED;

	struct block_t {
		block_t* m_next;
		DWORD m_bytesInBlock;
		BYTE* m_data;
	} block = { NULL, header->m_dataSize, (BYTE*)lpMap + header->m_headerSize };

	return block_list_to_dom<block_t>(&block, dom);
}

