#ifndef _4076C170_0BA8_436A_8E9D_F6DABEF1AF1F_
#define _4076C170_0BA8_436A_8E9D_F6DABEF1AF1F_

/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#pragma warn_unusedarg off

class CInstanceCache {
public:
	struct block_t {
		block_t* m_next;
		DWORD m_bytesInBlock;
		BYTE m_data[1024 * 32];

		block_t()
			: m_next(NULL), m_bytesInBlock(0)
		{
		}

		~block_t()
		{
		}
		static void* operator new(size_t size, void* ptr)
		{
			return ptr;
		}
		static void operator delete(void* deadObject, size_t size)
		{
			return CoTaskMemFree(deadObject);
		}
	};

	block_t m_first_block;
	block_t* m_current_block;

	void free_blocks(void);

	DWORD m_dataSize;

	virtual bool check_existance(const wchar_t* location)
	{
		return false;
	}

	HRESULT append(LPCVOID ptr, DWORD size, LPVOID* dataInCache = NULL);

	HRESULT AppendInstance(const wchar_t* filePath, const wchar_t* location, CDataElementCache& dataElementCache);
	HRESULT AppendInstance(const wchar_t* location, const char* scannedElements, size_t scannedElementsSize, bool explicitVR, CDataElementCache& dataElementCache);

	HRESULT Dump(const wchar_t* path);
	HRESULT GetDOM(IXMLDOMDocument* dom);

	CInstanceCache()
		: m_first_block(), m_current_block(&m_first_block),
		  m_dataSize(0)
		
	{}

	~CInstanceCache()
	{
		free_blocks();
	}

	static const char* sig;
	static const wchar_t* filename;
	struct header_t {
		char m_sig[4];
		DWORD m_foo;
		DWORD m_headerSize;
		DWORD m_dataSize;
	};
};

class CInstanceCacheFile : public CInstanceCache {
	HANDLE m_hFile;
	HANDLE m_hMap;
	LPVOID m_lpMap;
	HRESULT close(HRESULT hr = S_OK);

	virtual bool check_existance(const wchar_t* location);

public:

	HRESULT Open(const wchar_t* path);
	HRESULT Commit(bool bClose = false);
	void Close();


	CInstanceCacheFile()
		: m_hFile(INVALID_HANDLE_VALUE), m_hMap(NULL), m_lpMap(NULL)
	{
	}

	~CInstanceCacheFile()
	{
		close();
	}

	static HRESULT LoadDOM(const wchar_t* path, IXMLDOMDocument* dom);
};

#pragma warn_unusedarg reset

#endif
