/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"

//======================================================================
//
// Destructors
//
//======================================================================
//======================================================================
//
//======================================================================
AutoSysFreeString::~AutoSysFreeString()
 {
	if (bstr != NULL) {
		SysFreeString(bstr);
		bstr = NULL;
	}
}
//======================================================================
//
//======================================================================
AutoVariantClear::~AutoVariantClear()
{
	VariantClear(&m_val);
}
//======================================================================
//
//======================================================================
AutoCoUninitialize::~AutoCoUninitialize()
{
	CoUninitialize();
}
//======================================================================
//
//======================================================================
AutoWSACleanup::~AutoWSACleanup()
{
	WSACleanup();
}
//======================================================================
//
//======================================================================
AutoCloseFileHandle::~AutoCloseFileHandle()
{
	if (m_hndl != INVALID_HANDLE_VALUE) {
		CloseHandle(m_hndl);
		m_hndl = INVALID_HANDLE_VALUE;
	}
}
//======================================================================
//
//======================================================================
AutoCloseHandle::~AutoCloseHandle()
{
	if (m_hndl != NULL) {
		CloseHandle(m_hndl);
		m_hndl = NULL;
	}
}

//======================================================================
//
//======================================================================
AutoUnmapViewOfFile::~AutoUnmapViewOfFile()
{
	if (lpMap != NULL) {
		UnmapViewOfFile(lpMap);
		lpMap = NULL;
	}
}

//======================================================================
//
//======================================================================
AutoCloseChangeNotificationHandle::~AutoCloseChangeNotificationHandle()
{
	if (m_handle == INVALID_HANDLE_VALUE)
		return;

	FindCloseChangeNotification(m_handle);
}

//======================================================================
//
// Critical Section
//
//======================================================================
//======================================================================
//
//======================================================================
CStackCriticalSection::CStackCriticalSection()
	: crit(NULL)
{
}

//======================================================================
//
//======================================================================
CStackCriticalSection::CStackCriticalSection(CRITICAL_SECTION* cs)
	: crit(cs)
{
	if (crit != NULL)
		EnterCriticalSection(crit);
}
//======================================================================
//
//======================================================================
CStackCriticalSection::~CStackCriticalSection()
{
	if (crit != NULL)
		LeaveCriticalSection(crit);
}
//======================================================================
//
// Find File
//
//======================================================================
//======================================================================
//
//======================================================================
CStackFindFile::CStackFindFile()
{
	handle = FindFirstFile(L"*", &data);
}
//======================================================================
//
//======================================================================
CStackFindFile::CStackFindFile(const char* dir, const char* wild)
{
	int dir_wlen = multi_byte_to_wide_char(dir, -1, NULL, 0) - 1;
	int wild_wlen;
	if (wild != NULL) {
		wild_wlen = multi_byte_to_wide_char(wild, -1, NULL, 0) - 1;
	} else {
		wild = "*";
		wild_wlen = 1;
	}

	wchar_t* const pattern = (wchar_t*)_alloca(sizeof(wchar_t) * (dir_wlen + 1 + wild_wlen + 1));
	multi_byte_to_wide_char(dir, -1, pattern, dir_wlen);
	pattern[dir_wlen] = L'\\';
	multi_byte_to_wide_char(wild, -1, pattern + dir_wlen + 1, wild_wlen + 1);

	handle = FindFirstFile(pattern, &data);
}

//======================================================================
//
//======================================================================
CStackFindFile::CStackFindFile(const wchar_t* dir, const wchar_t* wild)
{
	int dir_wlen = lstrlen(dir);
	int wild_wlen;
	if (wild != NULL) {
		wild_wlen = lstrlen(wild);
	} else {
		wild = L"*";
		wild_wlen = 1;
	}

	wchar_t* const pattern = (wchar_t*)_alloca(sizeof(wchar_t) * (dir_wlen + 1 + wild_wlen + 1));
	strncat3(pattern, dir, dir_wlen, L'\\', wild, wild_wlen);

	handle = FindFirstFile(pattern, &data);
}


//======================================================================
//
//======================================================================
CStackFindFile::~CStackFindFile()
{
	if (handle != INVALID_HANDLE_VALUE)
		FindClose(handle);
}
//======================================================================
//
//======================================================================
bool CStackFindFile::next(void) {
	return FindNextFile(handle, &data);
}
