/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"

//======================================================================
//
// class CConsoleLog
//
//======================================================================
//======================================================================
//
//======================================================================
HRESULT CConsoleLog::Init(bool consoleMode, const wchar_t* logFilePathBase, SECURITY_ATTRIBUTES* sa)
{
	m_fileMode = !consoleMode;

	if (m_fileMode && logFilePathBase != NULL) {
		m_logFilePath = (wchar_t*)CoTaskMemAlloc(sizeof(wchar_t) * (lstrlen(logFilePathBase) + 5));
		lstrcpy(m_logFilePath, logFilePathBase);
		lstrcat(m_logFilePath, L".log");

		m_hFile = CreateFile(m_logFilePath, GENERIC_WRITE | GENERIC_READ, FILE_SHARE_READ, sa, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (m_hFile == INVALID_HANDLE_VALUE) {
			HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			SetStdHandle(STD_OUTPUT_HANDLE, m_hFile);
			return hr;
		}
	} else {
		m_hFile = GetStdHandle(STD_OUTPUT_HANDLE);
	}

	ini();

	return S_OK;
}
//======================================================================
//
//======================================================================
HRESULT CConsoleLog::Init(HANDLE hFile)
{
	m_fileMode = false;
	m_hFile = hFile;

	ini();

	return S_OK;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
void CConsoleLog::ini(void)
{
	InitializeCriticalSection(&m_crit);

	SYSTEMTIME time;
	GetLocalTime(&time);

	char buf[84 + 1];
	sprintf(buf, "<?xml version=\"1.0\" encoding=\"Shift_JIS\"?>\r\n<log start-time=\"%04u-%02u-%02dT%02d:%02d:%02d\">\r\n", 
		time.wYear, time.wMonth, time.wDay, time.wHour, time.wMinute, time.wSecond);

	DWORD bytesWritten;
	WriteFile(m_hFile, buf, 84, &bytesWritten, NULL);

	if (m_fileMode)
		WriteFile(m_hFile, "</log>\r\n", 8, &bytesWritten, NULL);
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
void CConsoleLog::fin(void)
{
	DeleteCriticalSection(&m_crit);

	if (m_hFile != INVALID_HANDLE_VALUE) {

		if (m_fileMode) {

			CloseHandle(m_hFile);

			SYSTEMTIME time;
			GetLocalTime(&time);
			const unsigned int timestrlen = 1 + 4 + 3*5 + 4 + 1;
			wchar_t timestr[timestrlen];
			wsprintf(timestr, L"-%04d.%02d.%02d.%02d.%02d.%02d.log", 
				time.wYear, time.wMonth, time.wDay, time.wHour, time.wMinute, time.wSecond);

			const int len = lstrlen(m_logFilePath) - 4;
			wchar_t* const backlogfile = (wchar_t*)_alloca(sizeof(wchar_t) * (len + timestrlen));
			lstrcpy(backlogfile, m_logFilePath);
			lstrcpy(backlogfile + len, timestr);

			MoveFile(m_logFilePath, backlogfile);

		} else {

			DWORD bytesWritten;
			WriteFile(m_hFile, "</log>\r\n", 8, &bytesWritten, NULL);

		}
	}

	if (m_logFilePath != NULL)
		CoTaskMemFree(m_logFilePath);
}

//======================================================================
//
//======================================================================
void CConsoleLog::writeErrorInfo(const SimpleErrorInfo& ei)
{
	CStackCriticalSection cs(&m_crit);

	if (m_fileMode)
		SetFilePointer(m_hFile, -8, NULL, FILE_END);

	DWORD bytesWritten;

	char buf[SimpleErrorInfo::kMessageBufferLen + 1];

	SYSTEMTIME time;
	GetLocalTime(&time);

	sprintf(buf, "<Error time=\"%04d/%02d/%02d-%02d:%02d:%02d\">\r\n",
		time.wYear, time.wMonth, time.wDay, time.wHour, time.wMinute, time.wSecond);
	WriteFile(m_hFile, buf, 36, &bytesWritten, NULL);

	const wchar_t* cp = ei.m_message;

	if (FAILED(ei.m_hr)) {
		sprintf(buf, "<SystemError HRESULT=\"%08X\">", ei.m_hr);
		WriteFile(m_hFile, buf, 32, &bytesWritten, NULL);
		
		const int len = wide_char_to_multi_byte(cp, -1, buf, sizeof(buf));
		WriteFile(m_hFile, buf, len, &bytesWritten, NULL);

		WriteFile(m_hFile, "</SystemError>\r\n", 16, &bytesWritten, NULL);

		cp += lstrlen(cp) + 1;
	}
	while (cp < ei.m_tail) {
		WriteFile(m_hFile, "<ApplicationError>", 18, &bytesWritten, NULL);

		const int len = wide_char_to_multi_byte(cp, -1, buf, sizeof(buf));
		WriteFile(m_hFile, buf, len, &bytesWritten, NULL);

		WriteFile(m_hFile, "</ApplicationError>\r\n", 21, &bytesWritten, NULL);

		cp += lstrlen(cp) + 1;
	}

	WriteFile(m_hFile, "</Error>\r\n</log>\r\n", (m_fileMode ? 10 + 8 : 10), &bytesWritten, NULL);
}

//======================================================================
//
//======================================================================
void CConsoleLog::writeText(const char* text)
{
	CStackCriticalSection cs(&m_crit);

	if (m_fileMode)
		SetFilePointer(m_hFile, -8, NULL, FILE_END);

	DWORD bytesWritten;
	WriteFile(m_hFile, text, strlen(text), &bytesWritten, NULL);

	if (m_fileMode)
		WriteFile(m_hFile, "</log>\r\n", 8, &bytesWritten, NULL); 
}
