/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"

//======================================================================
//
//
//
//======================================================================
//======================================================================
//
//======================================================================
SimpleErrorInfo::SimpleErrorInfo(const SimpleErrorInfo& rhs)
	: m_hr(rhs.m_hr)
{
	const int len = rhs.m_tail - rhs.m_message;
	if (len > 0)
		CopyMemory(m_message, rhs.m_message, sizeof(wchar_t) * len);
	m_tail = m_message + len;
}
//======================================================================
//
//======================================================================
SimpleErrorInfo& SimpleErrorInfo::operator=(const SimpleErrorInfo& rhs)
{
	m_hr = rhs.m_hr;
	const int len = rhs.m_tail - rhs.m_message;
	if (len > 0)
		CopyMemory(m_message, rhs.m_message, sizeof(wchar_t) * len);
	m_tail = m_message + len;

	return *this;
}

//======================================================================
//
//======================================================================
SimpleErrorInfo::SimpleErrorInfo(HRESULT hr)
	: m_hr(hr)
{
	const unsigned int len = FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM, NULL, m_hr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), m_message, kMessageBufferLen, NULL);
	m_tail = m_message + len + 1;
}
//======================================================================
void SimpleErrorInfo::set(HRESULT hr)
{
	m_hr = hr;
	const unsigned int len = FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM, NULL, m_hr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), m_message, kMessageBufferLen, NULL);
	m_tail = m_message + len + 1;
}

//======================================================================
//
//======================================================================
SimpleErrorInfo::SimpleErrorInfo(HRESULT hr, const wchar_t* formatStr, ...)
	: m_hr(hr)
{
	const unsigned int len = FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM, NULL, m_hr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), m_message, kMessageBufferLen, NULL);
	m_tail = m_message + len + 1;

	va_list args;
	va_start(args, formatStr);

	append_message(formatStr, &args);
}
//======================================================================
void SimpleErrorInfo::set(HRESULT hr, const wchar_t* formatStr, ...)
{
	m_hr = hr;
	const unsigned int len = FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM, NULL, m_hr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), m_message, kMessageBufferLen, NULL);
	m_tail = m_message + len + 1;

	va_list args;
	va_start(args, formatStr);

	append_message(formatStr, &args);
}

//======================================================================
//
//======================================================================
SimpleErrorInfo::SimpleErrorInfo(const wchar_t* formatStr, ...)
	: m_hr(S_OK), m_tail(m_message)
{
	va_list args;
	va_start(args, formatStr);

	append_message(formatStr, &args);
}
//======================================================================
void SimpleErrorInfo::set(const wchar_t* formatStr, ...)
{
	m_hr = S_OK;
	m_tail = m_message;

	va_list args;
	va_start(args, formatStr);

	append_message(formatStr, &args);
}

//======================================================================
//
//======================================================================
void SimpleErrorInfo::add(const wchar_t* formatStr, ...)
{
	va_list args;
	va_start(args, formatStr);

	append_message(formatStr, &args);
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
void SimpleErrorInfo::append_message(const wchar_t* formatStr, va_list* arg)
{
	const unsigned int len = FormatMessage(FORMAT_MESSAGE_FROM_STRING, formatStr, 0, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), m_tail, (m_message + kMessageBufferLen) - m_tail, arg);
	m_tail += len + 1;
}
//======================================================================
//
//======================================================================
void SimpleErrorInfo::reset()
{
	m_hr = S_OK;
	m_tail = m_message;
}

//======================================================================
//
//======================================================================
void SimpleErrorInfo::set_from_CTN_COND()
{
	m_hr = S_FALSE;
	m_tail = m_message;

	CONDITION cond;
	char* str;
	while (COND_TopConditionEx(&cond, &str) != COND_NORMAL) {
		wchar_t wmsg[256];
		multi_byte_to_wide_char(str, -1, wmsg, kMessageBufferLen);
		add(L"%1!08X!: %2", cond, wmsg);
	}
}

