/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "zip.h"

//========================================================================================
//
// File Mapping for ordinaly file
//
//========================================================================================
//========================================================================================
//
//========================================================================================
CStackFileMapping::CStackFileMapping(HANDLE fileHndl, const bool writable)
{
	init();

	m_dwFileSize = GetFileSize(fileHndl, &m_dwFileSizeHigh);
	if (m_dwFileSize == 0xFFFFFFFF) {
		m_dwLastErr = GetLastError();
		return;
	}

	m_hMap = CreateFileMapping(fileHndl, NULL, writable ? PAGE_READWRITE : PAGE_READONLY, 0, 0, NULL);
	if (m_hMap == NULL) {
		m_dwLastErr = GetLastError();
		return;
	}
	
	m_pMap = MapViewOfFile(m_hMap, writable ? FILE_MAP_WRITE : FILE_MAP_READ, 0, 0, 0);
	if (m_pMap == NULL) {
		m_dwLastErr = GetLastError();
		CloseHandle(m_hMap); m_hMap = NULL;
		return;
	}
}

//========================================================================================
//
//========================================================================================
CStackFileMapping::CStackFileMapping(const char* file)
{
	TOWC(file)
	open(_file_WC);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void CStackFileMapping::open(const wchar_t* file)
{
	init();

	DWORD flag = GetFileAttributes(file);
	if (flag == 0xFFFFFFFF) {
		m_dwLastErr = GetLastError();
		return;
	}

	m_hFile = CreateFile(file, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (m_hFile == INVALID_HANDLE_VALUE) {
		m_dwLastErr = GetLastError();
		return;
	}
	m_dwFileSize = GetFileSize(m_hFile, &m_dwFileSizeHigh);
	if (m_dwFileSize == 0 && m_dwFileSizeHigh == 0) {
		m_dwLastErr = ERROR_BAD_LENGTH;
		return;
	}

	m_hMap = CreateFileMapping(m_hFile, NULL, PAGE_READONLY, 0, 0, NULL);
	if (m_hMap == NULL) {
		m_dwLastErr = GetLastError();
		CloseHandle(m_hFile); m_hFile = INVALID_HANDLE_VALUE;
		return;
	}

	m_pMap = MapViewOfFile(m_hMap, FILE_MAP_READ, 0, 0, 0);
	if (m_pMap == NULL) {
		m_dwLastErr = GetLastError();
		CloseHandle(m_hMap); m_hMap = NULL;
		CloseHandle(m_hFile); m_hFile = INVALID_HANDLE_VALUE;
		return;
	}
}

//========================================================================================
//
//========================================================================================
CStackFileMapping::~CStackFileMapping()
{
	if (m_pMap != NULL)
		UnmapViewOfFile(m_pMap);
	if (m_hMap != NULL)
		CloseHandle(m_hMap);
	if (m_hFile != INVALID_HANDLE_VALUE)
		CloseHandle(m_hFile);
}
