/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static void unescape(char* cp, char* ep, char* buf, size_t bufSize)
{
	char* dp = buf;
	char* de = buf + bufSize - 1;
	while (cp < ep && dp < de) {
		char c = *cp;
		if (c == '%') {
			if (ep - cp >= 2 && cp[1] == '%') {
				c = '%';
				cp += 2;
			} else if (ep - cp >= 3 && isalnum(cp[1]) && isalnum(cp[2])) {
				c = (cp[1] >= 'A' ? ((cp[1] & 0xDF) - 'A') + 10 : (cp[1] - '0'));
				c *= 16;
				c += (cp[2] >= 'A' ? ((cp[2] & 0xDF) - 'A') + 10 : (cp[2] - '0'));
				cp += 3;
			} else {
				cp += 1;
			}
		} else {
			cp += 1;
		}
		*dp++ = c;
	}
	*dp = '\0';
}


//======================================================================
//
//======================================================================
void GetMultiQueryValue(const char* const queryString, tQueryParam* prm_tbl, const int numPrms)
{
	const char* cp = queryString;
	while (*cp != '\0') {
		const char* name;
		size_t name_len;
		const char* val;
		size_t val_len;
		cp = GetNextQueryPair(cp, &name, &name_len, &val, &val_len);
		if (name_len == 0)
			break;
		tQueryParam* prm = prm_tbl;
		for (int i = 0; i < numPrms; i++, prm++) {
			int len = strlen(prm->name);
			if (len == name_len && memcmp(name, prm->name, len) == 0) {
				prm->hasParam = true;
				prm->hasValue = val != NULL && val_len > 0;
				prm->val_ptr = val;
				prm->val_len = val_len;
				break;
			}
		}
	}
}
//======================================================================
int tQueryParam::strcmp(const char* str)
{
	int len = strlen(str);
	if (len != val_len)
		return val_len - len;

	return memcmp(val_ptr, str, len);
}
//======================================================================
int tQueryParam::atoi()
{
	char* const buf = (char*)_alloca(val_len + 1);
	memcpy(buf, val_ptr, val_len);
	buf[val_len] = '\0';
	return ::atoi(buf);
}
//======================================================================
float tQueryParam::atof()
{
	char* const buf = (char*)_alloca(val_len + 1);
	memcpy(buf, val_ptr, val_len);
	buf[val_len] = '\0';
	return ::atof(buf);
}

//======================================================================
//
//======================================================================
char* url_encode(char* const dst, const char* const src)
{
	char* dp = dst;
	const char* sp = src;
	while (*sp != '\0') {
		char c = *sp++;
		if (isalnum(c)
		 || c == '$' || c == '-' || c == '_' || c == '.' || c == '+'
		 || c == '!' || c == '*' || c == '\'' |  c == '(' || c == ')' || c == ',') {
		 	if (dst != NULL) {
				*dp++ = c;
			} else {
				dp++;
			}
		} else {
		 	if (dst != NULL) {
				*dp++ = '%';
				int i = (c >> 4) & 0x0F;
				*dp++ = i < 10 ? i + '0' : i - 10 + 'A';
				i = c & 0x0F;
				*dp++ = i < 10 ? i + '0' : i - 10 + 'A';
			} else {
				dp += 3;
			}
		}
	}
	if (dst != NULL)
		*dp = '\0';
	return dp;
}
//======================================================================
//
//======================================================================
char* c_encode(char* const dst, const char* const src)
{
	char* dp = dst;
	const char* sp = src;
	while (*sp != '\0') {
		unsigned char c = (unsigned char)*sp++;
		if (c >= ' ' && c != '\'' && c != '\"' && c != '\\') {
			if (dst != NULL) {
				*dp++ = (char)c;
			} else {
				dp++;
			}
		} else {
			char str[4];
			char* cp = str;
			*cp++ = '\\';
			switch (c) {
			case '\r':
				*cp++ = 'r';
				break;
			case '\n':
				*cp++ = 'n';
				break;
			case '\t':
				*cp++ = 't';
				break;
			case '\'':
				*cp++ = '\'';
				break;
			case '\"':
				*cp++ = '\"';
				break;
			case '\\':
				*cp++ = '\\';
				break;
			default:
				*cp++ = 'x';
				int i = (c >> 4) & 0x0F;
				*cp++ = i < 10 ? i + '0' : i - 10 + 'A';
				i = c & 0x0F;
				*cp++ = i < 10 ? i + '0' : i - 10 + 'A';
				break;
			}
			int len = cp - str;
			if (dst != NULL) {
				memcpy(dp, str, len);
				dp += len;
			} else {
				dp += len;
			}
		}
	}
	if (dst != NULL)
		*dp = '\0';
	return dp;
}

//======================================================================
//
//======================================================================
size_t unescape_url(const char* const str, const char* const end, char* buf)
{
	const char* cp = str;
	char* dp = buf;
	size_t length = 0;
	while (cp < end) {
		char c = *cp;
		if (c == '%') {
			if (end - cp >= 2 && cp[1] == '%') {
				c = '%';
				cp += 2;
			} else if (end - cp >= 3 && isalnum(cp[1]) && isalnum(cp[2])) {
				c = (cp[1] >= 'A' ? ((cp[1] & 0xDF) - 'A') + 10 : (cp[1] - '0'));
				c *= 16;
				c += (cp[2] >= 'A' ? ((cp[2] & 0xDF) - 'A') + 10 : (cp[2] - '0'));
				cp += 3;
			} else {
				return 0;
			}
		} else if (c == '+') {
			c = ' ';
			cp++;
		} else {
			cp++;
		}
		if (buf != NULL)
			*dp++ = c;
	}

	if (buf != NULL)
		*dp = '\0';
	return dp - buf;
}

//======================================================================
//
//======================================================================
const char* GetNextQueryPair(const char* const queryString, const char** name, size_t* name_len, const char** val, size_t* val_len)
{
	if (queryString == NULL || *queryString == '\0')
		return NULL;

	*name = NULL;
	*name_len = 0;
	*val = NULL;
	*val_len = NULL;

	const char* cp = queryString;
	while (*cp != '\0' && *cp != '&' && *cp != '=')
		cp++;
	*name = queryString;
	*name_len = cp - queryString;
	if (*cp == '\0')
		return cp;
	else if (*cp == '&')
		return cp;
	*val = ++cp;
	while (*cp != '\0' && *cp != '&')
		cp++;
	*val_len = cp - *val;
	if (*cp == '&')
		cp++;
	return cp;
}
//======================================================================
//
//======================================================================
bool FindQueryStringValue(const char* queryString, const char* varName, const char** val_ptr, size_t* val_len)
{
	if (queryString == NULL)
		return false;

	const int varNameLen = strlen(varName);

	const char* cp = queryString;
	while (*cp != '\0') {
		const char* name;
		size_t name_size;
		cp = GetNextQueryPair(cp, &name, &name_size, val_ptr, val_len);
		if (name_size == varNameLen && memcmp(name, varName, name_size) == 0)
			return true;
	}
	return false;
}

//======================================================================
//
//======================================================================
int istag(char* str)
{
	if (strlen(str) != 8)
		return false;
	for (char* cp = str; *cp != '\0'; cp++)
		if (!isxdigit(*cp))
			return false;

	return true;
}

#if defined(COND_IS_IN) && defined(DCM_OBJECTS_IS_IN)
//======================================================================
//
//
//
//======================================================================
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static unsigned long count_vm(const char* cp, size_t len)
{
	if (len == 0)
		return 0;

	unsigned long vm = 1;
	const char* const ep = cp + len;
	while (cp < ep) {
		if (*cp == '\\')
			vm++;
		cp++;
	}

	return vm;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static CONDITION string_AddElement_FD(DCM_OBJECT** obj, DCM_ELEMENT* element, const char* str, size_t len)
{
	element->multiplicity = count_vm(str, len);	
	element->length = element->multiplicity * 8;
	if (element->length > 0) {
		double* const va = (double*)_alloca(element->length);
		element->d.ot = (void*)va;

		double* vp = va;
		const char* const ep = str + len;
		while (true) {
			const char* cp = str;
			while (cp < ep && *cp != '\\')
				cp++;
			char buf[80];
			memcpy(buf, str, cp - str);
			buf[cp - str] = '\0';
			*vp++ = strtod(buf, NULL);
			if (*cp == '\0')
				break;
			str = ++cp;
		}
	}
	return DCM_AddElement(obj, element);
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static CONDITION string_AddElement_FL(DCM_OBJECT** obj, DCM_ELEMENT* element, const char* str, size_t len)
{
	element->multiplicity = count_vm(str, len);	
	element->length = element->multiplicity * 4;
	if (element->length > 0) {
		float* const va = (float*)_alloca(element->length);
		element->d.ot = (void*)va;

		float* vp = va;
		const char* const ep = str + len;
		while (true) {
			const char* cp = str;
			while (cp < ep && *cp != '\\')
				cp++;
			char buf[80];
			memcpy(buf, str, cp - str);
			buf[cp - str] = '\0';
			*vp++ = strtod(buf, NULL);
			if (*cp == '\0')
				break;
			str = ++cp;
		}
	}
	return DCM_AddElement(obj, element);
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static CONDITION string_AddElement_SL(DCM_OBJECT** obj, DCM_ELEMENT* element, const char* str, size_t len)
{
	element->multiplicity = count_vm(str, len);	
	element->length = element->multiplicity * 4;
	if (element->length > 0) {
		long* const va = (long*)_alloca(element->length);
		element->d.ot = (void*)va;

		long* vp = va;
		const char* const ep = str + len;
		while (true) {
			const char* cp = str;
			while (cp < ep && *cp != '\\')
				cp++;
			char buf[80];
			memcpy(buf, str, cp - str);
			buf[cp - str] = '\0';
			*vp++ = strtol(buf, NULL, 10);
			if (*cp == '\0')
				break;
			str = ++cp;
		}
	}
	return DCM_AddElement(obj, element);
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static CONDITION string_AddElement_UL(DCM_OBJECT** obj, DCM_ELEMENT* element, const char* str, size_t len)
{
	element->multiplicity = count_vm(str, len);	
	element->length = element->multiplicity * 4;
	if (element->length > 0) {
		unsigned long* const va = (unsigned long*)_alloca(element->length);
		element->d.ot = (void*)va;

		unsigned long* vp = va;
		const char* const ep = str + len;
		while (true) {
			const char* cp = str;
			while (cp < ep && *cp != '\\')
				cp++;
			char buf[80];
			memcpy(buf, str, cp - str);
			buf[cp - str] = '\0';
			*vp++ = strtoul(buf, NULL, 10);
			if (*cp == '\0')
				break;
			str = ++cp;
		}
	}
	return DCM_AddElement(obj, element);
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static CONDITION string_AddElement_SS(DCM_OBJECT** obj, DCM_ELEMENT* element, const char* str, size_t len)
{
	element->multiplicity = count_vm(str, len);	
	element->length = element->multiplicity * 2;
	if (element->length > 0) {
		short* const va = (short*)_alloca(element->length);
		element->d.ot = (void*)va;

		short* vp = va;
		const char* const ep = str + len;
		while (true) {
			const char* cp = str;
			while (cp < ep && *cp != '\\')
				cp++;
			char buf[80];
			memcpy(buf, str, cp - str);
			buf[cp - str] = '\0';
			*vp++ = strtol(buf, NULL, 10);
			if (*cp == '\0')
				break;
			str = ++cp;
		}
	}
	return DCM_AddElement(obj, element);
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static CONDITION string_AddElement_US(DCM_OBJECT** obj, DCM_ELEMENT* element, const char* str, size_t len)
{
	element->multiplicity = count_vm(str, len);	
	element->length = element->multiplicity * 2;
	if (element->length > 0) {
		unsigned short* const va = (unsigned short*)_alloca(element->length);
		element->d.ot = (void*)va;

		unsigned short* vp = va;
		const char* const ep = str + len;
		while (true) {
			const char* cp = str;
			while (cp < ep && *cp != '\\')
				cp++;
			char buf[80];
			memcpy(buf, str, cp - str);
			buf[cp - str] = '\0';
			*vp++ = strtoul(buf, NULL, 10);
			if (*cp == '\0')
				break;
			str = ++cp;
		}
	}
	return DCM_AddElement(obj, element);
}

//======================================================================
//
//======================================================================
CONDITION QueryStringToObject(const char* query_string, DCM_OBJECT** obj)
{
	CONDITION cond;

	*obj = NULL;
	cond = DCM_CreateObject(obj, DCM_ORDERLITTLEENDIAN);
	if (!CTN_SUCCESS(cond))
		return cond;

	if (query_string == NULL)
		return cond;

	const char* cp = query_string;
	while (*cp != '\0') {
		const char* name;
		size_t name_len;
		const char* val;
		size_t val_len;
		cp = GetNextQueryPair(cp, &name, &name_len, &val, &val_len);
		if (name_len == 0)
			continue;

		char varName[MAX_PATH];
		memcpy(varName, name, name_len);
		varName[name_len] = '\0';

		DCM_ELEMENT element;
		element.tag = strtoul(varName, NULL, 16);
		cond = DCM_LookupElement(&element);
		if (!CTN_SUCCESS(cond))
			continue;

		switch (element.representation) {
		case DCM_AE:			/* Application Entity */
		case DCM_AS:			/* Age string */
		case DCM_CS:			/* Control string */
		case DCM_DA:			/* Date */
		case DCM_DS:			/* Decimal string */
		case DCM_IS:			/* Integer string */
		case DCM_LO:			/* Long string */
		case DCM_LT:			/* Long text */
		case DCM_SH:			/* Short string */
		case DCM_ST:			/* Short text */
		case DCM_TM:			/* Time */
		case DCM_UI:			/* Unique identifier (UID) */
		case DCM_PN:			/* Person Name */
		case DCM_DT:			/* Date/Time */
			element.multiplicity = count_vm(val, val_len);
			element.length = val_len;			
			element.d.string = const_cast<char*>(val);
			cond = DCM_AddElement(obj, &element);
			break;

		case DCM_FD:			/* Floating double */
			cond = string_AddElement_FD(obj, &element, val, val_len);
			break;

		case DCM_FL:			/* Float */
			cond = string_AddElement_FL(obj, &element, val, val_len);
			break;

		case DCM_SL:			/* Signed long */
			cond = string_AddElement_SL(obj, &element, val, val_len);
			break;

		case DCM_UL:			/* Unsigned long */
			cond = string_AddElement_UL(obj, &element, val, val_len);
			break;

		case DCM_SS:			/* Signed short */
			cond = string_AddElement_SS(obj, &element, val, val_len);
			break;

		case DCM_US:			/* Unsigned short */
			cond = string_AddElement_SS(obj, &element, val, val_len);
			break;
		}
		if (!CTN_SUCCESS(cond))
			return cond;
	}

	return cond;
}
#endif
//======================================================================
//
//======================================================================
void FileTimeToDT(const FILETIME* ft, char* dt)
{
	char* cp = dt;
	DWORD dw = ft->dwHighDateTime;
	for (int i = 0; i < 8; i++) {
		DWORD msn = 0xF0000000 & dw;
		dw <<= 4;
		msn >>= 28;
		char c = msn < 10 ? '0' + msn : 'A' + (msn - 10);
		*cp++ = c;
	}
	dw = ft->dwLowDateTime;
	for (int i = 0; i < 3; i++) {
		DWORD msn = 0xF0000000 & dw;
		dw <<= 4;
		msn >>= 28;
		char c = msn < 10 ? '0' + msn : 'A' + (msn - 10);
		*cp++ = c;
	}
	*cp = '\0';
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static bool n_digits_expected(const char* cp, int n)
{
	while (n--) {
		if (!isdigit(*cp++))
			return false;
	}
	return true;		
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static int strntoi(const char* const src, const size_t len)
{
	char buf[8];
	memcpy(buf, src, len); buf[len] = '\0';
	return atoi(buf);
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static void date_time_to_st(const char* const date, const char* const time, SYSTEMTIME& st)
{
	const SYSTEMTIME base_st = {
		1601,
		1,
		0,
		1,
		0,
		0,
		0,
		0
	};
	st = base_st;

	const char* cp = date;
	if (!n_digits_expected(cp, 4))
		return;
	st.wYear = strntoi(cp, 4); cp += 4;
	while (!isdigit(*cp))
		cp++;

	if (!n_digits_expected(cp, 2))
		return;
	st.wMonth = strntoi(cp, 2); cp += 2;
	while (!isdigit(*cp))
		cp++;

	if (!n_digits_expected(cp, 2))
		return;
	st.wDay = strntoi(cp, 2); cp += 2;

	cp = time;

	if (!n_digits_expected(cp, 2))
		return;
	st.wHour = strntoi(cp, 2); cp += 2;

	if (!n_digits_expected(cp, 2))
		return;
	st.wMinute = strntoi(cp, 2); cp += 2;

	if (!n_digits_expected(cp, 2))
		return;
	st.wSecond = strntoi(cp, 2); cp += 2;
}
//======================================================================
//
//======================================================================
void DateTimeToFT(const char* const date, const char* const time, FILETIME* ft)
{
	SYSTEMTIME st;
	date_time_to_st(date, time, st);
	SystemTimeToFileTime(&st, ft);
}

//======================================================================
//
//======================================================================
void DateTimeToDT(const char* const date, const char* const time, char* dt)
{
	FILETIME ft;
	DateTimeToFT(date, time, &ft);
	FileTimeToDT(&ft, dt);
}

//======================================================================
//
//
//
//======================================================================
//======================================================================
//
//======================================================================
size_t tMultiStringConcatEntry::CalcLength()
{
	tMultiStringConcatEntry* tbl = this;
	size_t len = 0;
	while (tbl->str != NULL) {
		tbl->len = strlen(tbl->str);
		len += tbl->len;
		tbl++;
	}
	return len;
}
//======================================================================
//
//======================================================================
void tMultiStringConcatEntry::Concat(char* buf)
{
	tMultiStringConcatEntry* tbl = this;

	while (tbl->str != NULL) {
		memcpy(buf, tbl->str, tbl->len);
		buf += tbl->len;
		tbl++;
	}
	*buf = '\0';
}
//======================================================================
//
//======================================================================
BSTR tMultiStringConcatEntry::NewBSTR()
{
	tMultiStringConcatEntry* tbl = this;

	size_t cstrLen = tbl->CalcLength();
	char* const cstr = (char*)_alloca(cstrLen + 1);
	tbl->Concat(cstr);

	size_t bstrLen = multi_byte_to_wide_char(cstr, -1, NULL, 0);
	BSTR bstr = SysAllocStringLen(NULL, bstrLen);
	multi_byte_to_wide_char(cstr, -1, bstr, bstrLen);

	return bstr;
}

//======================================================================
//
//======================================================================
size_t tMultiStringConcatEntryW::CalcLength()
{
	tMultiStringConcatEntryW* tbl = this;
	size_t len = 0;
	while (tbl->str != NULL) {
		tbl->len = lstrlen(tbl->str);
		len += tbl->len;
		tbl++;
	}
	return len;
}
//======================================================================
//
//======================================================================
void tMultiStringConcatEntryW::Concat(wchar_t* buf)
{
	tMultiStringConcatEntryW* tbl = this;

	while (tbl->str != NULL) {
		memcpy(buf, tbl->str, sizeof(wchar_t) * tbl->len);
		buf += tbl->len;
		tbl++;
	}
	*buf = L'\0';
}
//======================================================================
//
//======================================================================
BSTR tMultiStringConcatEntryW::NewBSTR()
{
	tMultiStringConcatEntryW* tbl = this;

	size_t bstrLen = tbl->CalcLength();
	BSTR bstr = SysAllocStringLen(NULL, bstrLen);
	tbl->Concat(bstr);
	return bstr;
}

//======================================================================
//
//======================================================================
bool strncmpq(const wchar_t* s1, const wchar_t* s2, unsigned int len)
{
	unsigned int dwc = len >> 1;
	unsigned int mod = len & 1;

	s1 += dwc * 2;
	s2 += dwc * 2;

	if (mod != 0) {
		if (*s1 != *s2)
			return false;
	}

	DWORD* d1 = (DWORD*)s1;
	DWORD* d2 = (DWORD*)s2;

	while (dwc--) {
		if (*--d1 != *--d2)
			return false;
	}

	return true;
}

//======================================================================
//
//======================================================================
bool strncmpq(const char* s1, const char* s2, unsigned int len)
{
	unsigned int dwc = len >> 2;
	unsigned int mod = len & 3;

	s1 += dwc * 4;
	s2 += dwc * 4;

	if (mod != 0) {
		for (int i = 0; i < mod; i++) {
			if (s1[i] != s2[i]) {
				return false;
			}
		}
	}

	DWORD* d1 = (DWORD*)s1;
	DWORD* d2 = (DWORD*)s2;

	while (dwc--) {
		if (*--d1 != *--d2)
			return false;
	}

	return true;
}
