#ifndef _1BAD9A29_4AB1_4A33_944A_F1C38F7AED74_
#define _1BAD9A29_4AB1_4A33_944A_F1C38F7AED74_

/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#pragma warn_unusedarg off

//======================================================================
//
//
//
//======================================================================
class SimpleErrorInfo {
public:
	enum {
		kMessageBufferLen = 512 - 2 - 2
	};

	HRESULT m_hr;
	wchar_t* m_tail;
	wchar_t m_message[kMessageBufferLen + 1];

	SimpleErrorInfo()
		: m_hr(S_OK), m_tail(m_message) {}

	SimpleErrorInfo(const SimpleErrorInfo& rhs);
	SimpleErrorInfo& operator=(const SimpleErrorInfo& rhs);

	SimpleErrorInfo(HRESULT hr);
	SimpleErrorInfo(HRESULT hr, const wchar_t* formatStr, ...);
	SimpleErrorInfo(const wchar_t* formatStr, ...);

	void set(HRESULT hr);
	void set(HRESULT hr, const wchar_t* formatStr, ...);
	void set(const wchar_t* formatStr, ...);

	void add(const wchar_t* formatStr, ...);

	void reset();

#ifdef DICOM_IS_IN
	void set_from_CTN_COND();
#endif

private:
	void append_message(const wchar_t* formatStr, va_list* arg);
};

//======================================================================
//
//
//
//======================================================================
class CConsoleLog {
	CRITICAL_SECTION m_crit;
	HANDLE m_hFile;

	wchar_t* m_logFilePath;

	bool m_fileMode;

	void fin();
	void ini();

public:
	CConsoleLog()
		: m_hFile(INVALID_HANDLE_VALUE), m_fileMode(false), m_logFilePath(NULL)
	{}

	~CConsoleLog()
	{
		fin();
	}

	HRESULT Init(bool consoleMode, const wchar_t* logFilePathBase, SECURITY_ATTRIBUTES* sa = NULL);
	HRESULT Init(HANDLE hFile);

	void writeText(const char* text);
	void writeErrorInfo(const SimpleErrorInfo& ei);

};

//======================================================================
//
//
//
//======================================================================
//======================================================================
//
//======================================================================
template< class T >
class AutoComPtr {
	T* pIntf;
	AutoComPtr(const AutoComPtr<T>& rhs);
	AutoComPtr& operator=(const AutoComPtr<T>& rhs);
	static void* operator new(size_t);
	static void operator delete(void*, size_t);
public:
	AutoComPtr() : pIntf(NULL) {}
	T* operator->() { return pIntf; }
	T** r() { return &pIntf; }
	T* p() { return pIntf; }
	void reset() {
		if (pIntf != NULL) {
			pIntf->Release();
			pIntf = NULL;
		}
	}
	~AutoComPtr() {
		if (pIntf != NULL) {
			pIntf->Release();
		}
	}
};

//======================================================================
//
//
//
//======================================================================
//======================================================================
//
//======================================================================
template< class T >
class AutoComRelease {
	T** pIntf;
	AutoComRelease();
	AutoComRelease(const AutoComRelease<T>& rhs);
	AutoComRelease& operator=(const AutoComRelease<T>& rhs);
	static void* operator new(size_t);
	static void operator delete(void*, size_t);
public:
	AutoComRelease(T** rhs) : pIntf(rhs) {}
	~AutoComRelease() {
		if (pIntf != NULL && *pIntf != NULL) {
			(*pIntf)->Release();
		}
	}
};
//======================================================================
//
//======================================================================
class AutoSysFreeString {
	BSTR& bstr;
	AutoSysFreeString();
	AutoSysFreeString(const AutoSysFreeString& rhs);
	AutoSysFreeString& operator=(const AutoSysFreeString& rhs);
public:
	AutoSysFreeString(BSTR& rhs) : bstr(rhs) {}
	~AutoSysFreeString();
};

//======================================================================
//
//======================================================================
class AutoCoUninitialize {
	AutoCoUninitialize(const AutoCoUninitialize& rhs);
	AutoCoUninitialize& operator=(const AutoCoUninitialize& rhs);
public:
	AutoCoUninitialize() {}
	~AutoCoUninitialize();
};
//======================================================================
//
//======================================================================
class AutoWSACleanup {
	AutoWSACleanup(const AutoWSACleanup& rhs);
	AutoWSACleanup& operator=(const AutoWSACleanup& rhs);
public:
	AutoWSACleanup() {}
	~AutoWSACleanup();
};

//======================================================================
//
//======================================================================
class AutoVariantClear {
	VARIANT& m_val;
	AutoVariantClear();
	AutoVariantClear(const AutoVariantClear& rhs);
	AutoVariantClear& operator=(const AutoVariantClear& rhs);
public:
	AutoVariantClear(VARIANT& val, bool init = false)
		: m_val(val) {
		if (init)
			VariantInit(&m_val);
	}
	~AutoVariantClear();
};

//======================================================================
//
//======================================================================
class AutoCloseFileHandle {
	HANDLE& m_hndl;
public:
	AutoCloseFileHandle(HANDLE& hndl)
		: m_hndl(hndl) {}

	~AutoCloseFileHandle();
};
//======================================================================
//
//======================================================================
class AutoCloseHandle {
	HANDLE& m_hndl;
public:
	AutoCloseHandle(HANDLE& hndl)
		: m_hndl(hndl) { }

	~AutoCloseHandle();
};

//======================================================================
//
//======================================================================
class AutoUnmapViewOfFile {
	LPVOID& lpMap;
public:
	AutoUnmapViewOfFile(LPVOID& map_ptr) : lpMap(map_ptr) {}
	~AutoUnmapViewOfFile();
};

//======================================================================
//
//======================================================================
class AutoUnlockFileEx {
	HANDLE& m_hFile;
	DWORD m_nNumberOfBytesToUnlockLow;
	DWORD m_nNumberOfBytesToUnlockHigh;
	OVERLAPPED* m_ov;
public:
	AutoUnlockFileEx(HANDLE& hFile, DWORD nNumberOfBytesToUnlockLow, DWORD nNumberOfBytesToUnlockHigh, OVERLAPPED* ov)
		: m_hFile(hFile), m_nNumberOfBytesToUnlockLow(nNumberOfBytesToUnlockLow),m_nNumberOfBytesToUnlockHigh(nNumberOfBytesToUnlockHigh), m_ov(ov)
	{
	}
	~AutoUnlockFileEx()
	{
		UnlockFileEx(m_hFile, 0, m_nNumberOfBytesToUnlockLow, m_nNumberOfBytesToUnlockHigh, m_ov);
	}	
};

//======================================================================
//
//======================================================================
class AutoCloseChangeNotificationHandle {
	HANDLE& m_handle;
public:
	AutoCloseChangeNotificationHandle(HANDLE& handle)
		: m_handle(handle)
	{
	}
	~AutoCloseChangeNotificationHandle();
};

//======================================================================
//
//======================================================================
class AutoInterlockedExchange {
	LONG* m_pValue;
	LONG m_rValue;
public:
	AutoInterlockedExchange(LONG* pValue, LONG rValue)
		: m_pValue(pValue), m_rValue(rValue)
	{
	}
	~AutoInterlockedExchange()
	{
 		InterlockedExchange(m_pValue, m_rValue);
	}
};

//======================================================================
//
//======================================================================
class AutoInterlockedDecrement {
	LONG* m_pValue;
public:
	AutoInterlockedDecrement(LONG* pValue)
		: m_pValue(pValue)
	{
	}
	~AutoInterlockedDecrement()
	{
 		InterlockedDecrement(m_pValue);
	}
};

//======================================================================
//
//======================================================================
class AutoInterlockedCounter {
	LONG* m_pValue;
public:
	AutoInterlockedCounter(LONG* pValue)
		: m_pValue(pValue)
	{
		InterlockedIncrement(m_pValue);
	}
	~AutoInterlockedCounter()
	{
 		InterlockedDecrement(m_pValue);
	}
};



//======================================================================
//
//======================================================================
class AutoCloseAndDeleteFile {
	HANDLE& m_hFile;
	const wchar_t* m_path;
public:
	AutoCloseAndDeleteFile(HANDLE& hFile, const wchar_t* path)
		: m_hFile(hFile), m_path(path)
	{}
	~AutoCloseAndDeleteFile()
	{
		if (m_hFile == INVALID_HANDLE_VALUE)
			return;
		CloseHandle(m_hFile);
		if (m_path == NULL)
			return;
		DeleteFile(m_path);
	}
	void dontDelete()
	{
		m_path = NULL;
	}
};
	

//======================================================================
//
//======================================================================
class CWin32ErrorText {
	const wchar_t* const lpMsgBuf;

	static wchar_t* getmsg(DWORD err)
	{
		wchar_t* msg = NULL;
		const DWORD len = FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			err,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(wchar_t*)&msg,
			0,
			NULL);
		if (len == 0)
			return NULL;
		return msg;
	}

public:
	CWin32ErrorText(DWORD err = S_OK)
		: lpMsgBuf(getmsg(err == S_OK ? GetLastError() : err))
	{
	}

	~CWin32ErrorText() {
		if (lpMsgBuf != NULL)
			LocalFree((void*)lpMsgBuf);
	}

	const wchar_t* str(void) const { return lpMsgBuf; }
};

//======================================================================
//
// Stack Classes
//
//======================================================================
//======================================================================
// Critical Section
//======================================================================
class CStackCriticalSection {
	CRITICAL_SECTION* crit;
public:

	CStackCriticalSection();
	CStackCriticalSection(CRITICAL_SECTION* cs);

	~CStackCriticalSection();
};

//======================================================================
// File Mapping
//======================================================================
class CStackFileMapping {
	HANDLE m_hFile;
	HANDLE m_hMap;
	LPCVOID m_pMap;
	DWORD m_dwFileSize;
	DWORD m_dwFileSizeHigh;
	DWORD m_dwLastErr;

	void init(void) {
		m_hFile = INVALID_HANDLE_VALUE;
		m_hMap = NULL;
		m_pMap = NULL;
		m_dwFileSize = 0;
		m_dwFileSizeHigh = 0;
		m_dwLastErr = NO_ERROR;
	}

	void open(const wchar_t* file);

public:
	CStackFileMapping() { init(); };

	CStackFileMapping(const wchar_t* file) { open(file); }
	CStackFileMapping(const char* file);

	CStackFileMapping(HANDLE hndl, const bool writable = false);

	~CStackFileMapping();

	const void* mapPointer(void) const { return m_pMap; }
	size_t mapSize(void) const { return m_dwFileSize; }
	HANDLE fileHandle(void) const { return m_hFile;}
	DWORD lastError(void) const { return m_dwLastErr; }
};


//======================================================================
// Buffer
//======================================================================
template< int TBufferSize  >
class CStackBuffer {

	struct block_t {
		block_t* m_next;
		char m_buf[TBufferSize];
	};

	size_t m_data_size;
	size_t m_buf_size;

	block_t m_stack_block;
	block_t* m_buf_list_tail;

	HANDLE m_heap;
	block_t* Alloc()
	{
		if (m_heap != NULL)
			return (block_t*)HeapAlloc(m_heap, HEAP_NO_SERIALIZE, sizeof(block_t));
		else
			return (block_t*)HeapAlloc(GetProcessHeap(), 0, sizeof(block_t));
	}

	void Free(block_t* p)
	{
		if (m_heap != NULL)
			HeapFree(m_heap, HEAP_NO_SERIALIZE, p);
		else
			HeapFree(GetProcessHeap(), 0, p);
	}

	// protect unwanted construction
	CStackBuffer(const CStackBuffer& rhs);
	CStackBuffer& operator=(const CStackBuffer& rhs);

	void free_buffers()
	{
		if (m_buf_list_tail != &m_stack_block) {
			block_t* buf = m_stack_block.m_next;
			while (buf != NULL) {
				block_t* next = buf->m_next;
				Free(buf);
				buf = next;
			}
		}
	}

public:

	CStackBuffer(HANDLE heap = NULL)
		: m_buf_list_tail(&m_stack_block), m_data_size(0), m_buf_size(TBufferSize), m_heap(heap)
	{
		m_stack_block.m_next = NULL;
	}

	~CStackBuffer()
	{
		free_buffers();
	}

	void clear()
	{
		free_buffers();

		m_data_size = 0;
		m_buf_size = TBufferSize;
		m_buf_list_tail = &m_stack_block;
		m_stack_block.m_next = NULL;
	}

	void append(const char* data, size_t data_size)
	{
		while (data_size > 0) {
			size_t buf_remain = m_buf_size - m_data_size;
			size_t copy_size = data_size > buf_remain ? buf_remain : data_size;
			size_t buf_offset = m_data_size % TBufferSize;
			memcpy(m_buf_list_tail->m_buf + buf_offset, data, copy_size);

			m_data_size += copy_size;

			data_size -= copy_size;
			data += copy_size;
		
			if (data_size == 0)
				return;
		
			m_buf_list_tail->m_next = Alloc();
			m_buf_list_tail = m_buf_list_tail->m_next;
			m_buf_list_tail->m_next = NULL;
			m_buf_size += TBufferSize;
		}
	}

	void fetch(size_t offset, char* data, size_t data_size) const
	{
		if (offset >= m_data_size)
			return;

		size_t data_remain = m_data_size;
		const block_t* buf = &m_stack_block;
		while (offset >= TBufferSize) {
			buf = buf->m_next;
			offset -= TBufferSize;
			data_remain -= TBufferSize;
		}

		const size_t buf_remain = TBufferSize - offset;
		const size_t data_avail = buf_remain > data_remain ? data_remain : buf_remain;
		const size_t copy_size = data_size > data_avail ? data_avail : data_size;

		memcpy(data, buf->m_buf + offset, copy_size);

		data += copy_size;
		data_size -= copy_size;
		data_remain -= copy_size;

		while (data_size > 0 && data_remain > 0) {
			buf = buf->m_next;

			const size_t data_avail = TBufferSize > data_remain ? data_remain : TBufferSize;
			const size_t copy_size = data_size > data_avail ? data_avail : data_size;

			memcpy(data, buf->m_buf, copy_size);

			data += copy_size;
			data_size -= copy_size;
			data_remain -= copy_size;
		}
	}

	const char* peek(size_t offset, size_t& data_avail) const
	{
		if (offset >= m_data_size)
			return NULL;

		size_t data_remain = m_data_size;
		const block_t* buf = &m_stack_block;
		while (offset >= TBufferSize) {
			buf = buf->m_next;
			offset -= TBufferSize;
			data_remain -= TBufferSize;
		}

		const size_t buf_remain = TBufferSize - offset;
		data_avail = buf_remain > data_remain ? data_remain : buf_remain;

		return buf->m_buf + offset;
	}

	size_t data_size(void) const { return m_data_size; }
	size_t buf_size(void) const { return m_buf_size; }
	size_t chunk_size(void) const { return TBufferSize; }
};

//======================================================================
// Response Buffer
//======================================================================
class CStackResponseBuffer {
	CStackBuffer< 0x4000 > m_buffer;
public:

	CStackResponseBuffer(HANDLE heap = NULL)
		: m_buffer(heap) {}

	void add(const char* str, size_t len = 0);

	DWORD flush(EXTENSION_CONTROL_BLOCK* ecb, const char* content_type, const char* extra_header_fields = NULL);
	DWORD flush(FILE* fp, const char* content_type, const char* extra_header_fields = NULL);
};

//======================================================================
// Find File
//======================================================================
class CStackFindFile {
public:
	HANDLE handle;
	WIN32_FIND_DATA data;

	CStackFindFile();
	CStackFindFile(const char* dir, const char* wild = NULL);
	CStackFindFile(const wchar_t* dir, const wchar_t* wild = NULL);
	~CStackFindFile();

	bool next(void);
};
//======================================================================
//
//======================================================================
class CVirtualBuffer {
public:
	enum {
		kPageSize = 4096 * 2
	};

private:
	LPVOID m_pvBaseAddress;
	DWORD m_dwReservedSize;

	DWORD m_dwLogicalCommitedSize;
	DWORD m_dwDataSize;
	DWORD m_dwCommitedSize;

	DWORD m_dwLastError;

	CVirtualBuffer(const CVirtualBuffer& rhs);
	CVirtualBuffer& operator=(const CVirtualBuffer& rhs);

	LPVOID commit_memory(const DWORD dwRequestedSize);

public:

	CVirtualBuffer();
	~CVirtualBuffer();

	BOOL Reserve(const DWORD dwReserveSize, const DWORD dwPreCommit = 0);
	VOID Release(void);
	LPVOID Append(LPCVOID pvData, const DWORD dwDataSize);
	BOOL Commit(void);
	void Decommit(void);
	BOOL RollBack(void);
	BOOL Read(LPVOID* ppvData, DWORD* pdwDataSize);
	VOID Protect(void);
	DWORD GetOffset(LPVOID pv);
	LPVOID GetPointer(DWORD offset);
	DWORD GetDataSize(void);
	DWORD lastError(void) const;
};

//======================================================================
//
//======================================================================
class CVirtualBufferStream : public IStream {
	CVirtualBuffer m_vb;
	LONG m_cRef;
	ULONG m_offset;

	CVirtualBufferStream();

public:
	// Allocator
	void* operator new(size_t size);
	void operator delete(void* deadObject, size_t size);

	// constructors/destructors
	CVirtualBufferStream(DWORD size);
	~CVirtualBufferStream();

	// IUnknown
	STDMETHODIMP QueryInterface (REFIID iid, void **ppv);
	STDMETHODIMP_(ULONG) AddRef(void);
	STDMETHODIMP_(ULONG) Release(void);

	// IStream
	STDMETHODIMP Read(void *pv, ULONG cb, ULONG *pcbRead);
	STDMETHODIMP Write(VOID const *pv, ULONG cb, ULONG *pcbWritten);
	STDMETHODIMP Seek(LARGE_INTEGER dbMove, DWORD dwOrigin, ULARGE_INTEGER *pbNewPosition);

	STDMETHODIMP SetSize(ULARGE_INTEGER cbNewSize) { return E_FAIL; }
	STDMETHODIMP CopyTo(IStream *pstm, ULARGE_INTEGER cb, ULARGE_INTEGER *pcbRead, ULARGE_INTEGER *pcbWritten) { return E_FAIL; }
	STDMETHODIMP Commit(DWORD grfCommitFlags) { return E_FAIL; }
	STDMETHODIMP Revert(void) { return E_FAIL; }
	STDMETHODIMP LockRegion(ULARGE_INTEGER bOffset, ULARGE_INTEGER cb, DWORD dwLockType) { return E_FAIL; }
	STDMETHODIMP UnlockRegion(ULARGE_INTEGER bOffset, ULARGE_INTEGER cb, DWORD dwLockType) { return E_FAIL; }
	STDMETHODIMP Stat(STATSTG *pstatstg, DWORD grfStatFlag) { return E_FAIL; }
	STDMETHODIMP Clone(IStream **ppstm) { return E_FAIL; }
};

//======================================================================
//
//======================================================================
class CBlockBufferStream : public IStream {
public:
	enum {
		kBlockSize = 8192
	};

private:
	struct block_t {
		block_t* m_next;
		BYTE m_data[kBlockSize];

		block_t()
			: m_next(NULL)
		{
		}
	};

	LONG m_cRef;

	ULONG m_data_size;
	block_t m_block;

	block_t* m_current_mb;
	ULONG m_buf_offset;
	ULONG m_offset;

public:
	// Allocator
	void* operator new(size_t size);
	void operator delete(void* deadObject, size_t size);

	// constructors/destructors
	CBlockBufferStream();
	~CBlockBufferStream();

	// IUnknown
	STDMETHODIMP QueryInterface (REFIID iid, void **ppv);
	STDMETHODIMP_(ULONG) AddRef(void);
	STDMETHODIMP_(ULONG) Release(void);

	// IStream
	STDMETHODIMP Read(void *pv, ULONG cb, ULONG *pcbRead);
	STDMETHODIMP Write(VOID const *pv, ULONG cb, ULONG *pcbWritten);
	STDMETHODIMP Seek(LARGE_INTEGER dbMove, DWORD dwOrigin, ULARGE_INTEGER *pbNewPosition);

	STDMETHODIMP SetSize(ULARGE_INTEGER cbNewSize) { return E_FAIL; }
	STDMETHODIMP CopyTo(IStream *pstm, ULARGE_INTEGER cb, ULARGE_INTEGER *pcbRead, ULARGE_INTEGER *pcbWritten) { return E_FAIL; }
	STDMETHODIMP Commit(DWORD grfCommitFlags) { return E_FAIL; }
	STDMETHODIMP Revert(void) { return E_FAIL; }
	STDMETHODIMP LockRegion(ULARGE_INTEGER bOffset, ULARGE_INTEGER cb, DWORD dwLockType) { return E_FAIL; }
	STDMETHODIMP UnlockRegion(ULARGE_INTEGER bOffset, ULARGE_INTEGER cb, DWORD dwLockType) { return E_FAIL; }
	STDMETHODIMP Stat(STATSTG *pstatstg, DWORD grfStatFlag) { return E_FAIL; }
	STDMETHODIMP Clone(IStream **ppstm) { return E_FAIL; }
};

//======================================================================
// IStream implementation -- external File Handle, Stack Version (can't new)
//======================================================================
class CStackFileStream : public IStream {
	HANDLE& m_hFile;
	ULONG m_cRef;

	// protect 
	CStackFileStream();
	void* operator new(size_t size);
	void operator delete(void* deadObject, size_t size);

public:

	// constructor/destructor
	CStackFileStream(HANDLE& hFile);
	~CStackFileStream();

	// IUnknown
	STDMETHODIMP QueryInterface (REFIID iid, void **ppv);
	STDMETHODIMP_(ULONG) AddRef(void)  { return ++m_cRef; }
	STDMETHODIMP_(ULONG) Release(void) { return --m_cRef; }

	// IStream
	STDMETHODIMP Read(void *pv, ULONG cb, ULONG *pcbRead);
	STDMETHODIMP Write(VOID const *pv, ULONG cb, ULONG *pcbWritten);
	STDMETHODIMP Seek(LARGE_INTEGER dbMove, DWORD dwOrigin, ULARGE_INTEGER *pbNewPosition);

	STDMETHODIMP SetSize(ULARGE_INTEGER cbNewSize) { return E_FAIL; }
	STDMETHODIMP CopyTo(IStream *pstm, ULARGE_INTEGER cb, ULARGE_INTEGER *pcbRead, ULARGE_INTEGER *pcbWritten) { return E_FAIL; }
	STDMETHODIMP Commit(DWORD grfCommitFlags) { return E_FAIL; }
	STDMETHODIMP Revert(void) { return E_FAIL; }
	STDMETHODIMP LockRegion(ULARGE_INTEGER bOffset, ULARGE_INTEGER cb, DWORD dwLockType) { return E_FAIL; }
	STDMETHODIMP UnlockRegion(ULARGE_INTEGER bOffset, ULARGE_INTEGER cb, DWORD dwLockType) { return E_FAIL; }
	STDMETHODIMP Stat(STATSTG *pstatstg, DWORD grfStatFlag) { return E_FAIL; }
	STDMETHODIMP Clone(IStream **ppstm) { return E_FAIL; }
};

//======================================================================
//
//======================================================================
class fake_bstr_t {
	LPVOID m_buffer;
	DWORD m_buffer_len; // in count of wchar_t
public:

	fake_bstr_t(LPVOID pv = NULL, DWORD len = 0)
		: m_buffer(pv), m_buffer_len(len)
	{
	}

	~fake_bstr_t()
	{
		if (m_buffer != NULL)
			CoTaskMemFree(m_buffer);
	}

	BSTR bstr(void) const
	{
		return (BSTR)((char*)m_buffer + 4);
	}

	LPWSTR wstr(DWORD offset = 0) {
		return (LPWSTR)((char*)m_buffer + 4) + offset;
	}

	DWORD strlen(void) const
	{
		return *(DWORD*)m_buffer;
	}

	void setLength(DWORD len)
	{
		*(DWORD*)m_buffer = len;
	}

	void strcpy(const wchar_t* src)
	{
		const DWORD len = static_cast<DWORD>(lstrlen(src));
		CopyMemory(wstr(), src, sizeof(wchar_t) * (len + 1));
		setLength(len);
	}
		
	void calcLength(void)
	{
		setLength(static_cast<DWORD>(lstrlen(wstr())));
	}

	void ensure_buffer(DWORD requeted_len)
	{
		if (requeted_len > m_buffer_len) {
			LPVOID tmp = CoTaskMemAlloc(sizeof(WCHAR) * requeted_len);
			if (m_buffer != 0) {
				CopyMemory(tmp, m_buffer, m_buffer_len);
				CoTaskMemFree(m_buffer);
			}
			m_buffer = tmp;
			m_buffer_len = requeted_len;
		}
	}

};

//======================================================================
//
//
//
//======================================================================
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
typedef enum {
	eUnknown = 0,
	eInteger,
	eReal,
	eString,
	eTime,
	eBinary,
	ePixel = -1
} tValueType;
//---------------------------------------------------------------------
struct tQueryParam {
	const char*  name;		// [in]
	tValueType type;		// [in]; integer: 0, real: 1, pixel: -1, string: 2
	bool hasParam;			// [out];
	bool hasValue;			// [out];
	size_t val_len;			// [out];
	const char* val_ptr;	// [out];

	int strcmp(const char* str);
	void strcpy(char* str);
	int atoi();
	float atof();
};
//======================================================================
const char* GetNextQueryPair(const char* const queryString, const char** name, size_t* name_len, const char** val, size_t* val_len);
bool FindQueryStringValue(const char* queryString, const char* varName, const char** val_ptr, size_t* val_len);
void GetMultiQueryValue(const char* const queryString, tQueryParam* prm, const int numPrms);

char* url_encode(char* const dst, const char* const src);
char* c_encode(char* const dst, const char* const src);
size_t unescape_url(const char* const str, const char* const end, char* buf);

int istag(char* str);

//======================================================================
//
//======================================================================
void DateTimeToDT(const char* const date, const char* const time, char* dt);
void FileTimeToDT(const FILETIME* ft, char* dt);
void DateTimeToFT(const char* const date, const char* const time, FILETIME* ft);


//======================================================================
//
//======================================================================
void gather_files(HANDLE hFile, char* dir, char* name_pattern, unsigned long& numImages);

//======================================================================
//
//======================================================================
inline int multi_byte_to_wide_char(const char* src, int src_len, wchar_t* dst, int dst_len)
{
	return MultiByteToWideChar(CP_OEMCP, MB_PRECOMPOSED, src, src_len, dst, dst_len);
}
inline int wide_char_to_multi_byte(const wchar_t* src, int src_len, char* dst, int dst_len)
{
	return WideCharToMultiByte(CP_OEMCP, 0, src, src_len, dst, dst_len, NULL, NULL);
}

#define TOWC(x) const int _##x##_WC_len = multi_byte_to_wide_char(x, -1, NULL, 0) - 1; \
	const wchar_t* const _##x##_WC = (const wchar_t*)_alloca(sizeof(wchar_t) * (_##x##_WC_len + 1)); \
	multi_byte_to_wide_char(x, -1, const_cast<wchar_t*>(_##x##_WC), _##x##_WC_len + 1); \
	

#define TOAC(x) const int _##x##_AC_len = wide_char_to_multi_byte(x, -1, NULL, 0) - 1; \
	const char* const _##x##_AC = (char*)_alloca(sizeof(char) * (_##x##_AC_len + 1)); \
	wide_char_to_multi_byte(x, -1, const_cast<char*>(_##x##_AC), _##x##_AC_len + 1); \

//======================================================================
//
//======================================================================
struct tMultiStringConcatEntry {

	const char* str;
	size_t len;

	size_t CalcLength();
	void Concat(char* str_buf);
	BSTR NewBSTR(void);
};

//======================================================================
//
//======================================================================
struct tMultiStringConcatEntryW {

	const wchar_t* str;
	size_t len;

	size_t CalcLength(void);
	void Concat(wchar_t* str_buf);
	BSTR NewBSTR(void);
};

//======================================================================
//
//======================================================================
inline void strncat3(char* cp, const char* str1, size_t str1_len, char delm, const char* str2, size_t str2_len)
{
	memcpy(cp, str1, str1_len); cp += str1_len;
	*cp++ = delm;
	memcpy(cp, str2, str2_len); cp += str2_len;
	*cp = '\0';
}

//======================================================================
//
//======================================================================
inline void strncat3(wchar_t* cp, const wchar_t* str1, size_t str1_len, wchar_t delm, const wchar_t* str2, size_t str2_len)
{
	memcpy(cp, str1, sizeof(wchar_t) * str1_len); cp += str1_len;
	*cp++ = delm;
	memcpy(cp, str2, sizeof(wchar_t) * str2_len); cp += str2_len;
	*cp = '\0';
}

//======================================================================
//
//======================================================================

inline size_t round_size(size_t size)
{
	return ((size + 8 - 1) / 8) * 8;
}

//======================================================================
//
//======================================================================

bool strncmpq(const wchar_t* s1, const wchar_t* s2, unsigned int len);
bool strncmpq(const char* s1, const char* s2, unsigned int len);

//======================================================================
//
// XMLDOMUtils.cpp
//
//======================================================================
const SimpleErrorInfo* const _xduDontSetSimpleErrorInfo = reinterpret_cast<const SimpleErrorInfo*>(1);
#define xduDontSetSimpleErrorInfo const_cast<SimpleErrorInfo*>(_xduDontSetSimpleErrorInfo)
#define xduThrowSimpleErrorInfo static_cast<SimpleErrorInfo*>(NULL)

//======================================================================
// make attribute
//======================================================================
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, const wchar_t* value, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, const char* value, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, unsigned int value, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, double value, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduMakeAttribute(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* name, const SYSTEMTIME* time, SimpleErrorInfo* ei) throw (SimpleErrorInfo);

//======================================================================
// make empty element
//======================================================================
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo);

//======================================================================
// make element with text
//======================================================================
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, const wchar_t* value, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, const wchar_t* value, int valueLength, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, const char* value, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, const char* value, int valueLength, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, unsigned int value, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduMakeElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const wchar_t* name, double value, IXMLDOMElement** newElement, SimpleErrorInfo* ei) throw (SimpleErrorInfo);

//======================================================================
// add text to element
//======================================================================
HRESULT xduAddText(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* text, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduAddText(IXMLDOMDocument* dom, IXMLDOMElement* elm, const wchar_t* text, int textLength, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduAddText(IXMLDOMDocument* dom, IXMLDOMElement* elm, const char* text, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduAddText(IXMLDOMDocument* dom, IXMLDOMElement* elm, const char* text, int textLength, SimpleErrorInfo* ei) throw (SimpleErrorInfo);

//======================================================================
// make <ErrorInfo> element
//======================================================================
HRESULT xduMakeErrorInfoElement(IXMLDOMDocument* dom, IXMLDOMElement* parent, const SimpleErrorInfo& ei, SimpleErrorInfo* tei = xduDontSetSimpleErrorInfo) throw (SimpleErrorInfo);

//======================================================================
//
//======================================================================
HRESULT xduGetNodeTextIntrinsic(HRESULT hr, IXMLDOMNode* node, BSTR& bstr, SimpleErrorInfo* ei) throw (SimpleErrorInfo);
HRESULT xduGetNodeListTextIntrinsic(HRESULT hr, IXMLDOMNodeList* nodeList, BSTR& bstr, SimpleErrorInfo* ei) throw (SimpleErrorInfo);

template < class T >
HRESULT xduSelectNodesText(T* pnode, const wchar_t* xpath, BSTR& bstr, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	AutoComPtr<IXMLDOMNodeList> nodeList;
	HRESULT hr = pnode->selectNodes(const_cast<wchar_t*>(xpath), nodeList.r());
	return xduGetNodeListTextIntrinsic(hr, nodeList.p(), bstr, ei);
}

template < class T >
HRESULT xduSelectSingleNodeText(T* pnode, const wchar_t* xpath, BSTR& bstr, SimpleErrorInfo* ei) throw (SimpleErrorInfo)
{
	AutoComPtr<IXMLDOMNode> node;
	HRESULT hr = pnode->selectSingleNode(const_cast<wchar_t*>(xpath), node.r());
	return xduGetNodeTextIntrinsic(hr, node.p(), bstr, ei);
}

#ifdef DCM_OBJECTS_IS_IN
//======================================================================
//
//
//
//======================================================================
//======================================================================
//
//======================================================================
class Auto_DCM_CloseObject {
	DCM_OBJECT** m_object;
public:
	Auto_DCM_CloseObject(DCM_OBJECT** object) : m_object(object) {}
	~Auto_DCM_CloseObject() { DCM_CloseObject(m_object); }
};
//======================================================================
//
//======================================================================
CONDITION QueryStringToObject(const char* query_string, DCM_OBJECT** obj);
#endif

#pragma warn_unusedarg reset
#endif
