/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#pragma ANSI_strict off
#ifndef _ADOINT_H_
#include <ADOINT.H>
#endif
#pragma ANSI_strict reset


class CStudyRecordset;
class CStudyDBConnector;

class CStudyDB {
	friend class CStudyDBConnector;

	bool open(const BSTR connection_string);
	bool close(void);

	ADOConnection* m_conn;
	bool m_connected;

	static BSTR sm_table_name;

public:

	BSTR m_message;

	HRESULT m_hr;

	static BSTR make_insert_sql(
		const char* const volume_label,
		const char* const scp_title,
		const char* const scu_title,
		const char* const path,
		const char* const ref_file,
		const char* const pat_id,
		const char* const pat_name,
		const char* const std_date,
		const char* const std_time,
		const char* const std_id,
		const char* const modality);

	bool InsertStudy(
		bool& already_exists,
		const char* const volume_label,
		const char* const scp_title,
		const char* const scu_title,
		const char* const path,
		const char* const ref_file,
		const char* const pat_id,
		const char* const pat_name,
		const char* const std_date,
		const char* const std_time,
		const char* const std_id,
		const char* const modality
		);

	bool StudyExists(bool& exists, const char* volume_label, const char* path);
	bool study_exists(bool& exists, const BSTR sqlText);

	static BSTR make_check_sql(const char* volume_label, const char* path);
	static BSTR make_select_oldest_sql(const wchar_t* vol_label, int auto_delete_flag, const wchar_t* extra_cond = L"");

	CStudyDB();
	~CStudyDB();

	bool enabled(void) const { return m_conn != NULL; }
	bool connected(void) const { return m_connected; }

	bool SelectStudies(const BSTR sqlText, CStudyRecordset& srs); 
	bool ClearStudy(DWORD seq_no);

	bool Execute(const BSTR sql_command, ADORecordset** rsp = NULL);

	static const unsigned long study_tags[];

	static void Init(const tDWConfig* config);
};

class CStudyRecordset {
	ADORecordset* m_rs;
	ADOFields* m_fields;
	bool m_val_valid;
	VARIANT m_val;

	CStudyRecordset(const CStudyRecordset& rhs);
	CStudyRecordset& operator=(const CStudyRecordset& rhs);

public:

	HRESULT m_hr;

	CStudyRecordset();
	~CStudyRecordset();

	void SetReocrdset(ADORecordset* rs);

	bool End();
	bool MoveNext();
	VARIANT* Field(const char* colName);
};

class CStudyDBConnector {
	CStudyDB& m_studydb;

	CStudyDBConnector();
public:
	CStudyDBConnector(CStudyDB& studydb, const BSTR connection_string);
	~CStudyDBConnector();
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
template< class TADO >
static void getErrorInfo(TADO* obj, BSTR& message)
{
	if (message != NULL)
		SysFreeString(message);

	AutoComPtr<ADOErrors> errors;
	HRESULT hr = obj->get_Errors(errors.r());
	if (FAILED(hr))
		return;

	long errorCount = 0;
	errors->get_Count(&errorCount);

	VARIANT index;
	VariantInit(&index);
	index.vt = VT_I4;
	for (index.lVal = 0; index.lVal < errorCount; index.lVal++) {
		AutoComPtr<ADOError> error;
		hr = errors->get_Item(index, error.r());
		if (FAILED(hr))
			break;

		BSTR desc;
		hr = error->get_Description(&desc);
		if (FAILED(hr))
			break;
		if (message == NULL) {
			message = desc;
			continue;
		}
		AutoSysFreeString af_desc(desc);

		int mlen = SysStringLen(message);
		int dlen = SysStringLen(desc);

		BSTR new_msg = SysAllocStringLen(NULL, mlen + dlen);
		memcpy(new_msg, message, sizeof(wchar_t) * mlen);
		memcpy(new_msg + mlen, desc, sizeof(wchar_t) * dlen);
		new_msg[mlen + dlen] = L'\0';

		SysFreeString(message);
		message = new_msg;
	}
}
