/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "mbuf.h"

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
#define MBUF_COUNT 256

static mbuf_t* mbuf_pool;
static mbuf_t* mbuf_head;

static CRITICAL_SECTION mbuf_crit;

//========================================================================================
//
//========================================================================================
HRESULT mbuf_init(HANDLE hHeap)
{
	if (hHeap == NULL)
		hHeap = GetProcessHeap();

	mbuf_pool = (mbuf_t*)HeapAlloc(hHeap, 0, sizeof(mbuf_t) * MBUF_COUNT);
	if (mbuf_pool == NULL)
		return HRESULT_FROM_WIN32(GetLastError());

	for (int i = 0; i < MBUF_COUNT - 1; i++)
		mbuf_pool[i].next = &mbuf_pool[i + 1];
	mbuf_pool[MBUF_COUNT - 1].next = NULL;

	mbuf_head = mbuf_pool;

	InitializeCriticalSectionAndSpinCount(&mbuf_crit, 0x80000000 + 4096);

	return S_OK;
}

//========================================================================================
//
//========================================================================================
void mbuf_uninit(HANDLE hHeap)
{
	if (mbuf_pool == NULL)
		return;

	if (hHeap == NULL)
		hHeap = GetProcessHeap();
	HeapFree(hHeap, 0, mbuf_pool);
		
	DeleteCriticalSection(&mbuf_crit);
}

//========================================================================================
//
//========================================================================================
mbuf_t* mbuf_malloc(void)
{
	if (mbuf_pool == NULL)
		return NULL;

	CStackCriticalSection cs(&mbuf_crit);

	if (mbuf_head == NULL)
		return NULL;

	mbuf_t* mbuf = mbuf_head;
	mbuf_head = mbuf_head->next;

	mbuf->next = NULL;
	mbuf->size = 0;

	return mbuf;
}

//========================================================================================
//
//========================================================================================
void mbuf_free(mbuf_t* mbuf)
{
	if (mbuf_pool == NULL)
		return;

	CStackCriticalSection cs(&mbuf_crit);

	mbuf->next = mbuf_head;
	mbuf_head = mbuf;
}

//========================================================================================
//
//========================================================================================
int mbuf_free_count(void)
{
	if (mbuf_pool == NULL)
		return -1;

	CStackCriticalSection cs(&mbuf_crit);

	int count = 0;
	for (mbuf_t* mbuf = mbuf_head; mbuf != NULL; mbuf = mbuf->next)
		count++;

	return count;
}
