/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "cstudydb.h"
#include "dcmdict.h"

static bool gConsoleMode = false;

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void PrintMessages(const wchar_t** msg_tbl)
{
	int len = 0;
	for (const wchar_t** str = msg_tbl; *str != NULL; str++) {
		len += lstrlen(*str);
	}
	wchar_t* wstr = (wchar_t*)_alloca(sizeof(wchar_t) * (len + 1));
	*wstr = L'\0';
	for (const wchar_t** str = msg_tbl; *str != NULL; str++) {
		lstrcat(wstr, *str);
	}

	TOAC(wstr)
	DWORD bytesWritten;
	WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), _wstr_AC, _wstr_AC_len, &bytesWritten, NULL);

}
//========================================================================================
//
//
//
//========================================================================================
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
template < class tConfigObject >
static const tConfigObject* lookup_obj_by_label(const tConfigObject* tbl, int count, const char* label)
{
	const tConfigObject* obj = tbl;
	for (int i = 0; i < count; i++, obj++)
		if (strcmp(obj->label, label) == 0)
			return obj;
	return NULL;
}

//========================================================================================
//
//========================================================================================
const tVOLprm* tDWConfig::LookupVolumeByLabel(const char* const label)
{
	return lookup_obj_by_label<tVOLprm>(this->vol_tbl, this->num_vol, label);
}

//========================================================================================
//
//========================================================================================
const tSCPprm* tDWConfig::LookupSCPByLabel(const char* const label)
{
	return lookup_obj_by_label<tSCPprm>(this->scp_tbl, this->num_scp, label);
}

//========================================================================================
//
//========================================================================================
const tRemoteSCPprm* tDWConfig::LookupRemoteSCPByLabel(const char* const label)
{
	return lookup_obj_by_label<tRemoteSCPprm>(this->remotescp_tbl, this->num_remotescp, label);
}

//========================================================================================
//
//========================================================================================
const tSCUprm* tDWConfig::LookupSCUByLabel(const char* const label)
{
	return lookup_obj_by_label<tSCUprm>(this->scu_tbl, this->num_scu, label);
}
//========================================================================================
//
//
//
//========================================================================================
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
template < class tConfigObject >
static const tConfigObject* lookup_ae_by_title(const tConfigObject* tbl, const int count, const char* title)
{
	const tConfigObject* obj = tbl;
	for (int i = 0; i < count; i++, obj++)
		if (strcmp(obj->aetitle, title) == 0)
			return obj;
	return NULL;
}
//========================================================================================
//
//========================================================================================
const tSCPprm* tDWConfig::LookupSCPByTitle(const char* const title)
{
	return lookup_ae_by_title<tSCPprm>(this->scp_tbl, this->num_scp, title);
}

//========================================================================================
//
//========================================================================================
const tRemoteSCPprm* tDWConfig::LookupRemoteSCPByTitle(const char* const title)
{
	return lookup_ae_by_title<tRemoteSCPprm>(this->remotescp_tbl, this->num_remotescp, title);
}

//========================================================================================
//
//========================================================================================
const tSCUprm* tDWConfig::LookupSCUByTitle(const char* const title)
{
	return lookup_ae_by_title<tSCUprm>(this->scu_tbl, this->num_scu, title);
}

//========================================================================================
//
//
//
//========================================================================================
//========================================================================================
//
//========================================================================================
const tVOLprm* tDWConfig::LookupVolumeByLocation(const char* const loc, const char** cdir)
{
	const tVOLprm* vol = this->vol_tbl;
	for (int i = 0; i < this->num_vol; i++, vol++) {
		const char* voloc = vol->loc;
		int len = strlen(voloc);
		const char* const cp = loc + len;
		if (_strnicmp(voloc, loc, len) == 0
		 && (*cp == '/' || *cp == '\0')) {
		 	if (cdir != NULL)
		 		*cdir = *cp == '/' ? cp + 1 : cp;
		 	return vol;
		}
	}
	return NULL;
}

//========================================================================================
//
//
//
//========================================================================================
//========================================================================================
//
//========================================================================================
int tDWConfig::mapStoragePath(const char* loc, char* dir, const tVOLprm** volp)
{
	char* dp = dir;

	const char* chdir;
	if (volp != NULL)
		*volp = NULL;
	const tVOLprm* vol = LookupVolumeByLocation(loc, &chdir);
	if (vol == NULL)
		return 0;
	if (volp != NULL)
		*volp = vol;

	int len = strlen(vol->dir);
	if (dir != NULL)
		memcpy(dp, vol->dir, len);
	dp += len;

	if (dir != NULL)
		*dp = '\\';
	dp++;

	for (const char* cp = chdir; *cp != '\0'; cp++, dp++) {
		char c = *cp;
		if (c == '/')
			c = '\\';
		if (dir != NULL)
			*dp = toupper(c);
	}
	if (dir != NULL)
		*dp = '\0';

	return dp - dir;
}

//========================================================================================
//
//========================================================================================
int tDWConfig::mapStoragePathWC(const char* loc, wchar_t* dir, const tVOLprm** volp)
{
	wchar_t* dp = dir;

	const char* chdir;
	if (volp != NULL)
		*volp = NULL;
	const tVOLprm* vol = LookupVolumeByLocation(loc, &chdir);
	if (vol == NULL)
		return 0;
	if (volp != NULL)
		*volp = vol;

	int clen = strlen(vol->dir);
	int wlen = multi_byte_to_wide_char(vol->dir, clen, NULL, 0);
	if (dir != NULL)
		multi_byte_to_wide_char(vol->dir, clen, dp, wlen);
	dp += wlen;

	if (dir != NULL)
		*dp = L'\\';
	dp++;

	const char* lp = chdir;
	while (true) {
		const char* cp = strchr(lp, '/');
		if (cp == NULL) {
			wlen = multi_byte_to_wide_char(lp, -1, NULL, 0);
			if (dir != NULL)
				multi_byte_to_wide_char(lp, -1, dp, wlen);
			dp += wlen - 1;
			break;
		} else {
			wlen = multi_byte_to_wide_char(lp, cp - lp, NULL, 0);
			if (dir != NULL)
				multi_byte_to_wide_char(lp, cp - lp, dp, wlen);
			dp += wlen;
			if (dir != NULL)
				*dp = L'\\';
			dp++;
			lp = cp + 1;
		}
	}

	return dp - dir;
}

//========================================================================================
//
//========================================================================================
int tDWConfig::mapStoragePath(const wchar_t* loc, wchar_t* dir, const tVOLprm** volp)
{
	TOAC(loc)
	return mapStoragePathWC(_loc_AC, dir, volp);
}

//========================================================================================
//
// Registry Interface
//
//========================================================================================
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static HKEY openReg(DWORD access)
{
	HKEY hKey = NULL;
	long rv = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
	                       L"Software\\DIOWave",
	                       0,
	                       access,
	                       &hKey);
	if (rv == ERROR_FILE_NOT_FOUND && access == KEY_WRITE) {
		wchar_t* keyList[] = {
			L"SOFTWARE",
			L"DIOWave",
			NULL
		};
		HKEY hKeyNext;
		hKey = HKEY_LOCAL_MACHINE;
		wchar_t** keyStr = keyList;
		while (*keyStr) {
			DWORD result;
			rv = RegCreateKeyEx(hKey,
			                    *keyStr,
			                    0,
			                    NULL,
			                    REG_OPTION_NON_VOLATILE,
			                    KEY_WRITE,
			                    NULL,
			                    &hKeyNext,
			                    &result);
			if (rv != ERROR_SUCCESS) {
				return NULL;
			}
			rv = RegCloseKey(hKey);
			if (rv != ERROR_SUCCESS) {
				return NULL;
			}
			hKey = hKeyNext;
			keyStr++;
		}
		RegCloseKey(hKey);

		rv = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
	                      L"Software\\DIOWave",
	                      0,
	                      KEY_WRITE,
	                      &hKey);
	}
	if (rv != ERROR_SUCCESS)
		return NULL;

	return hKey;
} 

//===================================================================================
// Write Registry
//===================================================================================
bool RegisterConfigFilePath(const char* cfg_file)
{
	HKEY hKey = openReg(KEY_WRITE);
	if (hKey == NULL)
		return false;

	long rv;

	if (cfg_file != NULL) {
		if (strlen(cfg_file) <= 1) {
			rv = RegDeleteValue(hKey, REG_CONFIGFILE);
		} else {
			TOWC(cfg_file)
			wchar_t path[MAX_PATH];
			wchar_t* fname;
			DWORD len = GetFullPathName(_cfg_file_WC, sizeof(path), path, &fname);
			if (len <= 0 || len >= sizeof(path)) {
				if (gConsoleMode)
					printf("GetFullPathName(%s) failed\n", cfg_file);
				RegCloseKey(hKey);
				return false;
			}
			rv = RegSetValueEx(hKey, REG_CONFIGFILE, 0, REG_SZ, (BYTE*)path, lstrlen(path) + 1);
		}
		if (rv != ERROR_SUCCESS) {
			if (gConsoleMode) {
				char* lpMsgBuf;
				FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
				              NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
				              (LPTSTR)&lpMsgBuf, 0, NULL);
				
				printf("%s\n", lpMsgBuf);
				LocalFree(lpMsgBuf);
			}
			RegCloseKey(hKey);
			return false;
		}
	}

	rv = RegCloseKey(hKey);

	return true;
}

//===================================================================================
// Write Registry
//===================================================================================
void RemoveRegistry(void)
{
	RegDeleteKey(HKEY_LOCAL_MACHINE, L"Software\\DIOWave");
}

//-----------------------------------------------------------------------------------
// Read Registry
//-----------------------------------------------------------------------------------
static wchar_t* getConfigFilePath()
{
	HKEY hKey = openReg(KEY_READ);
	if (hKey == NULL)
		return NULL;

	wchar_t* cfg_file = NULL;

	DWORD valSize;
	DWORD valType;
	valSize = MAX_PATH;
	long rv = RegQueryValueEx(hKey, REG_CONFIGFILE, NULL, &valType, NULL, &valSize);
	if (rv == ERROR_SUCCESS) {
		cfg_file = (wchar_t*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, valSize);
		if (cfg_file != NULL) {
			RegQueryValueEx(hKey, REG_CONFIGFILE, NULL, &valType, (LPBYTE)cfg_file, &valSize);
		}
	}

	RegCloseKey(hKey);

	return cfg_file;
}

//===================================================================================
//
// Config file parser
//
//===================================================================================
//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static HRESULT get_node_text(IXMLDOMNode* pnode, wchar_t* path, BSTR* bstr)
{
	*bstr = NULL;

	HRESULT hr;

	IXMLDOMNodeList* nodeList = NULL;
	hr = pnode->selectNodes(path, &nodeList);
	if (FAILED(hr)) {
		if (gConsoleMode) {
			CWin32ErrorText et(hr);
			const wchar_t* msg_tbl[] = {
				L"FAILED: pnode->selectNodes(\"",
				path,
				L"\", ...);\r\n",
				et.str(),
				NULL
			};
			PrintMessages(msg_tbl);
		}
		return hr;
	}
	AutoComRelease<IXMLDOMNodeList> ac_nodeList(&nodeList);

	IXMLDOMNode* node = NULL;
	hr = nodeList->nextNode(&node);	// fetch first node
	if (FAILED(hr)) {
		if (gConsoleMode) {
			CWin32ErrorText et(hr);
			const wchar_t* msg_tbl[] = {
				L"FAILED: {",
				path,
				L"}->nextNode(...);\r\n",
				et.str(),
				NULL
			};
			PrintMessages(msg_tbl);
		}
		return hr;
	}
	if (node == NULL)
		return S_OK;
	AutoComRelease<IXMLDOMNode> ac_node(&node);

	hr = node->get_text(bstr);
	if (FAILED(hr)) {
		if (gConsoleMode) {
			CWin32ErrorText et(hr);
			const wchar_t* msg_tbl[] = {
				L"FAILED: {",
				path,
				L"}->get_text();\r\n",
				et.str(),
				NULL
			};
			PrintMessages(msg_tbl);
		}
		return hr;
	}

	return S_OK;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static unsigned short bstr2ushort(BSTR bstr)
{
	unsigned short result;
	VarUI2FromStr(bstr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), LOCALE_NOUSEROVERRIDE, &result);
	SysFreeString(bstr);
	return result;
}
//----------------------------------------------------------------------------------------
static unsigned long bstr2ulong(BSTR bstr)
{
	unsigned long result;
	VarUI4FromStr(bstr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), LOCALE_NOUSEROVERRIDE, &result);
	SysFreeString(bstr);
	return result;
}
//----------------------------------------------------------------------------------------
static long bstr2long(BSTR bstr)
{
	long result;
	VarI4FromStr(bstr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), LOCALE_NOUSEROVERRIDE, &result);
	SysFreeString(bstr);
	return result;
}
//----------------------------------------------------------------------------------------
static char* bstr2cstr(BSTR bstr)
{
	int blen = SysStringLen(bstr);
	int len = wide_char_to_multi_byte(bstr, blen, NULL, 0);
	char* cstr = (char*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, len + 1);
	wide_char_to_multi_byte(bstr, blen, cstr, len);
	cstr[len] = '\0';
	SysFreeString(bstr);
	return cstr;
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT tLabeledConfigObj::parse(IXMLDOMNode* node)
{
	BSTR bstr;
	HRESULT hr = get_node_text(node, L"@label", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->label = bstr2cstr(bstr);

	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tLabeledConfigObj::check(const wchar_t* const obj_name)
{
	bool err = false;

	//
	// Check Label
	//
	if (this->label == NULL) {
		err |= true;
		if (gConsoleMode) {
			wchar_t num[32];
			wsprintf(num, L"%d", this->index + 1);
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				num,
				L"th ",
				obj_name,
				L"has none or empty \"label\" attribute\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}

	return err;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tLabeledConfigObj::release()
{
	if (this->label != NULL)
		HeapFree(GetProcessHeap(), 0, this->label);
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
template < class tConfigObject >
bool tConfigObj_crossCheck(tConfigObject* obj_tbl, const int num_obj, const wchar_t* const obj_name)
{
	bool err = false;

	if (num_obj == 1)
		return err;

	tConfigObject* obj0 = obj_tbl;
	for (int i = 0; i < num_obj - 1; i++, obj0++) {
		tConfigObject* obj1 = obj0 + 1;
		for (int j = i + 1; j < num_obj; j++, obj1++) {
			if (strcmp(obj0->label, obj1->label) == 0) {
				err |= true;
				if (gConsoleMode) {
					wchar_t num1[32], num2[32];
					wsprintf(num1, L"%d", i + 1);
					wsprintf(num2, L"%d", j + 1);
					const wchar_t* msg_tbl[] = {
						L"CROSS CHECK: ",
						L"\"label\" attribute duplicates on <",
						obj_name,
						L"> of #",
						num1,
						L" and #",
						num2,
						L"\r\n",
						NULL
					};
					PrintMessages(msg_tbl);
				}
			}
		}
	}

	return err;
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT tDICOM_AE::parse(IXMLDOMNode* node)
{
	HRESULT hr = tLabeledConfigObj::parse(node);
	if (FAILED(hr))
		return hr;

	BSTR bstr;

	hr = get_node_text(node, L"ae-title", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->aetitle = bstr2cstr(bstr);

	return S_OK;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tDICOM_AE::check(const wchar_t* const obj_name)
{
	bool err = tLabeledConfigObj::check(obj_name);

	//
	// Check Label
	//
	if (this->aetitle == NULL) {
		err |= true;
		if (gConsoleMode) {
			const char* label = this->label;
			TOWC(label)
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<",
				obj_name,
				L" label=\"",
				_label_WC,
				L"\">/<ae-title>: empty or undefined\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}

	return err;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tDICOM_AE::release()
{
	tLabeledConfigObj::release();

	if (this->aetitle != NULL)
		HeapFree(GetProcessHeap(), 0, this->aetitle);
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
DWORD tVOLprm::logon()
{
	wchar_t* domain = L".";
	wchar_t* user = this->user;
	wchar_t* bs = StrChrW(user, L'\\');
	wchar_t* at = StrChrW(user, L'@');
	if (bs != NULL) {
		wchar_t* tmpstr = (wchar_t*)_alloca(sizeof(wchar_t) * (lstrlen(user) + 1));
		lstrcpy(tmpstr, user);
		*(tmpstr + (bs - user)) = L'\0';
		user = tmpstr + (bs - user) + 1;
		domain = tmpstr;
	} else if (at != NULL) {
		domain = NULL;
	}
	
	BSTR passwd = this->password == NULL ? (const BSTR)L"" : this->password;

#define _LOGON32_LOGON_NEW_CREDENTIALS_   9
#define _LOGON32_PROVIDER_WINNT50_    3

#if(_WIN32_WINNT >= 0x0500)
#if LOGON32_LOGON_NEW_CREDENTIALS != _LOGON32_LOGON_NEW_CREDENTIALS_ || LOGON32_PROVIDER_WINNT50 != _LOGON32_PROVIDER_WINNT50_
#error
#endif
#endif

	return LogonUser(user, domain, passwd, _LOGON32_LOGON_NEW_CREDENTIALS_, _LOGON32_PROVIDER_WINNT50_, &this->hToken) ? NO_ERROR : GetLastError();
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT tVOLprm::parse(IXMLDOMNode* node)
{
	HRESULT hr = tLabeledConfigObj::parse(node);
	if (FAILED(hr))
		return hr;

	BSTR bstr;

	hr = get_node_text(node, L"directory", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->dir = bstr2cstr(bstr);

	hr = get_node_text(node, L"location", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->loc = bstr2cstr(bstr);

	hr = get_node_text(node, L"auto-delete-flag", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->auto_delete_flag = bstr2long(bstr);

	hr = get_node_text(node, L"study-cache-flag", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->study_cache_flag = bstr2long(bstr);

	hr = get_node_text(node, L"free-space", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->free_space = bstr2long(bstr);

	hr = get_node_text(node, L"path-template", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->path_tmpl_str = bstr2cstr(bstr);

	hr = get_node_text(node, L"user", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->user = bstr;

	hr = get_node_text(node, L"password", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->password = bstr;

	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tVOLprm::check(tDWConfig* config)
{
	bool err = tLabeledConfigObj::check(L"VOLUME");

	const char* label = this->label;
	TOWC(label)

	//
	// Check Directory
	//
	const char* const dir = this->dir;
	TOWC(dir);		
	if (dir == NULL) {
		err |= true;
		if (gConsoleMode) {
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<VOLUME label=\"",
				_label_WC,
				L"\">/<directory>: empty or undefined\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	} else if (gConsoleMode) {
		DWORD flag = GetFileAttributes(_dir_WC);
		if (flag == 0xFFFFFFFF) {
			DWORD dwErr = GetLastError();
			err |= true;
			CWin32ErrorText et(dwErr);
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<VOLUME label=\"",
				_label_WC,
				L"\">/<directory>=",
				_dir_WC,
				L":\r\n",
				et.str(),
				NULL
			};
			PrintMessages(msg_tbl);
		} else if ((flag & FILE_ATTRIBUTE_DIRECTORY) == 0) {
			err |= true;
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<VOLUME label=\"",
				_label_WC,
				L"\">/<directory>=",
				_dir_WC,
				L": is not directory\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}

	//
	// Check Location
	//
	if (this->loc == NULL) {
		err |= true;
		if (gConsoleMode) {

			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<VOLUME label=\"",
				_label_WC,
				L"\">/<location>: is empty or undefined\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}

	//
	//
	//
	if (this->free_space == 0)
		this->free_space = kDefaultFreeSpace;


	//
	// Check PathTemplate
	//
	if (this->path_tmpl_str == NULL) {
		int len = strlen(kDefaultPathTemplate);
		this->path_tmpl_str = (char*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, len + 1);
		strcpy(this->path_tmpl_str, kDefaultPathTemplate);
	}

	err |= compile_path_template();

	return err;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tVOLprm::compile_path_template()
{
	bool err = false;

	CDcmDict dict;

	int item_count = 0;
	const char* cp = this->path_tmpl_str;
	while (*cp != '\0') {
		if (*cp == '\\') {
			cp++;
			item_count++;	
		} else if (*cp == '(') {
			const char* pp = strchr(++cp, ')');
			if (pp == NULL) {
				err |= true;
				break;
			}
			if (pp - cp != 8) {
				err |= true;
				if (gConsoleMode) {
					const char* path_tmpl_str = this->path_tmpl_str;
					TOWC(path_tmpl_str)
					const wchar_t* msg_tbl[] = {
						L"CHECK: ",
						L"Path Template Syntax Error\r\n",
						_path_tmpl_str_WC,
						NULL
					};
					PrintMessages(msg_tbl);
				}
				break;
			}
			char buf[9];
			memcpy(buf, cp, 8);
			buf[8] = '\0';
			cp = ++pp;
			DCM_TAG tag = strtoul(buf, NULL, 16);
			const DCMDICT* d = dict.Lookup(tag);
			if (d == NULL) {
				err |= true;
				if (gConsoleMode) {
					const char* path_tmpl_str = this->path_tmpl_str;
					TOWC(path_tmpl_str)
					const wchar_t* msg_tbl[] = {
						L"CHECK: ",
						L"Path Template Syntax Error\r\n",
						_path_tmpl_str_WC,
						NULL
					};
					PrintMessages(msg_tbl);
				}
				break;
			}
			item_count++;
		} else if (IsCharAlphaNumeric(*cp) || *cp == '_' || *cp == '-' || *cp == ' ') {
			cp++;
			while (*cp != '\0' && (IsCharAlphaNumeric(*cp) || *cp == '_' || *cp == '-' || *cp == ' '))
				cp++;
			item_count++;
		} else {
			err |= true;
			if (gConsoleMode) {
				const char* path_tmpl_str = this->path_tmpl_str;
				TOWC(path_tmpl_str)
				const wchar_t* msg_tbl[] = {
					L"CHECK: ",
					L"Path Template Syntax Error\r\n",
					_path_tmpl_str_WC,
					NULL
				};
				PrintMessages(msg_tbl);
			}
			break;
		}
	}

	if (err)
		return err;

	tPathTemplateItem* item_tbl = (tPathTemplateItem*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(tPathTemplateItem) * item_count);
	tPathTemplateItem* item = item_tbl;

	int depth = 1;
	cp = this->path_tmpl_str;
	while (*cp != '\0') {
		if (*cp == '\\') {
			cp++;
			item->type = tPathTemplateItem::eDelimiter;
			depth++;
			item++;
		} else if (*cp == '(') {
			const char* pp = strchr(++cp, ')');
			char buf[9];
			memcpy(buf, cp, 8);
			buf[8] = '\0';
			cp = ++pp;
			item->item.element.tag = strtoul(buf, NULL, 16);
			item->type = tPathTemplateItem::eElement;
			item++;
		} else if (IsCharAlphaNumeric(*cp) || *cp == '_' || *cp == '-' || *cp == ' ') {
			item->item.str.cp = cp;
			cp++;
			while (*cp != '\0' && (IsCharAlphaNumeric(*cp) || *cp == '_' || *cp == '-' || *cp == ' '))
				cp++;
			item->item.str.len = cp - item->item.str.cp;
			item->type = tPathTemplateItem::eString;
			item++;
		}
	}

	this->path_tmpl_item_count = item_count;
	this->path_tmpl_depth = depth;
	this->path_tmpl_item_tbl = item_tbl;

	return false;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tVOLprm::release()
{
	tLabeledConfigObj::release();

	HANDLE heap = GetProcessHeap();

	if (this->dir != NULL)
		HeapFree(heap, 0, this->dir);

	if (this->loc != NULL)
		HeapFree(heap, 0, this->loc);

	if (this->path_tmpl_str != NULL)
		HeapFree(heap, 0, this->path_tmpl_str);

	if (this->path_tmpl_item_tbl != NULL)
		HeapFree(heap, 0, this->path_tmpl_item_tbl);

	if (this->user != NULL)
		SysFreeString(this->user);

	if (this->password != NULL)
		SysFreeString(this->password);

	if (this->hToken != NULL)
		CloseHandle(hToken);

}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tVOLprm::crossCheck(tDWConfig* config)
{
	if (config->num_vol == 1)
		return false;

	bool err = tConfigObj_crossCheck<tVOLprm>(config->vol_tbl, config->num_vol, CFO_VOL);

	tVOLprm* vol0 = config->vol_tbl;
	for (int i = 0; i < config->num_vol - 1; i++, vol0++) {
		tVOLprm* vol1 = vol0 + 1;
		for (int j = i + 1; j < config->num_vol; j++, vol1++) {
			if (_stricmp(vol0->dir, vol1->dir) == 0) {
				err |= true;
				if (gConsoleMode) {
					const char* dir = vol0->dir;
					TOWC(dir)
					const char* label0 = vol0->label;
					TOWC(label0)
					const char* label1 = vol1->label;
					TOWC(label1)
					const wchar_t* msg_tbl[] = {
						L"CROSS CHECK: ",
						L"<directory>=",
						_dir_WC,
						L": of <VOLUME label=\"",
						_label0_WC,
						L"\"> and <VOLUME label=\"",
						_label1_WC,
						L"\" is the same\r\n",
						NULL
					};
					PrintMessages(msg_tbl);
				}
			}
			if (_stricmp(vol0->loc, vol1->loc) == 0) {
				err |= true;
				if (gConsoleMode) {
					const char* loc = vol0->loc;
					TOWC(loc)
					const char* label0 = vol0->label;
					TOWC(label0)
					const char* label1 = vol1->label;
					TOWC(label1)
					const wchar_t* msg_tbl[] = {
						L"CROSS CHECK: ",
						L"<location>=",
						_loc_WC,
						L": of <VOLUME label=\"",
						_label0_WC,
						L"\"> and <VOLUME label=\"",
						_label1_WC,
						L"\" is the same\r\n",
						NULL
					};
					PrintMessages(msg_tbl);
				}
			}	
		}
	}

	return err;
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT tSCPprm::parse(IXMLDOMNode* node)
{
	HRESULT hr = tDICOM_AE::parse(node);
	if (FAILED(hr))
		return hr;

	BSTR bstr;

	hr = get_node_text(node, L"port", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->port = bstr2ushort(bstr);

	hr = get_node_text(node, L"volume", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->volume_label = bstr2cstr(bstr);

	hr = get_node_text(node, L"transfer-syntax-flag", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->transfer_syntax_flag = bstr2long(bstr);

	hr = get_node_text(node, L"read-timeout", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->read_timeout = bstr2long(bstr);

	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tSCPprm::check(tDWConfig* config)
{
	bool err = tDICOM_AE::check(L"SCP");

	const char* label = this->label;
	TOWC(label)

	//
	// Check <port>
	//
	if (this->port < 0 || this->port > 0xFFFF) {
		err = true;
		if (gConsoleMode) {
			wchar_t num[32];
			wsprintf(num, L"%d", this->port);
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<SCP label=\"",
				_label_WC,
				L"\">/<port>=",
				num,
				L": bad number\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}

	//
	// Check <volume>
	//
	const char* volume_label = this->volume_label;
	TOWC(volume_label)
	if (this->volume_label == NULL || strlen(this->volume_label) == 0) {
		err = true;
		if (gConsoleMode) {
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<SCP label=\"",
				_label_WC,
				L"\">/<volume>=",
				_volume_label_WC,
				L": empty or undefined\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}
	const tVOLprm* vol = config->LookupVolumeByLabel(this->volume_label);
	if (vol == NULL) {
		err = true;
		if (gConsoleMode) {
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<SCP label=\"",
				_label_WC,
				L"\">/<volume>=",
				_volume_label_WC,
				L": not found in <VOLUME> elements\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	} else {
		this->volume = vol;
	}
	
	//
	// Check <read-timeout>
	//
	if (this->read_timeout == 0)
		this->read_timeout = kDefaultReadTimeout;

	return err;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tSCPprm::crossCheck(tDWConfig* config)
{
	if (config->num_scp == 1)
		return false;

	bool err = tConfigObj_crossCheck<tSCPprm>(config->scp_tbl, config->num_scp, L"SCP");

	tSCPprm* scp0 = config->scp_tbl;
	for (int i = 0; i < config->num_scp - 1; i++, scp0++) {
		tSCPprm* scp1 = scp0 + 1;
		for (int j = i + 1; j < config->num_scp; j++, scp1++) {
			if (scp0->port == scp1->port) {
				err |= true;
				if (gConsoleMode) {
					wchar_t num[32];
					wsprintf(num, L"%d", scp0->port);
					const char* label0 = scp0->label;
					TOWC(label0)
					const char* label1 = scp1->label;
					TOWC(label1)
					const wchar_t* msg_tbl[] = {
						L"CROSS CHECK: ",
						L"<port>=",
						num,
						L": of <SCP label=\"",
						_label0_WC,
						L"\"> and <SCP label=\"",
						_label1_WC,
						L"\" is the same\r\n",
						NULL
					};
					PrintMessages(msg_tbl);
				}
			}
		}
	}

	return err;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tSCPprm::release()
{
	tDICOM_AE::release();

	if (this->volume_label != NULL)
		HeapFree(GetProcessHeap(), 0, this->volume_label);
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT tRemoteSCPprm::parse(IXMLDOMNode* node)
{
	HRESULT hr = tDICOM_AE::parse(node);
	if (FAILED(hr))
		return hr;

	BSTR bstr;

	hr = get_node_text(node, L"host", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->host = bstr2cstr(bstr);

	hr = get_node_text(node, L"port", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->port = bstr2ushort(bstr);

	return S_OK;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tRemoteSCPprm::check(tDWConfig* config)
{
	bool err = tDICOM_AE::check(L"REMOTE-SCP");

	const char* label = this->label;
	TOWC(label)

	//
	// Check <host>
	//
	if (this->host == NULL || strlen(this->host) == 0) {
		err = true;
		if (gConsoleMode) {
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<REMOTE-SCP label=\"",
				_label_WC,
				L"\">/<host>: is empty or undefined\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}


	//
	// Check <port>
	//
	if (this->port < 0 || this->port > 0xFFFF) {
		err = true;
		if (gConsoleMode) {
			wchar_t num[32];
			wsprintf(num, L"%d", this->port);
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<REMOTE-SCP label=\"",
				_label_WC,
				L"\">/<port>=",
				num,
				L": bad number\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}

	return err;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tRemoteSCPprm::crossCheck(tDWConfig* config)
{
	if (config->num_remotescp == 1)
		return false;

	bool err = tConfigObj_crossCheck<tRemoteSCPprm>(config->remotescp_tbl, config->num_remotescp, L"REMOTE-SCP");

	return err;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tRemoteSCPprm::release()
{
	tDICOM_AE::release();

	if (this->host != NULL)
		HeapFree(GetProcessHeap(), 0, this->host);
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT tSCUprm::parse(IXMLDOMNode* node)
{
	HRESULT hr = tDICOM_AE::parse(node);
	if (FAILED(hr))
		return hr;

	BSTR bstr;

	hr = get_node_text(node, L"service-class", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->service_class_str = bstr2cstr(bstr);

	hr = get_node_text(node, L"read-timeout", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->read_timeout = bstr2long(bstr);

	return S_OK;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tSCUprm::check(tDWConfig* config)
{
	bool err = tDICOM_AE::check(L"SCU");

	const char* label = this->label;
	TOWC(label)

	//
	// check <service-class>
	//
	if (this->service_class_str == NULL || strlen(this->service_class_str) == 0) {
		err |= true;
		if (gConsoleMode) {
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<SCU label=\"",
				_label_WC,
				L"\">/<service-class>: is empty or undefined\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	} else {
		if (_stricmp("c-store", this->service_class_str) == 0) {
			this->service_class |= scCSTORE;
		} else if (_stricmp("c-find", this->service_class_str) == 0) {
			this->service_class |= scCFIND;
		} else if (_stricmp("c-move", this->service_class_str) == 0) {
			this->service_class |= scCMOVE;
		} else {
			err |= true;
			if (gConsoleMode) {
				const wchar_t* msg_tbl[] = {
					L"CHECK: ",
					L"<SCU label=\"",
					_label_WC,
					L"\">/<service-class>: should be one of C-STRE, C-FIND and C-MOVE\r\n",
					NULL
				};
				PrintMessages(msg_tbl);
			}
		}
	}

	//
	// check <read-timeout>
	//
	if (this->read_timeout == 0)
		this->read_timeout = kDefaultReadTimeout;

	return err;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tSCUprm::crossCheck(tDWConfig* config)
{
	if (config->num_scu == 1)
		return false;

	bool err = tConfigObj_crossCheck<tSCUprm>(config->scu_tbl, config->num_scu, L"SCU");
	
	return err;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tSCUprm::release()
{
	tDICOM_AE::release();

	if (this->service_class_str != NULL)
		HeapFree(GetProcessHeap(), 0, this->service_class_str);
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT tODBCprm::parse(IXMLDOMDocument* document)
{
	BSTR bstr;

	HRESULT hr = get_node_text(document, L"/DIOWaveConfiguration/DATABASE/data-source", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->dsn = bstr;

	hr = get_node_text(document, L"/DIOWaveConfiguration/DATABASE/table", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->table = bstr;

	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tODBCprm::check()
{
	bool err = false;
	
	if (this->dsn == NULL)
		return false;
/*
		err |= true;
		if (gConsoleMode) {
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"<DATABASE>/<data-soruce>: is empty or undefined\r\n",
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}
*/
	if (this->table == NULL) {
		this->table = SysAllocString(kDefaultTableName);
	}

	if (gConsoleMode) {
		CStudyDB studydb;
		if (FAILED(studydb.m_hr)) {
			err |= true;
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"FAILED: CStudyDB construction \r\n",
				studydb.m_message,
				NULL
			};
			PrintMessages(msg_tbl);
		}

		CStudyDBConnector conn(studydb, this->dsn);

		if (!studydb.connected()) {
			err |= true;
			const wchar_t* msg_tbl[] = {
				L"CHECK: ",
				L"FAILED: connect to \"",
				this->dsn,
				L"\"\r\n",
				studydb.m_message,
				NULL
			};
			PrintMessages(msg_tbl);
		}
	}

	return err;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tODBCprm::release(void)
{
	if (dsn != NULL)
		SysFreeString(dsn);

	if (table != NULL)
		SysFreeString(table);
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT tIISprm::parse(IXMLDOMDocument* document)
{
	BSTR bstr;

	HRESULT hr = get_node_text(document, L"/DIOWaveConfiguration/IIS/version", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		this->version = bstr2ulong(bstr);

	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
bool tIISprm::check()
{
	if (this->version == 0)
		this->version = kDefaultIISVersion;
	return false;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tIISprm::release(void)
{
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
template < class tConfigObject >
HRESULT parse_config_object(IXMLDOMDocument* document, wchar_t* path, tConfigObject** pObj, unsigned long* numObj)
{
	HRESULT hr;

	*pObj = NULL;
	*numObj = 0;

	IXMLDOMNodeList* nodeList = NULL;
	hr = document->selectNodes(path, &nodeList);
	if (FAILED(hr)) {
		if (gConsoleMode) {
			CWin32ErrorText et(hr);
			const wchar_t* msg_tbl[] = {
				L"FAILED: document->selectNodes(\"",
				path,
				L"\", &nodeList);\r\n",
				et.str(),
				NULL
			};
			PrintMessages(msg_tbl);
		}
		return hr;
	}
	AutoComRelease<IXMLDOMNodeList> ac_nodeList(&nodeList);

	long listLength = 0;
	nodeList->get_length(&listLength);
	if (listLength == 0)
		return S_OK;
	
	tConfigObject* obj_tbl = (tConfigObject*)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(tConfigObject) * listLength);
	unsigned long index = 0;
	while (true) {
		IXMLDOMNode* node = NULL;
		hr = nodeList->nextNode(&node);
		if (FAILED(hr)) {
			if (gConsoleMode) {
				CWin32ErrorText et(hr);
				const wchar_t* msg_tbl[] = {
					L"FAILED: nodeList->nextNode(&node);\r\n",
					et.str(),
					NULL
				};
				PrintMessages(msg_tbl);
			}
			return hr;
		}
		if (node == NULL)
			break;
		AutoComRelease<IXMLDOMNode> ac_node(&node);
		tConfigObject* obj = obj_tbl + index;
		memset(obj, 0, sizeof(tConfigObject));
		hr = obj->parse(node);
		if (FAILED(hr))
			return hr;
		obj->index = index++;
	}

	*pObj = obj_tbl;
	*numObj = index;
	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
template< class tConfigObject >
bool check_config_object(tDWConfig* config, tConfigObject* objTbl, unsigned long numObj)
{
	if (numObj == 0)
		return false;

	bool err = false;

	tConfigObject* obj = objTbl; 
	for (int index = 0; index < numObj; index++, obj++) {
		err |= obj->check(config);
		if (err && !gConsoleMode)
			break;
	}			
	if (err && !gConsoleMode)
		return err;
	err |= tConfigObject::crossCheck(config);
	if (err && !gConsoleMode)
		return err;

	return err;
}

//----------------------------------------------------------------------------------------
// Load as DOM
//----------------------------------------------------------------------------------------
static HRESULT load_config(const char* cfg_file, IXMLDOMDocument** pDocument)
{
	HRESULT hr;

	const wchar_t* cf;

	wchar_t* reg_cfg_file = NULL;

	if (cfg_file == NULL) {
		reg_cfg_file = getConfigFilePath();
		if (reg_cfg_file != NULL) {
			int len = lstrlen(reg_cfg_file);
			cf = (wchar_t*)_alloca(sizeof(wchar_t) * (len + 1));
			memcpy(const_cast<wchar_t*>(cf), reg_cfg_file, sizeof(wchar_t) * (len + 1));
			HeapFree(GetProcessHeap(), 0, reg_cfg_file);
		} else {
			cf = kDefaultConfigFile;
		}
	} else {
		TOWC(cfg_file)
		cf = _cfg_file_WC;
	}

	if (gConsoleMode) {
		const wchar_t* msg_tbl[] = {
			L"Config File = \"",
			cf,
			L"\"\r\n",
			NULL
		};
		PrintMessages(msg_tbl);
	}

	DWORD attr = GetFileAttributes(cf);
	if (attr == 0xFFFFFFFF) {
		hr = GetLastError();
		if (gConsoleMode) {
			CWin32ErrorText et(hr);
			const wchar_t* msg_tbl[] = {
				L"FAILED: GetFileAttributes(\"",
				cf,
				L"\");\r\n",
				et.str(),
				NULL
			};
			PrintMessages(msg_tbl);
		}
		return hr;
	}

	BSTR cf_bstr = SysAllocString(cf);
	AutoSysFreeString af_cf_bstr(cf_bstr);
	
	VARIANT var;
	VariantInit(&var);
	var.vt = VT_BSTR;
	var.bstrVal = cf_bstr;

	CLSID clsidDOMDocument;
	hr = CLSIDFromProgID(L"Msxml2.FreeThreadedDOMDocument", &clsidDOMDocument);
	if (FAILED(hr)) {
		if (gConsoleMode) {
			CWin32ErrorText et(hr);
			const wchar_t* msg_tbl[] = {
				 L"FAILED: CLSIDFromProgID(\"Msxml2.FreeThreadedDOMDocument\", ...);\r\n",
				 et.str(),
				 NULL
			};
			PrintMessages(msg_tbl);
		}
		return hr;
	}
	
	IXMLDOMDocument* document = NULL;
	hr = CoCreateInstance(clsidDOMDocument, NULL, CLSCTX_INPROC_SERVER, IID_IXMLDOMDocument, (void**)&document);
	if (FAILED(hr)) {
		if (gConsoleMode) {
			CWin32ErrorText et(hr);
			const wchar_t* msg_tbl[] = {
				L"FAILED: CoCreateInstance(clsidDOMDocument, NULL, CLSCTX_INPROC_SERVER, IID_IXMLDOMDocument2, (void**)&document);\r\n",
				et.str(),
				NULL
			};
			PrintMessages(msg_tbl);
		}
		return hr;
	}
	AutoComRelease<IXMLDOMDocument> ac_document(&document);

	VARIANT_BOOL loadSucceed = VARIANT_FALSE;
	hr = document->load(var, &loadSucceed);
	if (FAILED(hr) || loadSucceed == VARIANT_FALSE) {
		if (gConsoleMode) {
			CWin32ErrorText et(hr);
			const wchar_t* msg_tbl[] = {
				L"FAILED: document->load(\"",
				cf_bstr,
				L"\");\r\n",
				et.str(),
				NULL
			};
			PrintMessages(msg_tbl);
		}
		return hr;
	}

	*pDocument = document;
	document->AddRef();

	return S_OK;
}

//===================================================================================
// Parse Main Body
//===================================================================================
HRESULT ReadConfig(ptDWConfig& config, const char* cfg_file, const bool consoleMode)
{
	gConsoleMode = consoleMode;

	HRESULT hr;

	config = (ptDWConfig)HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, sizeof(tDWConfig));
	config->size = sizeof(tDWConfig);

	IXMLDOMDocument* document;
	hr = load_config(cfg_file, &document);
	if (FAILED(hr))
		return hr;
	AutoComRelease<IXMLDOMDocument> ac_document(&document);

	BSTR bstr = NULL;
	AutoSysFreeString afs_bstr(bstr);

	hr = get_node_text(document, L"/DIOWaveConfiguration/LOG/directory", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		config->logDir = bstr;

	hr = get_node_text(document, L"/DIOWaveConfiguration/LOG/rtlog", &bstr);
	if (FAILED(hr))
		return hr;
	if (bstr != NULL)
		config->rtlog = bstr2long(bstr);

	hr = config->odbc.parse(document);
	if (FAILED(hr))
		return hr;

	hr = config->iis.parse(document);
	if (FAILED(hr))
		return hr;

	hr = parse_config_object<tVOLprm>(document, L"/DIOWaveConfiguration/VOLUME", &config->vol_tbl, &config->num_vol);
	if (FAILED(hr)) {
		return hr;
	}
	hr = parse_config_object<tSCPprm>(document, L"/DIOWaveConfiguration/SCP", &config->scp_tbl, &config->num_scp);
	if (FAILED(hr)) {
		return hr;
	}
	hr = parse_config_object<tRemoteSCPprm>(document, L"/DIOWaveConfiguration/REMOTE-SCP", &config->remotescp_tbl, &config->num_remotescp);
	if (FAILED(hr)) {
		return hr;
	}
	hr = parse_config_object<tSCUprm>(document, L"/DIOWaveConfiguration/SCU", &config->scu_tbl, &config->num_scu);
	if (FAILED(hr)) {
		return hr;
	}

	bool err = false;

	err |= check_config_object<tVOLprm>(config, config->vol_tbl, config->num_vol);
	if (err && !gConsoleMode)
		return E_FAIL;

	err |= check_config_object<tSCPprm>(config, config->scp_tbl, config->num_scp);
	if (err && !gConsoleMode)
		return E_FAIL;

	err |= check_config_object<tRemoteSCPprm>(config, config->remotescp_tbl, config->num_remotescp);
	if (err && !gConsoleMode)
		return E_FAIL;

	err |= check_config_object<tSCUprm>(config, config->scu_tbl, config->num_scu);
	if (err && !gConsoleMode)
		return E_FAIL;

	//
	//
	//
	if (config->logDir == NULL) {
		config->logDir = SysAllocString(kDefaultLogDirectory);
	}
	//
	//
	//
	err |= config->odbc.check();

	//
	//
	//
	err |= config->iis.check();

	//
	// 
	//
	if (err)
		return E_FAIL;

	return S_OK;
}

//===================================================================================
// Parse Main Body
//===================================================================================
HRESULT ReleaseConfigData(tDWConfig* config)
{
	if (config->logDir != NULL)
		SysFreeString(config->logDir);

	config->odbc.release();
	config->iis.release();

	for (int i = 0; i < config->num_vol; i++)
		config->vol_tbl[i].release();

	for (int i = 0; i < config->num_scp; i++)
		config->scp_tbl[i].release();

	for (int i = 0; i < config->num_remotescp; i++)
		config->remotescp_tbl[i].release();

	for (int i = 0; i < config->num_scu; i++)
		config->scu_tbl[i].release();

	return S_OK;
}

