#ifndef DW_CONFIG_IS_IN
#define DW_CONFIG_IS_IN 1
/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

struct tDWConfig;

/* default */
#define kDefaultBaseDirectory "C:\\DIOwave"
// #define kDefaultConfigFile "C:\\DIOWave\\etc\\DIOWave.cf"
#define kDefaultConfigFile L"C:\\DIOWave\\etc\\DIOWaveConf.xml"
#define kDefaultLogDirectory L"C:\\DIOWave\\logs"
#define kDefaultHTMLDirectory "C:\\DIOWave\\html"
#define kDefaultTempDirectory "C:\\DIOWave\\tmp"
#define kDefaultStudyCacheDirectory L"C:\\DIOWave\\cache"
// #define kDefaultODBCDSN L"DICOMSTUDY"
#define kDefaultTableName L"DICOMSTUDY"
#define kDefaultReadTimeout (60 * 1000)
#define kDefaultIISVersion 400
#define kDIOStoreReqFileKey "DSR"
#define kDIOStoreReqFilePattern "DSR*.tmp"
#define kDIORetrieveReqFileKey "DRR"
#define kDefaultPathTemplate "(0020000D)\\(0020000E)\\(00080018)"
#define kDefaultFreeSpace (1024 * 1024 * 16)
/*
//----------------------------------------------------------------------------------------
// Registory Values in "HKEY_LOCAL_MACHINE\DIOWave"
//----------------------------------------------------------------------------------------
*/
#define REG_CONFIGFILE L"ConfigurationFile"

/*
//----------------------------------------------------------------------------------------
// Configuration Values
//----------------------------------------------------------------------------------------
*/

/* OBJECT */
#define CFO_EXIT "EXIT"
#define CFO_LOG "LOG"
#define CFO_HTML "HTML"
#define CFO_VOL L"VOLUME"
#define CFO_ODBC "ODBC"
#define CFO_SCP L"SCP"
#define CFO_REMOTE_SCP "REMOTE-SCP"
#define CFO_SCU "SCU"
#define CFO_IIS "IIS"

/* Common Properties */
#define CFP_COUNT "Count"
#define CFP_LABEL "Label"

/* Common Properties for DICOM Applications */
#define CFP_AETITLE "AETitle"
#define CFP_PORT "Port"
#define CFP_READTIMEOUT "ReadTimeout"
#define CFP_HOST "HOST"
#define CFP_SERVICE "ServiceClassFlag"

/* Properties for VOLUME */
#define CFP_DIR "Directory"
#define CFP_LOC "Location"
#define CFP_AUTODELETE "AutoDeleteFlag"
#define CFP_TEMPLATE "FilePathTemplate"
#define CFP_STUDYCACHE "StudyCacheFlag"

/* Propaties for SCP */
#define CFP_REMOVEPRIVATE "RemovePrivateFlag"
#define CFP_TRANSFERSYNTAX "TransferSyntaxFlag"
#define CFP_ACCEPTSCU "AcceptSCU"

/* Proparties for REMOTE-SCP */
#define CFP_GROUPLENGTH "GroupLengthFlag"

/* Properties for SCU */

/* Propaties for ODBC */
#define CFP_DSN "DSN"
#define CFP_TABLE "Table"

/* Properties for IIS */
#define CFP_VERSION "Version"

/* Properties for LOG */
#define CFP_DIO_IMAGE_LOG L"dio-image.log"

/* bit mask*/
/* transfer syntax */
enum {
	tsImplicitVRLittleEndian = 0,
	tsExplicitVRLittleEndian
};

/* bit mask for .service_class_flag */
enum {
	scCSTORE = 1,
	scCFIND = 2,
	scCMOVE = 4
};

/* AutoDeleteFlag */
enum {
	adNone = -1,
	adStudyDT = 0,
	adLastRecieve,
	adLastAccess
};
//
//
//
struct tLabeledConfigObj {

	HRESULT parse(IXMLDOMNode* node);
	bool check(const wchar_t* const obj_name);
	void release(void);
	static bool crossCheck(tLabeledConfigObj* obj_tbl, const int num_obj, const char* const obj_name);

	unsigned long index;
	char* label;	// Attribute: label
};

//
//
//
struct tPathTemplateItem {
	enum {
		eDelimiter = 1,
		eString = 2,
		eElement = 3
	} type;
	union {
		struct {
			const char* cp;
			unsigned long len;
		} str;
		struct {
			DCM_TAG tag;
		} element;
	} item;
};

//
// Element: VOLUME
//
struct tVOLprm : tLabeledConfigObj {

	HRESULT parse(IXMLDOMNode* node);
	bool check(tDWConfig* config);
	void release(void);
	static bool crossCheck(tDWConfig* config);

	char* dir; // Child Element: directory
	char* loc; // Child Element: location

	long auto_delete_flag; // Child Element: auto-delete-flag

	long study_cache_flag; // Child Element: study-cache-flag

	long free_space; // Child Element: free-space

	char* path_tmpl_str; // Child Element: path-template
	int path_tmpl_item_count;
	int path_tmpl_depth;
	tPathTemplateItem* path_tmpl_item_tbl;
	bool compile_path_template();

	DWORD logon();
	BSTR user;
	BSTR password;
	HANDLE hToken;
};


//
//
//
struct tDICOM_AE : tLabeledConfigObj {

	HRESULT parse(IXMLDOMNode* node);
	bool check(const wchar_t* const obj_name);
	void release(void);

	char* aetitle; // Child Element: ae-title
};

//
// Element: SCP
//
struct tSCPprm : tDICOM_AE {

	HRESULT parse(IXMLDOMNode* node);
	bool check(tDWConfig* config);
	void release(void);
	static bool crossCheck(tDWConfig* config);

	unsigned short port; // Child Element: port

	char* volume_label; // Child Element: volume
	const tVOLprm* volume;

	long transfer_syntax_flag; // Child Element: transfer-syntax-flag

	long read_timeout; // Child Element: read-timeout
};

//
// Element: REMOTE-SCP
//
struct tRemoteSCPprm : tDICOM_AE {

	HRESULT parse(IXMLDOMNode* node);
	bool check(tDWConfig* config);
	void release(void);
	static bool crossCheck(tDWConfig* config);

	char* host; // Child Element: host
	unsigned short port; // Child Element: port
};

//
// Element: SCU
//
struct tSCUprm : tDICOM_AE {

	HRESULT parse(IXMLDOMNode* node);
	bool check(tDWConfig* config);
	void release(void);
	static bool crossCheck(tDWConfig* config);

	char* service_class_str; // Child Element: service-class
	int service_class;
	long read_timeout; // Child Element: read-timeout
};

//
// Element: DATABASE
//
struct tODBCprm {
	HRESULT parse(IXMLDOMDocument* dom);
	bool check(void);
	void release(void);

	BSTR dsn; // Child Element: data-source
	BSTR table; // Child Element: table
};

//
// Element: IIS
//
struct tIISprm {
	HRESULT parse(IXMLDOMDocument* dom);
	bool check(void);
	void release(void);

	DWORD version; // Child Element: version
};

//
//
//
struct tDWConfig {
	DWORD size;
	
	/* LOG Object */
	BSTR logDir;
	DWORD rtlog;	// size if KB

	/* ODBC Object */
	tODBCprm odbc;

	/* IIS Object */
	tIISprm iis;

	/* VOLUME Objects */
	unsigned long num_vol;
	tVOLprm* vol_tbl;

	/* SCP Objects */
	unsigned long num_scp;
	tSCPprm* scp_tbl;

	/* RemoteSCP Objects */
	unsigned long num_remotescp;
	tRemoteSCPprm* remotescp_tbl;

	/* SCU Objects */
	unsigned long num_scu;
	tSCUprm* scu_tbl;	

	//
	const tVOLprm* LookupVolumeByLabel(const char* const label);
	const tSCPprm* LookupSCPByLabel(const char* const label);
	const tRemoteSCPprm* LookupRemoteSCPByLabel(const char* const label);
	const tSCUprm* LookupSCUByLabel(const char* const label);
	//
	const tSCPprm* LookupSCPByTitle(const char* const title);
	const tRemoteSCPprm* LookupRemoteSCPByTitle(const char* const title);
	const tSCUprm* LookupSCUByTitle(const char* const title);
	//
	const tVOLprm* LookupVolumeByLocation(const char* const loc, const char** cdir = NULL);
	//
	int mapStoragePath(const char* loc, char* path, const tVOLprm** vol = NULL);
	int mapStoragePath(const wchar_t* loc, wchar_t* path, const tVOLprm** vol = NULL);
	int mapStoragePathWC(const char* loc, wchar_t* path, const tVOLprm** vol = NULL);
};

typedef tDWConfig* ptDWConfig;

bool RegisterConfigFilePath(const char* cfg_file);
void RemoveRegistry(void);

HRESULT ReadConfig(ptDWConfig& config, const char* cfg_file = NULL, const bool consoleMode = false);
HRESULT ReleaseConfigData(tDWConfig* config);

/*----------------------------------------------------------------------------------------
// Service Class User Parameter Preamble
//--------------------------------------------------------------------------------------*/
struct tPIDFile {
	DWORD dwProcessId;
	char cRequestMap[MAX_PATH];
};

/*----------------------------------------------------------------------------------------
// Service Class User Parameter Preamble
//--------------------------------------------------------------------------------------*/
union tSCUPreamble {
	char preamble[128];
	struct {
		char cFileName[32];	/* req file name */
		void* next;
	    FILETIME ftCreationTime;
	    FILETIME ftLastAccessTime; 
	    FILETIME ftLastWriteTime;
	    DWORD dwProcessId;
	    DWORD dwDone;
	} header;
};

struct tSCUAssocParams {
	char scu_label[32];
	char scu_title[24];

	char target_label[32];
	char target_host[256];
	short target_port;
	char target_title[24];

	char class_uid[66];
	char transfer_uid[66];
};

struct tSCURequest {
	enum {
		eSCUSuccess = 0,
		eSCUStartWaiting = 1,
		eSCUBlocked = 2,
		eSCUStarted = 3,
		eSCUAbend = 0xFFFF,
	};
	tSCUPreamble preamble;
	tSCUAssocParams ap;
}; 

/*----------------------------------------------------------------------------------------
// Store Parmater (parameter for Storage Service Class User)
//--------------------------------------------------------------------------------------*/
struct tStoreImageRequest {
	DWORD size;
	DWORD options;
	CONDITION cond;
	char transfer_uid[68];
	char class_uid[68];
	char instance_uid[68];
	char path[1];
};

struct tStoreRequest : tSCURequest {
	enum {
		eStoreRefused = 0xA700,
		eStoreError1 = 0xA900,
		eStoreError2 = 0xC000,
		eStoreWarning1 = 0xB000,
		eStoreWarning2 = 0xB007,
		eStoreWarning3 = 0xB006,
		eStoreRunning = 0xFF00,
	};

	unsigned short status;
	unsigned long numImages;
	unsigned long numRemain;
	unsigned long numSucceeded;
	unsigned long numFailed;
	unsigned long numWarning;
    char errorComment[DICOM_LO_LENGTH + 1];

	char note[256];
	char message[1024];

	tStoreImageRequest images[1];
};

/*----------------------------------------------------------------------------------------
// Retrieve Parmater (parameter for Retrieve Service Class User)
//--------------------------------------------------------------------------------------*/

struct tRetrieveRequest  : tSCURequest {
	enum {
		eMovePending = 0xFF00,
		eMoveRefused1 = 0xA701,
		eMoveRefused2 = 0xA702,
		eMoveRefused3 = 0xA801,
		eMoveFailed1 = 0xA900,
		eMoveFailed2 = 0xC000,
		eMoveCancel = 0xFE00,
		eMoveWarning = 0xB000
	};
	tSCUPreamble preamble;
	tSCUAssocParams ap;
	char move_destination_title[24];

	// from MSG_C_MOVE_RESP
    unsigned short status;
    unsigned short remainingSubOperations;
    unsigned short completedSubOperations;
    unsigned short failedSubOperations;
    unsigned short warningSubOperations;
    char errorComment[DICOM_LO_LENGTH + 1];

	char note[256];
	char message[1024];

	char query_string[1];

};

#endif
