/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

BOOL WINAPI DllMain( HINSTANCE hInst, DWORD wDataSeg, LPVOID lpReserved );

//========================================================================================
// ServerSupportFunction
//========================================================================================
static BOOL WINAPI ServerSupportFunction(HCONN hConn,
                                         DWORD dwHSERRequest,
                                         LPVOID lpvBuffer,
                                         LPDWORD lpdwSize,
                                         LPDWORD lpdwDataType)
{
	if (dwHSERRequest != HSE_REQ_SEND_RESPONSE_HEADER)
		return true;
	return WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), lpdwDataType, *lpdwSize, lpdwSize, NULL);
}
//----------------------------------------------------------------------------------------
static BOOL WINAPI NullServerSupportFunction(HCONN hConn,
                                         DWORD dwHSERRequest,
                                         LPVOID lpvBuffer,
                                         LPDWORD lpdwSize,
                                         LPDWORD lpdwDataType)
{
	return true;
}

//========================================================================================
// WriteClient
//========================================================================================
static BOOL WINAPI WriteClient(HCONN ConnID,
                               LPVOID Buffer,
                               LPDWORD lpdwBytes,
                               DWORD dwReserved)
{
	return WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), Buffer, *lpdwBytes, lpdwBytes, NULL);
}
//----------------------------------------------------------------------------------------
static BOOL WINAPI NullWriteClient(HCONN ConnID,
                               LPVOID Buffer,
                               LPDWORD lpdwBytes,
                               DWORD dwReserved)
{
	return true;
}

//========================================================================================
// GetServerVariable
//========================================================================================
static BOOL WINAPI GetServerVariable(HCONN hConn,
                                      LPSTR lpszVariableName,
                                      LPVOID lpvBuffer,
                                      LPDWORD lpdwSize )
{
	*(char*)lpvBuffer = '\0';
	lpdwSize = 0;
	return true;
}
//========================================================================================
// Main
//========================================================================================
int main(int argc, char** argv)
{
	CoInitializeEx(NULL, COINIT_APARTMENTTHREADED );

	EXTENSION_CONTROL_BLOCK ecb;

	ecb.GetServerVariable = GetServerVariable;

#if 1
	char* pathInfo = "";
	char* queryString = "";
	char queryStringBuf[MAX_PATH] = "";

	if (argc > 1) {
		ecb.ServerSupportFunction = NullServerSupportFunction;
		pathInfo = argv[1];
		if (argc > 2) {
			argc -= 2;
			argv += 2;
			strcpy(queryStringBuf, *argv);
			while (--argc) {
				strcat(queryStringBuf, "&");
				strcat(queryStringBuf, *++argv);
			}
			queryString = queryStringBuf;
		}
	} else {
		ecb.ServerSupportFunction = ServerSupportFunction;
		char* cp = getenv("PATH_INFO");
		if (cp != NULL)
			pathInfo = cp;
		cp = getenv("QUERY_STRING");
		if (cp != NULL)
			queryString = cp;
	}

	ecb.WriteClient = WriteClient;
	ecb.lpszPathInfo = pathInfo;
	ecb.lpszQueryString = queryString;

	HSE_VERSION_INFO VerInfo;
	GetExtensionVersion(&VerInfo);

	HttpExtensionProc(&ecb);

	TerminateExtension(HSE_TERM_MUST_UNLOAD);

#endif


#if 0
// --- Multi Thread Check
	ecb.WriteClient = NullWriteClient;
	ecb.ServerSupportFunction = NullServerSupportFunction;

	EXTENSION_CONTROL_BLOCK ecbArray[100];
	char qsArray[100][80];
	HANDLE threadArray[100];

	for (int i = 0; i < 100; i++) {
		memcpy(&ecbArray[i], &ecb, sizeof(ecb));
		sprintf(qsArray[i], "%s&WC=%d", queryString, i);
		ecbArray[i].lpszQueryString = qsArray[i];
		threadArray[i] = CreateThread(NULL, 64 * 1024, (LPTHREAD_START_ROUTINE)HttpExtensionProc, &ecbArray[i], 0, NULL);
		if (threadArray[i] == NULL) {
			fprintf(stderr, "can't create thread for %d\n", i);
			exit(1);
		}
	}

	bool live = true;
	while (live) {
		Sleep(100);
		live = false;
		for (int i = 0; i < 100; i++) {
			DWORD ExitCode;
			if (threadArray[i] == NULL)
				continue;
			if (GetExitCodeThread(threadArray[i], &ExitCode)) {
				if (ExitCode == STILL_ACTIVE) {
					live |= true;
				} else {
					threadArray[i] = NULL;
					if (ExitCode == 1) {
						printf("[%d] ", i);
					} else {
						printf("\n[%d] = %d ", i, ExitCode);
					}
				}
			} else {
				printf("can't get thread exit status\n");
				exit(1);
			}
		}
	}

#endif

#if 0
// --- speed check loop
	if (argc < 6)
		return 0;

	HINSTANCE hLib = 0;
	PFN_GETEXTENSIONVERSION getExtensionVersion = 0;
	PFN_HTTPEXTENSIONPROC httpExtensionProc = 0;
	PFN_TERMINATEEXTENSION terminateExtension = 0;

	hLib = LoadLibrary("dio-image.isa");
	if (hLib == NULL) {
		fprintf(stderr, "can't load dio-image.dll");
		return 0;
	}

	httpExtensionProc = (PFN_HTTPEXTENSIONPROC)GetProcAddress(hLib, "HttpExtensionProc");
	getExtensionVersion = (PFN_GETEXTENSIONVERSION)GetProcAddress(hLib, "GetExtensionVersion");
	terminateExtension = (PFN_TERMINATEEXTENSION)GetProcAddress(hLib, "TerminateExtension");

	HSE_VERSION_INFO ver_info;
	getExtensionVersion(&ver_info);

	char* pathInfo = argv[1];
	char* queryStringBase = argv[2];
	int for_begin = atoi(argv[3]);
	int for_end = atoi(argv[4]);
	int for_step = atoi(argv[5]);

	char queryStringBuf[MAX_PATH];
	ecb.lpszPathInfo = pathInfo;
	ecb.lpszQueryString = queryStringBuf;
	ecb.WriteClient = NullWriteClient;
	ecb.ServerSupportFunction = NullServerSupportFunction;

	FILETIME tm0;
	GetSystemTimeAsFileTime(&tm0);

	int count = 0;
	for (int param = for_begin; param <= for_end; param += for_step, count++) {
		sprintf(queryStringBuf, "%s%d", queryStringBase, param);
		httpExtensionProc(&ecb);
	}

	FILETIME tm1;
	GetSystemTimeAsFileTime(&tm1);

	double t0 = (double)tm0.dwLowDateTime + (double)tm0.dwHighDateTime * (double)0x10000 * (double)0x10000;
	double t1 = (double)tm1.dwLowDateTime + (double)tm1.dwHighDateTime * (double)0x10000 * (double)0x10000;

	double duration_ms = (t1 - t0) / (1000.0  * 10.0);
	double duration_sec = duration_ms / 1000.0;

	printf("%g\t%d\t%s?%s^%d^%d^%d\n", (float)duration_sec, count, pathInfo, queryStringBase, for_begin, for_end, for_step);
	fprintf(stderr, "%g sec, %u images\n", (float)duration_sec, count);
	fprintf(stderr, "%g images/sec\n", (float)((double)count / duration_sec));
	fprintf(stderr, "%g ms/image\n", (float)(duration_ms / (double)count));

	terminateExtension(HSE_TERM_MUST_UNLOAD);

#endif

	CoUninitialize();

	return 0;
}
