/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "psf.h"

float fscale256 = 8.0;

//========================================================================================
//
// Point Spead Function for resampling for Grayscale
// 
//========================================================================================
//----------------------------------------------------------------------------------------
//
// Float
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------------------
inline unsigned char calc_psf1_float(const int& yi0, const int& yi1, const int& dy0,
                               const int& xi0, const int& xi1, const int& dx0,
                               unsigned char** buf_ptr2)
{

	const int height = (yi1 - yi0) * 2;
	const int width = xi1 - xi0;
	const int a = height + width;

	if (a == 0) {
		return *(*buf_ptr2 + xi0);
	} else if (a == 1) { // horizontai 2
		const float r0 = calc_rate_float(dx0, 0);
		const float r1 = calc_rate_float(dx0 + 256, 0);
	
		unsigned char* bp = *buf_ptr2 + xi0;
		const unsigned char c0 = *bp++;
		const unsigned char c1 = *bp;

		const float sum = r0 + r1;
		const float acm = c0 * r0 + c1 * r1;
		return (unsigned char)(acm / sum);
		
	} else if (a == 2) {
		const float r0 = calc_rate_float(0, dy0);
		const float r1 = calc_rate_float(0, dy0 + 256);

		const unsigned char c0 = *(*buf_ptr2++ + xi0);
		const unsigned char c1 = *(*buf_ptr2 + xi0);

		const float sum = r0 + r1;
		const float acm = c0 * r0 + c1 * r1;
		return (unsigned char)(acm / sum);

	} else if (a == 3) {
		const float r00 = calc_rate_float(dx0, dy0);
		const float r01 = calc_rate_float(dx0 + 256, dy0);
		const float r10 = calc_rate_float(dx0, dy0 + 256);
		const float r11 = calc_rate_float(dx0 + 256, dy0 + 256);

		unsigned char* bp = *buf_ptr2 + xi0;
		const unsigned char c00 = *bp++;
		const unsigned char c01 = *bp;
		buf_ptr2++;
		bp = *buf_ptr2 + xi0;
		const unsigned char c10 = *bp++;
		const unsigned char c11 = *bp;

		const float sum = r00 + r01 + r10 + r11;
		const float acm = c00 * r00 + c01 * r01 + c10 * r10 + c11 * r11;
		return (unsigned char)(acm / sum);
	} else {
		float acm = 0.0;
		float sum = 0.0;
	
		int dy = dy0;
		for (int i = yi1 - yi0;;) {
			int dx = dx0;
			unsigned char* bp = *buf_ptr2 + xi0;
			for (int j = xi1 - xi0;;) {
				float rate = calc_rate_float(dx, dy);
				acm += *bp * rate;
				sum += rate;
				if (j == 0)
					break;
				j--;
				bp++;
				dx += 256;
			}
			if (i == 0)
				break;
			i--;
			buf_ptr2++;
			dy += 256;
		}
	
		return (unsigned char)(acm / sum);
	}

}

//========================================================================================
//
//========================================================================================
void calc_psf1_row_float(unsigned char** buf_ptr2, const unsigned int pixel_width, unsigned char* outRowBuf, const unsigned int& output_width, const float& inv_scale, const int& view_x, const int& yi0, const int& yi1, const int& dy0)
{
	unsigned char* op = (unsigned char*)outRowBuf;

	float xi = view_x;
	for (int x = 0; x < output_width; x++, xi += inv_scale) {
		int xi0, xi1, dx0;
		calc_coord(xi, pixel_width, xi0, xi1, dx0);

		*op++ = calc_psf1_float(yi0, yi1, dy0, xi0, xi1, dx0, buf_ptr2);
	}
}

//----------------------------------------------------------------------------------------
//
// Integer
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline unsigned int calc_rate_int_1(const int& idx)
{
	return 256 - (idx < 0 ? -idx : idx);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline unsigned int calc_rate_int(const int& idx, const int& idy)
{
	return calc_rate_int_1(idx) + calc_rate_int_1(idy);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline unsigned char calc_psf1_int(const int yi0, const int yi1, const int dy0,
		const int xi0, const int xi1, const int dx0,
		unsigned char** buf_ptr2)
{
	int height = (yi1 - yi0) * 2;
	int width = xi1 - xi0;
	int a = height + width;

	if (a == 0) {
		return *(*buf_ptr2 + xi0);
	} else if (a == 1) { // horizontai 2
		const unsigned int r0 = calc_rate_int_1(dx0);
		const unsigned int r1 = calc_rate_int_1(dx0 + 256);
	
		unsigned char* bp = *buf_ptr2 + xi0;
		const unsigned char c0 = *bp++;
		const unsigned char c1 = *bp;

		const unsigned int sum = r0 + r1;
		const unsigned int acm = c0 * r0 + c1 * r1;
		return (unsigned char)(acm / sum);
		
	} else if (a == 2) {
		const unsigned int r0 = calc_rate_int_1(dy0);
		const unsigned int r1 = calc_rate_int_1(dy0 + 256);

		const unsigned char c0 = *(*buf_ptr2++ + xi0);
		const unsigned char c1 = *(*buf_ptr2 + xi0);

		const unsigned int sum = r0 + r1;
		const unsigned int acm = c0 * r0 + c1 * r1;
		return (unsigned char)(acm / sum);

	} else if (a == 3) {
		const unsigned int r00 = calc_rate_int(dx0, dy0);
		const unsigned int r01 = calc_rate_int(dx0 + 256, dy0);
		const unsigned int r10 = calc_rate_int(dx0, dy0 + 256);
		const unsigned int r11 = calc_rate_int(dx0 + 256, dy0 + 256);

		unsigned char* bp = *buf_ptr2 + xi0;
		const unsigned char c00 = *bp++;
		const unsigned char c01 = *bp;
		buf_ptr2++;
		bp = *buf_ptr2 + xi0;
		const unsigned char c10 = *bp++;
		const unsigned char c11 = *bp;

		const unsigned int sum = r00 + r01 + r10 + r11;
		const unsigned int acm = c00 * r00 + c01 * r01 + c10 * r10 + c11 * r11;
		return (unsigned char)(acm / sum);
	} else {
		unsigned int acm = 0;
		unsigned int sum = 0;
	
		int dy = dy0;
		for (int i = yi1 - yi0;;) {
			int dx = dx0;
			unsigned char* bp = *buf_ptr2 + xi0;
			for (int j = xi1 - xi0;;) {
				unsigned int rate = calc_rate_int(dx, dy);
				acm += *bp * rate;
				sum += rate;
				if (j == 0)
					break;
				j--;
				bp++;
				dx += 256;
			}
			if (i == 0)
				break;
			i--;
			buf_ptr2++;
			dy += 256;
		}
	
		return (unsigned char)(acm / sum);
	}
}

//========================================================================================
//
//========================================================================================
void calc_psf1_row_int(unsigned char** buf_ptr2, const unsigned int pixel_width, unsigned char* outRowBuf, const unsigned int& output_width, const float& inv_scale, const int& view_x, const int& yi0, const int& yi1, const int& dy0)
{
	unsigned char* op = (unsigned char*)outRowBuf;

	float xi = view_x;
	for (int x = 0; x < output_width; x++, xi += inv_scale) {
		int xi0, xi1, dx0;
		calc_coord(xi, pixel_width, xi0, xi1, dx0);

		*op++ = calc_psf1_int(yi0, yi1, dy0, xi0, xi1, dx0, buf_ptr2);
	}
}

