/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//----------------------------------------------------------------------------------------
//
//resampling
//
//----------------------------------------------------------------------------------------
/*
---- gen-hypot.c ----

#include <stdio.h>
#include <math.h>

main()
{
	printf("const float hypotTable[256 * 256] = {\n");

	for (int y = 0; y < 256; y++) {
		float y2 = y / 256.0;
		y2 *= y2;
		for (int x = 0; x < 256; x++) {
			float x2 = x / 256.0;
			x2 *= x2;
			float rate = 1.0 - sqrt(x2 + y2);
			if (rate < 0.0)
				rate = 0.0;
			printf("%e, ", rate);
			if ((x & 3) == 3)
				printf("\n");
		}
	}

	printf("};\n");
}
*/

//----------------------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------------------
inline void calc_coord_(const float xyi, int& xyi0, int& dxy0)
{
//	xyi0 = xfloor(xyi);
//	dxy0 = xceil(((float)xyi0 - xyi) * 256);

	extern unsigned short ceil_cw;
	extern unsigned short floor_cw;
	extern float fscale256;

	short cw_save;
	asm {
		fnstcw cw_save

		fld fscale256 // push

		fld xyi // push
		fldcw floor_cw
		frndint

		mov edi, xyi0
		fist dword ptr [edi]

		fsub xyi
		fscale

		fldcw ceil_cw
		mov edi, dxy0
		fistp dword ptr [edi] // pop

		fcomp // pop

		fldcw cw_save
	}
}
//----------------------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------------------
inline void calc_coord(const float& xyi, const unsigned int& pixel_size, int& xyi0, int& xyi1, int& dxy0)
{
	calc_coord_(xyi, xyi0, dxy0);
	if (dxy0 < 0 && xyi0 + 1 < pixel_size)
		xyi1 = xyi0 + 1;
	else
		xyi1 = xyi0;
}
//----------------------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------------------
inline float calc_rate_float(const int& idx, const int& idy)
{
	extern float hypotTable[256 * 256];
	return hypotTable[
	                  idx < 0 ? -idx : idx +
	                  (idy < 0 ? -idy : idy) * 256
	                 ];
}


//----------------------------------------------------------------------------------------
// Point Spead Function for resampling for RGB
//----------------------------------------------------------------------------------------
inline void calc_psf3(const int& yi0, const int& yi1, const int& dy0,
                      const int& xi0, const int& xi1, const int& dx0,
                      unsigned char** buf_ptr2,
                      unsigned char* rgb)
{
	float acm0 = 0.0;
	float acm1 = 0.0;
	float acm2 = 0.0;
	float sum = 0.0;

	int dy = dy0;
	for (int i = yi1 - yi0;;) {
		int dx = dx0;
		unsigned char* bp = *buf_ptr2 + 3 * xi0;
		for (int j = xi1 - xi0;;) {
			float rate = calc_rate_float(dx, dy);
			acm0 += *bp++ * rate;
			acm1 += *bp++ * rate;
			acm2 += *bp++ * rate;
			sum += rate;
			if (j == 0)
				break;
			j--;
			dx += 256;
		}
		if (i == 0)
			break;
		i--;
		buf_ptr2++;
		dy += 256;
	}

	rgb[0] = (unsigned char)(acm0 / sum);
	rgb[1] = (unsigned char)(acm1 / sum);
	rgb[2] = (unsigned char)(acm2 / sum);
}


typedef void (*calc_psf1_row_t)(unsigned char** buf_ptr2, const unsigned int pixel_width, unsigned char* outRowBuf, const unsigned int& output_width, const float& inv_scale, const int& view_x, const int& yi0, const int& yi1, const int& dy0);
void calc_psf1_row_int(unsigned char** buf_ptr2, const unsigned int pixel_width, unsigned char* outRowBuf, const unsigned int& output_width, const float& inv_scale, const int& view_x, const int& yi0, const int& yi1, const int& dy0);
void calc_psf1_row_float(unsigned char** buf_ptr2, const unsigned int pixel_width, unsigned char* outRowBuf, const unsigned int& output_width, const float& inv_scale, const int& view_x, const int& yi0, const int& yi1, const int& dy0);

