/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "voi_proc.h"

#if __option (mmx)
#define _USE_MMX_
#define _USE_FPU_
#endif

//----------------------------------------------------------------------------------------
//
// MMX VOI processor
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static const long long V_ZERO = 0;
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_prep_int_vectors(const voi_proc_c* const ctx)
{
#ifdef _USE_MMX_
	int denom = 256;
	int rescale_denom_shift = 8;
	const float abs_rescale = ctx->rescale < 0.0 ? -ctx->rescale : ctx->rescale;
	while ((int)(abs_rescale * (float)(denom * 2)) < 32767) {
		denom *= 2;
		rescale_denom_shift++;
	}
	const signed short i_rescale = (signed short)(ctx->rescale * denom);
	const int rescale_shift = ctx->shift * denom;

	denom = 256;
	int slope_denom_shift = 8;
	const float abs_slope = ctx->slope < 0.0 ? -ctx->slope : ctx->slope;
	while ((int)(abs_slope * (float)(denom * 2)) < 32767) {
		denom *= 2;
		slope_denom_shift++;
	}
	const signed short i_slope = (signed short)(ctx->slope * denom);

	asm {
		mov ecx, ctx

		// vectorize shl_count
		mov ebx, [ecx]shr_count
		mov eax, dword ptr [ebx] 
		movd mm0, eax
		movq [ecx]VQ_SHR_COUNT, mm0

		// vectorize i_rescale to ebp(0)
		xor eax, eax
		mov ax, i_rescale
		mov ebx, eax
		shl ebx, 16
		or eax, ebx
		movd mm0, eax
		movq mm1, mm0
		psllq mm1, 32
		por mm0, mm1
		movq [ecx]VW_I_RESCALE, mm0
	
		// vectorize rescale_shift to ebp(1)
		movd mm0, rescale_shift
		movq mm1, mm0
		psllq mm1, 32
		por mm0, mm1
		movq [ecx]VD_RESCALE_SHIFT, mm0

		// vectorize shift
		mov ebx, [ecx]shift
		movd mm0, dword ptr [ebx]
		movq mm1, mm0
		psllq mm1, 32
		por mm0, mm1
		movq [ecx]VD_SHIFT, mm0

		// vectorize rescale_denom_shift to ebp(2)
		movd mm0, rescale_denom_shift
		movq [ecx]VQ_RESCALE_DENOM_SHIFT, mm0
	
		// vectorize i_slope to ebp(3)
		xor eax, eax
		mov ax, i_slope
		mov ebx, eax
		shl ebx, 16
		or eax, ebx
		movd mm0, eax
		movq mm1, mm0
		psllq mm1, 32
		por mm0, mm1
		movq [ecx]VW_I_SLOPE, mm0
	
		// vectorize slope_denom_shift to ebp(4)
		movd mm0, slope_denom_shift
		movq [ecx]VQ_SLOPE_DENOM_SHIFT, mm0

		// vectorize signMask
		xor eax, eax
		mov ebx, [ecx]signMask
		mov ax, word ptr [ebx]
		mov ebx, eax
		shl ebx, 16
		or eax, ebx
		movd mm0, eax
		movq mm1, mm0
		psllq mm1, 32
		por mm0, mm1
		movq [ecx]VW_SIGN_MASK, mm0
		
		// vectorize signBitMask
		xor eax, eax
		mov ebx, [ecx]signBitMask
		mov ax, word ptr [ebx]
		mov ebx, eax
		shl ebx, 16
		or eax, ebx
		movd mm0, eax
		movq mm1, mm0
		psllq mm1, 32
		por mm0, mm1
		movq [ecx]VW_SIGN_BIT_MASK, mm0
	}
#endif
}

//----------------------------------------------------------------------------------------
// Inteter / Rescale / --- /   Signed / ----- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_rescale_noshr_signed_nomask VOI_PROC_ARG
{
#ifndef _USE_MMX_
	return voi_proc_float_rescale_noshr_signed_nomask(ctx, usp, ucp, width, skip);
#else

#define MM_0 mm0
#define MM_1 mm1
#define MM_TMP mm2

#define MM_VW_I_RESCALE mm3
#define MM_VD_RESCALE_SHIFT mm4
#define MM_VQ_RESCALE_DENOM_SHIFT mm5
#define MM_VW_I_SLOPE mm6
#define MM_VQ_SLOPE_DENOM_SHIFT mm7

	asm {
		mov eax, ctx
		movq MM_VW_I_RESCALE, [eax]VW_I_RESCALE
		movq MM_VD_RESCALE_SHIFT, [eax]VD_RESCALE_SHIFT
		movq MM_VQ_RESCALE_DENOM_SHIFT, [eax]VQ_RESCALE_DENOM_SHIFT
		movq MM_VW_I_SLOPE, [eax]VW_I_SLOPE
		movq MM_VQ_SLOPE_DENOM_SHIFT, [eax]VQ_SLOPE_DENOM_SHIFT
	}

	int lq = width / 4;
	int lr = width % 4;
	while (true) {
		int count;
		if (lq == 0) {
			if (lr == 0) {
				break;
			} else {
				count = lr;
				lr = 0;
			}
		} else {
			count = 4;
			lq--;
		}
		asm {
			// load 4 words to mm0
			mov esi, usp

			mov ebx, skip
			mov ecx, count
			cmp ecx, 4
			jl skipping_load
			cmp ebx, 1
			je direct_load_data
			mov cx, 4
		skipping_load:
			shl ebx, 1
			sub esp, 2 * 4
			mov edi, esp
		skipping_load_loop:
			mov ax, word ptr [esi]
			mov word ptr [edi], ax
			add esi, ebx
			add edi, 2
			dec cx
			jnz skipping_load_loop
			movq MM_0, qword ptr [esp]
			add esp, 2 * 4
			jmp calc
		direct_load_data:
			movq MM_0, qword ptr [esi]
/*	
		expand_sign:
			mov ebx, VW_SIGN_MASK
			movq MM_1, qword ptr [ebx]
			movd eax, MM_1
			or eax, eax
			jz calc

			movq MM_TMP, MM_0

			mov ebx, VW_SIGN_BIT_MASK
			pand MM_TMP, qword ptr [ebx]
			pcmpgtw MM_TMP, V_ZERO
			pand MM_TMP, MM_1
			por MM_0, MM_TMP
*/
		calc:
			// int uiv = sv * i_rescale;
			movq MM_1, MM_0
			pmullw MM_0, MM_VW_I_RESCALE
			pmulhw MM_1, MM_VW_I_RESCALE
			movq MM_TMP, MM_0
			punpcklwd MM_0, MM_1
			punpckhwd MM_TMP, MM_1
			movq MM_1, MM_TMP
		
			// uiv += rescale_shift;
			paddd MM_0, MM_VD_RESCALE_SHIFT
			paddd MM_1, MM_VD_RESCALE_SHIFT
		
			// uiv >>= rescale_denom_shift;
			psrad MM_0, MM_VQ_RESCALE_DENOM_SHIFT
			psrad MM_1, MM_VQ_RESCALE_DENOM_SHIFT
			packssdw MM_0, MM_1
		
			// uiv *= i_slope;
			movq MM_1, MM_0
			pmullw MM_0, MM_VW_I_SLOPE
			pmulhw MM_1, MM_VW_I_SLOPE
			movq MM_TMP, MM_0
			punpcklwd MM_0, MM_1
			punpckhwd MM_TMP, MM_1
			movq MM_1, MM_TMP
		
			// uiv >>= slope_denom_shift;
			psrad MM_0, MM_VQ_SLOPE_DENOM_SHIFT
			psrad MM_1, MM_VQ_SLOPE_DENOM_SHIFT

			// *cp = limit2uc(uiv);
			packssdw MM_0, MM_1
			packuswb MM_0, MM_0

			mov edi, ucp

			mov ecx, count
			cmp ecx, 4
			jl partial_write
			movd dword ptr [edi], MM_0
			jmp write_end
		partial_write:
			sub esp, 4
			mov esi, esp
			movd dword ptr [esi], MM_0
			cld
			rep movsb
			add esp, 4
		write_end:
		}
		ucp += count;
		usp += skip * count;
	}
#undef MM_0
#undef MM_1
#undef MM_TMP

#undef MM_VW_I_RESCALE
#undef MM_VD_RESCALE_SHIFT
#undef MM_VQ_RESCALE_DENOM_SHIFT
#undef MM_VW_I_SLOPE
#undef MM_VQ_SLOPE_DENOM_SHIFT

#endif
}

//----------------------------------------------------------------------------------------
// Inteter / ------- / --- /   Signed / ----- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_norescale_noshr_signed_nomask VOI_PROC_ARG
{
	return voi_proc_float_norescale_noshr_signed_nomask(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
// Inteter / Rescale / SHL /   Signed / ----- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_rescale_shr_signed_nomask VOI_PROC_ARG
{
	return voi_proc_float_rescale_shr_signed_nomask(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
// Inteter / ------- / SHL /   Signed / ----- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_norescale_shr_signed_nomask VOI_PROC_ARG
{
	return voi_proc_float_norescale_shr_signed_nomask(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
// Inteter / Rescale / --- /   Signed / mask 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_rescale_noshr_signed_mask VOI_PROC_ARG
{
	return voi_proc_float_rescale_noshr_signed_mask(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
// Inteter / ------- / --- /   Signed / mask 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_norescale_noshr_signed_mask VOI_PROC_ARG
{
	return voi_proc_float_norescale_noshr_signed_mask(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
// Inteter / Rescale / SHL /   Signed / mask 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_rescale_shr_signed_mask VOI_PROC_ARG
{
	return voi_proc_float_rescale_shr_signed_mask(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
// Inteter / ------- / SHL /   Signed / mask 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_norescale_shr_signed_mask VOI_PROC_ARG
{
	return voi_proc_float_norescale_shr_signed_mask(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
// Inteter / Rescale / --- / Unsigned / ----
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_rescale_noshr_unsigned VOI_PROC_ARG
{
	return voi_proc_float_rescale_noshr_unsigned(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
// Inteter / ------ / --- / Unsigned / ---- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_norescale_noshr_unsigned VOI_PROC_ARG
{
	return voi_proc_float_norescale_noshr_unsigned(ctx, usp, ucp, width, skip);
#if 0
#ifndef _USE_MMX_
	return voi_proc_float_norescale_noshr_unsigned(ctx, usp, ucp, width, skip);
#else

#define MM_0 mm0
#define MM_1 mm1
#define MM_TMP mm2

#define MM_VD_SHIFT mm3
#define MM_VQ_RESCALE_DENOM_SHIFT mm4
#define MM_VW_I_SLOPE mm5
#define MM_VQ_SLOPE_DENOM_SHIFT mm6

	asm {
		mov ecx, ctx
		movq MM_VD_SHIFT, [ecx]VD_SHIFT
		movq MM_VQ_RESCALE_DENOM_SHIFT, [ecx]VQ_RESCALE_DENOM_SHIFT
		movq MM_VW_I_SLOPE, [ecx]VW_I_SLOPE
		movq MM_VQ_SLOPE_DENOM_SHIFT, [ecx]VQ_SLOPE_DENOM_SHIFT
	}

	int lq = width / 4;
	int lr = width % 4;
	while (true) {
		int count;
		if (lq == 0) {
			if (lr == 0) {
				break;
			} else {
				count = lr;
				lr = 0;
			}
		} else {
			count = 4;
			lq--;
		}
		asm {
			// load 4 words to mm0
			mov esi, usp

			mov ebx, skip
			mov ecx, count
			cmp ecx, 4
			jl skipping_load
			cmp ebx, 1
			je direct_load_data
			mov cx, 4
		skipping_load:
			shl ebx, 1
			sub esp, 2 * 4
			mov edi, esp
		skipping_load_loop:
			mov ax, word ptr [esi]
			mov word ptr [edi], ax
			add esi, ebx
			add edi, 2
			dec cx
			jnz skipping_load_loop
			movq MM_0, qword ptr [esp]
			add esp, 2 * 4
			jmp calc
		direct_load_data:
			movq MM_0, qword ptr [esi]
/*	
		expand_sign:
			mov ebx, VW_SIGN_MASK
			movq MM_1, qword ptr [ebx]
			movd eax, MM_1
			or eax, eax
			jz calc

			movq MM_TMP, MM_0

			mov ebx, VW_SIGN_BIT_MASK
			pand MM_TMP, qword ptr [ebx]
			pcmpgtw MM_TMP, V_ZERO
			pand MM_TMP, MM_1
			por MM_0, MM_TMP
*/
		calc:
			// uiv += shift;
			pxor MM_TMP, MM_TMP
			movq MM_1, MM_0
			punpcklwd MM_0, MM_TMP
			punpckhwd MM_1, MM_TMP
			paddd MM_0, MM_VD_SHIFT
			paddd MM_1, MM_VD_SHIFT
			packssdw MM_0, MM_1

			// uiv *= i_slope;
			movq MM_1, MM_0
			pmullw MM_0, MM_VW_I_SLOPE
			pmulhw MM_1, MM_VW_I_SLOPE
			movq MM_TMP, MM_0
			punpcklwd MM_0, MM_1
			punpckhwd MM_TMP, MM_1
			movq MM_1, MM_TMP
		
			// uiv >>= slope_denom_shift;
			psrad MM_0, MM_VQ_SLOPE_DENOM_SHIFT
			psrad MM_1, MM_VQ_SLOPE_DENOM_SHIFT

			// *cp = limit2uc(uiv);
			packssdw MM_0, MM_1
			packuswb MM_0, MM_0

			mov edi, ucp

			mov ecx, count
			cmp ecx, 4
			jl partial_write
			movd dword ptr [edi], MM_0
			jmp write_end
		partial_write:
			sub esp, 4
			mov esi, esp
			movd dword ptr [esi], MM_0
			cld
			rep movsb
			add esp, 4
		write_end:
		}
		ucp += count;
		usp += skip * count;
	}
#undef MM_0
#undef MM_1
#undef MM_TMP

#undef MM_VD_SHIFT
#undef MM_VQ_RESCALE_DENOM_SHIFT
#undef MM_VW_I_SLOPE
#undef MM_VQ_SLOPE_DENOM_SHIFT

#endif
#endif
}

//----------------------------------------------------------------------------------------
// Inteter / Rescale / SHL / Unsigned / ---- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_rescale_shr_unsigned VOI_PROC_ARG
{
	return voi_proc_float_rescale_shr_unsigned(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
// Inteter / ------ / SHL / Unsigned / ---- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_int_norescale_shr_unsigned VOI_PROC_ARG
{
	return voi_proc_float_norescale_shr_unsigned(ctx, usp, ucp, width, skip);
}

//----------------------------------------------------------------------------------------
//
// Floating Point VOI processor
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//		FNSTCW word ptr cw
//		mov ax, 0xF3FF
//		and ax, word ptr cw
//		or ax, 0x0300
//		mov word ptr cw, ax
//		FLDCW word ptr cw
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline unsigned char limit2uc(const int& v)
{
	if (v > 255)
		return 255;
	if (v < 0)
		return 0;
	return (unsigned char)v;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline int float_norescale_signed(const signed short& sv, const int& shift, const float& slope)
{
	return (int)((sv + shift) * slope);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline int float_rescale_signed(const signed short& sv, const float& f_rescale_slope, const float& f_shift_slope)
{
	return (int)(sv * f_rescale_slope + f_shift_slope);
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline int float_norescale_unsigned(const unsigned short& sv, const int& shift, const float& slope)
{
	return (int)((sv + shift) * slope);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
inline int float_rescale_unsigned(const unsigned short& sv, const float& f_rescale_slope, const float& f_shift_slope)
{
	return (int)(sv * f_rescale_slope + f_shift_slope);
}

//----------------------------------------------------------------------------------------
// Float / Rescale / --- /   Signed / ----- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_rescale_noshr_signed_nomask VOI_PROC_ARG
{
#ifdef _USE_FPU_
	int iv;
	asm {
		mov eax, ctx
		FLD dword ptr [eax]f_rescale_slope
		FLD dword ptr [eax]f_shift_slope
		mov ecx, width
		mov edx, skip
		shl edx, 1
		mov esi, usp
		mov edi, ucp
	for_loop:
		movsx ax, word ptr [esi] // extend sign
		mov iv, eax
		FLD ST(1)
		FIMUL iv
		FADD ST(0), ST(1)
		FISTP iv

		xor eax, eax
		xor ebx, ebx
		add eax, iv
		cmovl eax, ebx
		mov bl, 0xFF
		cmp eax, ebx
		cmovg eax, ebx

		mov byte ptr [edi], al
		add esi, edx
		inc edi
		dec ecx
		jnz for_loop
		FSTP ST(0)
		FSTP ST(0)
	}
#else
	for (int i = 0; i < width; i++, usp += skip, ucp++) {
		signed short sv = *(signed short*)usp;
		int v = float_rescale_signed(sv, ctx->f_rescale_slope, ctx->f_shift_slope);
		*ucp = limit2uc(v);
	}
#endif
}

//----------------------------------------------------------------------------------------
// Float / ------- / --- /   Signed / ----- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_norescale_noshr_signed_nomask VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		signed short sv = *(unsigned short*)usp;
		int v = float_norescale_signed(sv, ctx->shift, ctx->slope);
		*ucp++ = limit2uc(v);
	}
}

//----------------------------------------------------------------------------------------
// Float / Rescale / SHL /   Signed / ----- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_rescale_shr_signed_nomask VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		signed short sv = *(signed short*)usp;
		sv >>= ctx->shr_count;
		int v = float_rescale_signed(sv, ctx->f_rescale_slope, ctx->f_shift_slope);
		*ucp++ = limit2uc(v);
	}
}

//----------------------------------------------------------------------------------------
// Float / ------- / SHL /   Signed / ----- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_norescale_shr_signed_nomask VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		signed short sv = *(signed short*)usp;
		sv >>= ctx->shr_count;
		int v = float_norescale_signed(sv, ctx->shift, ctx->slope);
		*ucp++ = limit2uc(v);
	}
}

//----------------------------------------------------------------------------------------
// Float / Rescale / --- /   Signed / mask 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_rescale_noshr_signed_mask VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		signed short sv = *(signed short*)usp;
		if (sv & ctx->signBitMask)
			sv |= ctx->signMask;
		int v = float_rescale_signed(sv, ctx->f_rescale_slope, ctx->f_shift_slope);
		*ucp++ = limit2uc(v);
	}
}

//----------------------------------------------------------------------------------------
// Float / ------- / --- /   Signed / mask 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_norescale_noshr_signed_mask VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		signed short sv = *(signed short*)usp;
		if (sv & ctx->signBitMask)
			sv |= ctx->signMask;
		int v = float_norescale_signed(sv, ctx->shift, ctx->slope);
		*ucp++ = limit2uc(v);
	}
}

//----------------------------------------------------------------------------------------
// Float / Rescale / SHL /   Signed / mask 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_rescale_shr_signed_mask VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		signed short sv = *(signed short*)usp;
		sv >>= ctx->shr_count;
		if (sv & ctx->signBitMask)
			sv |= ctx->signMask;
		int v = float_rescale_signed(sv, ctx->f_rescale_slope, ctx->f_shift_slope);
		*ucp++ = limit2uc(v);
	}
}

//----------------------------------------------------------------------------------------
// Float / ------- / SHL /   Signed / mask 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_norescale_shr_signed_mask VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		signed short sv = *(signed short*)usp;
		sv >>= ctx->shr_count;
		if (sv & ctx->signBitMask)
			sv |= ctx->signMask;
		int v = float_norescale_signed(sv, ctx->shift, ctx->slope);
		*ucp++ = limit2uc(v);
	}
}

//----------------------------------------------------------------------------------------
// Float / Rescale / --- / Unsigned / ----
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_rescale_noshr_unsigned VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		unsigned short sv = *(unsigned short*)usp;
		int v = float_rescale_unsigned(sv, ctx->f_rescale_slope, ctx->f_shift_slope);
		*ucp++ = limit2uc(v);
	}
}

//----------------------------------------------------------------------------------------
// Float / ------ / --- / Unsigned / ---- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_norescale_noshr_unsigned VOI_PROC_ARG
{
#ifdef _USE_FPU_
	int iv;
	asm {
		mov eax, ctx
		mov ebx, dword ptr [eax]slope
		FLD dword ptr [ebx]
		mov ebx, dword ptr[eax]shift
		FILD dword ptr [ebx]
		mov ecx, width
		mov edx, skip
		shl edx, 1
		mov esi, usp
		mov edi, ucp
	for_loop:
		movzx ax, word ptr [esi] // extend zero
		mov iv, eax
		FILD iv
		FADD ST(0), ST(1)
		FMUL ST(0), ST(2)
		FISTP iv

		xor eax, eax
		xor ebx, ebx
		add eax, iv
		cmovl eax, ebx
		mov bl, 0xFF
		cmp eax, ebx
		cmovg eax, ebx

		mov byte ptr [edi], al
		add esi, edx
		inc edi
		dec ecx
		jnz for_loop
		FSTP ST(0)
		FSTP ST(0)
	}
#else
	for (int i = 0; i < width; i++, usp += skip) {
		unsigned short sv = *(unsigned short*)usp;
		int v = float_norescale_unsigned(sv, ctx->shift, ctx->slope);
		*ucp++ = limit2uc(v);
	}
#endif
}

//----------------------------------------------------------------------------------------
// Float / Rescale / SHL / Unsigned / ---- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_rescale_shr_unsigned VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		unsigned short sv = *(unsigned short*)usp;
		sv >>= ctx->shr_count;
		int v = float_rescale_unsigned(sv, ctx->f_rescale_slope, ctx->f_shift_slope);
		*ucp++ = limit2uc(v);
	}
}

//----------------------------------------------------------------------------------------
// Float / ------ / SHL / Unsigned / ---- 
//----------------------------------------------------------------------------------------
void voi_proc_c::voi_proc_float_norescale_shr_unsigned VOI_PROC_ARG
{
	for (int i = 0; i < width; i++, usp += skip) {
		unsigned short sv = *(unsigned short*)usp;
		sv >>= ctx->shr_count;
		int v = float_norescale_unsigned(sv, ctx->shift, ctx->slope);
		*ucp++ = limit2uc(v);
	}
}

//========================================================================================
//
//========================================================================================
voi_proc_c::voi_proc_c(
		const bool& cacheVOILUT,
		const int& voi_precision,
		const bool& isSigned,
		const bool& doRescale,
		const unsigned int& a_shr_count,
		const unsigned short& a_signMask, const unsigned short& a_signBitMask,
		const float& a_rescale,
		const int& a_shift, const float& a_slope)
	: shr_count(a_shr_count),
	  signMask(a_signMask), signBitMask(a_signBitMask),
	  rescale(a_rescale),
	  shift(a_shift), slope(a_slope),
	  f_rescale_slope(a_rescale * a_slope), f_shift_slope(a_shift * a_slope)
{
	if (voi_precision == 1) { // float
		if (doRescale) {
			if (shr_count) {
				if (isSigned) {
					if (signMask) {
						proc_ptr = voi_proc_float_rescale_shr_signed_mask;
					} else {
						proc_ptr = voi_proc_float_rescale_shr_signed_nomask;
					}
				} else {
					proc_ptr = voi_proc_float_rescale_shr_unsigned;
				}
			} else {
				if (isSigned) {
					if (signMask) {
						proc_ptr = voi_proc_float_rescale_noshr_signed_mask;
					} else {
						proc_ptr = voi_proc_float_rescale_noshr_signed_nomask;
					}
				} else {
					proc_ptr = voi_proc_float_rescale_noshr_unsigned;
				}
			}
		} else {
			if (shr_count) {
				if (isSigned) {
					if (signMask) {
						proc_ptr = voi_proc_float_norescale_shr_signed_mask;
					} else {
						proc_ptr = voi_proc_float_norescale_shr_signed_nomask;
					}
				} else {
					proc_ptr = voi_proc_float_norescale_shr_unsigned;
				}
			} else {
				if (isSigned) {
					if (signMask) {
						proc_ptr = voi_proc_float_norescale_noshr_signed_mask;
					} else {
						proc_ptr = voi_proc_float_norescale_noshr_signed_nomask;
					}
				} else {
					proc_ptr = voi_proc_float_norescale_noshr_unsigned;
				}
			}
		}
	} else {
		if (doRescale) {
			if (shr_count) {
				if (isSigned) {
					if (signMask) {
						proc_ptr = &voi_proc_int_rescale_shr_signed_mask;
					} else {
						proc_ptr = voi_proc_int_rescale_shr_signed_nomask;
					}
				} else {
					proc_ptr = voi_proc_int_rescale_shr_unsigned;
				}
			} else {
				if (isSigned) {
					if (signMask) {
						proc_ptr = voi_proc_int_rescale_noshr_signed_mask;
					} else {
						proc_ptr = voi_proc_int_rescale_noshr_signed_nomask;
					}
				} else {
					proc_ptr = voi_proc_int_rescale_noshr_unsigned;
				}
			}
		} else {
			if (shr_count) {
				if (isSigned) {
					if (signMask) {
						proc_ptr = voi_proc_int_norescale_shr_signed_mask;
					} else {
						proc_ptr = voi_proc_int_norescale_shr_signed_nomask;
					}
				} else {
					proc_ptr = voi_proc_int_norescale_shr_unsigned;
				}
			} else {
				if (isSigned) {
					if (signMask) {
						proc_ptr = voi_proc_int_norescale_noshr_signed_mask;
					} else {
						proc_ptr = voi_proc_int_norescale_noshr_signed_nomask;
					}
				} else {
					proc_ptr = voi_proc_int_norescale_noshr_unsigned;
				}
			}
		}
		voi_prep_int_vectors(this);
	}
}
