/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//----------------------------------------------------------------------------------------
//
// VOI Proc
//
//----------------------------------------------------------------------------------------
class voi_proc_c;
	
#define VOI_PROC_ARG (const voi_proc_c* const ctx, unsigned short* usp, unsigned char* ucp, const unsigned int width, const int skip)

class voi_proc_c {
	typedef void (*proc_ptr_t) VOI_PROC_ARG;
	proc_ptr_t proc_ptr;

	// basic input params
	const unsigned int& shr_count;
	const unsigned short& signMask;
	const unsigned short& signBitMask;
	const float& rescale;
	const int& shift;
	const float& slope;

	// for MMX proc
	long long VQ_SHR_COUNT;
	long long VW_I_RESCALE;
	long long VD_RESCALE_SHIFT;
	long long VD_SHIFT;
	long long VQ_RESCALE_DENOM_SHIFT;
	long long VW_I_SLOPE;
	long long VQ_SLOPE_DENOM_SHIFT;
	long long VW_SIGN_BIT_MASK;
	long long VW_SIGN_MASK;

	static void voi_prep_int_vectors(const voi_proc_c* const ctx);

	// for float proc
	const float f_rescale_slope;
	const float f_shift_slope;

	static void voi_proc_int_rescale_noshr_signed_nomask VOI_PROC_ARG;
	static void voi_proc_int_norescale_noshr_signed_nomask VOI_PROC_ARG;
	static void voi_proc_int_rescale_shr_signed_nomask VOI_PROC_ARG;
	static void voi_proc_int_norescale_shr_signed_nomask VOI_PROC_ARG;
	static void voi_proc_int_rescale_noshr_signed_mask VOI_PROC_ARG;
	static void voi_proc_int_norescale_noshr_signed_mask VOI_PROC_ARG;
	static void voi_proc_int_rescale_shr_signed_mask VOI_PROC_ARG;
	static void voi_proc_int_norescale_shr_signed_mask VOI_PROC_ARG;
	static void voi_proc_int_rescale_noshr_unsigned VOI_PROC_ARG;
	static void voi_proc_int_norescale_noshr_unsigned VOI_PROC_ARG;
	static void voi_proc_int_rescale_shr_unsigned VOI_PROC_ARG;
	static void voi_proc_int_norescale_shr_unsigned VOI_PROC_ARG;

	static void voi_proc_float_rescale_noshr_signed_nomask VOI_PROC_ARG;
	static void voi_proc_float_norescale_noshr_signed_nomask VOI_PROC_ARG;
	static void voi_proc_float_rescale_shr_signed_nomask VOI_PROC_ARG;
	static void voi_proc_float_norescale_shr_signed_nomask VOI_PROC_ARG;
	static void voi_proc_float_rescale_noshr_signed_mask VOI_PROC_ARG;
	static void voi_proc_float_norescale_noshr_signed_mask VOI_PROC_ARG;
	static void voi_proc_float_rescale_shr_signed_mask VOI_PROC_ARG;
	static void voi_proc_float_norescale_shr_signed_mask VOI_PROC_ARG;
	static void voi_proc_float_rescale_noshr_unsigned VOI_PROC_ARG;
	static void voi_proc_float_norescale_noshr_unsigned VOI_PROC_ARG;
	static void voi_proc_float_rescale_shr_unsigned VOI_PROC_ARG;
	static void voi_proc_float_norescale_shr_unsigned VOI_PROC_ARG;

public:
	voi_proc_c(
		const bool& cacheVOILUT,
		const int& voi_precision,
		const bool& isSigned,
		const bool& doRescale,
		const unsigned int& shr_count,
		const unsigned short& signMask, const unsigned short& signBitMask,
		const float& rescale,
		const int& shift, const float& slope);

	void proc(unsigned short* sp, unsigned char* cp, const unsigned int& width, const int skip = 1) {
		proc_ptr(this, sp, cp, width, skip);
	}
};
