/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "cstudydb.h"
#include "dcmdict.h"
#include "CDcmScanner.h"

#include "StudyCache.h"

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
HRESULT CStudyCache::open(const wchar_t* studyDir)
{
	const int studyDirLen = lstrlen(studyDir);

	wchar_t* const dcFile = (wchar_t*)_alloca(sizeof(wchar_t) * (studyDirLen + 1 + lstrlen(CDataElementCache::filename) + 1));
	lstrcpy(dcFile, studyDir);
	lstrcat(dcFile, L"\\");
	lstrcat(dcFile, CDataElementCache::filename);

	HRESULT hr = m_dc.Open(dcFile);
	if (FAILED(hr)) {
		return hr;
	}

	wchar_t* const icFile = (wchar_t*)_alloca(sizeof(wchar_t) * (studyDirLen + 1 + lstrlen(CInstanceCache::filename) + 1));
	lstrcpy(icFile, studyDir);
	lstrcat(icFile, L"\\");
	lstrcat(icFile, CInstanceCache::filename);

	hr = m_ic.Open(icFile);
	if (FAILED(hr)) {
		m_ic.Close();
		return hr;
	}

	return S_OK;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
HRESULT CStudyCache::close()
{
	if (m_commit_immediately) {
		m_ic.Close();
		m_dc.Close();
	} else {
		const HRESULT ic_hr = m_ic.Commit(true);
		const HRESULT dc_hr = m_dc.Commit(true);
		if (FAILED(ic_hr))
			return ic_hr;
		if (FAILED(dc_hr))
			return dc_hr;
	}

	return S_OK;
}


//======================================================================
//
//======================================================================
HRESULT CStudyCache::append(const wchar_t* studyDir, const wchar_t* imageLoc, const char* scannedElements, size_t scannedElementsSize)
{
	int len = lstrlen(studyDir);
	if (m_studyDir == NULL) {

		HRESULT hr = open(studyDir);
		if (FAILED(hr))
			return hr;

		m_studyDir = (wchar_t*)CTN_MALLOC(sizeof(wchar_t) * (len + 1));
		lstrcpy(m_studyDir, studyDir);
		m_studyDirLen = len;
		
	} else if (lstrcmp(studyDir, m_studyDir) != 0) {

		HRESULT hr = close();
		if (FAILED(hr))
			return hr;

		hr = open(studyDir);
		if (FAILED(hr))
			return hr;

		if (len > m_studyDirLen) {
			CTN_FREE(m_studyDir);
			m_studyDir = (wchar_t*)CTN_MALLOC(sizeof(wchar_t) * (len + 1));
			m_studyDirLen = len;
		}
		lstrcpy(m_studyDir, studyDir);
	}

	HRESULT hr = m_ic.AppendInstance(imageLoc, scannedElements, scannedElementsSize, true, m_dc);
	if (FAILED(hr))
		return hr;

	if (m_commit_immediately) {
		hr = m_dc.Commit(false);
		if (FAILED(hr)) {
			m_ic.Close();
			m_dc.Close();
			return hr;
		}

		hr = m_ic.Commit(false);
		if (FAILED(hr)) {
			m_ic.Close();
			m_dc.Close();
			return hr;
		}
	}

	return S_OK;
}

//======================================================================
//
//======================================================================
CStudyCache::~CStudyCache()
{
	if (m_studyDir != NULL)
		CTN_FREE(m_studyDir);

	close();
}

