/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "rtlog.h"

struct logheader_t {
	DWORD size;
	DWORD head;	// oldest 
	DWORD tail; // beyond the last
	DWORD last; // newest -- for reader
	DWORD overlap; // overlap count;
	DWORD data;
};

struct log_t {
	unsigned short length;
	struct SHORT_SYTEMTIME {
		unsigned char month;
		unsigned char day;
		unsigned char hour;
		unsigned char minute;
		unsigned char second;
		unsigned char sub_sec;
	} time;
	DWORD data[1];
};

#define rlSocketLog	0x01FEFFFF
#define rlDULFSMLog	0x02FEFFFF
#define rlCondLog	0x03FEFFFF

static void print_log(const log_t* logp)
{
	const char* type = "";
	switch (logp->data[0]) {
	case rlSocketLog:
		printf("<socket");
		printf(" time=\"%02d/%02d-%02d:%02d:%02d.%02d\" ", logp->time.month, logp->time.day, logp->time.hour, logp->time.minute, logp->time.second, logp->time.sub_sec);
		switch (logp->data[1]) {
		case rlSocketAccept:
			in_addr* remoteAddr = (in_addr*)&logp->data[4];
			printf("action=\"accept\" handle=\"%d\" port=\"%d\" remote=\"%d.%d.%d.%d\" />\n", logp->data[2], logp->data[3], remoteAddr->S_un.S_un_b.s_b1, remoteAddr->S_un.S_un_b.s_b2, remoteAddr->S_un.S_un_b.s_b3, remoteAddr->S_un.S_un_b.s_b4);
			break;
		case rlSocketRead:
			printf("action=\"read\" handle=\"%d\" req=\"%d\" res=\"%d\" />\n", logp->data[2], logp->data[3], logp->data[4]);
			break;
		case rlSocketWrite:
			printf("action=\"write\" handle=\"%d\" req=\"%d\" res=\"%d\" />\n", logp->data[2], logp->data[3], logp->data[4]);
			break;
		case rlSocketClose:
			printf("action=\"close\" handle=\"%d\" />\n", logp->data[2]);
			break;
		}
		break;
	case rlDULFSMLog:
		printf("<dulfsm");
		printf(" time=\"%02d/%02d-%02d:%02d:%02d.%02d\" ", logp->time.month, logp->time.day, logp->time.hour, logp->time.minute, logp->time.second, logp->time.sub_sec);
		printf("handle=\"%d\" event=\"%d\" state=\"%d\" />\n", logp->data[1], logp->data[2], logp->data[3]);
		break;
	case rlCondLog:
		printf("<cond");
		printf(" time=\"%02d/%02d-%02d:%02d:%02d.%02d\" ", logp->time.month, logp->time.day, logp->time.hour, logp->time.minute, logp->time.second, logp->time.sub_sec);
		printf("cond=\"%X\">%s</cond>\n", logp->data[1], &logp->data[2]);
		break;
	}
}

int main(int argc, char** argv)
{
	HANDLE hFile = INVALID_HANDLE_VALUE;
	AutoCloseFileHandle ac_hFile(hFile);

	HANDLE hMap = NULL;
	AutoCloseHandle ac_hMap(hMap);

	if (argc < 2) {
		hMap = OpenFileMapping(FILE_MAP_READ, FALSE, L"DIO_STORAGE_RT_LOG");
		if (hMap == NULL) {
			DWORD err = GetLastError();
			printf("ERROR:%d\n", err);
			return err;
		}
		LPVOID lp = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
		if (lp == NULL) {
			DWORD err = GetLastError();
			printf("ERROR:%d\n", err);
			return err;
		}
		DWORD dwMapSize = *(DWORD*)lp;
		UnmapViewOfFile(lp);
		CloseHandle(hMap); hMap = NULL;
		hMap = OpenFileMapping(FILE_MAP_READ, FALSE, L"DIO_STORAGE_RT_LOG");
		if (hMap == NULL) {
			DWORD err = GetLastError();
			printf("ERROR:%d\n", err);
			return err;
		}
		LPVOID lpMap = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
		AutoUnmapViewOfFile um_lpMap(&lpMap);

		const char* const base_ptr = (char*)lpMap;
		logheader_t* logheader = (logheader_t*)lpMap;
		volatile DWORD* lastp = &logheader->last;

		DWORD last = *lastp;
		while (true) {
			while (last == *lastp) {
				Sleep(10);
			}
			while (last != *lastp) {
				log_t* logp = (log_t*)(base_ptr + last);
				if (logp->length == 0) {
					last = offsetof(logheader_t, data);
					logp = (log_t*)(base_ptr + last);
				}
				print_log(logp);
				if (logp->length == 0) {
					last = offsetof(logheader_t, data);
				} else {
					last += logp->length;
				}
			}
		}
		
	} else {
		const char* path = argv[1];
		TOWC(path)

		hFile = CreateFile(_path_WC, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE) {
			DWORD err = GetLastError();
			printf("ERROR:%d\n", err);
			return err;
		}
		DWORD dwMapSizeHigh;
		DWORD dwMapSize = GetFileSize(hFile, &dwMapSizeHigh);

		hMap = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, dwMapSize, NULL);
		if (hMap == NULL) {
			DWORD err = GetLastError();
			printf("ERROR:%d\n", err);
			return err;
		}
		LPVOID lpMap = MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
		AutoUnmapViewOfFile um_lpMap(&lpMap);

		const char* const base_ptr = (char*)lpMap;
		logheader_t* logheader = (logheader_t*)lpMap;

		printf("<?xml version=\"1.0\" encoding=\"shift_jis\" ?>\n<rtlog>\n");
		DWORD tail = logheader->tail;
		DWORD head = logheader->head;
		while (tail != head) {
			log_t* logp = (log_t*)(base_ptr + head);
			if (logp->length == 0) {
				head = offsetof(logheader_t, data);
				logp = (log_t*)(base_ptr + head);
			}
			print_log(logp);
			if (logp->length == 0) {
				head = offsetof(logheader_t, data);
			} else {
				head += logp->length;
			}
		}
		printf("</rtlog>\n");

	}

	return 0;
}