/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "cstudydb.h"
#include "dcmdict.h"
#include "CDcmScanner.h"

#include "StudyCache.h"

#include "rtlog.h"

#include "app_cond.h"

#include "dio-storage.h"

CONDITION storeToFile(DUL_ASSOCIATIONKEY** association, DUL_PRESENTATIONCONTEXT* ctx, MSG_C_STORE_REQ** storeRequest, DUL_ASSOCIATESERVICEPARAMETERS* service, tStorageParam* sp, CStudyDB& studydb, CStudyCache& studycache, unsigned short& resultCode, bool& send_response);

//========================================================================================
//
// StorageLog
//
//========================================================================================
//========================================================================================
//
//========================================================================================
StorageLog* StorageLog::self;
//========================================================================================
//
//========================================================================================
void StorageLog::print_assoc_open_info(const tStorageParam* sp)
{
	SYSTEMTIME time;
	GetLocalTime(&time);

	char buf[128]; // 70 + 20 + 20 + alpha;
	sprintf(buf, "<ASSOC event=\"open\" scp=\"%s\" scu=\"%s\" time=\"%04u-%02u-%02uT%02u:%02u:%02u.%03u\" />\r\n",
		sp->pSCPprm->label, sp->assoc.callingAPTitle,
		time.wYear, time.wMonth, time.wDay, time.wHour, time.wMinute, time.wSecond, time.wMilliseconds);

	self->writeText(buf);
}

//========================================================================================
//
//========================================================================================
void StorageLog::print_assoc_close_info(const tStorageParam* sp)
{
	char buf[2048];

	SYSTEMTIME time;
	GetLocalTime(&time);

	sprintf(buf, "<ASSOC event=\"close\" scp=\"%s\" scu=\"%s\" time=\"%04u-%02u-%02uT%02u:%02u:%02u.%03u\" >\r\n\
\t<ASSOC-SUMMARY commands=\"%u\" requests=\"%u\" recieved-bytes=\"%llu\" stored-files=\"%u\" stored-bytes=\"%llu\" overwritten-files=\"%u\" />\r\n",
		sp->pSCPprm->label,	sp->assoc.callingAPTitle,
		time.wYear, time.wMonth, time.wDay, time.wHour, time.wMinute, time.wSecond, time.wMilliseconds,
		sp->assoc.commandCount,
		sp->assoc.requestCount,
		sp->assoc.recievedBytes,
		sp->assoc.storedFiles,
		sp->assoc.storedBytes,
		sp->assoc.overwriteFiles);

	char* cp = buf + strlen(buf);

	if (!COND_Empty()) {
		CONDITION cond;
		char* str;
		while (COND_TopConditionEx(&cond, &str) != COND_NORMAL) {
			sprintf(cp, "\t<ASSOC-ERROR cond=\"%08X\">%s</ASSOC-ERROR>\r\n", cond, str);
			cp += strlen(cp);
			COND_PopCondition(FALSE);
		}
	}

	strcpy(cp, "</ASSOC>\r\n");

	self->writeText(buf);
}

//========================================================================================
//
//========================================================================================
void StorageLog::print_remove_study_info(const tStorageParam* sp, const wchar_t* path, unsigned long bytesDeleted)
{
	char buf[2048];

	TOAC(path)

	SYSTEMTIME time;
	GetLocalTime(&time);

	sprintf(buf, "<STUDY action=\"remove\" scp=\"%s\" scu=\"%s\" vol=\"%s\" path=\"%s\" bytes-deleted=\"%u\" time=\"%04u-%02u-%02uT%02u:%02u:%02u.%03u\" />\r\n",
		sp->pSCPprm->label, sp->assoc.callingAPTitle, sp->pSCPprm->volume->label,
		_path_AC,
		bytesDeleted,
		time.wYear, time.wMonth, time.wDay, time.wHour, time.wMinute, time.wSecond, time.wMilliseconds);

	self->writeText(buf);
}
//========================================================================================
//
//========================================================================================
void StorageLog::print_insert_study_info(const tStorageParam* sp, const char* path)
{
	SYSTEMTIME time;
	GetLocalTime(&time);

	char buf[2048];
	sprintf(buf, "<STUDY action=\"insert\" scp=\"%s\" scu=\"%s\" volume=\"%s\" path=\"%s\" time=\"%04u-%02u-%02uT%02u:%02u:%02u.%03u\" />\r\n",
		sp->pSCPprm->label, sp->assoc.callingAPTitle, sp->pSCPprm->volume->label,
		path,
		time.wYear, time.wMonth, time.wDay, time.wHour, time.wMinute, time.wSecond, time.wMilliseconds);

	self->writeText(buf);
}

//----------------------------------------------------------------------------------------
//
// VERIFY
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static CONDITION echoCallback(MSG_C_ECHO_REQ * request,
	     MSG_C_ECHO_RESP * response, void *ctx,
	     DUL_PRESENTATIONCONTEXT * pci)
{
	response->dataSetType = DCM_CMDDATANULL;
	response->messageIDRespondedTo = request->messageID;
	response->status = MSG_K_SUCCESS;
	strcpy(response->classUID, request->classUID);

	return SRV_NORMAL;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static CONDITION handleEchoRequest(DUL_ASSOCIATIONKEY** association, DUL_PRESENTATIONCONTEXT* ctx, MSG_C_ECHO_REQ** echoReq)
{
	MSG_C_ECHO_RESP echoResponse = {
		MSG_K_C_ECHO_RESP, 0, 0, DCM_CMDDATANULL, 0, ""
	};

	return SRV_CEchoResponse(association, ctx, echoReq, &echoResponse, echoCallback, NULL, "");
}

//----------------------------------------------------------------------------------------
//
// STORE
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static CONDITION handleStoreRequest(DUL_ASSOCIATIONKEY** association, DUL_PRESENTATIONCONTEXT* ctx, MSG_C_STORE_REQ** request, DUL_ASSOCIATESERVICEPARAMETERS* service, tStorageParam* sp, CStudyDB& studydb, CStudyCache& studycache)
{
	//
	// Setup Response
	//
    MSG_C_STORE_RESP response;
	memset(&response, 0, sizeof(response));
	response.messageIDRespondedTo = (*request)->messageID;
	response.type = MSG_K_C_STORE_RESP;
	strcpy(response.classUID, (*request)->classUID);
	strcpy(response.instanceUID, (*request)->instanceUID);
	response.conditionalFields = MSG_K_C_STORERESP_CLASSUID | MSG_K_C_STORERESP_INSTANCEUID;
	response.dataSetType = DCM_CMDDATANULL;
	//
	//
	//
	CONDITION store_cond = APP_NORMAL;
	CONDITION response_cond = APP_NORMAL;
	bool send_response;
	store_cond = storeToFile(association, ctx, request, service, sp, studydb, studycache, response.status, send_response);
	if (send_response) {
		//
		// Response Message
		//
		DCM_OBJECT* responseObject;
		response_cond = MSG_BuildCommand(&response, &responseObject);
		if (response_cond == MSG_NORMAL) {
			response_cond = SRV_SendCommand(association, ctx, &responseObject);
			DCM_CloseObject(&responseObject);
			if (response_cond != SRV_NORMAL)
				COND_PushCondition(SRV_RESPONSEFAILED, SRV_Message(SRV_RESPONSEFAILED), "handleStoreRequest");
		} else {
			COND_PushCondition(SRV_OBJECTBUILDFAILED, SRV_Message(SRV_OBJECTBUILDFAILED), "STORE Request", "handleStoreRequest");
		}
	}

	if (!CTN_SUCCESS(store_cond)) 
		return store_cond;
	else if (!CTN_SUCCESS(response_cond))
		return response_cond;
	else
		return APP_NORMAL;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static CONDITION sendCancelResponse(DUL_ASSOCIATIONKEY** association, DUL_PRESENTATIONCONTEXT* ctx, MSG_C_STORE_REQ** request)
{
	//
	// Setup Cancel Response
	//
    MSG_C_STORE_RESP response;
	memset(&response, 0, sizeof(response));
	response.messageIDRespondedTo = (*request)->messageID;
	response.type = MSG_K_C_STORE_RESP;
	response.conditionalFields = 0;
	response.dataSetType = DCM_CMDDATANULL;
	response.status = MSG_K_CANCEL;

	DCM_OBJECT* responseObject;
	CONDITION cond = MSG_BuildCommand(&response, &responseObject);
	if (cond == MSG_NORMAL) {
		cond = SRV_SendCommand(association, ctx, &responseObject);
		DCM_CloseObject(&responseObject);
		if (cond != SRV_NORMAL)
			COND_PushCondition(SRV_RESPONSEFAILED, SRV_Message(SRV_RESPONSEFAILED), "handleStoreRequest");
	} else {
		COND_PushCondition(SRV_OBJECTBUILDFAILED, SRV_Message(SRV_OBJECTBUILDFAILED), "STORE Request", "handleStoreRequest");
	}
	return cond;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
enum {
	rsltExit = 0,
	rsltError = -1,
	rsltNormal = 1
};
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
class Auto_DUL_DropAssociation {
	DUL_ASSOCIATIONKEY** m_assoc;
public:
	Auto_DUL_DropAssociation(DUL_ASSOCIATIONKEY** assoc) : m_assoc(assoc) {}
	~Auto_DUL_DropAssociation()
	{
	if (m_assoc != NULL)
		DUL_DropAssociation(m_assoc);
	}
};
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
class Auto_ClearServiceParameters {
	DUL_ASSOCIATESERVICEPARAMETERS* m_service;
public:
	Auto_ClearServiceParameters(DUL_ASSOCIATESERVICEPARAMETERS* service) : m_service(service) {}
	~Auto_ClearServiceParameters()
	{
	if (m_service != NULL)
		DUL_ClearServiceParameters(m_service);
	}
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
class Auto_MSG_Free {
	LPVOID* m_message;
public:
	Auto_MSG_Free(LPVOID* msg) : m_message(msg) {}
	~Auto_MSG_Free()
	{
		if (*m_message != NULL)
			MSG_Free(m_message);
	}
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
class Auto_print_assoc_close_info {
	const tStorageParam* m_sp;
public:
	Auto_print_assoc_close_info(const tStorageParam* sp)
		: m_sp(sp)
	{
	}
	~Auto_print_assoc_close_info()
	{
		StorageLog::print_assoc_close_info(m_sp);
	}
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static const char * const abstractSyntaxTable[] = {
	DICOM_SOPCLASSVERIFICATION,
	DICOM_SOPCLASSCOMPUTEDRADIOGRAPHY,     /* "1.2.840.10008.5.1.4.1.1.1" */
	DICOM_SOPCLASSCT,                      /* "1.2.840.10008.5.1.4.1.1.2" */
	DICOM_SOPCLASSHARDCOPYCOLORIMAGE,      /* "1.2.840.10008.5.1.1.30" */
	DICOM_SOPCLASSHARDCOPYGRAYSCALEIMAGE,  /* "1.2.840.10008.5.1.1.29" */
	DICOM_SOPCLASSMR,                      /* "1.2.840.10008.5.1.4.1.1.4" */
	DICOM_SOPCLASSNM,                      /* "1.2.840.10008.5.1.4.1.1.20" */
	DICOM_SOPCLASSPET,                     /* "1.2.840.10008.5.1.4.1.1.128" */

	DICOM_SOPRTIMAGESTORAGE,               /* "1.2.840.10008.5.1.4.1.1.481.1" */
	DICOM_SOPRTDOSESTORAGE,                /* "1.2.840.10008.5.1.4.1.1.481.2" */

	DICOM_SOPCLASSSECONDARYCAPTURE,        /* "1.2.840.10008.5.1.4.1.1.7" */

	DICOM_SOPCLASSUS,                      /* "1.2.840.10008.5.1.4.1.1.6.1" */
	DICOM_SOPCLASSUSMULTIFRAMEIMAGE,       /* "1.2.840.10008.5.1.4.1.1.3.1" */

	DICOM_SOPCLASSXRAYANGIO,               /* "1.2.840.10008.5.1.4.1.1.12.1" */
	DICOM_SOPCLASSXRAYFLUORO,              /* "1.2.840.10008.5.1.4.1.1.12.2" */

	DICOM_SOPCLASSDIGXRAYPRESENTATION,     /* "1.2.840.10008.5.1.4.1.1.1.1" */
	DICOM_SOPCLASSDIGXRAYPROCESSING,       /* "1.2.840.10008.5.1.4.1.1.1.1.1" */
	DICOM_SOPCLASSMAMMOXRPRESENTATION,     /* "1.2.840.10008.5.1.4.1.1.1.2" */
	DICOM_SOPCLASSMAMMOXRPROCESSING,       /* "1.2.840.10008.5.1.4.1.1.1.2.1" */
	DICOM_SOPCLASSINTRAORALPRESENTATION,   /* "1.2.840.10008.5.1.4.1.1.1.3" */
	DICOM_SOPCLASSINTRAORALPROCESSING,     /* "1.2.840.10008.5.1.4.1.1.1.3.1" */

	DICOM_SOPCLASSVLENDOSCOPIC,            /* "1.2.840.10008.5.1.4.1.1.77.1.1" */
	DICOM_SOPCLASSVLMICROSCOPIC,           /* "1.2.840.10008.5.1.4.1.1.77.1.2" */
	DICOM_SOPCLASSVLSLIDEMICROSCOPIC,      /* "1.2.840.10008.5.1.4.1.1.77.1.3" */
	DICOM_SOPCLASSVLPHOTOGRAPHIC,          /* "1.2.840.10008.5.1.4.1.1.77.1.4" */

	NULL
};
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static bool supportedClass(char *abstractSyntax)
{
	char** cpp = const_cast<char**>(abstractSyntaxTable);
	while (*cpp != NULL) {
		if (strcmp(abstractSyntax, *cpp) == 0)
			return TRUE;
		else
			cpp++;
	}
	return FALSE;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
/*
static void enqStorageStatQueue(tStorageParam* sp, const tStorageStat& ss)
{
	tStorageParam::stat_info_t& stat = sp->stat;

	int index = stat.ssCount % kStorageStatQueueSize;
	tStorageStat* cur_ss = stat.ssQueue + index;
	if (*cur_ss->path == '\0') {
		*cur_ss = ss;
		stat.ssCount++;
	} else if (strcmp(cur_ss->path, ss.path) == 0) {
		cur_ss->numBytes += ss.numBytes;
		cur_ss->numFiles += ss.numFiles;
	} else {
		stat.ssCount++;
		index = stat.ssCount % kStorageStatQueueSize;
		*cur_ss = ss;
	}



} else {
		
	}
	unsigned long index = stat.ssCount % kStorageStatQueueSize;
	tStorageStat* cur_ss = stat.ssQueue + index;
	if (*cur_ss->path == '\0') {
		*cur_ss = ss;
		stat->ssCount++;
	} else if (strcmp(cur_ss->path, ss.path) == 0) {
		cur_ss->numBytes += ss.numBytes;
		cur_ss->numFiles += ss.numFiles;
	}

	const unsigned int ss_in_queue = stat->ssCount < kStorageStatQueueSize ? stat->ssCount : kStorageStatQueueSize;
	unsigned long long sumBytes = 0;
	unsigned long maxBytes = 0;
	unsigned long sumFiles = 0;
	for (unsigned int i = 0; i < ss_in_queue; i++) {
		sumBytes += stat->ssQueue[i].numBytes;
		sumFiles += stat->ssQueue[i].numFiles;
		if (maxBytes < stat->ssQueue[i].numBytes)
			maxBytes = stat->ssQueue[i].numBytes;
	}
	stat->aveBytes = sumBytes / ss_in_queue;
	stat->aveFiles = sumFiles / ss_in_queue;
	stat->aveFileSize = sumBytes / sumFiles;
	stat->maxBytes = maxBytes;

	unsigned long index = ++stat->ssCount % kStorageStatQueueSize;
	tStorageStat* cur_ss = stat->ssQueue + index;
	*cur_ss = ss;
}
*/

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static int storageProvider(DUL_NETWORKKEY* network, tStorageParam* sp, CStudyDB& studydb)
{
	sp->assoc.reset();

	//
	// config vars
	//
	const tSCPprm* scp = sp->pSCPprm;
	const tVOLprm* vol = scp->volume;

	//
	// Initialize Service Params
	//
	DUL_ASSOCIATESERVICEPARAMETERS service;
	memset(&service, 0, sizeof(service));
	service.maxPDU = 0x8000;
	strcpy(service.calledImplementationClassUID, MIR_IMPLEMENTATIONCLASSUID);
	strcpy(service.calledImplementationVersionName, MIR_IMPLEMENTATIONVERSIONNAME);

	//
	// Wait for Association
	//
	DUL_ASSOCIATIONKEY* association = NULL;
	CONDITION cond = DUL_ReceiveAssociationRQ(&network, DUL_BLOCK, &service, &association);
	if (cond == DUL_NORMAL) {
		strcpy(sp->assoc.callingAPTitle, service.callingAPTitle);
		StorageLog::print_assoc_open_info(sp);
	} else {
		if (cond == DUL_EVENTSIGNALED) {
			COND_PushCondition(APP_STOP, APP_Message(APP_STOP));
			return rsltExit;
		} else {
			cond = COND_PushCondition(APP_SERVICEERROR, APP_Message(APP_SERVICEERROR),
			                          scp->label, "N/C", "DUL_ReceiveAssociationRQ()");
			return rsltError;
		}
	}
	Auto_DUL_DropAssociation auto_drop_association(&association);
	Auto_ClearServiceParameters clear_service(&service);
	Auto_print_assoc_close_info ap(sp);

	//
	// Check Association
	//
	cond = APP_NORMAL;
	DUL_ABORTITEMS abortItems;
	if (strcmp(service.applicationContextName, DICOM_STDAPPLICATIONCONTEXT) != 0) {
		abortItems.result = DUL_REJECT_PERMANENT;
		abortItems.source = DUL_ULSU_REJECT;
		abortItems.reason = DUL_ULSU_REJ_UNSUP_APP_CTX_NAME;
		cond = COND_PushCondition(APP_ILLEGALSERVICEPARAMETER, APP_Message(APP_ILLEGALSERVICEPARAMETER),
		                          "Application Context Name", service.applicationContextName);
	} else if (strlen(service.callingImplementationClassUID) == 0) {
		abortItems.result = DUL_REJECT_PERMANENT;
		abortItems.source = DUL_ULSU_REJECT;
		abortItems.reason = DUL_ULSU_REJ_NOREASON;
		cond = COND_PushCondition(APP_MISSINGSERVICEPARAMETER, APP_Message(APP_MISSINGSERVICEPARAMETER),
	                          "Requestor's ImplementationClassUID");
	} else if (_stricmp(scp->aetitle, service.calledAPTitle) != 0) {
		abortItems.result = DUL_REJECT_PERMANENT;
		abortItems.source = DUL_ULSU_REJECT;
		abortItems.reason = DUL_ULSU_REJ_UNREC_CALLED_TITLE;
		cond = COND_PushCondition(APP_ILLEGALSERVICEPARAMETER, APP_Message(APP_ILLEGALSERVICEPARAMETER),
	                  "AE Title", service.calledAPTitle);
	}
	
	if (cond != APP_NORMAL) {
		//
		// Reject Association
		//
		DUL_RejectAssociationRQ(&association, &abortItems);
		return rsltNormal;
	}

	//
	// Lookup Requested Service
	//
	DUL_PRESENTATIONCONTEXT* requestedCtx = (DUL_PRESENTATIONCONTEXT*)LST_Head(&service.requestedPresentationContext);
	if (requestedCtx == NULL) {
		cond = COND_PushCondition(APP_SERVICEERROR, APP_Message(APP_SERVICEERROR),
		                          scp->label, service.callingAPTitle, "Presentation Ctx List");
		return rsltError;
	}
	LST_Position(&service.requestedPresentationContext, requestedCtx);
	int classCount = 0;
	for (; requestedCtx != NULL; requestedCtx = (DUL_PRESENTATIONCONTEXT*)LST_Next(&service.requestedPresentationContext)) {
		if (supportedClass(requestedCtx->abstractSyntax)) {
			DUL_SC_ROLE scRole;
			switch (requestedCtx->proposedSCRole) {
			case DUL_SC_ROLE_DEFAULT:
				scRole = DUL_SC_ROLE_DEFAULT;
				break;
			case DUL_SC_ROLE_SCU:
				scRole = DUL_SC_ROLE_SCU;
				break;
			case DUL_SC_ROLE_SCP:
				scRole = DUL_SC_ROLE_DEFAULT;
				break;
			case DUL_SC_ROLE_SCUSCP:
				scRole = DUL_SC_ROLE_SCU;
				break;
			default:
				scRole = DUL_SC_ROLE_DEFAULT;
				break;
			}

			static const char* const non_compressed[] = {
				DICOM_TRANSFERLITTLEENDIANEXPLICIT,
				DICOM_TRANSFERLITTLEENDIAN,
				NULL
			};
			static const char* const all_supported[] = {
				DICOM_TRANSFERJPEGLOSSLESSPROCFIRSTORDERREDICT,
				DICOM_TRANSFERJPEGLOSSLESSPROC14,
				DICOM_TRANSFERJPEGEXTENDEDPROC2AND4,
				DICOM_TRANSFERJPEGBASELINEPROCESS1,
				DICOM_TRANSFERLITTLEENDIANEXPLICIT,
				DICOM_TRANSFERLITTLEENDIAN,
				NULL
			};
			const char** transferSyntaxTable = const_cast<const char**>(scp->transfer_syntax_flag == 1 ? non_compressed : all_supported);
			cond = SRV_AcceptServiceClassEx(requestedCtx, scRole, &service, const_cast<char**>(abstractSyntaxTable), const_cast<char**>(transferSyntaxTable));
			if (cond == SRV_NORMAL)
				classCount++;
		} else {
			cond = SRV_RejectServiceClass(requestedCtx, DUL_PRESENTATION_REJECT_USER, &service);
			if (cond != SRV_NORMAL) {
				cond = COND_PushCondition(APP_SERVICEERROR, APP_Message(APP_SERVICEERROR),
				                          scp->label, service.callingAPTitle, "SRV_RejectServiceClass()");
				return rsltError;
			}
		}
	}

	if (classCount == 0) {
		//
		// No Accepted Service Class
		//
		COND_PushCondition(APP_ILLEGALSERVICEPARAMETER, APP_Message(APP_ILLEGALSERVICEPARAMETER),
	                  "No Syntax Accepted", "N/A");
		abortItems.result = DUL_REJECT_PERMANENT;
		abortItems.source = DUL_ULSP_PRESENTATION_REJECT;
		abortItems.reason = DUL_ULSP_PRES_REJ_TEMP_CONGEST;
		cond = DUL_RejectAssociationRQ(&association, &abortItems);
		return rsltNormal;
	}

	//
	// Acknowledge Association Request
	//
	cond = DUL_AcknowledgeAssociationRQ(&association, &service);
	if (cond != DUL_NORMAL) {
		cond = COND_PushCondition(APP_SERVICEERROR, APP_Message(APP_SERVICEERROR),
		                          scp->label, service.callingAPTitle, "DUL_AcknowledgeAssociationRQ()");
		return rsltError;
	}

	//
	// Study Management
	//
	CStudyCache studycache(true);
	CStudyDBConnector studydb_conn(studydb, gDWConfig->odbc.dsn);

	//
	// Response Each Request Commands
	//
	int result = rsltNormal;
	while (true) {

		//
		// Force to Stop
		//
		if (WaitForSingleObject(sp->wsp.hStopEvent, 0) == WAIT_OBJECT_0) {
			COND_PushCondition(APP_STOP, APP_Message(APP_STOP));
			DUL_ReleaseAssociation(&association);
			result = rsltExit;
			break;		
		}

		void* message = NULL;
		Auto_MSG_Free free_msg(&message);

		MSG_TYPE messageType;
		DUL_PRESENTATIONCONTEXTID ctxID;
		CONDITION cond = SRV_ReceiveCommand(&association, &service, DUL_BLOCK, 0, &ctxID, (unsigned short*)NULL, &messageType, &message);
		if (cond == SRV_PEERREQUESTEDRELEASE) {
			DUL_AcknowledgeRelease(&association);
			COND_PopCondition(FALSE);
			break;
		} else if (cond == SRV_PEERABORTEDASSOCIATION) {
			cond = COND_PushCondition(APP_SERVICEERROR, APP_Message(APP_SERVICEERROR),
			                          scp->label, service.callingAPTitle, "peer abort in SRV_ReceiveCommand()");
			result = rsltError;
			break;
		} else if (!CTN_SUCCESS(cond)) {
			cond = COND_PushCondition(APP_SERVICEERROR, APP_Message(APP_SERVICEERROR),
			                          scp->label, service.callingAPTitle, "failed in SRV_ReceiveCommand()");
			result = rsltError;
			break;
		}
		sp->assoc.commandCount++;

		//
		// Matching Presentation Context ID
		//
		DUL_PRESENTATIONCONTEXT* ctx = (DUL_PRESENTATIONCONTEXT*)LST_Head(&service.acceptedPresentationContext);
		if (ctx != NULL)
			LST_Position(&service.acceptedPresentationContext, ctx);
		for (; ctx != NULL; ctx = (DUL_PRESENTATIONCONTEXT*)LST_Next(&service.acceptedPresentationContext))
			if (ctx->presentationContextID == ctxID)
				break;

		if (NULL != ctx && MSG_K_C_ECHO_REQ == messageType) {
			//
			// Echo Requested
			//
			handleEchoRequest(&association, ctx, (MSG_C_ECHO_REQ**)&message);
		} else if (NULL != ctx && MSG_K_C_STORE_REQ == messageType) {
			//
			// Store Requested
			//
			sp->assoc.requestCount++;

/*
			//
			// Force to Stop
			//
			if (WaitForSingleObject(sp->wsp.hStopEvent, 0) == WAIT_OBJECT_0) {
				COND_PushCondition(APP_STOP, APP_Message(APP_STOP));
				sendCancelResponse(&association, ctx, (MSG_C_STORE_REQ**)&message);
				result = rsltExit;
				continue;				
			}
*/

			//
			// Service to Request
			//
			cond = handleStoreRequest(&association, ctx, (MSG_C_STORE_REQ**)&message, &service, sp, studydb, studycache);
			if (cond == DUL_EVENTSIGNALED) {
				result = rsltExit;
				break;				
			} else if (!CTN_SUCCESS(cond)) {
				result = rsltError;
				break;
			}
		} else {
			//
			// Got Bad Request
			//
			if (NULL == ctx) {
				COND_PushCondition(APP_BADREQUEST, "[WARN|Presentation Context (%d) had not accepted (Command: %d)]", (int)ctxID, (int)messageType);
			} else {
				COND_PushCondition(APP_BADREQUEST, "[WARN|Unexpected Command: %d]", (int)messageType);
			}
		}
	}

	return result;
}

//========================================================================================
//
//
//
//========================================================================================
struct tRTLogger {
	HANDLE hFile;
	HANDLE hMap;
	LPVOID lpMap;
	BSTR bsLogFile;
	DWORD dwLastError;

	void init(void);
	void finish(void);

	CRITICAL_SECTION crit;

	static void logWriter(LPVOID data, DWORD size);

	struct logheader_t {
		DWORD size;
		DWORD head;	// oldest 
		DWORD tail; // beyond the last
		DWORD last; // newest -- for reader
		DWORD overlap; // ;
		DWORD data;
	};

} rtlogger;
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tRTLogger::init(void)
{
	ZeroMemory(this, sizeof(tRTLogger));
	hFile = INVALID_HANDLE_VALUE;
	InitializeCriticalSection(&crit);

	bsLogFile = SysAllocStringLen(NULL, SysStringLen(gDWConfig->logDir) + lstrlen(L"\\dio-storage-rt.log"));
	lstrcpy(bsLogFile, gDWConfig->logDir);
	lstrcat(bsLogFile, L"\\dio-storage-rt.log");

	hFile = CreateFile(bsLogFile, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		dwLastError = GetLastError();
		return;
	}

	DWORD dwMapSize = gDWConfig->rtlog * 1024;

	SECURITY_ATTRIBUTES sa = { sizeof(sa), NULL, FALSE };

	HANDLE hToken;
	if (OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken)) {
		AutoCloseHandle ac_hToken(hToken);
		DWORD daclBufLen = 0;
		GetTokenInformation(hToken, TokenDefaultDacl, NULL, 0, &daclBufLen);
		if (daclBufLen != 0) {
			PTOKEN_DEFAULT_DACL dacl = (PTOKEN_DEFAULT_DACL)_alloca(daclBufLen);
			if (GetTokenInformation(hToken, TokenDefaultDacl, dacl, daclBufLen, &daclBufLen)) {
				ACL_SIZE_INFORMATION acl_size;
				if (GetAclInformation(dacl->DefaultDacl, &acl_size, sizeof(acl_size), AclSizeInformation)) {
					DWORD newAclSize = sizeof(ACL);
					for (int i = 0; i < acl_size.AceCount; i++) {
						ACE_HEADER* ace;
						GetAce(dacl->DefaultDacl, i, (LPVOID*)&ace);
						newAclSize += ace->AceSize;
					}
					SID sid = {SID_REVISION, 1, SECURITY_WORLD_SID_AUTHORITY, SECURITY_INTERACTIVE_RID};
					newAclSize += sizeof(ACCESS_ALLOWED_ACE) + GetLengthSid(&sid);
					PACL newAcl = (PACL)_alloca(newAclSize);
					if (InitializeAcl(newAcl, newAclSize, ACL_REVISION)) {
						for (int i = 0; i < acl_size.AceCount; i++) {
							ACE_HEADER* ace;
							GetAce(dacl->DefaultDacl, i, (LPVOID*)&ace);
							if (!AddAce(newAcl, ACL_REVISION, i, ace, ace->AceSize)) {
								break;
							}
						}
						if (AddAccessAllowedAce(newAcl, ACL_REVISION, GENERIC_READ, &sid)) {
							SECURITY_DESCRIPTOR sd;
							InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
							if (SetSecurityDescriptorDacl(&sd, TRUE, newAcl, FALSE)) {
								sa.lpSecurityDescriptor = &sd;
							}
						}
					}
				}
			}
		}
	}
	DWORD err = GetLastError();

	hMap = CreateFileMapping(hFile, &sa, PAGE_READWRITE, 0, dwMapSize, L"DIO_STORAGE_RT_LOG");
	if (hMap == NULL) {
		dwLastError = GetLastError();
		CloseHandle(hFile); hFile = INVALID_HANDLE_VALUE;
		return;
	}

	lpMap = MapViewOfFile(hMap, FILE_MAP_WRITE, 0, 0, 0);
	if (lpMap == NULL) {
		dwLastError = GetLastError();
		CloseHandle(hMap); hMap = NULL;
		CloseHandle(hFile); hFile = INVALID_HANDLE_VALUE;
		return;
	}

	ZeroMemory(lpMap, dwMapSize);
	logheader_t* header = (logheader_t*)lpMap;
	header->size = dwMapSize;

	rtlog_writer = logWriter;
	rtlog_on = 1;
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tRTLogger::finish(void)
{
	DeleteCriticalSection(&crit);

	if (lpMap != NULL)
		UnmapViewOfFile(lpMap);
	if (hMap != NULL)
		CloseHandle(hMap);
	if (hFile != INVALID_HANDLE_VALUE)
		CloseHandle(hFile);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void tRTLogger::logWriter(LPVOID data, DWORD size)
{
	CStackCriticalSection cs(&rtlogger.crit);

	struct log_t {
		unsigned short length;
		struct SHORT_SYTEMTIME {
			unsigned char month;
			unsigned char day;
			unsigned char hour;
			unsigned char minute;
			unsigned char second;
			unsigned char sub_sec;
		} time;
		DWORD data;
	} log;

	SYSTEMTIME time;
	GetLocalTime(&time);
	log.time.month = time.wMonth;
	log.time.day = time.wDay;
	log.time.hour = time.wHour;
	log.time.minute = time.wMinute;
	log.time.second = time.wSecond;
	log.time.sub_sec = time.wMilliseconds / 10;

	log.length = offsetof(log_t, data) + size;

	DWORD last = 0;
	logheader_t* header = (logheader_t*)rtlogger.lpMap;
	if (header->tail == 0) { // first time
		DWORD ofs = offsetof(logheader_t, data);
		header->head = ofs;
		last = ofs;
		header->tail = ofs + log.length;
	} else {
		if (header->tail + log.length >= header->size) {
			char* ptr = (char*)header;
			ptr += header->tail;
			*(unsigned short*)ptr = 0;
			header->overlap++;

			last = offsetof(logheader_t, data);
			header->tail = last + log.length;
			header->head = offsetof(logheader_t, data);
			while (header->head < header->tail) {
				char* ptr = (char*)header;
				ptr += header->head;
				unsigned short len = ((log_t*)ptr)->length;
				header->head += len;
			}
		} else {
			last = header->tail;
			header->tail = last + log.length;

			if (header->head < header->tail && header->overlap) {
				while (header->head < header->tail) {
					char* ptr = (char*)header;
					ptr += header->head;
					unsigned short len = ((log_t*)ptr)->length;
					if (len == 0) {
						header->head = offsetof(logheader_t, data);
						break;
					} else {
						header->head += len;
					}
				}
			}
		}
	}

	char* ptr = (char*)header;
	ptr += last;

	CopyMemory(ptr, &log, offsetof(log_t, data));
	ptr += offsetof(log_t, data);
	CopyMemory(ptr, data, size);

	header->last = last;

}

//========================================================================================
//
//========================================================================================
void InitRTLog(void)
{
	if (gDWConfig->rtlog == 0)
		return;

	rtlogger.init();
}
//========================================================================================
//
//========================================================================================
void FinRTLog(void)
{
	if (gDWConfig->rtlog == 0)
		return;

	rtlogger.finish();
}


//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static DWORD thread_proc(LPVOID lpParam)
{
    HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
    if (FAILED(hr)) {
		return 1;
	}
	AutoCoUninitialize acu;

	// dummy call to create TLSs
	Get_CTN_HEAP_HANDLE();
	COND_Empty();

	tStorageParam* sp = (tStorageParam*)lpParam;

	DUL_NETWORKKEY* network = NULL;
	CONDITION cond = DUL_InitializeNetworkEx(DUL_NETWORK_TCP, DUL_AEACCEPTOR, (void*)&sp->pSCPprm->port, DUL_TIMEOUT, DUL_ORDERBIGENDIAN, (void*)&sp->wsp, &network);
	if (cond != DUL_NORMAL) {
		return 1;
	}

	CStudyDB studydb;

	while (true) {
		int result = storageProvider(network, sp, studydb);
		COND_PopCondition(TRUE);
		if (result == rsltExit)
			break;
	}

	DUL_DropNetworkEx(&network);

	return 0;
} 
//========================================================================================
//
//========================================================================================
DWORD WINAPI StorageProviderThread(LPVOID lpParam)
{
	_endthreadex(thread_proc(lpParam));
	return 0;
}

