/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "cstudydb.h"
#include "dcmdict.h"
#include "CDcmScanner.h"

#include "StudyCache.h"

#include "rtlog.h"

#include "app_cond.h"

#include "dio-storage.h"

#ifdef  MIR_IMPLEMENTATIONCLASSUID
#undef  MIR_IMPLEMENTATIONCLASSUID
#define	MIR_IMPLEMENTATIONCLASSUID		"1.2.840.113654.2.3.1995.3.0.0.1"
#endif

#ifdef	MIR_IMPLEMENTATIONVERSIONNAME
#undef	MIR_IMPLEMENTATIONVERSIONNAME
#define	MIR_IMPLEMENTATIONVERSIONNAME		"MIRCTN19FEB2001 MODIFIED FOR DIOWAVE "
#endif

CONDITION storeToFile(DUL_ASSOCIATIONKEY** association, DUL_PRESENTATIONCONTEXT* ctx, MSG_C_STORE_REQ** storeRequest, DUL_ASSOCIATESERVICEPARAMETERS* service, tStorageParam* sp, CStudyDB& studydb, CStudyCache& studycache, unsigned short& resultCode, bool& send_response);

//----------------------------------------------------------------------------------------
// extern
//----------------------------------------------------------------------------------------
extern "C" {
	CONDITION SRVPRV_ReadNextPDV(DUL_ASSOCIATIONKEY ** association, DUL_BLOCKOPTIONS block, int timeout, DUL_PDV * pdv);
}
const DWORD kDefaultDiskSpaceMargine = 16 * 1024 * 1024;

//----------------------------------------------------------------------------------------
////
//// STORE
////
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
// study Management
//
//----------------------------------------------------------------------------------------
static CRITICAL_SECTION* vol_crit_tbl = NULL;
//========================================================================================
//
//========================================================================================
void InitVolumeMangement()
{
	if (vol_crit_tbl == NULL) {
		vol_crit_tbl = (CRITICAL_SECTION*)GlobalAlloc(GMEM_FIXED, sizeof(CRITICAL_SECTION) * gDWConfig->num_vol);
		for (int i = 0; i < gDWConfig->num_vol; i++)
			InitializeCriticalSection(vol_crit_tbl + i);
	}
}
//========================================================================================
//
//========================================================================================
void FinVolumeManagement()
{
	if (vol_crit_tbl != NULL) {
		for (int i = 0; i < gDWConfig->num_vol; i++)
			DeleteCriticalSection(vol_crit_tbl + i);
		GlobalFree(vol_crit_tbl);
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static bool deleteDirectoryContent(const wchar_t* dir_path, size_t dir_path_len, const WIN32_FIND_DATA& find_data, DWORD& bytesDeleted)
{
	static bool deleteDirectoryContents(const wchar_t* dir_path, DWORD& bytesDeleted);

	if ((FILE_ATTRIBUTE_DIRECTORY & find_data.dwFileAttributes) != 0 && *find_data.cFileName == L'.')
		return true;

	const int cFileName_len = lstrlen(find_data.cFileName);

	wchar_t* const file = (wchar_t*)_alloca(sizeof(wchar_t) * (dir_path_len + 1 + cFileName_len + 1));
	strncat3(file, dir_path, dir_path_len, L'\\', find_data.cFileName, cFileName_len);

	if ((FILE_ATTRIBUTE_DIRECTORY & find_data.dwFileAttributes) != 0) {
		if (!deleteDirectoryContents(file, bytesDeleted)) {
			return false;
		}
		if (!RemoveDirectory(file)) {
			DWORD dwErr = GetLastError();
			TOAC(file)
			COND_PushWin32Error(APP_WIN32ERROR, dwErr, "RemoveDirectory", _file_AC);
			return false;
		}
	} else {
		if (!DeleteFile(file)) {
			DWORD dwErr = GetLastError();
			TOAC(file)
			COND_PushWin32Error(APP_WIN32ERROR, dwErr, "DeleteFile", _file_AC);
			return false;
		}
		bytesDeleted += find_data.nFileSizeLow;
	}

	return true;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static bool deleteDirectoryContents(const wchar_t* dir_path, DWORD& bytesDeleted)
{
	CStackFindFile find(dir_path);
	if (find.handle == INVALID_HANDLE_VALUE)
		return false;

	const size_t dir_path_len = lstrlen(dir_path);

	//---------------------
	// File Scan Loop
	//---------------------
	do {
		if (!deleteDirectoryContent(dir_path, dir_path_len, find.data, bytesDeleted)) {
			return false;
		}
	} while (find.next());

	return true;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static int delete_study_directory(const tStorageParam* sp, CStudyDB& studydb, const wchar_t* vol_dir, const wchar_t* path, DWORD seq_no)
{
	size_t vol_dir_len = lstrlen(vol_dir);
	size_t path_len = lstrlen(path);

	wchar_t* const dir_path = (wchar_t*)_alloca(sizeof(wchar_t) * (vol_dir_len + 1 + path_len + 1));
	strncat3(dir_path, vol_dir, vol_dir_len, L'\\', path, path_len);

	wchar_t* const delete_tmp = (wchar_t*)_alloca(sizeof(wchar_t) * (vol_dir_len + 1 + 10 + 1));
	strncat3(delete_tmp, vol_dir, vol_dir_len, L'\\', L"delete_tmp", 10);

	if (MoveFile(dir_path, delete_tmp) == 0) {
		const DWORD dwErr = GetLastError();
		TOAC(dir_path)
		COND_PushWin32Error(APP_WIN32ERROR, dwErr, "MoveFile", _dir_path_AC);
		return 1;
	}

	if (!studydb.ClearStudy(seq_no)) {
		wchar_t* message = studydb.m_message;
		TOAC(message)
		COND_PushCondition(APP_ODBCERROR, APP_Message(APP_ODBCERROR), _message_AC);
		return -1;
	}

	unsigned long bytesDeleted = 0;
	if (!deleteDirectoryContents(delete_tmp, bytesDeleted))
		return -2;

	if (RemoveDirectory(delete_tmp) == 0) {
		const WORD dwErr = GetLastError();
		TOAC(delete_tmp)
		COND_PushWin32Error(APP_WIN32ERROR, dwErr, "RemoveDirectory", _delete_tmp_AC);
		return -3;
	}

	StorageLog::print_remove_study_info(sp, path, bytesDeleted);

	return 0;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static bool checkDiskSpece(const ULARGE_INTEGER& bytesAvail, const DWORD margineBytes)
{
	return !(bytesAvail.HighPart == 0 && bytesAvail.LowPart < margineBytes);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static bool ensure_disk_space(CStudyDB& studydb, const tStorageParam* sp, const wchar_t* vol_path, long free_space, unsigned short& resultCode)
{
	const tSCPprm* scp = sp->pSCPprm;
	const tVOLprm* vol = scp->volume;

	//
	// Lock Volume
	//
	CStackCriticalSection cs(vol_crit_tbl + vol->index);

	// Get Volume Disk Space
	ULARGE_INTEGER bytesAvail, totalBytes;
	if (!GetDiskFreeSpaceEx(vol_path, &bytesAvail, &totalBytes, NULL)) {
		COND_PushWin32Error(APP_WIN32ERROR, GetLastError(), "GetDiskFreeSpaceEx", vol->dir);
		return true;
	}

	if (vol->auto_delete_flag < 0 || !studydb.connected()) {
		//
		// no volume management
		//
		if (checkDiskSpece(bytesAvail, free_space)) {
			return true;
		} else {
			COND_PushCondition(APP_DISKFULL, "Disk Full: VOLUME=%s MARGIN=%u", vol->label, free_space);
			resultCode = MSG_K_RESOURCELIMITATION;
			return false;
		}
	} else if (checkDiskSpece(bytesAvail, free_space)) {
		return true;
	}

	const char* const volume_label = vol->label;
	TOWC(volume_label)

	const char* const vol_dir = vol->dir;
	TOWC(vol_dir)

	BSTR extra_cond = SysAllocStringLen(NULL, 0);
	AutoSysFreeString af_extra_cond(extra_cond);

	BSTR select = CStudyDB::make_select_oldest_sql(_volume_label_WC, vol->auto_delete_flag, extra_cond);
	AutoSysFreeString af_select(select);

	bool err = false;
	do {

		DWORD seq_no = 0;
		BSTR path = NULL;
		AutoSysFreeString af_path(path);
		{
			AutoComPtr<ADORecordset> rs;

			if (!studydb.Execute(select, rs.r())) {
				wchar_t* message = studydb.m_message;
				TOAC(message)
				COND_PushCondition(APP_ODBCERROR, APP_Message(APP_ODBCERROR), _message_AC);
				err = true;
				break;
			}

			CStudyRecordset studyrs;
			studyrs.SetReocrdset(rs.p());

			if (studyrs.End()) {
				TOAC(select)
				COND_PushCondition(APP_ODBCERROR, "%s returns no record", _select_AC);
				err = true;
				rs->Close();
				break;
			}

			VARIANT* vp;

			vp = studyrs.Field("SEQ_NO");
			if (vp == NULL || vp->vt != VT_I4) {
				COND_PushCondition(APP_ODBCERROR, "SEQ_NO is NULL or not VT_I4");
				err = true;
				rs->Close();
				break;
			}
			seq_no = vp->lVal;

			vp = studyrs.Field("PATH");
			if (vp == NULL || vp->vt != VT_BSTR) {
				COND_PushCondition(APP_ODBCERROR, "PATH is NULL or not VT_BSTR");
				err = true;
				rs->Close();
				break;
			}
			path = SysAllocString(vp->bstrVal);

			rs->Close();
		}

		const int r = delete_study_directory(sp, studydb, _vol_dir_WC, path, seq_no);
		if (r < 0) {
			err = true;
			break;
		} else if (r > 0) {
			wchar_t str[64];
			wsprintf(str, L" AND SEQ_NO <> %u", seq_no);
			const int len = SysStringLen(extra_cond) + lstrlen(str);
			BSTR tmp_extra_cond = SysAllocStringLen(NULL, len);
			lstrcpy(tmp_extra_cond, extra_cond);
			lstrcat(tmp_extra_cond, str);
			SysFreeString(extra_cond);
			extra_cond = tmp_extra_cond;
			SysFreeString(select);
			select = CStudyDB::make_select_oldest_sql(_volume_label_WC, vol->auto_delete_flag, extra_cond);
		}

		GetDiskFreeSpaceEx(vol_path, &bytesAvail, &totalBytes, NULL);
	} while (!checkDiskSpece(bytesAvail, free_space));

	if (err) {
		GetDiskFreeSpaceEx(vol_path, &bytesAvail, &totalBytes, NULL);
		if (checkDiskSpece(bytesAvail, free_space)) {
			return true;
		} else {
			COND_PushCondition(APP_DISKFULL, "Failed to Free Disk Space and Finally Disk Full: VOLUME=%s MARGIN=%u", vol->label, free_space);
			resultCode = MSG_K_RESOURCELIMITATION;
			return false;
		}
	} else {
		return true;
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static BSTR make_insert_sql(const char* const scp_title, const char* const scu_title, const char* const volume_label, const char* const path, const char* const ref_file, const char* const elements, size_t elementsLength)
{
	struct {
		DCM_TAG tag;
		const char* str;
		size_t len;
	} element_tbl[] = {
		{ 0x00080020, "", 0 }, // 0 std_date
		{ 0x00080030, "", 0 }, // 1 std_time
		{ 0x00080060, "", 0 }, // 2 modality
		{ 0x00100010, "", 0 }, // 3 pat_name
		{ 0x00100020, "", 0 }, // 4 pat_id
		{ 0x00200010, "", 0 }, // 5 std_id
		{ 0xFFFFFFFF, "", 0 }
	};

	CDcmScannerF scanner(elements, elements + elementsLength);
	const char* cp = elements;

	int index = 0;

	while (true) {
		tDcmElementHdr hdr;
		const int r = scanner.NextElement(cp, hdr, true);
		if (r != CDcmScannerF::S_CONTINUE)
			break;
		DCM_TAG ltag = hdr.ltag();

		while (element_tbl[index].tag != 0xFFFFFFFF && ltag > element_tbl[index].tag)
			index++;
		if (element_tbl[index].tag == 0xFFFFFFFF)
			break;

		if (ltag == element_tbl[index].tag) {
			const unsigned long len = hdr.value_length();
			if (len > 0) {
				const unsigned char* sp = (const unsigned char*)hdr.value_ptr(cp);
				const unsigned char* ep = sp + len;

				while (sp < ep && *sp <= ' ')
					sp++;
				while (sp < ep && *(ep - 1) <= ' ')
					ep--;

				element_tbl[index].str = (char*)sp;
				element_tbl[index].len = ep - sp;
			}
			index++;
		}
	
		cp = hdr.next_ptr(cp);
	}

	const char* sp;
	char* dp;

	int i = 0;
	char* const std_date = (char*)_alloca(element_tbl[i].len + 1);
	sp = element_tbl[i].str;
	dp = std_date;
	for (int ci = 0; ci < element_tbl[i].len; ci++) {
		char c = *sp++;
		if (isdigit(c))
			*dp++ = c;
	}
	*dp = '\0';

	i++;
	char* const std_time = (char*)_alloca(element_tbl[i].len + 1);
	sp = element_tbl[i].str;
	dp = std_time;
	for (int ci = 0; ci < element_tbl[i].len; ci++) {
		char c = *sp++;
		if (isdigit(c) || c == '.')
			*dp++ = c;
	}
	*dp = '\0';

	i++;
	char* const modality = (char*)_alloca(element_tbl[i].len + 1);
	memcpy(modality, element_tbl[i].str, element_tbl[i].len);
	modality[element_tbl[i].len] = '\0';

	i++;
	char* const pat_name = (char*)_alloca(element_tbl[i].len + 1);
	memcpy(pat_name, element_tbl[i].str, element_tbl[i].len);
	pat_name[element_tbl[i].len] = '\0';

	i++;
	char* const pat_id = (char*)_alloca(element_tbl[i].len + 1);
	memcpy(pat_id, element_tbl[i].str, element_tbl[i].len);
	pat_id[element_tbl[i].len] = '\0';

	i++;
	char* const std_id = (char*)_alloca(element_tbl[i].len + 1);
	memcpy(std_id, element_tbl[i].str, element_tbl[i].len);
	std_id[element_tbl[i].len] = '\0';

	return CStudyDB::make_insert_sql(volume_label, scp_title, scu_title, path, ref_file, pat_id, pat_name, std_date, std_time, std_id, modality);
}

//----------------------------------------------------------------------------------------
//
// study Directory Constructor
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static int build_full_path(const char* vol_dir, char** path_components, int num_components, wchar_t* full_path, int wlen)
{
	if (wlen == 0) {
		wlen = multi_byte_to_wide_char(vol_dir, -1, NULL, 0) - 1;

		char** cpp = path_components;
		for (int i = 0; i < num_components; i++) {
			char* pcp = *cpp++;
			wlen += 1 + (multi_byte_to_wide_char(pcp, -1, NULL, 0) - 1);
		}

		return wlen;
	}

	wchar_t* cp = full_path;
	const wchar_t* const ep = cp + wlen + 1;

	cp += multi_byte_to_wide_char(vol_dir, -1, cp, ep - cp) - 1;

	char** cpp = path_components;
	for (int i = 0; i < num_components; i++) {
		char* pcp = *cpp++;
		*cp++ = L'\\';
		cp += (multi_byte_to_wide_char(pcp, -1, cp, ep - cp) - 1);
	}

	return wlen;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static CONDITION create_image_directory(const wchar_t* image_dir)
{
	int wlen = lstrlen(image_dir);
	wchar_t* const path_buf = (wchar_t*)_alloca(sizeof(wchar_t) * (wlen + 1));
	lstrcpy(path_buf, image_dir);

	wchar_t* ep = path_buf + wlen;
	wchar_t* cp = ep;
	while (true) {
		DWORD attr = GetFileAttributes(path_buf);
		if (attr != 0xFFFFFFFF && (attr & FILE_ATTRIBUTE_DIRECTORY) != 0)
			break;
		cp--;
		while (cp > path_buf && *cp != L'\\')
			cp--;
		if (cp == path_buf) {
			return COND_PushCondition(APP_STOREERROR, APP_Message(APP_STOREERROR), "create_image_directory()");
		}
		*cp = L'\0';
	}

	while (cp < ep) {
		*cp = L'\\';
		if (!CreateDirectory(path_buf, NULL)) {
			DWORD dwErr = GetLastError();
			TOAC(image_dir)
			return COND_PushWin32Error(APP_WIN32ERROR, dwErr, "CreateDirectory", _image_dir_AC);
		}
		while (cp < ep && *cp != L'\0')
			cp++;
		if (cp == ep)
			break;
	}

	return APP_NORMAL;
}
//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static CONDITION writeMetaHeader(HANDLE hFile, const char* scannedElements, size_t scannedElementsSize, const char* transfer_syntax)
{
	static bool inited = false;
	static struct {
		char zero[128];
		char dicm[4];
	} preamble;
	static size_t impl_uid_len, impl_ver_name_len;
	const char* impl_uid = MIR_IMPLEMENTATIONCLASSUID;
	const char* impl_ver_name = MIR_IMPLEMENTATIONVERSIONNAME;

	if (!inited) {
		ZeroMemory(preamble.zero, sizeof(preamble.zero));
		memcpy(preamble.dicm, "DICM", sizeof(preamble.dicm));

		impl_uid_len = strlen(impl_uid);
		if (impl_uid_len & 1)
			impl_uid_len++;	// UID elements padding is NULL, so alread has NULL

		impl_ver_name_len = strlen(impl_ver_name);
		if (impl_ver_name_len & 1) {
			impl_ver_name_len--;	// already single space is padded
		}

		inited = true;
	}

	DWORD bytesWritten;
	if (!WriteFile(hFile, &preamble, sizeof(preamble), &bytesWritten, NULL)) {
		return COND_PushWin32Error(APP_WIN32ERROR, GetLastError(), "WriteFile", "META HEADER - PREAMBLE");
	}

	size_t class_uid_len, instance_uid_len;
	const char* class_uid = NULL;
	const char* instance_uid = NULL;

	CDcmScannerF scanner(scannedElements, scannedElements + scannedElementsSize);
	const char* sp = scannedElements;
	tDcmElementHdr hdr;
	int r;
	while ((r = scanner.NextElement(sp, hdr, true)) == CDcmScannerF::S_CONTINUE) {
		const DCM_TAG tag = hdr.ltag();
		if (tag == DCM_IDSOPCLASSUID) {
			class_uid = hdr.value_ptr(sp);
			class_uid_len = hdr.value_length();
		} else if (tag == DCM_IDSOPINSTANCEUID) {
			instance_uid = hdr.value_ptr(sp);
			instance_uid_len = hdr.value_length();
			break;
		} 
		sp = hdr.next_ptr(sp);
	}

	size_t transfer_syntax_len = strlen(transfer_syntax);
	if (transfer_syntax_len & 1)
		transfer_syntax_len++;

	const size_t FileMetaInfoVersion_size = 4 + 2 + 2 + 4 + 2;	// OB: tag + vr + pad + len + val;
	const size_t SOPClassUID_size = 4 + 2 + 2 + class_uid_len;
	const size_t SOPInstanceUID_size = 4 + 2 + 2 + instance_uid_len;
	const size_t TransferSyntaxUID_size = 4 + 2 + 2 + transfer_syntax_len;
	const size_t ImplClassUID_size = 4 + 2 + 2 + impl_uid_len;
	const size_t ImplVerName_size = 4 + 2 + 2 + impl_ver_name_len;

	const unsigned long group_length = FileMetaInfoVersion_size + SOPClassUID_size + SOPInstanceUID_size + TransferSyntaxUID_size + ImplClassUID_size + ImplVerName_size;
	const unsigned long meta_length = 4 + 2 + 2 + 4 + group_length;

	char* const meta = (char*)_alloca(meta_length);
	char* cp = meta;

	union header_t {
		struct {
			unsigned short tag[2];
			char vr[2];
			unsigned short length;
		} s;
		struct {
			unsigned short tag[2];
			char vr[2];
			unsigned short foo;
			unsigned long length;
		} l;
	} *h;

	h = (header_t*)cp;
	h->s.tag[0] = DCM_GROUPFILEMETA;
	h->s.tag[1] = 0x0000;
	h->s.vr[0] = 'U';
	h->s.vr[1] = 'L';
	h->s.length = 4;
	cp += 8;
	*(unsigned long*)cp = group_length;
	cp += 4;

	h = (header_t*)cp;
	h->l.tag[0] = DCM_GROUPFILEMETA;
	h->l.tag[1] = 0x0001;
	h->l.vr[0] = 'O';
	h->l.vr[1] = 'B';
	h->l.foo = 0;
	h->l.length = 2;
	cp += 12;
	*cp++ = 0;
	*cp++ = 1;

	h = (header_t*)cp;
	h->s.tag[0] = DCM_GROUPFILEMETA;
	h->s.tag[1] = 0x0002;
	h->s.vr[0] = 'U';
	h->s.vr[1] = 'I';
	h->s.length = class_uid_len;
	cp += 8;
	memcpy(cp, class_uid, class_uid_len);
	cp += class_uid_len;

	h = (header_t*)cp;
	h->s.tag[0] = DCM_GROUPFILEMETA;
	h->s.tag[1] = 0x0003;
	h->s.vr[0] = 'U';
	h->s.vr[1] = 'I';
	h->s.length = instance_uid_len;
	cp += 8;
	memcpy(cp, instance_uid, instance_uid_len);
	cp += instance_uid_len;

	h = (header_t*)cp;
	h->s.tag[0] = DCM_GROUPFILEMETA;
	h->s.tag[1] = 0x0010;
	h->s.vr[0] = 'U';
	h->s.vr[1] = 'I';
	h->s.length = transfer_syntax_len;
	cp += 8;
	memcpy(cp, transfer_syntax, transfer_syntax_len);
	cp += transfer_syntax_len;

	h = (header_t*)cp;
	h->s.tag[0] = DCM_GROUPFILEMETA;
	h->s.tag[1] = 0x0012;
	h->s.vr[0] = 'U';
	h->s.vr[1] = 'I';
	h->s.length = impl_uid_len;
	cp += 8;
	memcpy(cp, impl_uid, impl_uid_len);
	cp += impl_uid_len;

	h = (header_t*)cp;
	h->s.tag[0] = DCM_GROUPFILEMETA;
	h->s.tag[1] = 0x0013;
	h->s.vr[0] = 'S';
	h->s.vr[1] = 'H';
	h->s.length = impl_ver_name_len;
	cp += 8;
	memcpy(cp, impl_ver_name, impl_ver_name_len);
	cp += impl_ver_name_len;

	if (!WriteFile(hFile, meta, cp - meta, &bytesWritten, NULL)) {
		return COND_PushWin32Error(APP_WIN32ERROR, GetLastError(), "WriteFile", "META HEADER - META Group Elements");
	}

	return APP_NORMAL;

}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static char hex(int i)
{
	if (i < 10)
		return i + '0';
	else
		return i + 'A';
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static int elementToString(const char* val, size_t val_len, bool isString, char* buf)
{	
	char* bp = buf;
	const char* cp = val;

	if (isString) {
		for (int i = 0; i < val_len; i++) {
			char c = *cp++;
			if ((c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9') || (c == '.')) {
				if (buf != NULL)
					*bp++ = c;
				else
					bp++;
			} else if (c == '\0') {
				break;	// UID
			} else {
				if (buf != NULL) {
					*bp++ = '_';
					*bp++ = hex((c & 0xF0) << 4);
					*bp++ = hex(c & 0x0F);
				} else {
					bp += 3;
				}
			}
			
		}
		if (buf != NULL)
			*bp = '\0';
		return bp - buf;
	} else {
		if (buf != NULL)
			return val_len * 3;
		for (int i = 0; i < val_len; i++) {
			char c = *cp++;
			*bp++ = '_';
			*bp++ = hex((c & 0xF0) << 4);
			*bp++ = hex(c & 0x0F);
		}
		*bp = '\0';
		return bp - buf;
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static CONDITION build_path_components(char* path_buf, size_t path_buf_size, char** path_component_tbl, const tVOLprm* vol, const char* scannedElements, size_t scannedElementsSize, unsigned short& resultCode)
{
	char* cp = path_buf;
	const char* const ep = path_buf + path_buf_size;

	*path_component_tbl++ = cp;

	for (int i = 0; i < vol->path_tmpl_item_count; i++) {
		const tPathTemplateItem* const tmpl_item = vol->path_tmpl_item_tbl + i;
		if (tmpl_item->type == tPathTemplateItem::eDelimiter) {

			if (ep - cp <= 1) {
		    	resultCode = MSG_K_PROCESSINGFAILURE;
				return COND_PushCondition(APP_STOREERROR, APP_Message(APP_STOREERROR), "File Path too long");
			}

			*cp++ = '\0';
			*path_component_tbl++ = cp;

			continue;

		} else if (tmpl_item->type == tPathTemplateItem::eString) {

			if (ep - cp <= tmpl_item->item.str.len) {
		    	resultCode = MSG_K_PROCESSINGFAILURE;
				return COND_PushCondition(APP_STOREERROR, APP_Message(APP_STOREERROR), "File Path too long");
			}

			memcpy(cp, tmpl_item->item.str.cp, tmpl_item->item.str.len);
			cp += tmpl_item->item.str.len;

			continue;

		} else if (tmpl_item->type == tPathTemplateItem::eElement) {

			const DCM_TAG tag = tmpl_item->item.element.tag;
	
			CDcmScannerF scanner(scannedElements, scannedElements + scannedElementsSize);
			const char* sp = scannedElements;
			tDcmElementHdr hdr;
			int r;
			while ((r = scanner.NextElement(sp, hdr, true)) == CDcmScannerF::S_CONTINUE) {
				if (hdr.ltag() == tag)
					break;
				sp = hdr.next_ptr(sp);
			}
			if (hdr.ltag() != tag) {
				resultCode = MSG_K_MISSINGATTRIBUTEVALUE;
				char msg[256];
				sprintf(msg, "File Path Element (%08X) not found", tag);
				return COND_PushCondition(APP_STOREERROR, APP_Message(APP_STOREERROR), msg);
			}

			int length = elementToString(hdr.value_ptr(sp), hdr.value_length(), hdr.isString(), NULL);
			if (ep - cp <= length) {
				resultCode = MSG_K_PROCESSINGFAILURE;
				return COND_PushCondition(APP_STOREERROR, APP_Message(APP_STOREERROR), "File Path too long");
			}

			elementToString(hdr.value_ptr(sp), hdr.value_length(), hdr.isString(), cp);
			cp += length;

			continue;

		} else {
			resultCode = MSG_K_PROCESSINGFAILURE;
			return COND_PushCondition(APP_STOREERROR, APP_Message(APP_STOREERROR), "Bad Path Template");
		}
	}

	*cp = '\0';

	return APP_NORMAL;
}

//----------------------------------------------------------------------------------------
//
//
//
//----------------------------------------------------------------------------------------
class CAssocScanner : public CDcmScannerS {
public:
	bool m_complete;
	unsigned long m_pixelLength;
	CStackBuffer<0x4000> m_buffer;

private:
	virtual void Event_DataElement(
			const tDcmElementHdr& hdr,
			const char* const vp,
			const unsigned long value_avail,
			const unsigned long value_offset,
			const bool explicitVR,
			const DCMDICT* const dict);

public:
	CAssocScanner(HANDLE heap);
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
CAssocScanner::CAssocScanner(HANDLE heap)
	: m_buffer(heap), m_complete(false), m_pixelLength(0)
{
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
void CAssocScanner::Event_DataElement(
		const tDcmElementHdr& hdr,
		const char* const vp,
		const unsigned long value_avail,
		const unsigned long value_offset,
		const bool explicitVR,
		const DCMDICT* const dict)
{
	if (m_complete)
		return;

	// only interest in top level attributes, not elements in sequence
	if (sq_level() != 0)
		return;

	const DCM_TAG tag = hdr.ltag();

	// ignore private group
	if (tag & DCM_MAKETAG(1, 0))
		return;
		
	// complete when hits pixel data
	if (tag == DCM_PXLPIXELDATA) {
		m_pixelLength = hdr.value_length();
		m_complete = true;
		return;
	}

	// ignore unknown element	
	if (hdr.m_vr == NULL)
		return;

	// ignore BLE (OB, OW, SQ)
	if (hdr.isBLE() || hdr.isSQ())
		return;

	// convert to explict VR
	if (value_offset == 0) { // converts explicit VR
		m_buffer.append((const char*)&hdr.m_tag[0], 2);
		m_buffer.append((const char*)&hdr.m_tag[1], 2);
		m_buffer.append(hdr.m_vr, 2);
		unsigned short length = hdr.value_length();
		m_buffer.append((const char*)&length, 2);
	}
	m_buffer.append(vp, value_avail);
}
//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
CONDITION storeToFile(DUL_ASSOCIATIONKEY** association, DUL_PRESENTATIONCONTEXT* ctx, MSG_C_STORE_REQ** storeRequest, DUL_ASSOCIATESERVICEPARAMETERS* service, tStorageParam* sp, CStudyDB& studydb, CStudyCache& studycache, unsigned short& resultCode, bool& send_response)
{
	CONDITION cond;
	send_response = true;

	//
	// config vars
	//
	const tSCPprm* scp = sp->pSCPprm;
	const tVOLprm* vol = scp->volume;

	//
	//
	//
	const bool explicitVR = strcmp(ctx->acceptedTransferSyntax, DICOM_TRANSFERLITTLEENDIAN) != 0;

	//
	// fetch header
	//
	DUL_PDV pdv;
	CStackBuffer<0x8000> buffer(Get_CTN_HEAP_HANDLE());
	CAssocScanner scanner(Get_CTN_HEAP_HANDLE());
	do {
		cond = SRVPRV_ReadNextPDV(association, DUL_BLOCK, 0, &pdv);
		sp->assoc.recievedBytes += pdv.fragmentLength;
		if (cond == DUL_EVENTSIGNALED) {
			resultCode = MSG_K_CANCEL;
			send_response = true;
			return cond;
		} else if (!CTN_SUCCESS(cond)) {
			send_response = false;
			return COND_PushCondition(SRV_RESPONSEFAILED, SRV_Message(SRV_RESPONSEFAILED), "storeToFile() while fetching header");
		} else if (pdv.pdvType != DUL_DATASETPDV) {
			send_response = false;
			return COND_PushCondition(SRV_UNEXPECTEDPDVTYPE, SRV_Message(SRV_UNEXPECTEDPDVTYPE), (int)pdv.pdvType, "storeToFile() while fetching header");
		}

		buffer.append(const_cast<const char*>(pdv.data), pdv.fragmentLength);
		scanner.Read(const_cast<const char*>(pdv.data), pdv.fragmentLength, explicitVR);

	} while (!scanner.m_complete && !pdv.lastPDV);

	//
	// Get Scanned Elements as Straight buffer
	//
	size_t scannedElementsSize = scanner.m_buffer.data_size();
	size_t bufsize;
	const char* scannedElements = scanner.m_buffer.peek(0, bufsize);
	if (scannedElementsSize > bufsize) {
		scannedElements = (const char*)_alloca(scannedElementsSize);
		scanner.m_buffer.fetch(0, const_cast<char*>(scannedElements), scannedElementsSize);
	}

	//
	//
	//
	const char* const vol_dir = vol->dir;
	TOWC(vol_dir)

	//
	// Ensure Disk Spece
	//
	if (!ensure_disk_space(studydb, sp, _vol_dir_WC, vol->free_space, resultCode)) {
		return APP_NORMAL;
	}

	//
	// Build Recieve Temporary File Path
	//
	wchar_t tempname[32];
	wsprintf(tempname, L"\\.SCP-%d.tmp", scp->index);
	const wchar_t* const temp_file_path = (const wchar_t*)_alloca(sizeof(wchar_t) * (_vol_dir_WC_len + lstrlen(tempname) + 1));
	CopyMemory(const_cast<wchar_t*>(temp_file_path), _vol_dir_WC, sizeof(wchar_t) * _vol_dir_WC_len);
	lstrcpy(const_cast<wchar_t*>(temp_file_path) + _vol_dir_WC_len, tempname);

	unsigned long storedBytes = 0;
	unsigned long storedFiles = 1;
	unsigned long overwriteFiles = 0;
	{
		//
		// Create (Temporary) File
		//
		HANDLE hFile = CreateFile(temp_file_path, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile == INVALID_HANDLE_VALUE) {
			TOAC(temp_file_path)
			COND_PushWin32Error(APP_WIN32ERROR, GetLastError(), "CreateFile", vol->dir);
			cond = COND_PushCondition(APP_STOREERROR, APP_Message(APP_STOREERROR), "CreateFile()");
			resultCode = MSG_K_PROCESSINGFAILURE;
			return cond;
		}
		AutoCloseAndDeleteFile ac_hFile(hFile, temp_file_path);

		//
		// Write Meta Header
		//
		cond = writeMetaHeader(hFile, scannedElements, scannedElementsSize, ctx->acceptedTransferSyntax);
		if (!CTN_SUCCESS(cond)) {
			resultCode = MSG_K_PROCESSINGFAILURE;
			return cond;
		}

		//
		// Flush Buffer
		//
		size_t buffer_offset = 0;
		while (true) {
			size_t bufsize;
			const char* chunk = buffer.peek(buffer_offset, bufsize);
			if (chunk == NULL)
				break;
			buffer_offset += bufsize;

			DWORD bytesWritten;
			if (!WriteFile(hFile, chunk, bufsize, &bytesWritten, NULL)) {
				DWORD dwErr = GetLastError();
				resultCode = MSG_K_C_STORE_OUTOFRESOURCES;
				return COND_PushWin32Error(APP_WIN32ERROR, dwErr, "WriteFile", "Flush Buffer");
			}
		}

		//
		// Flush Remain PDVs
		//
		while (!pdv.lastPDV) {
			cond = SRVPRV_ReadNextPDV(association, DUL_BLOCK, 0, &pdv);
			sp->assoc.recievedBytes += pdv.fragmentLength;
			if (cond == DUL_EVENTSIGNALED) {
				resultCode = MSG_K_CANCEL;
				send_response = true;
				return cond;
			} else if (!CTN_SUCCESS(cond)) {
				send_response = false;
				return COND_PushCondition(SRV_RESPONSEFAILED, SRV_Message(SRV_RESPONSEFAILED), "storeToFile() while flushing data");
			} else if (pdv.pdvType != DUL_DATASETPDV) {
				send_response = false;
				return COND_PushCondition(SRV_UNEXPECTEDPDVTYPE, SRV_Message(SRV_UNEXPECTEDPDVTYPE), (int)pdv.pdvType, "storeToFile() while flushing data");
			}

			DWORD bytesWritten;
			if (!WriteFile(hFile, pdv.data, pdv.fragmentLength, &bytesWritten, NULL)) {
				DWORD dwErr = GetLastError();
				resultCode = MSG_K_C_STORE_OUTOFRESOURCES;
				return COND_PushWin32Error(APP_WIN32ERROR, dwErr, "WriteFile", "Flush Remain PDVs");
			}
		}

		storedBytes = GetFileSize(hFile, NULL); 
		ac_hFile.dontDelete();
	}

	//
	// Construct Path Components
	//
	int path_depth = vol->path_tmpl_depth;
	char** path_component_tbl = (char**)_alloca(sizeof(char*) * path_depth);
	char path_buf[2048]; // may be too long, ok enough
	cond = build_path_components(path_buf, sizeof(path_buf), path_component_tbl, vol, scannedElements, scannedElementsSize, resultCode);
	if (!CTN_SUCCESS(cond)) {
		return cond;
	}

	//
	// Image Dirctory Path
	//
	int wlen = build_full_path(vol->dir, path_component_tbl, path_depth - 1, NULL, 0);
	wchar_t* const image_dir = (wchar_t*)_alloca(sizeof(wchar_t) * (wlen + 1));
	build_full_path(vol->dir, path_component_tbl, path_depth - 1, image_dir, wlen);

	//
	// Image File Path
	//
	const char* const imagefile = path_component_tbl[path_depth - 1];
	TOWC(imagefile)
	wlen = lstrlen(image_dir);
	wchar_t* const image_path = (wchar_t*)_alloca(sizeof(wchar_t) * (wlen + 1 + _imagefile_WC_len + 1));
	strncat3(image_path, image_dir, wlen, L'\\', _imagefile_WC, _imagefile_WC_len);

	//
	// Lock Volume
	//
	CStackCriticalSection cs(vol_crit_tbl + vol->index);

	//
	// Create Iamge Directory
	//
	DWORD attr = GetFileAttributes(image_dir);
	if (attr == 0xFFFFFFFF) {
		cond = create_image_directory(image_dir);
		if (!CTN_SUCCESS(cond)) {
	    	resultCode = MSG_K_PROCESSINGFAILURE;
			return cond;
		}
	} else if ((attr & FILE_ATTRIBUTE_DIRECTORY) == 0) {
		TOAC(image_dir)
	    resultCode = MSG_K_PROCESSINGFAILURE;
		return COND_PushCondition(SRV_FILECREATEFAILED, "file exists for directory %s", image_dir);
	} else {
		//
		// Check existing
		//
		WIN32_FILE_ATTRIBUTE_DATA fileAttr;
		if (GetFileAttributesEx(image_path, GetFileExInfoStandard, &fileAttr)) {
			storedBytes = storedBytes - fileAttr.nFileSizeLow;
			storedFiles = 0;
			overwriteFiles++;
		}
	}

	//
	// Move Temporary File To DICOM File
	//
	if (!MoveFileEx(temp_file_path, image_path, MOVEFILE_REPLACE_EXISTING)) {
		DWORD dwErr = GetLastError();
		resultCode = MSG_K_PROCESSINGFAILURE;
		TOAC(image_path)
		COND_PushWin32Error(APP_WIN32ERROR, dwErr, "MoveFileEx", _image_path_AC);
		return COND_PushCondition(APP_STOREERROR, APP_Message(APP_STOREERROR), "Move Temporary file to DICOM file");
	}	

	//
	// make study path to location (URI)
	//
	const char* study_loc = path_component_tbl[0];
	TOWC(study_loc)

	wchar_t* const studyDir = (wchar_t*)_alloca(sizeof(wchar_t) * (_vol_dir_WC_len + _study_loc_WC_len + 1));
	strncat3(studyDir, _vol_dir_WC, _vol_dir_WC_len, L'\\', _study_loc_WC, _study_loc_WC_len);

	const bool study_changed = sp->last_study_loc == NULL || strcmp(study_loc, sp->last_study_loc) != 0;

	//
	// make image path to location (URI)
	//
	int i = 1;
	int len = strlen(path_component_tbl[i++]);
	while (i < path_depth)
		len += 1 + strlen(path_component_tbl[i++]);
	char* const file_loc = (char*)_alloca(len + 1);
	i = 1;
	strcpy(file_loc, path_component_tbl[i++]);
	while (i < path_depth) {
		strcat(file_loc, "/");
		strcat(file_loc, path_component_tbl[i++]);
	}
	TOWC(file_loc)

	//
	// Insert entry to Study Cache Files
	//
	if (vol->study_cache_flag != 0)
		studycache.append(studyDir, _file_loc_WC, scannedElements, scannedElementsSize);

	//
	// Insert study information to the database
	//
	if (study_changed && studydb.connected()) {

		if (sp->last_study_loc != NULL) {
			CTN_FREE(sp->last_study_loc);
			SysFreeString(sp->check_sql);
			sp->last_study_loc = NULL;
			sp->check_sql = NULL;
		}

		if (sp->last_study_loc == NULL) {
			len = strlen(study_loc);
			sp->last_study_loc = (char*)CTN_MALLOC(len + 1);
			memcpy(sp->last_study_loc, study_loc, len + 1);
			sp->check_sql = CStudyDB::make_check_sql(vol->label, study_loc);
		}

		bool exists;
		if (studydb.study_exists(exists, sp->check_sql)) {
			if (!exists) {
				BSTR insert_sql = make_insert_sql(scp->label, service->callingAPTitle, vol->label, study_loc, file_loc, scannedElements, scannedElementsSize);
				AutoSysFreeString af_insert_sql(insert_sql);
				if (studydb.Execute(insert_sql)) {
					StorageLog::print_insert_study_info(sp, study_loc);
				} else {
					wchar_t* message = studydb.m_message;
					TOAC(message)
					COND_PushCondition(APP_ODBCERROR, APP_Message(APP_ODBCERROR), _message_AC);
				}
			}
		} else {
			wchar_t* message = studydb.m_message;
			TOAC(message)
			COND_PushCondition(APP_ODBCERROR, APP_Message(APP_ODBCERROR), _message_AC);
		}
	}

	//
	// Record statistics
	//
	sp->assoc.storedFiles += storedFiles;
	sp->assoc.storedBytes += storedBytes;
	sp->assoc.overwriteFiles += overwriteFiles;

	resultCode = MSG_K_SUCCESS;
	return APP_NORMAL;
}
