/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "DcmDict.h"
#include "CDcmScanner.h"
#include "UtilClasses.h"

#include "RequestFileWatcher.h"

//======================================================================
//
//======================================================================
//
// send-dicom-svc.cpp
//
BOOL ReportStatusToSCMgr(DWORD dwCurrentState, DWORD dwWin32ExitCode, DWORD dwWaitHint);

//======================================================================
//
//======================================================================
static CLSID clsidDomDoc;

//======================================================================
//
//
//
//======================================================================
class MyConsoleLog : public CConsoleLog {
public:

	void Write(const SimpleErrorInfo& ei)
	{
		writeErrorInfo(ei);
	}

	void Write(const BSTR domlogPath)
	{
		TOAC(domlogPath)

		const DWORD len = 19 + _domlogPath_AC_len + 6;
		char* buf = (char*)_alloca(len + 1);
		sprintf(buf, "<Session log-file=\"%s\"/>\r\n", _domlogPath_AC);

		writeText(buf);
	}
};
//======================================================================
MyConsoleLog* ConLog;

//======================================================================
//
// class DOMLog
//
//======================================================================
class DOMLog {
	IXMLDOMDocument* m_dom;

	static const wchar_t* const logFilePathBase;

public:

	DOMLog();
	~DOMLog();
	
	IXMLDOMDocument* dom() const { return m_dom; } 

	void Flush(HANDLE hFile = INVALID_HANDLE_VALUE) throw (SimpleErrorInfo);

};
//======================================================================
//
//======================================================================
const wchar_t* const DOMLog::logFilePathBase = L"C:\\DIOWave\\logs\\send-dicom-svc-";
//======================================================================
//
//======================================================================
DOMLog::DOMLog()
	: m_dom(NULL)
{
	try {
		AutoComPtr<IXMLDOMDocument> dom;
		HRESULT hr = CoCreateInstance(clsidDomDoc, NULL, CLSCTX_INPROC_SERVER, IID_IXMLDOMDocument, (void**)dom.r());
		if (FAILED(hr))
			throw SimpleErrorInfo(hr, L"CoCreateInstance(clsidDomDoc, NULL, CLSCTX_INPROC_SERVER, IID_IXMLDOMDocument)");

		BSTR bstr = SysAllocString(L"C-STORE-LOG");
		AutoSysFreeString af_bstr(bstr);

		AutoComPtr<IXMLDOMElement> doc;
		hr = dom->createElement(bstr, doc.r());
		if (FAILED(hr))
			throw SimpleErrorInfo(hr, L"IXMLDOMDocument::createElement(%1)", bstr);

		hr = dom->putref_documentElement(doc.p());
		if (FAILED(hr))
			throw SimpleErrorInfo(hr, L"IXMLDOMDocument::putref_documentElement()");

		dom->AddRef();
		m_dom = dom.p();

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"RequestFileWatcher::make_domdoc()");
		throw;
	}
}
//======================================================================
//
//======================================================================
DOMLog::~DOMLog()
{
	if (m_dom != NULL)
		m_dom->Release();
}

//======================================================================
//
//======================================================================
void DOMLog::Flush(HANDLE hFile) throw (SimpleErrorInfo)
{
	if (hFile == INVALID_HANDLE_VALUE) {
		SYSTEMTIME time;
		GetLocalTime(&time);

		const int len = lstrlen(logFilePathBase);
		BSTR bstr = SysAllocStringLen(NULL, len + 19 + 4);
		AutoSysFreeString af_bstr(bstr);

		lstrcpy(bstr, logFilePathBase);
		wsprintf(bstr + len, L"%04d.%02d.%02d.%02d.%02d.%02d.xml", 
			time.wYear, time.wMonth, time.wDay, time.wHour, time.wMinute, time.wSecond);

		VARIANT var;
		VariantInit(&var);

		var.vt = VT_BSTR;
		var.bstrVal = bstr;

		HRESULT hr = m_dom->save(var);
		if (FAILED(hr)) {
			SimpleErrorInfo ei(hr, L"IXMLDOMDocument::save(%1)", bstr);
			ei.add(L"DOMLog::Flush()");
			throw ei;
		}
		ConLog->Write(bstr);
	} else {
		CStackFileStream stream(hFile);

		VARIANT var;
		VariantInit(&var);
		var.vt = VT_UNKNOWN;
		stream.QueryInterface(IID_IUnknown, (void**)&var.punkVal);

		HRESULT hr = m_dom->save(var);
		if (FAILED(hr)) {
			SimpleErrorInfo ei(hr, L"IXMLDOMDocument::save()");
			ei.add(L"DOMLog::Flush()");
			throw ei;
		}
	}
}

//======================================================================
//
// class RequestFileWatcher::config_t
//
//======================================================================
//======================================================================
//
//======================================================================
RequestFileWatcher::config_t RequestFileWatcher::config_t::defaultConfig = { NULL, 3 * 1000, 256 * 1024 * 1024, 1024, NULL, 104, "DIO_WAVE", NULL };
//======================================================================
//
//======================================================================
HRESULT RequestFileWatcher::config_t::New(IXMLDOMNode* configNode, config_t** pConfig, SimpleErrorInfo& ei)
{
	config_t cf;
	CopyMemory(&cf, &defaultConfig, sizeof(config_t));

	BSTR bstr = NULL;
	AutoSysFreeString af_bstr(bstr);

	HRESULT hr;
	int blen, len;

	hr = xduSelectSingleNodeText(configNode, L"directory", bstr, &ei);
	if (FAILED(hr)) {
		ei.add(L"xduSelectSingleNodeText(directory)");
		return hr;
	} else if (hr == S_OK) {
		blen = SysStringLen(bstr);
		cf.m_watchDirectory = (wchar_t*)_alloca(sizeof(wchar_t) * (blen + 1));
		lstrcpy(cf.m_watchDirectory, bstr);
	}

	hr = xduSelectSingleNodeText(configNode, L"host", bstr, &ei);
	if (FAILED(hr)) {
		ei.add(L"xduSelectSingleNodeText(host)");
		return hr;
	} else if (hr == S_OK) {
		blen = SysStringLen(bstr);
		len = wide_char_to_multi_byte((const wchar_t*)bstr, blen, NULL, 0);
		cf.m_host = (char*)_alloca(len + 1);
		wide_char_to_multi_byte((const wchar_t*)bstr, blen, cf.m_host, len);
		cf.m_host[len] = '\0';
	}

	hr = xduSelectSingleNodeText(configNode, L"scu-title", bstr, &ei);
	if (FAILED(hr)) {
		ei.add(L"xduSelectSingleNodeText(scu-title)");
		return hr;
	} else if (hr == S_OK) {
		blen = SysStringLen(bstr);
		len = wide_char_to_multi_byte((const wchar_t*)bstr, blen, NULL, 0);
		cf.m_scu_title = (char*)_alloca(len + 1);
		wide_char_to_multi_byte((const wchar_t*)bstr, blen, cf.m_scu_title, len);
		cf.m_scu_title[len] = '\0';
	}

	hr = xduSelectSingleNodeText(configNode, L"scp-title", bstr, &ei);
	if (FAILED(hr)) {
		ei.add(L"xduSelectSingleNodeText(scp-title)");
		return hr;
	} else if (hr == S_OK) {
		blen = SysStringLen(bstr);
		len = wide_char_to_multi_byte((const wchar_t*)bstr, blen, NULL, 0);
		cf.m_scp_title = (char*)_alloca(len + 1);
		wide_char_to_multi_byte((const wchar_t*)bstr, blen, cf.m_scp_title, len);
		cf.m_scp_title[len] = '\0';
	}

	hr = xduSelectSingleNodeText(configNode, L"timeout", bstr, &ei);
	if (FAILED(hr)) {
		ei.add(L"xduSelectSingleNodeText(timeout)");
		return hr;
	} else if (hr == S_OK) {
		VarUI4FromStr(bstr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), LOCALE_NOUSEROVERRIDE, &cf.m_timeOut);
	}

	hr = xduSelectSingleNodeText(configNode, L"store-limit-bytes", bstr, &ei);
	if (FAILED(hr)) {
		ei.add(L"xduSelectSingleNodeText(store-limit-bytes)");
		return hr;
	} else if (hr == S_OK) {
		VarUI4FromStr(bstr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), LOCALE_NOUSEROVERRIDE, &cf.m_storeLimitBytes);
	}

	hr = xduSelectSingleNodeText(configNode, L"store-limit-files", bstr, &ei);
	if (FAILED(hr)) {
		ei.add(L"xduSelectSingleNodeText(store-limit-files)");
		return hr;
	} else if (hr == S_OK) {
		VarUI4FromStr(bstr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), LOCALE_NOUSEROVERRIDE, &cf.m_storeLimitFiles);
	}

	hr = xduSelectSingleNodeText(configNode, L"port", bstr, &ei);
	if (FAILED(hr)) {
		ei.add(L"xduSelectSingleNodeText(port)");
		return hr;
	} else if (hr == S_OK) {
		VarUI2FromStr(bstr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), LOCALE_NOUSEROVERRIDE, &cf.m_port);
	}

	if (cf.m_watchDirectory == NULL || cf.m_host == NULL || cf.m_scp_title == NULL) {
		ei.set(L"required config element (directory, host, scp-title) not defined");
		return E_FAIL;
	}

	const DWORD dwAttr = GetFileAttributes(cf.m_watchDirectory);
	if (dwAttr == 0xFFFFFFFF) {
		hr = HRESULT_FROM_WIN32(GetLastError());
		ei.set(hr, L"GetFileAttributes(%1)", cf.m_watchDirectory);
		ei.add(L"RequestFileWatcher::Init()");
		return hr;
	} else if ((dwAttr & FILE_ATTRIBUTE_DIRECTORY) == 0) {
		ei.set(L"%1 is not a directory", cf.m_watchDirectory);
		ei.add(L"RequestFileWatcher::Init()");
		return hr;
	}

	len = sizeof(config_t) + sizeof(wchar_t) * (lstrlen(cf.m_watchDirectory) + 1 ) + strlen(cf.m_host) + 1 + strlen(cf.m_scu_title) + 1 + strlen(cf.m_scp_title) + 1;

	config_t* config = (config_t*)CoTaskMemAlloc(len);
	CopyMemory(config, &cf, sizeof(config_t));
	char* cp = (char*)(config + 1);

	len = lstrlen(cf.m_watchDirectory);
	CopyMemory(cp, cf.m_watchDirectory, sizeof(wchar_t) * (len + 1));
	config->m_watchDirectory = (wchar_t*)cp;
	cp += sizeof(wchar_t) * (len + 1);

	len = strlen(cf.m_host);
	CopyMemory(cp, cf.m_host, len + 1);
	config->m_host = cp;
	cp += len + 1;

	len = strlen(cf.m_scu_title);
	CopyMemory(cp, cf.m_scu_title, len + 1);
	config->m_scu_title = cp;
	cp += len + 1;

	len = strlen(cf.m_scp_title);
	CopyMemory(cp, cf.m_scp_title, len + 1);
	config->m_scp_title = cp;

	*pConfig = config;

	return S_OK;
}

//======================================================================
//
// class RequestFileWatcher
//
//======================================================================
//======================================================================
//
//======================================================================
HRESULT RequestFileWatcher::ScanRequestFiles(const wchar_t* watchDirectory, IXMLDOMDocument* dom, unsigned long& numFilesToStore) throw (SimpleErrorInfo)
{
	try {
		HANDLE hChangeNotify = FindFirstChangeNotification(watchDirectory, FALSE, FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_LAST_WRITE);
		if (hChangeNotify == INVALID_HANDLE_VALUE) {
			const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			m_ei.set(hr, L"FindFirstChangeNotification()");
			m_ei.add(L"RequestFileWatcher::ScanRequestFiles(%1)", watchDirectory);
			return hr;
		}
		AutoCloseChangeNotificationHandle ac_hChangeNotify(hChangeNotify);

		if (!FindNextChangeNotification(hChangeNotify)) {
			const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			m_ei.set(hr, L"FindNextChangeNotification()");
			m_ei.add(L"RequestFileWatcher::ScanRequestFiles(%1)", watchDirectory);
			return hr;
		}

		unsigned long scannedRequestFileCount = 0;
		unsigned long long numBytesToStore = 0;

		while (true) {
			const HRESULT hr = scanRequestFiles(watchDirectory, scannedRequestFileCount, numFilesToStore, numBytesToStore, dom);
			if (FAILED(hr)) {
				m_ei.add(L"RequestFileWatcher::ScanRequestFiles(%1)", watchDirectory);
				return hr;
			}

			if (numBytesToStore > m_config->m_storeLimitBytes
			 || numFilesToStore > m_config->m_storeLimitFiles)
				return S_OK;

			HANDLE handles[2] = { hChangeNotify, m_hStopEvent };
			const DWORD waitFlag = WaitForMultipleObjects(2, handles, FALSE, scannedRequestFileCount > 0 ? m_config->m_timeOut : INFINITE);
			if (waitFlag == WAIT_OBJECT_0) {
				if (!FindNextChangeNotification(hChangeNotify)) {
					const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
					m_ei.set(hr, L"FindNextChangeNotification()");
					m_ei.add(L"RequestFileWatcher::ScanRequestFiles(%1)", watchDirectory);
					return hr;
				}
				continue;
			} else if (waitFlag == WAIT_OBJECT_0 + 1) {
				return S_FALSE;
			} else if (waitFlag == WAIT_TIMEOUT) {
				return S_OK;
			} else {
				const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
				m_ei.set(hr, L"WaitForMultipleObjects()");
				m_ei.add(L"RequestFileWatcher::ScanRequestFiles(%1)", watchDirectory);
				return hr;
			}
		}
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"RequestFileWatcher::ScanRequestFiles(%1)", watchDirectory);
		throw;
	}

	return S_OK;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
HRESULT RequestFileWatcher::scanRequestFiles(const wchar_t* watchDirectory, DWORD& scannedRequestFileCount, unsigned long& numFilesToStore, unsigned long long& numBytesToStore, IXMLDOMDocument* dom) throw (SimpleErrorInfo)
{
	try {
		CStackFindFile find(watchDirectory);
		if (find.handle == INVALID_HANDLE_VALUE) {
			const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			m_ei.set(hr, L"FindFirstFile()");
			m_ei.add(L"RequestFileWatcher::scanRequestFiles(%1)", watchDirectory);
			return hr;
		}

		const int dirLen = lstrlen(watchDirectory);
		wchar_t* const path = (wchar_t*)_alloca(sizeof(wchar_t) * (dirLen + 1 + MAX_PATH + 1));
		CopyMemory(path, watchDirectory, sizeof(wchar_t) * dirLen);
		*(path + dirLen) = L'\\';
		wchar_t* const fileName = path + dirLen + 1;

		do {
			if (*find.data.cFileName == L'.' || find.data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				continue;

			scannedRequestFileCount++;

			lstrcpy(fileName, find.data.cFileName);

			const HRESULT hr = readRequestFile(path, find.data, dom, numFilesToStore, numBytesToStore);
			if (hr != S_OK)
				return hr;

			if (numBytesToStore > m_config->m_storeLimitBytes
			 || numFilesToStore > m_config->m_storeLimitFiles)
				return S_OK;

		} while (find.next());

		return S_OK;
	
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"RequestFileWatcher::scanRequestFiles(%1)", watchDirectory);
		throw;
	}
	return E_UNEXPECTED;
} 

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
HRESULT RequestFileWatcher::readRequestFile(const wchar_t* file, const WIN32_FIND_DATA& data, IXMLDOMDocument* dom, unsigned long& numFilesToStore, unsigned long long& numBytesToStore) throw (SimpleErrorInfo)
{
	try {
		HRESULT hr;

		const wchar_t* const tag = L"REQUEST";

		BSTR bstr = SysAllocString(tag);
		AutoSysFreeString af_bstr(bstr);
		AutoComPtr<IXMLDOMElement> reqFileElm;
		hr = dom->createElement(bstr, reqFileElm.r());
		if (FAILED(hr))
			throw SimpleErrorInfo(hr, L"IXMLDOMDocument::createElement(<%1>)", tag);

		AutoComPtr<IXMLDOMElement> doc;
		hr = dom->get_documentElement(doc.r());
		if (FAILED(hr))
			throw SimpleErrorInfo(hr, L"IXMLDOMDocument::get_documentElement()");
		hr = doc->appendChild(reqFileElm.p(), NULL);
		if (FAILED(hr))
			throw SimpleErrorInfo(hr, L"IXMLDOMElement::appendChild(<%1>)", tag);

		xduMakeAttribute(dom, reqFileElm.p(), L"file-path", file, xduThrowSimpleErrorInfo);

		if (data.nFileSizeLow == 0) {
			SimpleErrorInfo ei(L"file is empty");
			ei.add(L"RequestFileWatcher::readRequestFile(\"%1\")", file);
			xduMakeErrorInfoElement(dom, reqFileElm.p(), ei, NULL);
			return S_OK;
		}

		HANDLE hFile = CreateFile(file, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_FLAG_DELETE_ON_CLOSE, NULL);
		if (hFile == INVALID_HANDLE_VALUE) {
			hr = HRESULT_FROM_WIN32(GetLastError());
			if (hr == HRESULT_FROM_WIN32(ERROR_SHARING_VIOLATION))
				return S_OK;
			SimpleErrorInfo ei(hr, L"CreateFile()");
			ei.add(L"RequestFileWatcher::readRequestFile(\"%1\")", file);
			xduMakeErrorInfoElement(dom, reqFileElm.p(), ei, NULL);
			return hr;
		}
		AutoCloseFileHandle ac_hFile(hFile);

		CStackFileMapping map(hFile);
		if (map.lastError() != NO_ERROR) {
			hr = HRESULT_FROM_WIN32(map.lastError());
			SimpleErrorInfo ei(hr, L"CStackFileMapping()");
			ei.add(L"RequestFileWatcher::readRequestFile(\"%1\")", file);
			xduMakeErrorInfoElement(dom, reqFileElm.p(), ei, NULL);
			return hr;
		}

		const char* const block_begin = static_cast<const char*>(map.mapPointer());
		const char* const block_end = block_begin + map.mapSize();

		xduMakeElement(dom, reqFileElm.p(), L"file-content", block_begin, static_cast<int>(map.mapSize()), NULL, xduThrowSimpleErrorInfo);

		wchar_t* requestedPath = NULL;
		const char* cp = block_begin;
		while (true) {
			if (*cp != 'F') {
				while (cp < block_end && *cp != '\n')
					cp++;
				if (cp == block_end) {
					SimpleErrorInfo ei(L"tag \"F\" not found");
					ei.add(L"RequestFileWatcher::readRequestFile(\"%1\")", file);
					xduMakeErrorInfoElement(dom, reqFileElm.p(), ei, NULL);
					return S_OK;
				}
				cp++;
				continue;
			}
			cp++;
			while (cp < block_end && *cp <= ' ')
				cp++;
			if (cp == block_end) {
				SimpleErrorInfo ei(L"Empty \"F\" tag");
				ei.add(L"RequestFileWatcher::readRequestFile(\"%1\")", file);
				xduMakeErrorInfoElement(dom, reqFileElm.p(), ei, NULL);
				return S_OK;
			}
			const char* const sp = cp;
			while (cp < block_end && *cp > ' ')
				cp++;
			const int len = cp - sp;
			int wclen = multi_byte_to_wide_char(sp, len, NULL, 0);
			requestedPath = (wchar_t*)_alloca(sizeof(wchar_t) * (wclen + 1));
			multi_byte_to_wide_char(sp, len, requestedPath, wclen);
			requestedPath[wclen] = L'\0';
			break;
		}

		xduMakeAttribute(dom, reqFileElm.p(), L"requested-path", requestedPath, xduThrowSimpleErrorInfo);

		const DWORD dwAttr = GetFileAttributes(requestedPath);
		if (dwAttr == 0xFFFFFFFF) {
			hr = HRESULT_FROM_WIN32(GetLastError());
			SimpleErrorInfo ei(hr);
			xduMakeErrorInfoElement(dom, reqFileElm.p(), ei, NULL);
			return S_OK;
		} else if ((dwAttr & FILE_ATTRIBUTE_DIRECTORY) == 0) {
			SimpleErrorInfo ei(L"not directory");
			xduMakeErrorInfoElement(dom, reqFileElm.p(), ei, NULL);
			return S_OK;
		}

		hr = collect_files(dom, reqFileElm.p(), requestedPath, numFilesToStore, numBytesToStore);
		if (FAILED(hr)) {
			m_ei.add(L"RequestFileWatcher::readRequestFile(%1)", file);
			return hr;
		}
	} catch (SimpleErrorInfo& e) {
		e.add(L"RequestFileWatcher::readRequestFile(%1)", file);
		throw;
	}

	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT RequestFileWatcher::collect_files(IXMLDOMDocument* dom, IXMLDOMElement* reqFileElm, const wchar_t* directory, unsigned long& numFilesToStore, unsigned long long& numBytesToStore) throw (SimpleErrorInfo)
{
	try {
		HRESULT hr;

		const wchar_t* const tag = L"FL";

		BSTR bstr = SysAllocString(tag);
		AutoSysFreeString af_bstr(bstr);

		AutoComPtr<IXMLDOMElement> fileList;
		hr = dom->createElement(bstr, fileList.r());
		if (FAILED(hr))
			throw SimpleErrorInfo(hr, L"IXMLDOMDocument::createElement(\"%1\")", tag);

		hr = reqFileElm->appendChild(fileList.p(), NULL);
		if (FAILED(hr))
			throw SimpleErrorInfo(hr, L"IXMLDOMElement::appendChild(\"%1\")", tag);
			
		hr = collect_files(dom, directory, fileList.p(), numFilesToStore, numBytesToStore);
		if (FAILED(hr)) {
			m_ei.add(L"RequestFileWatcher::collect_files(%1)", directory);
			return hr;
		}

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"RequestFileWatcher::collect_files(%1)", directory);
		throw;
	}

	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT RequestFileWatcher::collect_files(IXMLDOMDocument* dom, const wchar_t* directory, IXMLDOMElement* fileList, unsigned long& numFilesToStore, unsigned long long& numBytesToStore) throw (SimpleErrorInfo)
{
	try {
		CStackFindFile find(directory);
		if (find.handle == INVALID_HANDLE_VALUE) {
			const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
			m_ei.set(hr, L"FindFirstFile()");
			m_ei.add(L"RequestFileWatcher::collect_files(%1)", directory);
			return hr;
		}

		const int dirLen = lstrlen(directory);
		wchar_t* const path = (wchar_t*)_alloca(sizeof(wchar_t) * (dirLen + 1 + MAX_PATH + 1));
		CopyMemory(path, directory, sizeof(wchar_t) * dirLen);
		*(path + dirLen) = L'\\';
		wchar_t* const fileName = path + dirLen + 1;

		do {
			if (*find.data.cFileName == L'.')
				continue;

			lstrcpy(fileName, find.data.cFileName);

			if (find.data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
				collect_files(dom, path, fileList, numFilesToStore, numBytesToStore);
			} else {
				collect_file(dom, path, find.data, fileList, numFilesToStore, numBytesToStore);
			}
		} while (find.next());

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"RequestFileWatcher::collect_files(%1)", directory);
		throw;
	}

	return S_OK;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void uid2wstr(const char* cp, int len, wchar_t* wstr)
{
	wchar_t* wp = wstr;
	const char* const ep = cp + len;
	while (cp < ep && *cp != '\0')
		*wp++ = *cp++;
	*wp = L'\0';
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
HRESULT RequestFileWatcher::collect_file(IXMLDOMDocument* dom, const wchar_t* file, const WIN32_FIND_DATA& data, IXMLDOMElement* fileList, unsigned long& numFilesToStore, unsigned long long& numBytesToStore) throw (SimpleErrorInfo)
{
	try {
		AutoComPtr<IXMLDOMElement> fileElement;
		xduMakeElement(dom, fileList, L"F", fileElement.r(), xduThrowSimpleErrorInfo);

		xduMakeAttribute(dom, fileElement.p(), L"path", file, xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, fileElement.p(), L"file-size", static_cast<unsigned int>(data.nFileSizeLow), xduThrowSimpleErrorInfo);

		if (data.nFileSizeLow < DCM_PREAMBLELENGTH + 4) {
			SimpleErrorInfo ei(L"file size is small");
			xduMakeErrorInfoElement(dom, fileElement.p(), ei, NULL);
			return S_OK;
		}

		CStackFileMapping map(file);
		if (map.lastError() != NO_ERROR) {
			HRESULT hr = HRESULT_FROM_WIN32(map.lastError());
			SimpleErrorInfo ei(hr, L"CStackFileMapping()");
			xduMakeErrorInfoElement(dom, fileElement.p(), ei, NULL);
			return S_OK;
		}

		const char* const block_begin = static_cast<const char*>(map.mapPointer());
		const char* const block_end = block_begin + map.mapSize();

		if (memcmp(block_begin + DCM_PREAMBLELENGTH, "DICM", 4) != 0) {
			SimpleErrorInfo ei(L"\"DICM\" signature not found");
			xduMakeErrorInfoElement(dom, fileElement.p(), ei, NULL);
			return S_OK;
		}

		CDcmScannerF scanner(block_begin, block_end);
		tDcmElementHdr hdr;

		const char* cp = block_begin + DCM_PREAMBLELENGTH + 4;
		bool explicitVR;

		wchar_t transfer_syntax_uid[DICOM_UI_LENGTH + 1];
		wchar_t sop_instance_uid[DICOM_UI_LENGTH + 1];
		wchar_t sop_class_uid[DICOM_UI_LENGTH + 1];
		wchar_t study_instance_uid[DICOM_UI_LENGTH + 1];

		while (true) {
			int sr = scanner.NextElement(cp, hdr, true);
			if (sr != CDcmScannerF::S_CONTINUE) {
				SimpleErrorInfo ei(L"CDcmScannerF::NextElement() failed in MetaHeader at offset %1!d!", cp - block_begin);
				xduMakeErrorInfoElement(dom, fileElement.p(), ei, NULL);
				return S_OK;
			}
			if (hdr.m_tag[0] != DCM_GROUPFILEMETA)
				break;
			if (hdr.m_tag[1] == DCM_TAG_ELEMENT(DCM_METATRANSFERSYNTAX)) {
				uid2wstr(hdr.value_ptr(cp), hdr.value_length(), transfer_syntax_uid);
		
				char uid[DICOM_UI_LENGTH + 1];
				CopyMemory(uid, hdr.value_ptr(cp), hdr.value_length());
				uid[hdr.value_length()] = '\0';
				
				if (strcmp(uid, DICOM_TRANSFERLITTLEENDIAN) == 0) {
					explicitVR = false;
				} else if (strcmp(uid, DICOM_TRANSFERLITTLEENDIANEXPLICIT) == 0) {
					explicitVR = true;
				} else if (strcmp(uid, DICOM_TRANSFERJPEGBASELINEPROCESS1) == 0) {
					explicitVR = true;
				} else if (strcmp(uid, DICOM_TRANSFERJPEGEXTENDEDPROC2AND4) == 0) {
					explicitVR = true;
				} else if (strcmp(uid, DICOM_TRANSFERJPEGLOSSLESSPROC14) == 0) {
					explicitVR = true;
				} else if (strcmp(uid, DICOM_TRANSFERJPEGLOSSLESSPROCFIRSTORDERREDICT) == 0) {
					explicitVR = true;
				} else if (strcmp(uid, "PNG") == 0) {
					explicitVR = true;
				} else {
					SimpleErrorInfo ei(L"Transfer Syntax : %1 not supported", transfer_syntax_uid);
					xduMakeErrorInfoElement(dom, fileElement.p(), ei, NULL);
					return S_OK;
				}
			}
			cp = hdr.next_ptr(cp);
		}

		const unsigned int metaSize = cp - block_begin;

		while (true) {
			int sr = scanner.NextElement(cp, hdr, explicitVR);
			if (sr < CDcmScannerF::S_CONTINUE) {
				SimpleErrorInfo ei(L"CDcmScannerF::NextElement() failed in Data Set at offset %1!d!", cp - block_begin);
				xduMakeErrorInfoElement(dom, fileElement.p(), ei, NULL);
				return S_OK;
			}

			const DCM_TAG ltag = hdr.ltag();
			const unsigned long length = hdr.value_length();
			if (ltag == DCM_PXLPIXELDATA) // Ɖ摜܂œǂ߂鎖mF
				break;
			else if (ltag == DCM_IDSOPCLASSUID)
				uid2wstr(hdr.value_ptr(cp), hdr.value_length(), sop_class_uid);
			else if (ltag == DCM_IDSOPINSTANCEUID)
				uid2wstr(hdr.value_ptr(cp), hdr.value_length(), sop_instance_uid);
			else if (ltag == DCM_RELSTUDYINSTANCEUID)
				uid2wstr(hdr.value_ptr(cp), hdr.value_length(), study_instance_uid);

			cp = hdr.next_ptr(cp);
		}

		numFilesToStore++;
		numBytesToStore += data.nFileSizeLow - metaSize;
		
		xduMakeAttribute(dom, fileElement.p(), L"id", static_cast<unsigned int>(numFilesToStore), NULL);

		AutoComPtr<IXMLDOMElement> dicom;
		xduMakeElement(dom, fileElement.p(), L"DICOM", dicom.r(), xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, dicom.p(), L"meta-size", metaSize, xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, dicom.p(), L"TransferSyntaxUID", transfer_syntax_uid, xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, dicom.p(), L"SOPClassUID", sop_class_uid, xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, dicom.p(), L"SOPInstanceUID", sop_instance_uid, xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, dicom.p(), L"StudyInstanceUID", study_instance_uid, xduThrowSimpleErrorInfo);

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"RequestFileWatcher::collect_file(%1)", file);
		throw;
	}

	return S_OK;
}

//----------------------------------------------------------------------------------------
// Automatic Cleanup
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
class Auto_DUL_DropNetwork {
	DUL_NETWORKKEY** network;
public:
	Auto_DUL_DropNetwork(DUL_NETWORKKEY** key) : network(key) {}
	~Auto_DUL_DropNetwork() {
		DUL_DropNetworkEx(network);
	}
};
//----------------------------------------------------------------------------------------
class Auto_DUL_ClearServiceParameters {
	DUL_ASSOCIATESERVICEPARAMETERS* service;
public:
	Auto_DUL_ClearServiceParameters(DUL_ASSOCIATESERVICEPARAMETERS* asp) : service(asp) {}
	~Auto_DUL_ClearServiceParameters() {
		DUL_ClearServiceParameters(service);
	}
};

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void bstr2uid(BSTR bstr, char* uid)
{
	const wchar_t* wp = bstr;
	const wchar_t* const ep = bstr + SysStringLen(bstr);
	while (wp < ep && *wp != L'\0')
		*uid++ = *wp++;
	*uid = '\0';
}

//
//----------------------------------------------------------------------------------------
typedef struct {
	const DUL_SC_ROLE role;
	const char* const text;
}   SC_MAP;
const SC_MAP scMap[] = {
    {DUL_SC_ROLE_DEFAULT, "Default"},
    {DUL_SC_ROLE_SCU, "SCU"},
    {DUL_SC_ROLE_SCP, "SCP"},
    {DUL_SC_ROLE_SCUSCP, "SCUSCP"},
};
//----------------------------------------------------------------------------------------
static const char* lookup_role(const unsigned int roleIndex)
{
	const SC_MAP* mp = scMap;
	for (int i = 0; i < DIM_OF(scMap); i++)
		if (roleIndex == mp->role)
			return mp->text;

	return "invalid";
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void set_presentation_context_info(IXMLDOMDocument* dom, IXMLDOMElement* info, const wchar_t* name, LST_HEAD** l)
{
    if (*l == NULL)
		return;

    DUL_PRESENTATIONCONTEXT* ctx = (DUL_PRESENTATIONCONTEXT*)LST_Head(l);
    if (ctx == NULL)
		return;

	try {
	    LST_Position(l, ctx);
	    while (ctx != NULL) {

			AutoComPtr<IXMLDOMElement> elm;
			xduMakeElement(dom, info, name, elm.r(), xduThrowSimpleErrorInfo);

			xduMakeAttribute(dom, elm.p(), L"presentationContextID", static_cast<unsigned int>(ctx->presentationContextID), xduThrowSimpleErrorInfo);
			xduMakeAttribute(dom, elm.p(), L"result", static_cast<unsigned int>(ctx->result), xduThrowSimpleErrorInfo);
			xduMakeElement(dom, elm.p(), L"abstractSyntax", ctx->abstractSyntax, NULL, xduThrowSimpleErrorInfo);
			xduMakeElement(dom, elm.p(), L"proposedSCRole", lookup_role(ctx->proposedSCRole), NULL, xduThrowSimpleErrorInfo);
			xduMakeElement(dom, elm.p(), L"acceptedSCRole", lookup_role(ctx->acceptedSCRole), NULL, xduThrowSimpleErrorInfo);
			if (ctx->proposedTransferSyntax != NULL) {
				DUL_TRANSFERSYNTAX* transfer = (DUL_TRANSFERSYNTAX*)LST_Head((void**)&ctx->proposedTransferSyntax);
				if (transfer != NULL) {
					LST_Position(&ctx->proposedTransferSyntax, transfer);
					while (transfer != NULL) {
						xduMakeElement(dom, elm.p(), L"proposedTransferSyntax", transfer->transferSyntax, NULL, xduThrowSimpleErrorInfo);
						transfer = (DUL_TRANSFERSYNTAX*)LST_Next((void**)&ctx->proposedTransferSyntax);
					}
				}
			}
			
			xduMakeElement(dom, elm.p(), L"acceptedTransferSyntax", ctx->acceptedTransferSyntax, NULL, xduThrowSimpleErrorInfo);

			ctx = (DUL_PRESENTATIONCONTEXT*)LST_Next(l);
		}
		return;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"set_presentation_context_info(%1)", name);
		throw;
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static void set_association_params(IXMLDOMDocument* dom, IXMLDOMElement* associationElement, DUL_ASSOCIATESERVICEPARAMETERS* params)
{
	try {
		AutoComPtr<IXMLDOMElement> info;
		xduMakeElement(dom, associationElement, L"association-params", info.r(), xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"applicationContextName", params->applicationContextName, NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"callingAPTitle", params->callingAPTitle, NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"calledAPTitle", params->calledAPTitle, NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"respondingAPTitle", params->respondingAPTitle, NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"maxPDU", static_cast<unsigned int>(params->maxPDU), NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"peerMaxPDU", static_cast<unsigned int>(params->peerMaxPDU), NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"callingPresentationAddress", params->callingPresentationAddress, NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"calledPresentationAddress", params->calledPresentationAddress, NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"callingImplementationClassUID", params->callingImplementationClassUID, NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"callingImplementationVersionName", params->callingImplementationVersionName, NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"calledImplementationClassUID", params->calledImplementationClassUID, NULL, xduThrowSimpleErrorInfo);
		xduMakeElement(dom, info.p(), L"calledImplementationVersionName", params->calledImplementationVersionName, NULL, xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, info.p(), L"result", static_cast<unsigned int>(params->result), xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, info.p(), L"resultSource", static_cast<unsigned int>(params->resultSource), xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, info.p(), L"diagnostic", static_cast<unsigned int>(params->diagnostic), xduThrowSimpleErrorInfo);
		set_presentation_context_info(dom, info.p(), L"requestedPresentationContext", &params->requestedPresentationContext);
		set_presentation_context_info(dom, info.p(), L"acceptedPresentationContext", &params->acceptedPresentationContext);
		return;
	} catch (SimpleErrorInfo& ei) {
		ei.add(L"set_association_params()");
		throw;
	}
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
static HRESULT set_ctn_error(IXMLDOMDocument* dom, IXMLDOMElement* elm)
{
	HRESULT hr;

	CONDITION cond;
	char* str;
	while (COND_TopConditionEx(&cond, &str) != COND_NORMAL) {

		AutoComPtr<IXMLDOMElement> errElm;
		xduMakeElement(dom, elm, L"ctn-error", str, errElm.r(), xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, errElm.p(), L"cond", static_cast<unsigned int>(cond), xduThrowSimpleErrorInfo);

		COND_PopCondition(FALSE);
	}

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT RequestFileWatcher::SendFiles(DUL_NETWORKKEY* network, tWaitSocketParam& wsp, IXMLDOMDocument* dom) throw (SimpleErrorInfo)
{
	try {
		HRESULT ret_hr = S_OK;

		HRESULT hr;

		AutoComPtr<IXMLDOMElement> sessionElement;
		xduMakeElement(dom, NULL, L"SESSION", sessionElement.r(), xduThrowSimpleErrorInfo);
		xduMakeAttribute(dom, sessionElement.p(), L"start-time", (const SYSTEMTIME*)NULL, xduThrowSimpleErrorInfo);

		AutoComPtr<IXMLDOMElement> associationElement;

		//
		// Initialize Service Params
		//
		DUL_ASSOCIATIONKEY* association = NULL;
		DUL_ASSOCIATESERVICEPARAMETERS service;
		ZeroMemory(&service, sizeof(service));
		strcpy(service.applicationContextName, DICOM_STDAPPLICATIONCONTEXT);
	    sprintf(service.calledPresentationAddress, "%s:%u", m_config->m_host, m_config->m_port);
		strcpy(service.callingAPTitle, m_config->m_scu_title);
		strcpy(service.calledAPTitle, m_config->m_scp_title);
		strcpy(service.callingImplementationClassUID, MIR_IMPLEMENTATIONCLASSUID);
		strcpy(service.callingImplementationVersionName, MIR_IMPLEMENTATIONVERSIONNAME);
		service.maxPDU = 0x8000;
		Auto_DUL_ClearServiceParameters auto_clear(&service);

		char last_class_uid[DICOM_UI_LENGTH + 1] = "";
		char last_transfer_uid[DICOM_UI_LENGTH + 1] = "";

		BSTR bstr = SysAllocStringLen(NULL, 32);
		AutoSysFreeString af_bstr(bstr);
		
		AutoComPtr<IXMLDOMNodeList> list;
		SysReAllocString(&bstr, L"/C-STORE-LOG/REQUEST/FL/F");
		hr = dom->selectNodes(bstr, list.r());
		if (FAILED(hr))
			throw SimpleErrorInfo(hr, L"IXMLDOMDocument::selectNodes(%1)", bstr);

		while (true) {

			if (WaitForSingleObject(m_hStopEvent, 0) == WAIT_OBJECT_0) {
				AutoComPtr<IXMLDOMElement> stop;
				xduMakeElement(dom, sessionElement.p(), L"Stopped", stop.r(), xduThrowSimpleErrorInfo);
				xduMakeAttribute(dom, stop.p(), L"time", (const SYSTEMTIME*)NULL, xduThrowSimpleErrorInfo);
				break;
			}

			AutoComPtr<IXMLDOMNode> fileNode;
			hr = list->nextNode(fileNode.r());
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMNodeList::nextNode()");
			else if (hr == S_FALSE)
				break;


			SysReAllocString(&bstr, L"DICOM");
			AutoComPtr<IXMLDOMNode> dicomNode;
			hr = fileNode->selectSingleNode(bstr, dicomNode.r());
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMElement::selectSingleNode(%1)", bstr);
			else if (hr == S_FALSE)
				continue;

			AutoComPtr<IXMLDOMElement> fileElement;
			hr = fileNode->QueryInterface(IID_IXMLDOMElement, (void**)fileElement.r());
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMNodeList::QueryInterface(IID_IXMLDOMElement)");

			AutoComPtr<IXMLDOMElement> dicomElement;
			hr = dicomNode->QueryInterface(IID_IXMLDOMElement, (void**)dicomElement.r());
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMNodeList::QueryInterface(IID_IXMLDOMElement)");

			VARIANT var;
			VariantInit(&var);

			SysReAllocString(&bstr, L"path");
			hr = fileElement->getAttribute(bstr, &var);
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMElement::getAttribute(%1)", bstr);
			BSTR path = var.bstrVal;
			AutoSysFreeString af_path(path);

			SysReAllocString(&bstr, L"file-size");
			hr = fileElement->getAttribute(bstr, &var);
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMElement::getAttribute(%1)", bstr);
			hr = VariantChangeType(&var, &var, 0, VT_UI4);
			if (FAILED(hr)) {
				SimpleErrorInfo ei(hr, L"VariantChangeType(VT_UI4)");
				ei.add(L"attribute: file-size");
				throw ei;
			}
			const DWORD fileSize = var.ulVal;

			SysReAllocString(&bstr, L"id");
			hr = fileElement->getAttribute(bstr, &var);
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMElement::getAttribute(%1)", bstr);
			hr = VariantChangeType(&var, &var, 0, VT_UI4);
			if (FAILED(hr)) {
				SimpleErrorInfo ei(hr, L"VariantChangeType(VT_UI4)");
				ei.add(L"attribute: id");
				throw ei;
			}
			const DWORD fileID = var.ulVal;

			SysReAllocString(&bstr, L"TransferSyntaxUID");
			hr = dicomElement->getAttribute(bstr, &var);
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMElement::getAttribute(%1)", bstr);
			char transfer_syntax_uid[DICOM_UI_LENGTH + 1];
			bstr2uid(var.bstrVal, transfer_syntax_uid);
			VariantClear(&var);

			SysReAllocString(&bstr, L"SOPClassUID");
			hr = dicomElement->getAttribute(bstr, &var);
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMElement::getAttribute(%1)", bstr);
			char sop_class_uid[DICOM_UI_LENGTH + 1];
			bstr2uid(var.bstrVal, sop_class_uid);
			VariantClear(&var);

			SysReAllocString(&bstr, L"SOPInstanceUID");
			hr = dicomElement->getAttribute(bstr, &var);
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMElement::getAttribute(%1)", bstr);
			char sop_instance_uid[DICOM_UI_LENGTH + 1];
			bstr2uid(var.bstrVal, sop_instance_uid);
			VariantClear(&var);

			SysReAllocString(&bstr, L"meta-size");
			hr = dicomElement->getAttribute(bstr, &var);
			if (FAILED(hr))
				throw SimpleErrorInfo(hr, L"IXMLDOMElement::getAttribute(%1)", bstr);
			hr = VariantChangeType(&var, &var, 0, VT_UI4);
			if (FAILED(hr)) {
				SimpleErrorInfo ei(hr, L"VariantChangeType(VT_UI4)");
				ei.add(L"attribute: meta-size");
				throw ei;
			}
			const DWORD metaSize = var.ulVal;

			if (strcmp(last_class_uid, sop_class_uid) != 0
			 || strcmp(last_transfer_uid, transfer_syntax_uid) != 0) {
				if (association != NULL) {
					DUL_ReleaseAssociation(&association);
					Sleep(250);
					DUL_DropAssociation(&association);
					association = NULL;

					xduMakeAttribute(dom, associationElement.p(), L"finish-time", (const SYSTEMTIME*)NULL, xduThrowSimpleErrorInfo);
					associationElement.reset();
				}
				DUL_ClearServiceParameters(&service);

				CONDITION cond = SRV_RequestServiceClassEx(sop_class_uid, transfer_syntax_uid, DUL_SC_ROLE_SCU, &service);
				if (CTN_SUCCESS(cond))
					cond = DUL_RequestAssociation(&network, &service, &association);

				xduMakeElement(dom, sessionElement.p(), L"ASSOCIATION", associationElement.r(), xduThrowSimpleErrorInfo);
				xduMakeAttribute(dom, associationElement.p(), L"start-time", (const SYSTEMTIME*)NULL, xduThrowSimpleErrorInfo);
				set_association_params(dom, associationElement.p(), &service);

				if (!CTN_SUCCESS(cond)) {
					set_ctn_error(dom, associationElement.p());
					if (cond == DUL_EVENTSIGNALED)
						ret_hr = S_FALSE;
					break;
				}

				strcpy(last_class_uid, sop_class_uid);
				strcpy(last_transfer_uid, transfer_syntax_uid);
			}

			AutoComPtr<IXMLDOMElement> store;
			xduMakeElement(dom, associationElement.p(), L"STORE", store.r(), xduThrowSimpleErrorInfo);
			xduMakeAttribute(dom, store.p(), L"file-id", static_cast<unsigned int>(fileID), xduThrowSimpleErrorInfo);

			SYSTEMTIME startTime;
			GetLocalTime(&startTime);
			xduMakeAttribute(dom, store.p(), L"start-time", &startTime, xduThrowSimpleErrorInfo);

			CStackFileMapping map(path);
			if (map.lastError() != NO_ERROR) {
				HRESULT hr = HRESULT_FROM_WIN32(map.lastError());
				SimpleErrorInfo ei(hr, L"CStackFileMapping");
				xduMakeErrorInfoElement(dom, store.p(), ei, xduThrowSimpleErrorInfo);
				continue;
			}

			const char* const dataset_ptr = (const char*)map.mapPointer() + metaSize;
			const DWORD dataset_size = map.mapSize() - metaSize;

			MSG_C_STORE_REQ request;
			ZeroMemory(&request, sizeof(request));
			request.type = MSG_K_C_STORE_REQ;
			request.messageID = SRV_MessageIDOut();
			request.dataSetType = DCM_CMDDATAIMAGE;
			strcpy(request.classUID, sop_class_uid);
			strcpy(request.instanceUID, sop_instance_uid);
			strcpy(request.moveAETitle, "");
		
			MSG_C_STORE_RESP response;
			const CONDITION cond = SRV_CStoreRequestEx(&association, &service, &request, &response, dataset_ptr, dataset_size);

			if (CTN_SUCCESS(cond)) {
				SYSTEMTIME finishTime;
				GetLocalTime(&finishTime);

				ULARGE_INTEGER startFt;
				SystemTimeToFileTime(&startTime, (LPFILETIME)&startFt);
				ULARGE_INTEGER finishFt;
				SystemTimeToFileTime(&finishTime, (LPFILETIME)&finishFt);

				double elapsedTime = (double)(finishFt.QuadPart - startFt.QuadPart) / (double)(10 * 1000 * 1000);

				xduMakeAttribute(dom, store.p(), L"elapsed-time", elapsedTime, xduThrowSimpleErrorInfo);
			} else {
				AutoComPtr<IXMLDOMElement> respElm;
				xduMakeElement(dom, store.p(), L"C-STORE-RESPONSE", respElm.r(), xduThrowSimpleErrorInfo);

				MSG_STATUS_DESCRIPTION* statusDescription;
				MSG_StatusLookupEx(response.status, response.type, &statusDescription);
				xduMakeElement(dom, respElm.p(), L"description", statusDescription->description, NULL, xduThrowSimpleErrorInfo);
				xduMakeAttribute(dom, respElm.p(), L"status", static_cast<unsigned int>(response.status), xduThrowSimpleErrorInfo);

			    if (response.conditionalFields & MSG_K_C_STORERESP_ERRORCOMMENT)
			    	xduMakeElement(dom, respElm.p(), L"errorComment", response.errorComment, NULL, xduThrowSimpleErrorInfo);
				if (response.conditionalFields & MSG_K_C_STORERESP_CLASSUID)
			    	xduMakeElement(dom, respElm.p(), L"classUID", response.classUID, NULL, xduThrowSimpleErrorInfo);
				if (response.conditionalFields & MSG_K_C_STORERESP_INSTANCEUID)
			    	xduMakeElement(dom, respElm.p(), L"instanceUID", response.instanceUID, NULL, xduThrowSimpleErrorInfo);

				set_ctn_error(dom, store.p());
				if (cond == DUL_EVENTSIGNALED)
					ret_hr = S_FALSE;
				break;
			}
		}

		DUL_ReleaseAssociation(&association);
		Sleep(250);
		DUL_DropAssociation(&association);

		if (associationElement.p() != NULL)
			xduMakeAttribute(dom, associationElement.p(), L"finish-time", (const SYSTEMTIME*)NULL, xduThrowSimpleErrorInfo);

		xduMakeAttribute(dom, sessionElement.p(), L"finish-time", (const SYSTEMTIME*)NULL, xduThrowSimpleErrorInfo);

		return ret_hr;

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"RequestFileWatcher::SendFiles()");
		throw;
	}

	return E_UNEXPECTED;
}

//======================================================================
//
//======================================================================
DWORD WINAPI RequestFileWatcher::ThreadProc(LPVOID lpParam)
{
	RequestFileWatcher* w = (RequestFileWatcher*)lpParam;
	_endthreadex(w->thread_proc());
	return 0;
}

//----------------------------------------------------------------------------------------
//
//----------------------------------------------------------------------------------------
DWORD RequestFileWatcher::thread_proc()
{
	HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
	if (FAILED(hr)) {
		m_ei.set(hr, L"CoInitializeEx(NULL, COINIT_MULTITHREADED)");
		ConLog->Write(m_ei);
		return hr;
	}
	AutoCoUninitialize ac;

	// dummy call to create TLSs
	Get_CTN_HEAP_HANDLE();
	COND_Empty();

	//
	// Initialize Socket Parameter
	//
	tWaitSocketParam wsp;
	wsp.dwAcceptTimeout = INFINITE;
	wsp.dwReadSocketTimeout = 60 * 1000;	// 5 min.
	wsp.hAcceptEvent = CreateEvent(NULL, FALSE, FALSE, NULL); // Auto Reset Event
	wsp.hCloseEvent = CreateEvent(NULL, TRUE, TRUE, NULL);  // Manual Reset Event, Signaled
	wsp.hStopEvent = m_hStopEvent;
	AutoCloseHandle ac_hAcceptEvent(wsp.hAcceptEvent);
	AutoCloseHandle ac_hCloseEvent(wsp.hCloseEvent);

	//
	// Initialize DUL
	//
	DUL_NETWORKKEY* network = NULL;
	CONDITION cond = DUL_InitializeNetworkEx(DUL_NETWORK_TCP, DUL_AEREQUESTOR, NULL, DUL_TIMEOUT, DUL_ORDERBIGENDIAN, (void*)&wsp, &network);
	if (!CTN_SUCCESS(cond)) {
		m_ei.set_from_CTN_COND();
		ConLog->Write(m_ei);
		return E_FAIL;
	}
	Auto_DUL_DropNetwork auto_drop(&network);

	while (true) {
		try {
			DOMLog domlog;
			
			unsigned long numFilesToStore = 0;
			hr = ScanRequestFiles(m_config->m_watchDirectory, domlog.dom(), numFilesToStore);
			if (hr == S_OK && numFilesToStore > 0)
				hr = SendFiles(network, wsp, domlog.dom());

			if (hr == S_FALSE)
				return S_OK;

			domlog.Flush();
	
		} catch (SimpleErrorInfo& ei) {
			ei.add(L"MSXML (DOM) Error Occured");
			ConLog->Write(ei);
		}
	}

	return S_OK;
}
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HRESULT readConfig(const wchar_t* configPath, IXMLDOMNodeList** nodeList, SimpleErrorInfo& ei)
{
	AutoComPtr<IXMLDOMDocument> dom;
	HRESULT hr = CoCreateInstance(clsidDomDoc, NULL, CLSCTX_INPROC_SERVER, IID_IXMLDOMDocument, (void**)dom.r());
	if (FAILED(hr)) {
		ei.set(hr, L"CoCreateInstance(Msxml2.FreeThreadedDOMDocument, IID_IXMLDOMDocument)");
		return hr;
	}

	hr = dom->put_async(VARIANT_FALSE);
	if (FAILED(hr)) {
		ei.set(hr, L"IXMLDOMDocument::put_async(VARIANT_FALSE)");
		return hr;
	}

	VARIANT var;
	VariantInit(&var);

	BSTR bstr = SysAllocString(configPath);
	AutoSysFreeString af_bstr(bstr);

	var.vt = VT_BSTR;
	var.bstrVal = bstr;

	VARIANT_BOOL isSuccessful;
	hr = dom->load(var, &isSuccessful);
	if (FAILED(hr)) {
		ei.set(hr, L"IXMLDOMDocument::load(%1)", configPath);
		return hr;
	}

	AutoComPtr<IXMLDOMNodeList> list;
	hr = dom->selectNodes(L"/config/param", list.r());
	if (FAILED(hr)) {
		ei.set(hr, L"IXMLDOMDocument::selectNodes(PARAM)");
		return hr;
	}

	list->AddRef();
	*nodeList = list.p();

	return S_OK;
}

//======================================================================
//
//======================================================================
HRESULT RequestFileWatcher::Main(bool consoleMode, HANDLE hStopEvent, const wchar_t* const configPath)
{
	MyConsoleLog clog;
	clog.Init(consoleMode, L"C:\DIOWave\\logs\\send-dicom-svc");
	ConLog = &clog;

	SimpleErrorInfo ei;

	HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
	if (FAILED(hr)) {
		ei.set(hr, L"CoInitializeEx(NULL, COINIT_MULTITHREADED)");
		ConLog->Write(ei);
		return hr;
	}
	AutoCoUninitialize ac;

	//
	//
	//
	hr = CLSIDFromProgID(L"Msxml2.FreeThreadedDOMDocument", &clsidDomDoc);
	if (FAILED(hr)) {
		ei.set(hr, L"CLSIDFromProgID(Msxml2.FreeThreadedDOMDocument)");
		ConLog->Write(ei);
		return hr;
	}

	//
	//
	//
	Init_CTN_HEAP();

	//
	// Initialize to prepare COND facilitiy as Multi Theaded Aware
	//
	COND_Initialize();

	//
	// Initialize CTN DUL Facility
	//	
	WSADATA wsaData;
	if (WSAStartup(WINSOCK_VERSION, &wsaData)) {
		hr = HRESULT_FROM_WIN32(GetLastError());
		ei.set(hr, L"WSAStartup(%1!u!)", WINSOCK_VERSION);
		ConLog->Write(ei);
		return hr;
	}
	AutoWSACleanup ac_wsa;
	DUL_Initialize();

	//
	// load config
	//
	AutoComPtr<IXMLDOMNodeList> configNodes;
	hr = readConfig(configPath, configNodes.r(), ei);
	if (FAILED(hr)) {
		ei.add(L"readConfig(%1)", configPath);
		ConLog->Write(ei);
		return hr;
	}

	long nodeCount;
	configNodes->get_length(&nodeCount);
	if (nodeCount == 0) {
		ei.set(L"emply configration");
		ConLog->Write(ei);
		return E_UNEXPECTED;
	}

	RequestFileWatcher* watchers = (RequestFileWatcher*)_alloca(sizeof(RequestFileWatcher) * nodeCount);
	ZeroMemory(watchers, sizeof(RequestFileWatcher) * nodeCount);

	unsigned int index = 0;
	while (true) {
		AutoComPtr<IXMLDOMNode> configNode;
		hr = configNodes->nextNode(configNode.r());
		if (FAILED(hr)) {
			ei.set(hr, L"IXMLDOMNodeList::nextNode()");
			ConLog->Write(ei);
			break;
		}	
		else if (hr == S_FALSE)
			break;

		RequestFileWatcher* w = watchers + index;

		hr = w->Init(configNode.p());
		if (FAILED(hr)) {
			ConLog->Write(w->m_ei);
			break;
		}

		w->m_hStopEvent = hStopEvent;
		w->m_hThread = (HANDLE)_beginthreadex((void*)NULL, (unsigned)0, (unsigned (__stdcall *)(void *))RequestFileWatcher::ThreadProc, (void*)w, (unsigned)0, (unsigned int*)&w->m_dwThreadId);
		if (w->m_hThread == NULL) {
			hr = HRESULT_FROM_WIN32(GetLastError());
			ei.set(hr, L"_beginthreadex()");
			ConLog->Write(ei);
			break;
		}
		index++;
	}
	
	HANDLE* hWaitHandles = (HANDLE*)_alloca(sizeof(HANDLE) * (index + 1));
	ZeroMemory(hWaitHandles, sizeof(HANDLE) * (index + 1));
	hWaitHandles[0] = hStopEvent;
	unsigned long threadCount = 0;
	for (unsigned int i = 0; i < index; i++) {
		if (watchers[i].m_hThread != NULL) {
			hWaitHandles[++threadCount] = watchers[i].m_hThread;
		}
	}

	if (FAILED(hr)) {
		SetEvent(hStopEvent);
		WaitForMultipleObjects(threadCount, hWaitHandles + 1, TRUE, 60 * 1000);
		return hr;
	}

	ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 1000);
	//---------------------------------------------
	//
	const DWORD dwWaitFlag = WaitForMultipleObjects(threadCount + 1, hWaitHandles, FALSE, INFINITE);
	//
	//---------------------------------------------
	if (dwWaitFlag == WAIT_OBJECT_0) {
		ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 60 * 1000);
		hr = S_OK;
	} else if (dwWaitFlag == WAIT_FAILED) {
		hr = HRESULT_FROM_WIN32(GetLastError());
		ei.set(hr, L"WaitForMultipleObjects(%1!d! + 1, hWaitHandles, FALSE, INFINITE)");
		ConLog->Write(ei);
		return hr;
	} else {
		int i = dwWaitFlag - WAIT_OBJECT_0 - 1;
		if (i >= 0 && i < threadCount) {
			watchers[i].m_ei.add(L"thread unexpectedly terminated");
			ConLog->Write(watchers[i].m_ei);
			hr = watchers[i].m_ei.m_hr;
		}
		SetEvent(hStopEvent);
	}

	const DWORD dwWaitFlag2 = WaitForMultipleObjects(threadCount, hWaitHandles + 1, TRUE, 60 * 1000);
	if (dwWaitFlag2 == WAIT_TIMEOUT) {
		ConLog->Write(L"WaitForMultipleObjects(,60 * 1000): Thread Cleanup Timeout");
	} else if (dwWaitFlag2 == WAIT_FAILED) {
		const HRESULT hr = HRESULT_FROM_WIN32(GetLastError());
		ei.set(hr, L"WaitForMultipleObjects(): Thread cleanup Failed");
		ConLog->Write(ei);
	}

	return hr;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static HANDLE s_hStopEvent = NULL;
//----------------------------------------------------------------------
//
//----------------------------------------------------------------------
static BOOL WINAPI controlHandler(DWORD dwCtrlType)
{
	switch (dwCtrlType) {
	case CTRL_BREAK_EVENT:  // use Ctrl+C or Ctrl+Break to simulate
	case CTRL_C_EVENT:      // SERVICE_CONTROL_STOP in debug mode
		if (s_hStopEvent != NULL)
			SetEvent(s_hStopEvent);
		return TRUE;
	
	}
	return FALSE;
}

//======================================================================
//
//======================================================================
HRESULT RequestFileWatcher::SingleShot(RequestFileWatcher::config_t* config, const wchar_t* const requestedPath)
{
	m_config = config;

	MyConsoleLog clog;
	clog.Init(GetStdHandle(STD_ERROR_HANDLE));
	ConLog = &clog;

	SimpleErrorInfo ei;

	HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);
	if (FAILED(hr)) {
		ei.set(hr, L"CoInitializeEx(NULL, COINIT_MULTITHREADED)");
		ConLog->Write(ei);
		return hr;
	}
	AutoCoUninitialize ac;

	//
	//
	//
	hr = CLSIDFromProgID(L"Msxml2.FreeThreadedDOMDocument", &clsidDomDoc);
	if (FAILED(hr)) {
		ei.set(hr, L"CLSIDFromProgID(Msxml2.FreeThreadedDOMDocument)");
		ConLog->Write(ei);
		return hr;
	}

	//
	//
	//
	s_hStopEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	AutoCloseHandle ac_hStopEvent(s_hStopEvent);
	SetConsoleCtrlHandler(controlHandler, TRUE);
	m_hStopEvent = s_hStopEvent;
	
	//
	//
	//
	Init_CTN_HEAP();
	Get_CTN_HEAP_HANDLE();

	//
	// Initialize to prepare COND facilitiy as Multi Theaded Aware
	//
	COND_Initialize();
	COND_Empty();

	//
	// Initialize CTN DUL Facility
	//	
	WSADATA wsaData;
	if (WSAStartup(WINSOCK_VERSION, &wsaData)) {
		hr = HRESULT_FROM_WIN32(GetLastError());
		ei.set(hr, L"WSAStartup(%1!u!)", WINSOCK_VERSION);
		ConLog->Write(ei);
		return hr;
	}
	AutoWSACleanup ac_wsa;
	DUL_Initialize();

	//
	// Initialize Socket Parameter
	//
	tWaitSocketParam wsp;
	wsp.dwAcceptTimeout = INFINITE;
	wsp.dwReadSocketTimeout = 60 * 1000;	// 5 min.
	wsp.hAcceptEvent = CreateEvent(NULL, FALSE, FALSE, NULL); // Auto Reset Event
	wsp.hCloseEvent = CreateEvent(NULL, TRUE, TRUE, NULL);  // Manual Reset Event, Signaled
	wsp.hStopEvent = m_hStopEvent;
	AutoCloseHandle ac_hAcceptEvent(wsp.hAcceptEvent);
	AutoCloseHandle ac_hCloseEvent(wsp.hCloseEvent);

	//
	// Initialize DUL
	//
	DUL_NETWORKKEY* network = NULL;
	CONDITION cond = DUL_InitializeNetworkEx(DUL_NETWORK_TCP, DUL_AEREQUESTOR, NULL, DUL_TIMEOUT, DUL_ORDERBIGENDIAN, (void*)&wsp, &network);
	if (!CTN_SUCCESS(cond)) {
		m_ei.set_from_CTN_COND();
		ConLog->Write(m_ei);
		return E_FAIL;
	}
	Auto_DUL_DropNetwork auto_drop(&network);

	try {
		DOMLog domlog;

		unsigned long numFilesToStore = 0;
		unsigned long long numBytesToStore = 0;

		AutoComPtr<IXMLDOMElement> reqFileElm;
		xduMakeElement(domlog.dom(), NULL, L"REQUEST", reqFileElm.r(), NULL);

		HRESULT hr = collect_files(domlog.dom(), reqFileElm.p(), requestedPath, numFilesToStore, numBytesToStore);

		if (hr == S_OK && numFilesToStore > 0)
			hr = SendFiles(network, wsp, domlog.dom());

		domlog.Flush(GetStdHandle(STD_OUTPUT_HANDLE));

	} catch (SimpleErrorInfo& ei) {
		ei.add(L"MSXML (DOM) Error Occured");
		ConLog->Write(ei);
	}

	return S_OK;
}

