/*
DIOWave Visual Storage.
Copyright (C) 2002, Takaya Sakusabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "UtilClasses.h"
#include "dw-config.h"
#include "dcmdict.h"
#include "CDcmScanner.h"
#include "DataElementCache.h"
#include "InstanceCache.h"

#include "dio-study.dll.h"

//======================================================================
//
//======================================================================
static tDWConfig* gDWConfig = NULL;
//========================================================================================
//
// ISAPI Functions
//
//========================================================================================
//========================================================================================
// GetExtensionVersion
//========================================================================================
BOOL WINAPI GetExtensionVersion(HSE_VERSION_INFO  *pVer)
{
    pVer->dwExtensionVersion = MAKELONG( HSE_VERSION_MINOR,
                                         HSE_VERSION_MAJOR );
    strncpy( pVer->lpszExtensionDesc,
              "DIOWave Study Data <dio-study.isa> Ver. 0.9",
               HSE_MAX_EXT_DLL_NAME_LEN );

	const HRESULT hr = InitializeStudyCacheBuffer(TRUE);

	return SUCCEEDED(hr);
}

//========================================================================================
// TerminateExtension
//========================================================================================
BOOL WINAPI TerminateExtension(DWORD dwFlags)
{
	UninitializeStudyCacheBuffer();

	return TRUE;
}
//========================================================================================
// HttpExtensionProc
//========================================================================================
DWORD WINAPI HttpExtensionProc(EXTENSION_CONTROL_BLOCK *pECB)
{
	const char* pathInfo = pECB->lpszPathInfo;
	TOWC(pathInfo)

	LPVOID dataElmPtr;
	unsigned long dataElmSize;
	LPVOID instDirPtr;
	unsigned long instDirSize;
	LPVOID ctx;
	const HRESULT hr = MapStudyCacheBuffer(_pathInfo_WC, &dataElmPtr, &dataElmSize, &instDirPtr, &instDirSize, &ctx);
	if (FAILED(hr)) {
		HSE_SEND_HEADER_EX_INFO header_ex_info;
		header_ex_info.pszStatus = "500 Server Internal Error";
		header_ex_info.pszHeader = "Content-Type: text/plain\r\nContent-Length: 8\r\n\r\n";
		header_ex_info.cchStatus = 25;
		header_ex_info.cchHeader = 45;
		header_ex_info.fKeepConn = FALSE;
		if (!pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER_EX,
				                      &header_ex_info, NULL, NULL)) {
			pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
			return HSE_STATUS_ERROR;
		}
		DWORD bytesWrite = 8;
		char buf[12];
		sprintf(buf, "%X", hr);
		if (!pECB->WriteClient(pECB->ConnID, buf, &bytesWrite, HSE_IO_SYNC))
			return HSE_STATUS_ERROR;
		else
			return HSE_STATUS_SUCCESS;
	}

	const int content_length = sizeof(dataElmSize) + dataElmSize + sizeof(instDirSize) + instDirSize;
	char httpHeader[256];
	sprintf(httpHeader, "Content-Type: application/octet-stream\r\nContent-Length: %d\r\n\r\n", content_length);
	const int httpHeaderLen = strlen(httpHeader);

	HSE_SEND_HEADER_EX_INFO header_ex_info;
	static const char* const ok = "200 OK";
	header_ex_info.pszStatus = ok;
	header_ex_info.pszHeader = httpHeader;
	header_ex_info.cchStatus = strlen(ok);
	header_ex_info.cchHeader = httpHeaderLen;
	header_ex_info.fKeepConn = TRUE;
	if (!pECB->ServerSupportFunction(pECB->ConnID, HSE_REQ_SEND_RESPONSE_HEADER_EX,
				                      &header_ex_info, NULL, NULL)) {
		UnmapStudyCacheBuffer(ctx);
		pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
		return HSE_STATUS_ERROR;
	}

	DWORD bytesWrite;

	bytesWrite = sizeof(dataElmSize);
	if (!pECB->WriteClient(pECB->ConnID, &dataElmSize, &bytesWrite, HSE_IO_SYNC)) {
		UnmapStudyCacheBuffer(ctx);
		pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
		return HSE_STATUS_ERROR;
	}

	bytesWrite = dataElmSize;
	if (!pECB->WriteClient(pECB->ConnID, dataElmPtr, &bytesWrite, HSE_IO_SYNC)) {
		UnmapStudyCacheBuffer(ctx);
		pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
		return HSE_STATUS_ERROR;
	}

	bytesWrite = sizeof(instDirSize);
	if (!pECB->WriteClient(pECB->ConnID, &instDirSize, &bytesWrite, HSE_IO_SYNC)) {
		UnmapStudyCacheBuffer(ctx);
		pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
		return HSE_STATUS_ERROR;
	}

	bytesWrite = instDirSize;
	if (!pECB->WriteClient(pECB->ConnID, instDirPtr, &bytesWrite, HSE_IO_SYNC)) {
		UnmapStudyCacheBuffer(ctx);
		pECB->dwHttpStatusCode = HSE_STATUS_ERROR;
		return HSE_STATUS_ERROR;
	}

	UnmapStudyCacheBuffer(ctx);

	return HSE_STATUS_SUCCESS_AND_KEEP_CONN;
}	

	
	
	
	